// DumpFrame.cpp : Defines the entry point for the console application.
//

// uses TINE protocol, acquires 1 frame from VSV3 component with output feed and dumps it to disk;
// writing the header out in human-readable format as .txt file
// [writing the image out in full binary format (appended bits)]not yet
// writing the image bits out as text file 

#include "stdafx.h"


char *ReturnFormatString(int aImageFormat)
{
	if (aImageFormat == CF_IMAGE_FORMAT_GRAY)
		return "CF_IMAGE_FORMAT_GRAY";
	else if (aImageFormat == CF_IMAGE_FORMAT_RGB)
		return "CF_IMAGE_FORMAT_RGB";
	else if (aImageFormat == CF_IMAGE_FORMAT_RGBA)
		return "CF_IMAGE_FORMAT_RGBA";
	else if (aImageFormat == CF_IMAGE_FORMAT_RGB15)
		return "CF_IMAGE_FORMAT_RGB15";
	else if (aImageFormat == CF_IMAGE_FORMAT_RGB16)
		return "CF_IMAGE_FORMAT_RGB16";
	else if (aImageFormat == CF_IMAGE_FORMAT_JPEG)
		return "CF_IMAGE_FORMAT_JPEG";
	else if (aImageFormat == CF_IMAGE_FORMAT_TIFF)
		return "CF_IMAGE_FORMAT_TIFF";
	else if (aImageFormat == CF_IMAGE_FORMAT_YUV411)
		return "CF_IMAGE_FORMAT_YUV411";
	else if (aImageFormat == CF_IMAGE_FORMAT_YUV422)
		return "CF_IMAGE_FORMAT_YUV422";
	else if (aImageFormat == CF_IMAGE_FORMAT_YUV444)
		return "CF_IMAGE_FORMAT_YUV444";
	else if (aImageFormat == CF_IMAGE_FORMAT_HUFFYUV)
		return "CF_IMAGE_FORMAT_HUFFYUV";
	else if (aImageFormat == CF_IMAGE_FORMAT_BMP)
		return "CF_IMAGE_FORMAT_BMP";
	else if (aImageFormat == CF_IMAGE_FORMAT_BAYER)
		return "CF_IMAGE_FORMAT_BAYER";

	else return "UNKNOWN!";
}


char *ReturnFlagsString(int aImageFlags)
{
	static char returnBuffer[200];

	returnBuffer[0] = 0;

	if (aImageFlags&CF_IMAGE_FLAG_LITTLE_ENDIAN_BYTE_ORDER)
		strcat(returnBuffer, "LITTLE_ENDIAN");
	else
		strcat(returnBuffer, "BIG_ENDIAN");

	if (aImageFlags&CF_IMAGE_FLAG_IMAGE_LOSSLESS)
		strcat(returnBuffer, " LOSSLESS");

	if (aImageFlags&CF_IMAGE_FLAG_HORIZONTAL_FLIP_180)
		strcat(returnBuffer, " HORIZONTAL_FLIP_180");

	if (aImageFlags&CF_IMAGE_FLAG_VERTICAL_FLIP_180)
		strcat(returnBuffer, " VERTICAL_FLIP_180");

	return returnBuffer;
}



int main(int argc, char* argv[])
{
	char tinedevicestringbuffer[400];
	char tinepropertyname[64]={"Frame"};
	char filenamebase[250];
	char destfilename[255];

	if (argc < 3 || argc > 4)
	{
		printf("DumpFrame - no or too many parameters specified\n\n");
		printf("Usage:\n");
		printf("          DumpFrame.exe outfile vsv3_component_tine_path [property_name]\n\n");
		printf("  e.g.    DumpFrame.exe test /TEST/SGP_PRSGE1/Output Frame.Sched\n");
		printf("          DumpFrame.exe grab /TEST/SGP_PRSGE1/Output\n");
		printf("          DumpFrame.exe probe /TEST/SGP_PRSGE1/\n");
		printf("          DumpFrame.exe basic SGP_PRSGE1/Output\n\n");
		printf(" Remarks: For the outfile, any proper filename is supported, the\n");
		printf(" 		  extension must not be specified!\n");
		printf("		  If first slash is omitted, default context is used.\n");
		printf("          If last character is slash, Output is used.\n");
		printf("          If optional parameter is omitted, Frame is used.\n");
		

		exit(1);
	}

	filenamebase[sizeof(filenamebase)-1] = 0;
	strncpy(filenamebase, argv[1], sizeof(filenamebase)-1);

	tinedevicestringbuffer[sizeof(tinedevicestringbuffer)-1] = 0;
	strncpy(tinedevicestringbuffer, argv[2], sizeof(tinedevicestringbuffer)-1);

	// unsafe!
	if (strcmp(&tinedevicestringbuffer[strlen(tinedevicestringbuffer)-1], "/") == 0) strcat(tinedevicestringbuffer, "Output");

	if (argc == 4) 
	{
		tinepropertyname[63] = 0;
		strncpy(tinepropertyname, argv[3], 63);
	}


	DIMAGE	dimage; // (!)
	unsigned int imageFrameBufferSize = MAX_TRANSPORT_SIZE;
	unsigned char *pTineVideoBuffer = NULL;

	pTineVideoBuffer = new unsigned char[imageFrameBufferSize];
	if (pTineVideoBuffer == NULL)
	{
		printf("Out of memory error.");
		exit(2);
	}

	
	dimage.frameBufferSize = imageFrameBufferSize;
	dimage.frameBuffer = pTineVideoBuffer;

	DTYPE	dout;
	dout.dFormat = CF_IMAGE;
    dout.dArrayLength = MAX_TRANSPORT_SIZE;
    dout.data.vptr = &dimage;
	
	printf("DumpFrame: (1) Querying TINE %s/%s for 1 video frame.\n", tinedevicestringbuffer, tinepropertyname);

	// check link by executing it once!
	int ret = ExecLinkEx( tinedevicestringbuffer, tinepropertyname, &dout, NULL, CA_READ, 2000);
	if (ret!=0)
	{
		printf("Error querying property. TINE Error code: %d\n", ret);
		exit(3);
	}

	// write out that shite

	strcpy(destfilename, filenamebase);
	strcat(destfilename, ".hdr.txt");

	FILE *fp = fopen(destfilename, "wt");
	if (fp == NULL) 
	{
		printf("Error opening file \"%s\" for writing. Abort.\n", destfilename);
		exit(4);
	}

// a UINT32 b== nth position (starting by 1)
#define TAGOUT(a,b) ( (char) ((a>>((b-1)*8))&0xff) )

	fprintf(fp, "Information of vsv3 image header\n");
	fprintf(fp, "================================\n\n");
	fprintf(fp, "DIMAGE.sourceHeader\n");
	fprintf(fp, "   baseTag=%c%c%c%c\n",	TAGOUT(dimage.sourceHeader.baseTag,1),
											TAGOUT(dimage.sourceHeader.baseTag,2),
											TAGOUT(dimage.sourceHeader.baseTag,3),
											TAGOUT(dimage.sourceHeader.baseTag,4));
	fprintf(fp, "   cameraPortId=%d\n",dimage.sourceHeader.cameraPortId);

	fprintf(fp, "   versionTag=%u\n", dimage.sourceHeader.versionTag);
	fprintf(fp, "   totalLength=%u bytes\n", dimage.sourceHeader.totalLength);
	fprintf(fp, "   timestampSeconds=%u sec since epoch\n", dimage.sourceHeader.timestampSeconds);
	fprintf(fp, "   timestampMicroseconds=%u msec offset\n", dimage.sourceHeader.timestampMicroseconds);
	fprintf(fp, "   cameraPortName=\"%s\"\n", dimage.sourceHeader.cameraPortName);
	fprintf(fp, "\nDIMAGE.frameHeader\n");
	
	fprintf(fp, "   sourceWidth  = %d pixels\n", dimage.frameHeader.sourceWidth );
	fprintf(fp, "   sourceHeight = %d pixels\n", dimage.frameHeader.sourceHeight );
	fprintf(fp, "   aoiWidth     = %d pixels\n", dimage.frameHeader.aoiWidth );
	fprintf(fp, "   aoiHeight    = %d pixels\n", dimage.frameHeader.aoiHeight );
	fprintf(fp, "   xStart = %d offset in pixels\n", dimage.frameHeader.xStart );
	fprintf(fp, "   yStart = %d offset in pixels\n", dimage.frameHeader.yStart );
	fprintf(fp, "   bytesPerPixel = %d bytes\n", dimage.frameHeader.bytesPerPixel );
	fprintf(fp, "   effectiveBPP = %d bits\n", dimage.frameHeader.effectiveBitsPerPixel );
	fprintf(fp, "   horizontalBinning = %d to 1 merges\n", dimage.frameHeader.horizontalBinning );
	fprintf(fp, "   verticalBinning = %d to 1 merges\n", dimage.frameHeader.verticalBinning );
	fprintf(fp, "   sourceFormat = %s\n", ReturnFormatString(dimage.frameHeader.sourceFormat) );
	fprintf(fp, "   imageFormat = %s\n\n", ReturnFormatString(dimage.frameHeader.imageFormat) );
	fprintf(fp, "   frameNumber = %u\n", dimage.frameHeader.frameNumber );
	fprintf(fp, "   eventNumber = %u\n", dimage.frameHeader.eventNumber );
	fprintf(fp, "   xScale = %f (pixel per millimetre)\n", (double) dimage.frameHeader.xScale );
	fprintf(fp, "   yScale = %f (pixel per millimetre)\n", (double) dimage.frameHeader.yScale );
	fprintf(fp, "   imageRotation = %f (degrees, +CW -CCW)\n", (double) dimage.frameHeader.imageRotation );
	fprintf(fp, "   fspare1 = %f\n", (double) dimage.frameHeader.fspare1 );
	fprintf(fp, "   fspare2 = %f\n", (double) dimage.frameHeader.fspare2 );
	fprintf(fp, "   fspare3 = %f\n", (double) dimage.frameHeader.fspare3 );
	fprintf(fp, "   imageFlags = %s\n", ReturnFlagsString(dimage.frameHeader.imageFlags) );
	fprintf(fp, "   ispare1 = %d\n", dimage.frameHeader.ispare1 );
	fprintf(fp, "   ispare2 = %d\n", dimage.frameHeader.ispare2 );
	fprintf(fp, "   ispare3 = %d\n", dimage.frameHeader.ispare3 );
	fprintf(fp, "   appendedFrameSize = %d bytes\n", dimage.frameHeader.appendedFrameSize );

	fprintf(fp, "\nDIMAGE.frameBufferSize=%u bytes\n", dimage.frameBufferSize);	
	fprintf(fp, "DIMAGE.frameBuffer=0x%x memory location\n", dimage.frameBuffer);	
	fclose(fp);

	strcpy(destfilename, filenamebase);
	strcat(destfilename, ".frm.txt");

	fp = fopen(destfilename, "wt");
	if (fp == NULL) 
	{
		printf("Error opening file \"%s\" for writing. Abort.\n", destfilename);
		exit(4);
	}

	fprintf(fp, "Information of vsv3 frame buffer bits\n");
	fprintf(fp, "=====================================\n\n");
	fprintf(fp, "Contents: %u bytes (is: appendedFrameSize)\n\n", dimage.frameHeader.appendedFrameSize);
	fprintf(fp, "--\n\n");
	fprintf(fp, "         1:");
	unsigned char *buffer = (unsigned char *) dimage.frameBuffer;
	for (int i=0;i<dimage.frameHeader.appendedFrameSize;i++)
	{
		fprintf(fp, ".%02x", buffer[i]);
		if ((i+1)%20==0) fprintf(fp, "\n%010d:",i+1);
	}
	fclose(fp);

	delete [] pTineVideoBuffer;
	
	return 0;
}

