#ifndef THREADER_H_
#define THREADER_H_

#include "tinetype.h"
#if defined(__cplusplus) && !defined(FORCE_CPP)
extern "C" {
#endif

#ifdef MULTITHREADED
typedef struct 
{
  void (*tsk)(void);
  int rate;
  int hasReference;
  void *elRef;
  void *EqmRef;
} BkgThreadFcn;
extern TINE_EXPORT int useMultiThreadedBackgroundTasks;
extern TINE_EXPORT int useMultiThreadedEquipmentFunctions;
extern TINE_EXPORT int useMultiThreadedStockFunctions;
extern int hStkThread;
int bkgCreateThread(BkgThreadFcn *tsk);
#define thread_dbglog(txt) {\
    dbglog("[T1]: %s %s %s -> %s",\
    cl->contract.EqmName,cl->contract.EqmDeviceName,\
    cl->contract.EqmProperty,(txt));\
  }

#if defined(UNIX)

#define THREAD_TYPE void *
#define THREAD_PARAM void *
#include "pthread.h"
typedef pthread_mutex_t * MXHANDLE;
extern pthread_attr_t gBkgThreadAttr;
void *bkgThreadTask(void *lpvThreadParam);
#define threadExit(t) pthread_cancel((pthread_t)t)
#define ReleaseMutex(mx) pthread_mutex_unlock(mx)

#elif defined(WIN32)

#define THREAD_TYPE DWORD WINAPI 
#define THREAD_PARAM LPVOID
typedef HANDLE MXHANDLE;
extern DWORD dwBkgThreadId;
DWORD WINAPI bkgThreadTask(LPVOID lpvThreadParam);
#define threadExit(t) TerminateThread((HANDLE)t,0)
#define bkgThreadResume(t) ResumeThread((HANDLE)t)

#endif /* UNIX or WIN32 */

THREAD_TYPE stkThreadTask(THREAD_PARAM lpvThreadParam);
THREAD_TYPE eqmThreadTask(THREAD_PARAM lpvThreadParam);
int eqmCreateThread(THREAD_PARAM lpvThreadParam);
int eqmInitMutexSet(void *lpvMutexParam);
int stkInitMutexSet(void);

extern MXHANDLE hSystemInitMutex;
extern MXHANDLE hSystemServerMutex;
extern MXHANDLE hSystemKernelMutex;
extern MXHANDLE hSystemClientMutex;
extern MXHANDLE hLinkTblMutex;
extern MXHANDLE hLinkQueueMutex;
extern MXHANDLE hLogMutex;

extern int gSystemMutexSetInitialized;

int InitMutex(MXHANDLE *mx);
int InitSystemMutexSet(void);
int CloseSystemMutexSet(void);
int WaitForMutex(MXHANDLE mx,int to);
int ReleaseSystemMutex(MXHANDLE mx);
int CloseSystemMutex(MXHANDLE mx);

int stkCreateThread(THREAD_PARAM lpvThreadParam);

#define WaitForSystemMutex(mx,to) WaitForMutex(mx,to)

#define MTX_PRIMARY 0
#define MTX_SECONDARY 1
#define MTX_CARRIER 2

#else /* not MULTITHREADED */

extern int hSystemInitMutex;
extern int hSystemServerMutex;
extern int hSystemKernelMutex;
extern int hSystemClientMutex;
extern int hLinkTblMutex;
extern int hLinkQueueMutex;
extern int hLogMutex;

#endif /* MULTITHREADED */

#if defined(__cplusplus) && !defined(FORCE_CPP)
}
#endif

#endif /* THREADER_H_ */
