// IniEx_XML.cpp: implementation of the CIniEx XML wrapper class.
//
//////////////////////////////////////////////////////////////////////

#include "StdAfx.h"



/** \file IniEx_XML.cpp
	\brief Implements XML configuration file handling. 
	\sa IniEx_XML.h
    \author Stefan Weisse (SW) / DESY (adaption, refactoring, extending, move to XML)
	\author Koray Gecici (SW downloaded basic implementation from WWW in 2002 - codeguru.com)
	\date   2002-2008
*/


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////
const char CIniEx::WrapElement[]={"CIniExWrap"};

CIniEx::CIniEx(int aUnusedParameter)
{
	char szPath[MAX_PATH_LENGTH]; 
	char *pos;

	MAKE_COMPILER_HAPPY(aUnusedParameter);

	m_writeWhenChange=false;
	m_makeBackup=false;
	//m_NoCaseSensitive=true;
	m_Changed=false;

	m_ExePath[0] = 0;
	m_BackupFileName[0] =0;
	m_FileName[0] = 0;

	if( platformGetFullExecutablePath(szPath, MAX_PATH_LENGTH )== false ) memset(szPath, 0, MAX_PATH_LENGTH);	
    
	if ((pos = (char *) strrchr((const char *) szPath, PLF_PATH_SEPARATOR_CHAR)) != NULL) *pos = 0;

	if (strlen(szPath) == 0) 
		copywholestring(m_ExePath, "");
	else
		strncpy(m_ExePath, szPath, MAX_PATH_LENGTH-1);

}

CIniEx::~CIniEx()
{
	if (m_writeWhenChange)
		WriteFile(m_makeBackup);

	ResetContent();
}


// noCaseSensitive is always set to false internally, as XML files are case sensitive!
bool CIniEx::OpenAtExeDirectory(const char *pFileName,bool writeWhenChange,	bool createIfNotExist, bool noCaseSensitive,
							bool makeBackup)
{
	char filePath[MAX_PATH_LENGTH];
	
	copywholestring(filePath, m_ExePath);

	if (isRightmostChar( filePath, PLF_PATH_SEPARATOR_CHAR) == false) addstr(filePath,PLF_PATH_SEPARATOR_CHAR_AS_STRING);
	addstr(filePath, pFileName);

	return Open(filePath,writeWhenChange,createIfNotExist,noCaseSensitive,makeBackup);
}


// noCaseSensitive is always set to false internally, as XML files are case sensitive!
bool CIniEx::Open(const char *pFileName, bool writeWhenChange, bool createIfNotExist, bool noCaseSensitive,
				  bool makeBackup )
{
	//if there already was an ini file open for this instance
	//we have to save it and delete member variable contents
	
	if (!IsEmpty(m_FileName)) 
	{
		WriteFile();
		ResetContent();
	}

	MAKE_COMPILER_HAPPY(noCaseSensitive);
	//m_NoCaseSensitive=noCaseSensitive;

	m_writeWhenChange=writeWhenChange;
	m_makeBackup=makeBackup;
	
	bool loadOkay = m_Doc.LoadFile(pFileName);

	if ((loadOkay == false) && (createIfNotExist == true))
	{
		// check whether file can be opened, so it is there but maybe malformed XML, which
		// would not fit into the case of 'createIfNotExist'

		FILE *fp = fopen(pFileName, "rb");
		if (fp != NULL) // file is there, so another error
		{
			fclose(fp); 
			return false; 
		}
		else
		{
			ResetContent(); // initialize
			copywholestring(m_FileName,pFileName);
			return true;
		}
	}

	// file was loaded, so check whether it is a "good" file
	bool good = ValidateWrappedIniExInXML();
	if (good == true) 
		copywholestring(m_FileName,pFileName);
	else
		copywholestring(m_FileName,"");

	return good;
}


//returns backup file name
//if you didn't want backup (when openning file) it returns ""
char *CIniEx::WriteFile(bool makeBackup/*=false*/)
{
	//return m_BackupFileName;

	static char retStr[MAX_PATH_LENGTH];
	
	if (!m_Changed) 
		return "";
		
	char tmpFileName[MAX_PATH_LENGTH];
	
	copywholestring(tmpFileName, m_FileName);
	copywholestring(retStr, "");
	
	if (makeBackup)
	{
		if ( IsEmpty(m_BackupFileName) )
		{
			FindBackupFile();
		}

		// SW! wrap windows function Feb 13, 2008
		platformCopyFile( m_FileName, m_BackupFileName, false);
	}


	if (m_Doc.SaveFile( m_FileName) == true)
		m_Changed = false; 

	return m_BackupFileName;
}

bool CIniEx::GetWriteWhenChange(void)
{
	return m_writeWhenChange;
}


void CIniEx::SetWriteWhenChange(bool WriteWhenChange)
{
	m_writeWhenChange=WriteWhenChange;
}


//void CIniEx::SetBackupFileName(const char *backupFile)
//{
//	copywholestring(m_BackupFileName,backupFile);
//}

void CIniEx::FindBackupFile(void)
{
	// for convenience this is unimplemented at the moment
	// two lines replace to minor functionality (only 1 last backup file)
	
	/*char filePath[MAX_PATH_LENGTH];
	//char ext[MAX_PATH_LENGTH];

	WIN32_FIND_DATA ffData;
	BOOL bContinue=true;
	
	copywholestring(filePath, m_FileName);

	int nPlace=reversefind(filePath, '.');

	DeleteString(filePath,nPlace,strlen(filePath)-nPlace);
	addstr(filePath,PLF_ALLFILES_WILDCARD);

	int extNo=0;
	char *p;

	findfile

	HANDLE handle=FindFirstFile(filePath,&ffData);

	while (bContinue)
	{
		// SW! wrap windows function! Feb 13, 2008
		bContinue=FindNextFile(handle,&ffData);
		p=ffData.cFileName;
		p+=strlen(ffData.cFileName)-3;
		if (atoi(p)>extNo) extNo=atoi(p);
	}
	
	sprintf(m_BackupFileName, "%s.%.3d",m_FileName,extNo+1);*/
	
	int extNo=0;
	sprintf(m_BackupFileName, "%s.%.3d",m_FileName,extNo+1);
	
}


void CIniEx::ResetContent()
{
	// clean memory-mapped XML document
	// recreate basic outmost <CIniEx::WrapElement></CIniEx::WrapElement> section in XML
	m_Doc.Clear();

	TiXmlDeclaration * decl = new TiXmlDeclaration( "1.0", "", "" );
	TiXmlElement * element = new TiXmlElement( CIniEx::WrapElement );
	//TiXmlText * text = new TiXmlText( "World" );
	//element->LinkEndChild( text );
	m_Doc.LinkEndChild( decl );
	m_Doc.LinkEndChild( element );

	copywholestring(m_FileName,"");

	m_Changed=false;
}

/*
//Removes key and it's value from given section
bool CIniEx::RemoveKey(const char *Section,const char *Key)
{
	int nIndex=LookupSection(Section);
	if (nIndex==-1) return false;

	int nKeyIndex=LookupKey(nIndex,Key);
	if (nKeyIndex==-1) return false;

	RemoveAt(&(m_Keys[nIndex]),nKeyIndex);
	RemoveAt(&(m_Values[nIndex]),nKeyIndex);

	m_Changed=true;

	return true;
}

//Removes key and it's value 
bool CIniEx::RemoveKey(const char *Key)
{
	return RemoveKey("",Key);
}


//Removes given section(including all keys and values)
//return false when given section not found
//It won't couse memory leak because when deleting object
//the msize (malloc size) checking
bool CIniEx::RemoveSection(const char *Section)
{
	int nIndex=LookupSection(Section);
	if (nIndex==-1) return false;

	//m_Keys[nIndex]->RemoveAll();
	//m_Values[nIndex]->RemoveAll();

	delete [] m_Keys[nIndex];
	delete [] m_Values[nIndex];

	m_Keys[nIndex]=NULL;
	m_Values[nIndex]=NULL;

	RemoveAt(&m_Sections,nIndex);
	m_SectionNo--;
	m_Changed=true;
	
	return true;
}
*/

/*
int CIniEx::CompareStrings(const char *str1, const char *str2)
{
	if ((str1 == NULL) || (str2 == NULL)) return -1;
	if ((str1[0] == '#') || (str2[0] == '#'))
		return -1; // strings must be definetly not equal
	
	if (m_NoCaseSensitive)
		return stricmp((char *) str1, (char *)str2);
	else
		return strcmp(str1, str2);
}
*/

// unsure whether this is necessary at all, set to private in .h file now
char *CIniEx::GetValue(const char *Key)
{
	// section = ""
	static const char *DefaultValue="";

	TiXmlHandle docHandle( &m_Doc );
	
	TiXmlElement* child = docHandle.FirstChild( CIniEx::WrapElement ).FirstChild( Key ).ToElement();
	if (child != NULL)
	{
		char *str = (char *) child->GetText();
		if (str != NULL) return str;
	}

	return (char *) DefaultValue;
}


char *CIniEx::GetValue(const char *Section,const char *Key,const char *DefaultValue/*=""*/)
{
	TiXmlHandle docHandle( &m_Doc );
	
	TiXmlElement* child = docHandle.FirstChild( CIniEx::WrapElement ).FirstChild( Section ).FirstChild( Key ).ToElement();
	if (child != NULL)
	{
		char *str = (char *) child->GetText();
		if (str != NULL) return str;
	}

	return (char *) DefaultValue;
}


int CIniEx::GetValueInt(const char *Section,const char *Key,int defval /*DefaultValue=""*/)
{
	char tmp[MAX_STRING_LENGTH];
	sprintf(tmp, "%d", defval);
	
	int ret = atoi( (const char *) GetValue(Section, Key, tmp) );
	return( ret );
}

double CIniEx::GetValueDouble(const char *Section,const char *Key,double defval)
{
	char tmp[MAX_STRING_LENGTH];
	sprintf(tmp, "%f", defval);
	
	double ret = atof( (const char *) GetValue(Section, Key, tmp) );
	return( ret );
}


//sets for <Key>Value</Key> inside topmost section <CIniEx::WrapElement></CIniEx::WrapElement>
// unsure whether this is necessary at all, set to private in .h file now
void CIniEx::SetValue(const char *Key,const char *Value)
{
	//file opened?
	assert(!IsEmpty(m_FileName));

	TiXmlHandle docHandle1( &m_Doc );
	TiXmlHandle docHandle2( &m_Doc );

	TiXmlElement* root_element = docHandle1.FirstChild( CIniEx::WrapElement ).ToElement();
	assert( root_element != NULL );

	TiXmlElement* old_key_element = docHandle2.FirstChild( CIniEx::WrapElement ).FirstChild( Key ).ToElement();

	TiXmlElement* new_key_element = new TiXmlElement( Key );
	new_key_element->LinkEndChild( new TiXmlText( Value ) );

	if (old_key_element == NULL)
	{
		root_element->LinkEndChild( new_key_element );
	}
	else
	{
		root_element->ReplaceChild( old_key_element, *new_key_element);
	}

	m_Changed = true;
}

//writes <Key>value</Key> inside given <section> </section>
void CIniEx::SetValue(const char *Section,const char *Key,const char *Value)
{
	//file opened?
	assert(!IsEmpty(m_FileName));

	TiXmlHandle docHandle( &m_Doc );
	TiXmlHandle docHandle1( &m_Doc );
	TiXmlHandle docHandle2( &m_Doc );

	assert(strlen(Section) != 0 );

	TiXmlElement* section_element = docHandle.FirstChild( CIniEx::WrapElement ).FirstChild( Section ).ToElement();
	if (section_element == NULL)
	{
		section_element = new TiXmlElement( Section );

		TiXmlElement* root_element = docHandle1.FirstChild( CIniEx::WrapElement ).ToElement();

		assert( root_element != NULL );
		
		root_element->LinkEndChild( section_element );

		m_Changed = true;
	}

	TiXmlElement* old_key_element = docHandle2.FirstChild( CIniEx::WrapElement ).FirstChild( Section ).FirstChild( Key ).ToElement();

	TiXmlElement* new_key_element = new TiXmlElement( Key );
	new_key_element->LinkEndChild( new TiXmlText( Value ) );

	if (old_key_element == NULL)
	{
		section_element->LinkEndChild( new_key_element );
	}
	else
	{
		section_element->ReplaceChild( old_key_element, *new_key_element);
	}

	m_Changed = true;
}

void CIniEx::SetValueInt(const char *Section, const char *Key, int value )
{
	char tmp[MAX_STRING_LENGTH];

	sprintf(tmp, "%d", value);
	SetValue(Section, Key, tmp);
}

void CIniEx::SetValueDouble(const char *Section, const char *Key, double value )
{
	char tmp[MAX_STRING_LENGTH];

	sprintf(tmp, "%f", value);
	SetValue(Section, Key, tmp);
}

// ------------------------------------------- private ---------------------------
bool	CIniEx::ValidateWrappedIniExInXML()
{
	TiXmlHandle docHandle( &m_Doc );
	
	// validation means that outmost tag is CIniExWrap and (due to wellformed XML),
	// must only exist one single time
	//
	// TinyXML does not check for this special wellformedness (it allows
	// many docroot elements), so we have to do it

	TiXmlElement *child = docHandle.Child(CIniEx::WrapElement,0).ToElement();
	TiXmlElement *child2 = docHandle.Child(CIniEx::WrapElement,1).ToElement();

	if (child2 != NULL) return false;
	if (child == NULL) return false;

	return true;
}

TiXmlDocument *CIniEx::GetInternalXMLDocumentHandle()
{
	return &m_Doc;
}