// lossless orientation changes

#include "StdAfx.h"

const INT32 NUM_ORIENTATION_CHANGES = 8;
const char *ORIENTATION_CHANGE[] = {"NoChange","RotationBy90CW","RotationBy180","RotationBy270CW","MirrorAlongHorizontalAxis",
									"MirrorAlongVerticalAxis","RotationBy90CWThenMirrorAlongHorizontalAxis","RotationBy90CWThenMirrorAlongVerticalAxis"};

const INT32 ORIENTATION_CHANGE_NOTHING				= 0;
const INT32 ORIENTATION_CHANGE_ROT90CW				= 1;
const INT32 ORIENTATION_CHANGE_ROT180				= 2;
const INT32 ORIENTATION_CHANGE_ROT270CW				= 3;
const INT32 ORIENTATION_CHANGE_MIRRORHORIZ			= 4;
const INT32 ORIENTATION_CHANGE_MIRRORVERT			= 5;
const INT32 ORIENTATION_CHANGE_ROT90CWANDMIRRHORIZ	= 6;
const INT32 ORIENTATION_CHANGE_ROT90CWANDMIRRVERT	= 7;


// FIXME "inverse" proper imageFlags handling
// FIXME "inverse" image could have been processed before, so adjust imageFlags, rotation properly accordingly!

// FIXME: test 3bytesperpixel orientation change handling (untested)

bool bOrientationInitialized;
UINT8 *pTemporaryOrientationBuffer=NULL;

bool InitializeOrientation(void)
{
	if (bOrientationInitialized == true) return true;

	pTemporaryOrientationBuffer = new UINT8[CDIMAGE::MAX_TRANSPORT_SIZE];
	if (pTemporaryOrientationBuffer == NULL) return false;

	bOrientationInitialized = true;
	return true;
}


bool DeinitializeOrientation(void)
{
	if (bOrientationInitialized == false) return true;

	if (pTemporaryOrientationBuffer != NULL) delete [] pTemporaryOrientationBuffer;
	pTemporaryOrientationBuffer = NULL;

	bOrientationInitialized = false;
	return true;
}

void swapDimensionsAndInvolved( DIMAGE &aSwapDImage )
{
	INT32 tmp;
	float tmpScale;
	
	tmp = aSwapDImage.frameHeader.sourceWidth;
	aSwapDImage.frameHeader.sourceWidth = aSwapDImage.frameHeader.sourceHeight;
	aSwapDImage.frameHeader.sourceHeight = tmp;

	tmp = aSwapDImage.frameHeader.xStart;
	aSwapDImage.frameHeader.xStart = aSwapDImage.frameHeader.yStart;
	aSwapDImage.frameHeader.yStart = tmp;

	tmp = aSwapDImage.frameHeader.aoiWidth;
	aSwapDImage.frameHeader.aoiWidth = aSwapDImage.frameHeader.aoiHeight;
	aSwapDImage.frameHeader.aoiHeight = tmp;

	tmp = aSwapDImage.frameHeader.horizontalBinning;
	aSwapDImage.frameHeader.horizontalBinning = aSwapDImage.frameHeader.verticalBinning;
	aSwapDImage.frameHeader.verticalBinning = tmp;

	tmpScale = aSwapDImage.frameHeader.xScale;
	aSwapDImage.frameHeader.xScale = aSwapDImage.frameHeader.yScale;
	aSwapDImage.frameHeader.yScale = tmpScale;
}

UINT32 getAppendedWidth( const DIMAGE &aImg )
{
	if (aImg.frameHeader.aoiWidth > 0) 
		return aImg.frameHeader.aoiWidth;
	else
		return aImg.frameHeader.sourceWidth;
}


// 0 == error
UINT32 getAppendedHeight( const DIMAGE &aImg) 
{
	if (aImg.frameHeader.aoiHeight > 0) 
		return aImg.frameHeader.aoiHeight;
	else
		return aImg.frameHeader.sourceHeight;
}

// aImageBlock == CF_IMAGE+appended bits; aSourceBits == no CF_IMAGE, just bits
bool ROT90CWneu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = aImageBlock->frameBuffer;

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}
	else if (physbypp == 3)
	{
		// FIXME untested yet!
		const UINT8 *source = aSourceBits;
		UINT8 *dest = aImageBlock->frameBuffer;

		for (int x=0;x<widthbefore;x++)
		{
			int startPX = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[3*startPX];
				*dest++ = source[3*startPX+1];
				*dest++ = source[3*startPX+2];
				
				startPX -= widthbefore;
			}
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int start = (heightbefore-1)*widthbefore+x;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[start];
				start -= widthbefore;
			}
		}
	}
	else 
	{
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	// flip width and height at 90CW change
	swapDimensionsAndInvolved( *aImageBlock );
	
	aImageBlock->frameHeader.imageRotation = +90.0; // 90 clockwise(+)

	return true;
}


bool ROT180neu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		source += widthbefore*heightbefore;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;
		source += widthbefore*heightbefore;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}
	else if (physbypp == 3)
	{
		// FIXME untested yet!		
		const UINT8 *source = aSourceBits;
		
		unsigned int sourcePxPos = widthbefore*heightbefore;

		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = source[3*sourcePxPos];
			*dest++ = source[3*sourcePxPos+1];
			*dest++ = source[3*sourcePxPos+2];
			sourcePxPos--;
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		source += widthbefore*heightbefore;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		for (int n=0;n<widthbefore*heightbefore;n++)
		{
			*dest++ = *source--;
		}
	}
	else
	{
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	// on rotation by 180 degrees there is no change in width and height
	aImageBlock->frameHeader.imageRotation = +180.0; // 180

	return true;
}

bool ROT270CWneu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}
	else if (physbypp == 3)
	{
		// FIXME untested yet!
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int startPX = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[3*startPX];
				*dest++ = source[3*startPX+1];
				*dest++ = source[3*startPX+2];

				startPX += widthbefore;
			}
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int start = x;
			for (int y=0;y<heightbefore;y++)
			{
				*dest++ = source[start];
				start += widthbefore;
			}
		}
	}
	else
	{
		// no proper transformation found; do no change!
		// SW! maybe putting the source bits back is obsolete?
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	// flip width and height at 270CW change
	swapDimensionsAndInvolved( *aImageBlock );

	aImageBlock->frameHeader.imageRotation = +270.0; // 270 clockwise(+)

	return true;
}

bool MIRRORHORIZneu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if ((physbypp>=1) && (physbypp <= 4))
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int y=0;y<heightbefore;y++)
		{
			const UINT8 *srcline = source + ((heightbefore-y-1)*widthbefore)*physbypp;
			UINT8 *dstline = dest + (y*widthbefore)*physbypp;
			memcpy( dstline, srcline, widthbefore*physbypp);
		}
	}
	else
	{
		// no proper transformation found; do no change!
		// SW! maybe putting the source bits back is obsolete?
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	if (aImageBlock->frameHeader.imageFlags&CF_IMAGE_FLAG_HORIZONTAL_FLIP_180)
		aImageBlock->frameHeader.imageFlags |= ~CF_IMAGE_FLAG_HORIZONTAL_FLIP_180;
	else
		aImageBlock->frameHeader.imageFlags |= CF_IMAGE_FLAG_HORIZONTAL_FLIP_180;

	return true;
}

bool MIRRORVERTneu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;
	
	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}
	else if (physbypp == 3)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int deststartPX = widthbefore-x-1;
			int srcstartPX = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[3*deststartPX] = source[3*srcstartPX];    
				dest[3*deststartPX+1] = source[3*srcstartPX+1];    
				dest[3*deststartPX+2] = source[3*srcstartPX+2];    
				
				srcstartPX += widthbefore;
				deststartPX += widthbefore;
			}
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		for (int x=0;x<widthbefore;x++)
		{
			int deststart = widthbefore-x-1;
			int srcstart = x;
			for (int y=0;y<heightbefore;y++)
			{
				dest[deststart] = source[srcstart];    
				srcstart += widthbefore;
				deststart += widthbefore;
			}
		}
	}
	else
	{
		// no proper transformation found; do no change!
		// SW! maybe putting the source bits back is obsolete?
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	if (aImageBlock->frameHeader.imageFlags&CF_IMAGE_FLAG_VERTICAL_FLIP_180)
		aImageBlock->frameHeader.imageFlags |= ~CF_IMAGE_FLAG_VERTICAL_FLIP_180;
	else
		aImageBlock->frameHeader.imageFlags |= CF_IMAGE_FLAG_VERTICAL_FLIP_180;

	return true;

}

bool ROT90CWANDMIRRHORIZneu( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}
	else if (physbypp == 3)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int posPX = widthbefore*(heightbefore-1)+x;
			int subPX = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[3*posPX];
				*dest++ = source[3*posPX+1];
				*dest++ = source[3*posPX+2];

				posPX -= subPX;
			}
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		for (int x=widthbefore-1;x>=0;x--)
		{
			int pos = widthbefore*(heightbefore-1)+x;
			int sub = widthbefore;
			for (int y=heightbefore-1;y>=0;y--)
			{
				*dest++ = source[pos];
				pos -= sub;
			}
		}
	}
	else
	{
		// no proper transformation found; do no change!
		// SW! maybe putting the source bits back is obsolete?
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	// flip width and height at 90CW change
	swapDimensionsAndInvolved( *aImageBlock );	

	aImageBlock->frameHeader.imageRotation = +90.0; // 90 clockwise(+)

	if (aImageBlock->frameHeader.imageFlags&CF_IMAGE_FLAG_HORIZONTAL_FLIP_180)
		aImageBlock->frameHeader.imageFlags |= ~CF_IMAGE_FLAG_HORIZONTAL_FLIP_180;
	else
		aImageBlock->frameHeader.imageFlags |= CF_IMAGE_FLAG_HORIZONTAL_FLIP_180;

	return true;
}

// rotate 90 degrees clockwise and mirror vertical, too
bool ROT90CWANDMIRRVERT( DIMAGE *aImageBlock, const UINT8 *aSourceBits )
{
	assert( aImageBlock != NULL);
	assert( aSourceBits != NULL);

	INT32 widthbefore = getAppendedWidth( *aImageBlock );
	INT32 heightbefore = getAppendedHeight( *aImageBlock );
	INT32 physbypp = aImageBlock->frameHeader.bytesPerPixel;

	if (physbypp == 1)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		// R90CWMIRRVERT - aus jeder Zeile oben wird ne spalte links

		for (int y=0;y<heightbefore;y++)
		{
			int destpos = y;
			int destadd = heightbefore;

			for (int x=0;x<widthbefore;x++)
			{
				dest[destpos] = *source++;
				destpos += destadd;
			}
		}
	}
	else if (physbypp == 2)
	{
		const UINT16 *source = (const UINT16 *) aSourceBits;
		UINT16 *dest = (UINT16 *) (aImageBlock->frameBuffer);

		// R90CWMIRRVERT - aus jeder Zeile oben wird ne spalte links

		for (int y=0;y<heightbefore;y++)
		{
			int destpos = y;
			int destadd = heightbefore;

			for (int x=0;x<widthbefore;x++)
			{
				dest[destpos] = *source++;
				destpos += destadd;
			}
		}
	}
	else if (physbypp == 3)
	{
		const UINT8 *source = aSourceBits;
		UINT8 *dest = (aImageBlock->frameBuffer);

		for (int y=0;y<heightbefore;y++)
		{
			int destposPX = y;
			int destaddPX = heightbefore;

			for (int x=0;x<widthbefore;x++)
			{
				dest[3*destposPX] = *source++;
				dest[3*destposPX+1] = *source++;
				dest[3*destposPX+2] = *source++;
				destposPX += destaddPX;
			}
		}
	}
	else if (physbypp == 4)
	{
		const UINT32 *source = (const UINT32 *) aSourceBits;
		UINT32 *dest = (UINT32 *) (aImageBlock->frameBuffer);

		// R90CWMIRRVERT - aus jeder Zeile oben wird ne spalte links
		for (int y=0;y<heightbefore;y++)
		{
			int destpos = y;
			int destadd = heightbefore;

			for (int x=0;x<widthbefore;x++)
			{
				dest[destpos] = *source++;
				destpos += destadd;
			}
		}
	}
	else
	{
		// no proper transformation found; do no change!
		// SW! maybe putting the source bits back is obsolete?
		memcpy(aImageBlock->frameBuffer, aSourceBits, aImageBlock->frameHeader.appendedFrameSize);
		return false;
	}

	// flip width and height at 90CW change
	swapDimensionsAndInvolved( *aImageBlock );	

	aImageBlock->frameHeader.imageRotation = +90.0; // 90 clockwise(+)

	if (aImageBlock->frameHeader.imageFlags&CF_IMAGE_FLAG_VERTICAL_FLIP_180)
		aImageBlock->frameHeader.imageFlags |= ~CF_IMAGE_FLAG_VERTICAL_FLIP_180;
	else
		aImageBlock->frameHeader.imageFlags |= CF_IMAGE_FLAG_VERTICAL_FLIP_180;

	return true;
}


// in case of false the aImageBlock was NOT changed!
bool doOrientationChange( DIMAGE *aImageBlock, UINT32 aOrient )
{
	bool retval = false;

	assert( aImageBlock != NULL);
	assert( aOrient < NUM_ORIENTATION_CHANGES);

	if (bOrientationInitialized == false) return false;
	
	if (aOrient == ORIENTATION_CHANGE_NOTHING) return true;

	if ((aImageBlock->frameHeader.imageFormat != CDIMAGE::FORMAT_GRAY)	&&
		(aImageBlock->frameHeader.imageFormat != CDIMAGE::FORMAT_RGB)	&&
		(aImageBlock->frameHeader.imageFormat != CDIMAGE::FORMAT_RGBA)	&&
		(aImageBlock->frameHeader.imageFormat != CDIMAGE::FORMAT_RGB15) &&
		(aImageBlock->frameHeader.imageFormat != CDIMAGE::FORMAT_RGB16))	return false;

	memcpy(pTemporaryOrientationBuffer, aImageBlock->frameBuffer, aImageBlock->frameHeader.appendedFrameSize);

	switch( aOrient )
	{
		case ORIENTATION_CHANGE_ROT90CW:
				retval = ROT90CWneu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_ROT180:
				retval = ROT180neu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_ROT270CW:
				retval = ROT270CWneu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_MIRRORHORIZ:
				retval = MIRRORHORIZneu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_MIRRORVERT:
				retval = MIRRORVERTneu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_ROT90CWANDMIRRHORIZ:
				retval = ROT90CWANDMIRRHORIZneu( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		case ORIENTATION_CHANGE_ROT90CWANDMIRRVERT:
				retval = ROT90CWANDMIRRVERT( aImageBlock, pTemporaryOrientationBuffer );
				break;
		
		default:break;
	}

	return retval;
}


