#include "StdAfx.h"

#undef APPV_DATA
#undef APPV_MAJOR
#undef APPV_MINOR
#undef APPV_REVISION
#define APPV_DATE "Tue Feb 26 16:00:00 CEST ABCD 2008"
#define APPV_MAJOR 0
#define APPV_MINOR 9
#define APPV_REVISION 001


/**	\file rpcserv.cpp
	\brief Implements main() function and startup of the TINE server.

  	This is a changed version of the main TINE server startup file. It contains main() function which will
	start up the TINE server. Grabber specific initialisation is done inside the TINE server initialization.
	This TINE skeleton was improved by adding a win32 function TineServerConsoleHandler which does proper
	reaction on console (window) events like CTRL-C, ...

	\sa tine_server.cpp
    \author Stefan Weisse / DESY 
    \author Philip Duval / DESY (initial file)
	\date   2008
*/

void PostSystemInit(void);
void PreSystemInit(int argc, const char *argv[]);

extern void DoLog(const char *aFormat, ... );
extern void QuitCleanupFunction(int exitcode);

/**
	Windows Console server control handling. In order to do graceful exit of TINE server on certain console 
	(window) events, this special function was implemented and registered to Windows API inside main() function.

	For function prototype explanation see MSDN docs (-> SetConsoleCtrlHandler() ).
*/
BOOL WINAPI TineServerConsoleHandler(DWORD CEvent)
{

    switch(CEvent)
    {
		case CTRL_C_EVENT:
			DoLog("ConsoleEvent: CTRL-C hit. Application will quit gracefully.");
			QuitCleanupFunction(0);
			break;
		
		case CTRL_BREAK_EVENT:
			return TRUE;
			//DoLog("ConsoleEvent: CTRL-BREAK hit");
			break;
		
		case CTRL_CLOSE_EVENT:
			DoLog("ConsoleEvent: Server being closed. Application will quit gracefully.");
			QuitCleanupFunction(0);
			break;

		case CTRL_LOGOFF_EVENT:
			DoLog("ConsoleEvent: User is logging off. Application will quit gracefully.");
			QuitCleanupFunction(0);
			break;
	
		case CTRL_SHUTDOWN_EVENT:
			DoLog("ConsoleEvent: Machine is going to shut down. Application will quit gracefully.");
			QuitCleanupFunction(0);
			break;

		default: 
			DoLog("ConsoleEvent: Unknown event!");
			return TRUE;
			break;

    }
    
	return FALSE;
}

/*
extern "C" {
TINE_EXPORT void SetClientBurstLimit(int limit);
}
*/
int main(int argc, const char *argv[])
{
	int cc;

	MAKE_COMPILER_HAPPY(argc);
	MAKE_COMPILER_HAPPY(argv);

	AssignFecApplicationInfo(APPV_MAJOR,APPV_MINOR,APPV_REVISION);
	
	//StartupDebug = 0;
	
	if (SetConsoleCtrlHandler((PHANDLER_ROUTINE)TineServerConsoleHandler,TRUE)==FALSE)
	{
		// unable to install handler... 
		// display message to the user
		DoLog("Warning: Console handler could not be installed!");
	}

	//
	tineDebug=0;
	SetBurstLimit(6000); // 500
	SetCycleDelay(0); // 0

	PreSystemInit(argc, argv);
  
	/* init RPC server: */
	if ((cc=SystemInit(TRUE)) != 0) { printf("init error: %s\n",erlst[cc]); exit(1); }

	/* call FEC specific initialization: */
	PostSystemInit();
  
	for (;;)
	{
		SystemCycle(TRUE);
	}

	haltServer();
	return 0;
}
