//: TLink.java
// TODO: get rid of some of the hard-wired numbers below, such as '200' for the addr cache limit
package de.desy.tine.addrUtils;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.InetAddress;
import java.net.UnknownHostException;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.client.TLink;
import de.desy.tine.csvUtils.csv;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.startup.TInitializer;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.types.NAME16;

// TODO:  change all of the hardwired 200s and 100s below ...
public final class TSrvEntry
{
  private static TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
  private static final String fecHdr = "FecName, FecNetwork, FecNode, IPAddr, PortOffset, TineProtocol";
  private static final String srvHdr = "Name, FecName, EqpModule, Context, SubSystem";
  private static String crlf;
  private static String cacheFilePath;
  private static final int CACHE_TABLE_LENGTH = 1000;
  public static final String SRVEQM_NAME = "_SRV__";
  public static final String SRVFEC_NAME = "NETWORK";
  public static final String SRVEXP_NAME = "NETWORK";
  private static BufferedReader fecR = null;
  private static BufferedReader ensR = null;
  private static BufferedReader srvR = null;
  private static BufferedReader fecCacheR = null;
  private static BufferedWriter fecCacheW = null;
  private static BufferedWriter srvCacheW = null;
  private static String fecCacheFile = null;
  private static String srvCacheFile = null;
  private static boolean initialized = false;
  private static boolean initializing = false;
  private static int ensSrvCacheIndex = -1;
  private static int ensDevCacheIndex = -1;
  public String devName;
  public String eqmName;
  public String fecName;
  public String cntName;
  static NameCache[] devNameCache = new NameCache[CACHE_TABLE_LENGTH];
  static String[] modNameCache = new String[CACHE_TABLE_LENGTH];
  static TFecEntry[] fecAddrCache = new TFecEntry[CACHE_TABLE_LENGTH];
  static int modNameCacheLength = 0;
  static int fecAddrCacheLength = 0;
  static int NameCacheLength = 0;
  static int numConfiguredNameServers = 0;
  static int currentConfiguredNameServer = 0;
  public static String currentConfiguredNameServerTag;
  public TFecEntry fecAddr;
  private csv srvFile;
  static int debugLevel = 0;
  public boolean isLegacy = false;
  public static void reset()
  {
    // leave the configured ENS and GENS in the table ..
    if (modNameCacheLength > 2) modNameCacheLength = 2;
    if (fecAddrCacheLength > 2) fecAddrCacheLength = 2;
    if (NameCacheLength > 2) NameCacheLength = 2;
  }
  public String toString(String srvName,String ctxName)
  {
    if (eqmName == null) return new String("module does not exist");
    StaticAddress(srvName,ctxName);
    String modstr = new String("");
    modstr += "Local Equipment Module : " + eqmName.trim();
    modstr += "\nFront End Computer :" + fecName.trim();
    if (fecAddr != null && fecAddr.fecHost != null)
    {
      modstr += "\nIP Address : " + fecAddr.fecHost.getHostAddress();
      modstr += "\nPort Offset : " + fecAddr.fecPortOffset;
      modstr += "\nHost Name : " + fecAddr.fecHost.getHostName();
    }
    return modstr;
  }
  void StaticAddress(String srvName,String ctxName)
  {
    int i;
    if (srvName == null || ctxName == null) 
    {
      eqmName = ""; fecName = ""; cntName = "";
      return;
    }
    for (i=0; i<NameCacheLength; i++) 
    {
      if (ctxName.length() > 0 && 
          ctxName.compareToIgnoreCase("DEFAULT") != 0 &&
          ctxName.compareToIgnoreCase(devNameCache[i].cntName) != 0) continue;
      if (srvName.compareToIgnoreCase(devNameCache[i].devName) == 0)
      {
        eqmName = modNameCache[devNameCache[i].modNameIndex];
        fecName = fecAddrCache[devNameCache[i].srvNameIndex].fecName;
        fecAddr = fecAddrCache[devNameCache[i].srvNameIndex];
    	return;
      }
    }
    // acquire name info from ens or file:
    if (srvName.startsWith("ENS"))
    {
      eqmName = new String("ENSEQM");
      fecName = new String("ENS");
      cntName = new String("SITE");
    }
    if (srvName.compareTo("NAMEQRY") == 0)
    {
      eqmName = new String("ENSEQM");
      fecName = new String("ENS");
      cntName = new String("SITE");
    }
    if (srvName.startsWith("GENS"))
    {
      eqmName = new String("GRPEQM");
      fecName = new String("GENS");
      cntName = new String("SITE");
    }
    if (eqmName == null) return;
    if (modNameCacheLength >= CACHE_TABLE_LENGTH) return;
    if (fecAddrCacheLength >= CACHE_TABLE_LENGTH) return;
    if (NameCacheLength < CACHE_TABLE_LENGTH) devNameCache[NameCacheLength] = new NameCache();
    for (i=0; i<modNameCacheLength; i++) 
      if (eqmName.compareTo(modNameCache[i]) == 0) break;
    if (i == modNameCacheLength) modNameCache[modNameCacheLength++] = eqmName;
    devNameCache[NameCacheLength].modNameIndex = i;
    for (i=0; i<fecAddrCacheLength; i++) 
      if (fecName.compareTo(fecAddrCache[i].fecName) == 0) break;
    if (i == fecAddrCacheLength)
    {
      fecAddrCache[fecAddrCacheLength] = new TFecEntry(fecName);
      fecAddr = fecAddrCache[fecAddrCacheLength++];
      if (fecName.startsWith("ENS")) numConfiguredNameServers++;
    }
    devNameCache[NameCacheLength].srvNameIndex = i;
    devNameCache[NameCacheLength].devName = srvName;
    devNameCache[NameCacheLength].cntName = cntName;
    if (NameCacheLength < CACHE_TABLE_LENGTH) NameCacheLength++;
  }
  static int addNameServerToServerCache(TFecEntry sa)
  {
    if (fecAddrCacheLength > CACHE_TABLE_LENGTH || 
        NameCacheLength > CACHE_TABLE_LENGTH || 
        numConfiguredNameServers > 10) return -1;
    if (debugLevel > 1) DbgLog.log("addNameServerToServerCache","Adding ENS " + numConfiguredNameServers + " to address cache");
    if (ensSrvCacheIndex == -1) ensSrvCacheIndex = fecAddrCacheLength;
    if (ensDevCacheIndex == -1) ensDevCacheIndex = NameCacheLength;
    fecAddrCache[fecAddrCacheLength] = sa;
    devNameCache[NameCacheLength] = new NameCache();
    devNameCache[NameCacheLength].srvNameIndex = fecAddrCacheLength;
    devNameCache[NameCacheLength].modNameIndex = modNameCacheLength;
    devNameCache[NameCacheLength].devName = sa.fecName;
    devNameCache[NameCacheLength].cntName = "SITE";
    modNameCache[modNameCacheLength] = "ENSEQM";
    fecAddrCacheLength++;
    modNameCacheLength++;
    NameCacheLength++;
    numConfiguredNameServers++;
    if (numConfiguredNameServers == 1)  // been called at least once ...
    {
      currentConfiguredNameServerTag = new String(sa.fecName);
    }
    return 0;
  }
  static int addGroupServerToServerCache(TFecEntry sa)
  { // TODO: meld this cut-and-paster with the above routine ...
    if (fecAddrCacheLength > CACHE_TABLE_LENGTH || NameCacheLength > CACHE_TABLE_LENGTH || numConfiguredNameServers > 10)
      return -1;
    if (debugLevel > 1) DbgLog.log("addGroupServerToServerCache","Adding GENS " + numConfiguredNameServers + " to address cache");
    fecAddrCache[fecAddrCacheLength] = sa;
    devNameCache[NameCacheLength] = new NameCache();
    devNameCache[NameCacheLength].srvNameIndex = fecAddrCacheLength;
    devNameCache[NameCacheLength].modNameIndex = modNameCacheLength;
    devNameCache[NameCacheLength].devName = sa.fecName;
    devNameCache[NameCacheLength].cntName = "SITE";
    modNameCache[modNameCacheLength] = "GRPEQM";
    fecAddrCacheLength++;
    NameCacheLength++;
    modNameCacheLength++;
    return 0;
  }
  static int addNameServer(String ip)
  {
    TFecEntry sa;
    SrvAddr da;
    da = new SrvAddr();
    da.fecName = new String("ENS#" + numConfiguredNameServers);
    da.eqmContext = new String("SITE");
    da.eqmName = new String("ENSEQM");
    da.ipAddr = new String(ip);
    da.portOffset = 0;
    da.inetProtocol = TTransport.UDP; // 2;
    da.tineProtocol = TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
    sa = new TFecEntry(da);
    if (!sa.fecRegistered) return -1;
    addNameServerToServerCache(sa);
    // now add the GENS ...
    da.fecName = new String("GENS#" + numConfiguredNameServers);
    da.eqmName = new String("GRPEQM");    
    da.portOffset = 101;
    sa = new TFecEntry(da);
    if (!sa.fecRegistered) return -1;
    addGroupServerToServerCache(sa);
    return 0;
  }
  static synchronized void SetAddrFileLocation()
  {
    try
    {
      if (fecR.ready() && srvR.ready())
      {
        if (TLinkFactory.debugLevel > 1) DbgLog.log("SetAddrFileLocation","static database readers ready");
      }
    }
    catch (Exception e)
    {
      fecR = initializer.getAddressResource();
      srvR = initializer.getEquipmentResource();
    }
    if (initialized) return;
    if (initializing) return;
    initializing = true;
//    fecR = initializer.getAddressResource();
//    srvR = initializer.getEquipmentResource();
    
    String ensName;
    TFecEntry sa;
    currentConfiguredNameServerTag = new String("ENS");  // default
    
    for (int i=0; i<5; i++)  // don't allow more than 5 configured name servers ...
    {
      ensName = new String("ENS#" + i);	
      if (debugLevel > 1) DbgLog.log("SetAddrFileLocation","ENS: looking for " + ensName);
      ensR = initializer.getHostsResource();
      if (ensR != null)   
      {
        sa = new TFecEntry(ensR,ensName);
        if (!sa.fecRegistered)
        { // maybe a simple name is entered in the cshosts.csv file ?
          sa = new TFecEntry(ensR,"ENS");
        }
      }
      else 
      { // try to discover it 
        findNameServerOnNetwork();
        if (numConfiguredNameServers > 0) break;
        findNameServerFromDNS();
        if (numConfiguredNameServers > 0) break;
        // not found on network: use a hard-wired address
        sa = new TFecEntry("ENS");
      }
      if (!sa.fecRegistered) 
      { 
        if (debugLevel > 1) DbgLog.log("SetAddrFileLocation",ensName + " Not Registered");
        continue;
      } 
      addNameServerToServerCache(sa);
      if (ensR == null) break;
    }

    if (File.separatorChar == '/')
    { // unix or mac
      cacheFilePath = "/var/tmp";
    }
    else if (File.separatorChar == '\\')
    { // windows
      cacheFilePath = System.getenv("SystemDrive");
      if (cacheFilePath == null) cacheFilePath = "C:";
      crlf = "\r\n";
    }
    else
    {
      cacheFilePath = ".";
    }
    
    cacheFilePath += File.separator + "tine" + File.separator + "cache" + File.separator;
    srvCacheFile = cacheFilePath + "eqpdbase.csv";
    fecCacheFile = cacheFilePath + "fecaddr.csv";

    initialized = true;
    initializing = false;
}
  public TSrvEntry()
  {
    eqmName = ""; devName = ""; cntName = ""; fecName = "";
    SetAddrFileLocation();
  }
  private static boolean isInSrvFile(SrvAddr srv)
  {
    boolean hasEntry = false;
    csv csvf = new csv(srvCacheFile);
    try
    {
      String s, hdr;
      int srv_col = -1,eqn_col = -1 ,fec_col = -1,ctx_col = -1;
      boolean done = false;
      while((s=csvf.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!done)
        {
          hdr = s;
          if ((srv_col=csvf.findcol(hdr,"NAME")) < 0) throw new NoSuchFieldException();
          if ((eqn_col=csvf.findcol(hdr,"EQPMODULE")) < 0) throw new NoSuchFieldException();
          if ((fec_col=csvf.findcol(hdr,"FECNAME")) < 0) throw new NoSuchFieldException();
          if ((ctx_col=csvf.findcol(hdr,"CONTEXT")) < 0) throw new NoSuchFieldException();
          done = true;
          continue;
        }
        if (csvf.namcmp(srv.expName,s,srv_col) != 0) continue;
        if (csvf.namcmp(srv.eqmName,s,eqn_col) != 0) continue;
        if (csvf.namcmp(srv.fecName,s,fec_col) != 0) continue;
        if (csvf.namcmp(srv.eqmContext,s,ctx_col) != 0) continue;
        hasEntry = true;
        break;
      }
    }
    catch (Exception e)
    { // database corrupt or not found
      hasEntry = false;
    }
    finally
    {
      csvf.close();
    }
    return hasEntry;
  }
  private static boolean isInFecFile(SrvAddr srv)
  {
    boolean hasEntry = false;
    csv csvf = new csv(fecCacheFile);
    try
    {
      String s, hdr;
      int fec_col = -1,ip_col = -1 ,port_col = -1,prot_col = -1;
      boolean done = false;
      while((s=csvf.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!done)
        {
          hdr = s;
          if ((fec_col=csvf.findcol(hdr,"FEC_NAME")) < 0) throw new NoSuchFieldException();
          if ((ip_col=csvf.findcol(hdr,"IP_ADDR")) < 0) throw new NoSuchFieldException();
          if ((port_col=csvf.findcol(hdr,"PORT_OFFSET")) < 0) throw new NoSuchFieldException();
          prot_col = csvf.findcol(hdr,"TINE_PROTOCOL");
          done = true;
          continue;
        }
        if (csvf.namcmp(srv.fecName,s,fec_col) != 0) continue;
        if (csvf.namcmp(srv.ipAddr,s,ip_col) != 0) continue;
        if (srv.portOffset != Integer.parseInt(csvf.colptr(port_col,s))) continue;
        if (prot_col < 0) continue;
        if (srv.tineProtocol != Integer.parseInt(csvf.colptr(prot_col,s))) continue;
        hasEntry = true;
        break;
      }
    }
    catch (Exception e)
    { // database corrupt or not found
      hasEntry = false;
    }
    finally
    {
      csvf.close();
    }
    return hasEntry;
  }
  private static boolean addToSrvFile(SrvAddr srv)
  {
    boolean hasEntry = false;
    boolean needHeader = true;
    csv csvr = new csv(srvCacheFile);
    try
    {
      File csvw = new File(cacheFilePath);
      csvw.mkdirs();
      srvCacheW = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(srvCacheFile + ".tmp")));
      String s, hdr;
      int srv_col = -1,ctx_col = -1;
      boolean done = false;
      while((s=csvr.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!done)
        {
          hdr = s;
          if ((srv_col=csvr.findcol(hdr,"NAME")) < 0) throw new NoSuchFieldException();
          if ((ctx_col=csvr.findcol(hdr,"CONTEXT")) < 0) throw new NoSuchFieldException();
          srvCacheW.write(srvHdr.trim() + crlf);
          needHeader = false;
          done = true;
          continue;
        }
        if (csvr.namcmp(srv.expName,s,srv_col) == 0 &&
            csvr.namcmp(srv.eqmContext,s,ctx_col) == 0)
        { // found the entry -> update the particulars
          s = srv.expName.trim() + ", " + srv.fecName.trim() + ", " + srv.eqmName.trim()
              + ", " +  srv.eqmContext.trim() + ", " + srv.subSystem.trim();
          hasEntry = true;
        }
        srvCacheW.write(s + crlf);
      }
      if (needHeader) srvCacheW.write(srvHdr.trim() + crlf);
      if (!hasEntry)
      { // it's a new one !
        s = srv.expName.trim() + "," + srv.fecName.trim() + "," + srv.eqmName.trim()
            + "," +  srv.eqmContext.trim() + "," + srv.subSystem.trim();
        srvCacheW.write(s + crlf);        
      }
      srvCacheW.close();
      csvr.close();
      csvw = new File(srvCacheFile);
      csvw.delete();
      File srvf = new File(srvCacheFile + ".tmp");
      srvf.renameTo(new File(srvCacheFile));
    }
    catch (Exception e)
    { // database corrupt or not found
      hasEntry = false;
    }
    finally
    {
      csvr.close();
    }
    return hasEntry;
  }
  private static boolean addToFecFile(SrvAddr srv)
  {
    boolean needHeader = true;
    boolean hasEntry = false;
    csv csvr = new csv(fecCacheFile);
    try
    {
      File csvw = new File(cacheFilePath);
      csvw.mkdirs();
      fecCacheW = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(fecCacheFile + ".tmp")));
      String s, hdr;
      int fec_col = -1;
      boolean done = false;
      while((s=csvr.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!done)
        {
          hdr = s;
          if ((fec_col=csvr.findcol(hdr,"FEC_NAME")) < 0) throw new NoSuchFieldException();
          fecCacheW.write(fecHdr.trim() + crlf);
          done = true;
          needHeader = false;
          continue;
        }
        if (csvr.namcmp(srv.fecName,s,fec_col) == 0)
        {
          s = srv.fecName.trim() + ",00000000,000000000000," + srv.ipAddr.trim() + "," + 
              srv.portOffset + "," + srv.tineProtocol;  
          hasEntry = true;
        }
        fecCacheW.write(s + crlf);
      }
      if (needHeader) fecCacheW.write(fecHdr.trim() + crlf);
      if (!hasEntry)
      { // it's a new one !
        s = srv.fecName.trim() + ",00000000,000000000000," + srv.ipAddr.trim() + "," + 
            srv.portOffset + "," + srv.tineProtocol;  
        fecCacheW.write(s + crlf);        
      }
      csvr.close();
      fecCacheW.close();
      csvw = new File(fecCacheFile);
      csvw.delete();
      File fecf = new File(fecCacheFile + ".tmp");
      fecf.renameTo(new File(fecCacheFile));
    }
    catch (Exception e)
    { // database corrupt or not found
      hasEntry = false;
    }
    finally
    {
      csvr.close();
    }
    return hasEntry;
  }
  public static void addAddressToCacheFile(SrvAddr srv)
  {
    if (srv == null) return;
    if (!isInSrvFile(srv)) addToSrvFile(srv);
    if (!isInFecFile(srv)) addToFecFile(srv);    
  }
  private int fillinAddressFromFileCache(String srvName,String ctxName)
  {
    int cc = TErrorList.non_existent;
    try
    {
      String s, hdr;
      int i, nam_col = -1,eqn_col = -1 ,srv_col = -1,ctxt_col = -1;
      boolean done = false;
      while((s=srvFile.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!done)
        {
          hdr = s;
          if ((nam_col=srvFile.findcol(hdr,"NAME")) < 0) throw new NoSuchFieldException();
          if ((eqn_col=srvFile.findcol(hdr,"EQPMODULE")) < 0) throw new NoSuchFieldException();
          if ((srv_col=srvFile.findcol(hdr,"FECNAME")) < 0) throw new NoSuchFieldException();
          if ((ctxt_col=srvFile.findcol(hdr,"CONTEXT")) < 0) throw new NoSuchFieldException();
          done = true;
          continue;
        }
        if (srvFile.namcmp(srvName,s,nam_col) == 0)
        {
            eqmName = srvFile.colptr(eqn_col,s);
            fecName = srvFile.colptr(srv_col,s);
            cntName = srvFile.colptr(ctxt_col,s);
            if (ctxName != null && ctxName.compareToIgnoreCase(cntName) != 0) continue;
            if (modNameCacheLength >= CACHE_TABLE_LENGTH) break;
            if (fecAddrCacheLength >= CACHE_TABLE_LENGTH) break;
            if (NameCacheLength < CACHE_TABLE_LENGTH) devNameCache[NameCacheLength] = new NameCache();
            for (i=0; i<modNameCacheLength; i++) 
              if (eqmName.compareTo(modNameCache[i]) == 0) break;
            if (i == modNameCacheLength) modNameCache[modNameCacheLength++] = eqmName;
            devNameCache[NameCacheLength].modNameIndex = i;
            for (i=0; i<fecAddrCacheLength; i++) 
              if (fecName.compareToIgnoreCase(fecAddrCache[i].fecName) == 0) break;
            if (i == fecAddrCacheLength)
            { // it's a new one
              fecCacheR = new BufferedReader(new InputStreamReader(new FileInputStream(fecCacheFile)));
              fecAddrCache[fecAddrCacheLength] = new TFecEntry(fecCacheR,fecName);
              if (!fecAddrCache[fecAddrCacheLength].fecRegistered)
              { // try static cache
                fecAddrCache[fecAddrCacheLength] = new TFecEntry(fecR,fecName);
              }
              if (fecAddrCache[fecAddrCacheLength].fecRegistered)
              { // okay to increment the cache table
                fecAddr = fecAddrCache[fecAddrCacheLength++];
              }
              else
              { // not found anywhere !
                fecAddr = null;
                cc = TErrorList.non_existent_fec;
                break;
              }
            }
            devNameCache[NameCacheLength].srvNameIndex = i;
            devNameCache[NameCacheLength].devName = srvName;
            devNameCache[NameCacheLength].cntName = ctxName;
            if (NameCacheLength < CACHE_TABLE_LENGTH) NameCacheLength++;
            // found it
            cc = 0;
            break;
        }
      }
    }
    catch (Exception e)
    { // database corrupt or not found
      cc = TErrorList.database_not_loaded;
    }
    finally
    {
      srvFile.close();
    }
    return cc;
  }
  int getAddressFromFileCache(String srvName,String ctxName)
  {
    String lkupname;
    SetAddrFileLocation();
    if (debugLevel > 1) DbgLog.log("getAddressFromFileCache","FILE lookup: looking for \\" + ctxName + "\\" + srvName);
    if (srvName.startsWith("ENS") && numConfiguredNameServers > 0)
    {
      lkupname = currentConfiguredNameServerTag;
    }
    else if (srvName.startsWith("GENS") && numConfiguredNameServers > 0)
    {
      lkupname = "G" + currentConfiguredNameServerTag;
    }
    else
    {
      lkupname = srvName;
    }
    for (int i=0; i<NameCacheLength; i++)
    {
      if (ctxName.length() > 0 && 
          ctxName.compareToIgnoreCase("DEFAULT") != 0 &&
          ctxName.compareToIgnoreCase(devNameCache[i].cntName) != 0) continue;
      if (lkupname.compareToIgnoreCase(devNameCache[i].devName) == 0)
      {
        eqmName = modNameCache[devNameCache[i].modNameIndex];
        fecName = fecAddrCache[devNameCache[i].srvNameIndex].fecName;
        fecAddr = fecAddrCache[devNameCache[i].srvNameIndex];
        return 0;
      }
    }
    // try the dynamic cache
    srvFile = new csv(srvCacheFile);
    int cc = fillinAddressFromFileCache(srvName,ctxName);
    if (cc != 0)
    { // not found, try the static cache
      srvFile = new csv(srvR);
      cc = fillinAddressFromFileCache(srvName,ctxName);
    }
    return cc;
  }
  public TSrvEntry(String fname,String srvName,String ctxName)
  {
    getAddressFromFileCache(srvName,ctxName);
  }  
  public TSrvEntry(String srvName,String ctxName)
  {
    getAddress(srvName,ctxName);
  }
  public synchronized void getAddress(String srvName,String ctxName)
  {
    int i;
    debugLevel = TLinkFactory.debugLevel;
    if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","TModAddress() has been called from factory");
    if (srvName.compareToIgnoreCase("GLOBAL") == 0) return;
    SetAddrFileLocation();
    
    if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","looking for /" + ctxName + "/" + srvName);

    boolean isENSRequest = (srvName.startsWith("ENS") && ctxName.compareToIgnoreCase("DEFAULT") == 0);
    if (isENSRequest)
    {
      if (numConfiguredNameServers == 0)
      { // if I'm looking for the ENS and it's not configured, add static address to cache
        if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","add static ENS address");
        StaticAddress(srvName,ctxName);
        return;
      }
    }
    String dnam = srvName;
    if (srvName.compareTo("GENS") == 0) 
    {
      dnam = "G" + currentConfiguredNameServerTag;
      if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","setting GENS entry to " + dnam);
      int idx=0;
      if ((idx=dnam.indexOf('#')) != -1) dnam = dnam.substring(0, idx);
    }
    
    for (i=0; i<NameCacheLength; i++) 
    { // check address cache for requested device server
      //DbgLog.log("getAddress","cache : cnt " + devNameCache[i].cntName + " dev " + devNameCache[i].devName);
      if (ctxName.length() > 0 && 
          ctxName.compareToIgnoreCase("DEFAULT") != 0 &&
          ctxName.compareToIgnoreCase(devNameCache[i].cntName) != 0) continue;
      if (dnam.compareToIgnoreCase(devNameCache[i].devName) == 0)
      {
        eqmName = modNameCache[devNameCache[i].modNameIndex];
        fecName = fecAddrCache[devNameCache[i].srvNameIndex].fecName;
        fecAddr = fecAddrCache[devNameCache[i].srvNameIndex];
        isLegacy = fecAddr.getTineProtocol() < TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
        if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","found it in cache");
        return;
      }
    }

    if (isENSRequest)
    { // ENS request made it this far without a hit, so return canonical entry
      i = ensDevCacheIndex == -1 ? 0 : ensDevCacheIndex;
      eqmName = modNameCache[devNameCache[i].modNameIndex];
      fecName = fecAddrCache[devNameCache[i].srvNameIndex].fecName;
      fecAddr = fecAddrCache[devNameCache[i].srvNameIndex];
      return;      
    }
    if (srvName.compareTo(SRVEXP_NAME) == 0)
    { // asking the NETWORK for the address
      if (NameCacheLength < CACHE_TABLE_LENGTH && modNameCacheLength < CACHE_TABLE_LENGTH && fecAddrCacheLength < CACHE_TABLE_LENGTH)
      { // add to cache ...
        SrvAddr srv = new SrvAddr();
        srv.ipAddr = TLinkFactory.getInstance().getInitializer().getNetCastAddress();
        srv.portOffset = 0;
        srv.fecName = SRVFEC_NAME; srv.eqmContext = "";
        srv.eqmName = SRVEQM_NAME; srv.expName = SRVEXP_NAME;
        fecAddrCache[fecAddrCacheLength] = new TFecEntry(srv);
        modNameCache[modNameCacheLength] = SRVEQM_NAME;
        devNameCache[NameCacheLength] = new NameCache();
        devNameCache[NameCacheLength].cntName = "";
        devNameCache[NameCacheLength].devName = SRVEXP_NAME;
        devNameCache[NameCacheLength].modNameIndex = modNameCacheLength;
        devNameCache[NameCacheLength].srvNameIndex = fecAddrCacheLength;
        fecAddr = fecAddrCache[fecAddrCacheLength];
        fecName = fecAddrCache[fecAddrCacheLength].fecName;
        eqmName = SRVEQM_NAME;
        NameCacheLength++; modNameCacheLength++; fecAddrCacheLength++;
      }
      return;
    }
    if (numConfiguredNameServers == 0)
    { // I don't have a name server, so there's no point in looking
      if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","No configured ENS");
      return;
    }
    if (getAddressFromENS(srvName,ctxName) != 0)
    {
      getAddressFromFileCache(srvName,ctxName);
    }
  }
  private static synchronized SrvAddr getSrvAddrFromENS(String dname,String cname)
  {
    String tagName;
    int cc = 0;
    tagName = new String(currentConfiguredNameServerTag + "/");
    tagName = cname.length() > 0 ? tagName.concat(cname) : tagName.concat("#2");
    SrvAddr srv = new SrvAddr();
    TDataType d_null = new TDataType();
    TDataType d_srv = new TDataType(srv.toByteArray(),"");
    TLink ma = TLinkFactory.getInstance().simpleLink(tagName,dname,d_srv,d_null,(short)1);
    if (ma == null)
    {
      if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","cannot create link to ENS");
      return null;          
    }
    cc = ma.execute(200,true);
    ma.cancel();
    if (cc == 0)
    {
      srv.toStruct();
      if (srv.tineProtocol == 0) srv.tineProtocol = TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
      return srv;
    }
    if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","link to ENS : " + TErrorList.toString(cc));
    return null;          
  }
  private synchronized int getAddressFromENS(String srvName,String ctxName)
  {
    if (TLinkFactory.getInstance().isRunningStandAlone()) 
      return TErrorList.not_accepted;
    int i, n = 0;
    do 
    {
      SrvAddr srv = getSrvAddrFromENS(srvName,ctxName);
      if (srv != null)
      {
        if (modNameCacheLength >= CACHE_TABLE_LENGTH) return TErrorList.resources_exhausted;
        if (fecAddrCacheLength >= CACHE_TABLE_LENGTH) return TErrorList.resources_exhausted;
        if (NameCacheLength < CACHE_TABLE_LENGTH) devNameCache[NameCacheLength] = new NameCache();
        eqmName = srv.eqmName;
        for (i=0; i<modNameCacheLength; i++)
        { // is it already in the tables ? 
          if (eqmName.compareTo(modNameCache[i]) == 0) break;
        }
        if (i == modNameCacheLength) 
        { // didn't find it
          if (modNameCacheLength >= CACHE_TABLE_LENGTH-1)
          { // and there's no room to grow
            // TODO: recycle some of the older addresses ? (why does an app need to talk to 1000 different servers ?)
            return TErrorList.resources_exhausted;
          }
          modNameCache[modNameCacheLength++] = eqmName;
        }
        devNameCache[NameCacheLength].modNameIndex = i;
        fecName = srv.fecName;
        for (i=0; i<fecAddrCacheLength; i++)
        { // is it already in the tables ?
          if (fecName.compareToIgnoreCase(fecAddrCache[i].fecName) == 0) break;
        }
        if (i == fecAddrCacheLength)
        { // didn't find it
          if (fecAddrCacheLength >= CACHE_TABLE_LENGTH-1) 
          { // and there's no room to grow
            // TODO: recycle some of the older addresses ? (why does an app need to talk to 1000 different servers ?)
            return TErrorList.resources_exhausted;
          }
          fecAddrCache[fecAddrCacheLength++] = new TFecEntry(srv);
        }
        fecAddr = fecAddrCache[i];
        devNameCache[NameCacheLength].srvNameIndex = i;
        devNameCache[NameCacheLength].devName = srvName;
        devNameCache[NameCacheLength].cntName = ctxName;
        isLegacy = srv.tineProtocol < TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
        if (NameCacheLength < CACHE_TABLE_LENGTH) NameCacheLength++;
        addAddressToCacheFile(srv); // add address to local file cache
        if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","ENS success");
        return 0;
      }
      currentConfiguredNameServer = (currentConfiguredNameServer+1) % numConfiguredNameServers;
      currentConfiguredNameServerTag = new String("ENS#" + currentConfiguredNameServer);
      if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","trying ENS " + currentConfiguredNameServerTag);
    } while (++n < numConfiguredNameServers);
    if (TLinkFactory.debugLevel > 1) DbgLog.log("getAddress","ens failure");
    return TErrorList.address_unknown;
  }
  private static int findNameServerFromDNS()
  {
    int cc = 0;
    if (numConfiguredNameServers > 0) return 0;
    try
    {
      InetAddress iaddr = java.net.Inet4Address.getByName("tineens");
      addNameServer(iaddr.getHostAddress());
    }
    catch (UnknownHostException e)
    {
      cc = TErrorList.host_not_resolved;
    }
    return cc;
  }
  private static int findNameServerOnNetwork()
  {
    int cc = 0;
    FECAddr fec = new FECAddr();
    SrvAddr srv = new SrvAddr();
    if (numConfiguredNameServers > 0) return 0;
    if ((cc=findServerOnNetwork("", "ENSEQM", "", fec, srv)) == 0)
    {
      addNameServer(fec.ipAddr);
    }
    return cc;
  }
  private static int findServerOnNetwork(String context,String eqmName,String exportName,FECAddr fec,SrvAddr srv)
  {
    int p,cc=0;
    NAME16[] n16in = new NAME16[3];
    NAME16[] n16out = new NAME16[5];
    
    n16in[0] = new NAME16(context);
    n16in[1] = new NAME16(eqmName);
    n16in[2] = new NAME16(exportName);
    
    TDataType dout = new TDataType(n16out);
    TDataType din = new TDataType(n16in);

    TLink tl = new TLink(SRVEXP_NAME,"SRVADDR",dout,din,TAccess.CA_READ);
    cc = tl.execute(500, true);
    if (cc == 0)
    {
      if (fec != null)
      {
        p = Integer.valueOf(n16out[0].getName());
        fec.fecName = n16out[1].getName();
        fec.portOffset = p;
        System.arraycopy(tl.linkPeer.getAddress(), 0, fec.ipHAddr, 0, 4);
        fec.ipAddr = tl.linkPeer.getHostAddress();
      }
      if (srv != null)
      {
        srv.fecName = n16out[1].getName();
        srv.eqmContext = n16out[2].getName();
        srv.eqmName = n16out[3].getName();
        srv.expName = n16out[4].getName();
      }
    }
    return cc;
  }
  public TFecEntry getFecAddr()
  {
    return fecAddr;
  }
  public String getContextName()
  {
    return cntName;
  }
  public String getDeviceName()
  {
    return devName;
  }
  public String getEqmName()
  {
    return eqmName;
  }
  public String getFecName()
  {
    return fecName;
  }
}