package de.desy.tine.console;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.SystemColor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import javax.swing.ButtonGroup;
import javax.swing.JButton;
import javax.swing.JEditorPane;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JRadioButton;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JToolBar;
import de.desy.tine.client.TLinkFactory;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.equipment.TEquipmentModuleFactory;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;

public class TConsole
{
  private PrintStream stdout;
  private PrintStream stderr;
  private PrintStream prevOut;
  private PrintStream prevErr;
  private JFrame consoleFrame;
  private JTextArea textArea;
  private JScrollPane scrollPane;
  private static TConsole instance = new TConsole();
  public static TConsole getInstance() { return instance; }
  private TConsole()
  {
    storeDefaultStreamReferences();
    createOutputStreams();
    createOutputWindow();
  }
  public void show()
  {
    System.setOut(stdout);
    System.setErr(stderr);
    int dl = TLinkFactory.getOutputDebugLevel();
    if (dl >= 0 && dl < dbgButton.length) dbgButton[dl].setSelected(true);
    consoleFrame.setVisible(true);
  }
  public void hide()
  {
    restoreStreams();
    consoleFrame.setVisible(false);
  }
  private static final int MAX_TEXT_WINDOW_SIZE = 65000;
  private static final int TEX_WINDOW_CHOP_SIZE = 5000;
  private void createOutputStreams()
  {
    OutputStream out = new OutputStream()
    {
      @Override
      public void write(int b) throws IOException
      {
        String str = new String(new int[] {b}, 0, 1);
        textArea.append(str);
        if ( b == '\n')
        {
          textArea.setCaretPosition(textArea.getDocument().getLength());
          //scrollPane.getVerticalScrollBar().setValue(scrollPane.getVerticalScrollBar().getMaximum());
          String txt = textArea.getText();
          int len = txt.length();
          if (len > MAX_TEXT_WINDOW_SIZE)
          { // don't let this grow too big !
            textArea.setText(txt.substring(TEX_WINDOW_CHOP_SIZE, len));
          }
        }
      }
    };
    stdout = new PrintStream(out);
    OutputStream err = new OutputStream()
    {
      @Override
      public void write(int b) throws IOException
      {
        String str = new String(new int[] {b}, 0, 1);
        textArea.append(str);
        if ( b == '\n') 
          textArea.setCaretPosition(textArea.getDocument().getLength());
          //scrollPane.getVerticalScrollBar().setValue(scrollPane.getVerticalScrollBar().getMaximum());
      }
    };
    stderr = new PrintStream(err);
  }
  private JButton clientsButton = null;
  JButton getClientsButton()
  {
    if (clientsButton == null)
    {
      clientsButton = new JButton("clients");
      clientsButton.addActionListener(new ActionListener() 
      {
        public void actionPerformed(ActionEvent e)
        {
          TEquipmentModuleFactory.getInstance().dumpClientTable();
        }        
      });
    }
    return clientsButton;
  }
  private JButton contractsButton = null;
  JButton getContractsButton()
  {
    if (contractsButton == null)
    {
      contractsButton = new JButton("contracts");
      contractsButton.addActionListener(new ActionListener() 
      {
        public void actionPerformed(ActionEvent e)
        {
          TEquipmentModuleFactory.getInstance().dumpContractTable();
        }        
      });
    }
    return contractsButton;
  }
  private JButton connectionsButton = null;
  JButton getConnectionsButton()
  {
    if (connectionsButton == null)
    {
      connectionsButton = new JButton("connections");
      connectionsButton.addActionListener(new ActionListener() 
      {
        public void actionPerformed(ActionEvent e)
        {
          TLinkFactory.dumpLinkTable();
        }        
      });
    }
    return connectionsButton;
  }
  private JButton messagesButton = null;
  JButton getMessagesButton()
  {
    if (messagesButton == null)
    {
      messagesButton = new JButton("messages");
      messagesButton.addActionListener(new ActionListener() 
      {
        public void actionPerformed(ActionEvent e)
        {
          MsgLog.MsgEntry[] msgs = MsgLog.getMessages(25);
          System.out.println("\nLast 25 Messages:\n");
          String msg;
          for (MsgLog.MsgEntry m: msgs)
          {
            msg = m.getMsg();
            if (m.getCode() != 0) msg += " ("+TErrorList.getErrorString(m.getCode())+")";
            System.out.println(msg+" at "+TDataTime.toString(m.getTime()));
          }
        }        
      });
    }
    return messagesButton;
  }
  private JButton filterButton = null;
  JButton getFilterButton()
  {
    if (filterButton == null)
    {
      filterButton = new JButton("set debug filter: ");
      filterButton.addActionListener(new ActionListener() 
      {
        public void actionPerformed(ActionEvent e)
        {
          String text = getFilterPane().getText();
          if (text == null || text.length() == 0) text = null;
          DbgLog.setFilter(text);
        }        
      });
    }
    return filterButton;
  }
  private JEditorPane filterPane = null;
  JEditorPane getFilterPane()
  {
    if (filterPane == null)
    {
      filterPane = new JEditorPane();
      filterPane.setBackground(SystemColor.control);
      filterPane.setSize(new Dimension(80,25));
      filterPane.setMaximumSize(new Dimension(80,25));
      filterPane.setToolTipText("enter text to filter on (empty = no filter)");
    }
    return filterPane;
  }
  private ButtonGroup dbgGroup = null;
  private JRadioButton[] dbgButton = null;
  private JRadioButton getDebugButton(final int index)
  {
    if (dbgButton == null) 
    {
      dbgButton = new JRadioButton[5];
      dbgGroup = new ButtonGroup();
    }
    if (dbgButton[index] == null)
    {
      dbgButton[index] = new JRadioButton("" + (index > 0 ? new String(""+index) : "off"));
      dbgButton[index].addActionListener(new ActionListener()
      {
        public void actionPerformed(ActionEvent e)
        {
          TLinkFactory.setOutputDebugLevel(index);
        }
      });
      dbgGroup.add(dbgButton[index]);
    }
    return dbgButton[index];
  }
  private JLabel dbgLabel;
  private JToolBar toolBar = null;
  private JToolBar getToolBar()
  {
    if (toolBar == null)
    {
      toolBar = new JToolBar();
      toolBar.add(getClientsButton());
      toolBar.add(getContractsButton());
      toolBar.add(getConnectionsButton());
      toolBar.add(getMessagesButton());
      dbgLabel = new JLabel();
      dbgLabel.setText("debug level:");
      toolBar.add(dbgLabel);
//      toolBar.add(getClientsButton());
      toolBar.add(getDebugButton(0));
      toolBar.add(getDebugButton(1));
      toolBar.add(getDebugButton(2));
      toolBar.add(getDebugButton(3));
      toolBar.add(getDebugButton(4));
      toolBar.add(getFilterButton());
      toolBar.add(getFilterPane());
    }
    return toolBar;
  }
  public void addWindowListener(WindowListener l)
  {
    if (consoleFrame != null) consoleFrame.addWindowListener(l);
  }
  public void removeWindowListener(WindowListener l)
  {
    if (consoleFrame != null) consoleFrame.removeWindowListener(l);
  }
  private void createOutputWindow()
  {
    textArea = new JTextArea();
    textArea.setRows(10);
    String text = TEquipmentModuleFactory.getInstance().getFecName();
    if (text == null) text = "";
    consoleFrame = new JFrame(text + " Debugging console");
    consoleFrame.setLayout(new BorderLayout());
    scrollPane = new JScrollPane(textArea);
    consoleFrame.add(scrollPane, BorderLayout.CENTER);
    consoleFrame.addWindowListener(new ConsoleWindowListener());
    consoleFrame.setBounds(0, 0, 750, 750);
    consoleFrame.add(getToolBar(), BorderLayout.NORTH);
  }
  private void restoreStreams()
  {
    System.setOut(prevOut);
    System.setErr(prevErr);
    DbgLog.setFilter(null);
  }
  private void storeDefaultStreamReferences()
  {
    prevOut = System.out;
    prevErr = System.err;
  }
  private class ConsoleWindowListener implements WindowListener
  {
    public void windowOpened(WindowEvent e)
    {
    }
    public void windowClosing(WindowEvent e)
    {
      restoreStreams();
    }
    public void windowClosed(WindowEvent e)
    {
    }
    public void windowIconified(WindowEvent e)
    {
    }
    public void windowDeiconified(WindowEvent e)
    {
    }
    public void windowActivated(WindowEvent e)
    {
    }
    public void windowDeactivated(WindowEvent e)
    {
    }
  }
}
