package de.desy.tine.csvUtils;
import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.server.logger.MsgLog;

public final class csv
{
  private BufferedReader in;
  public csv(BufferedReader reader)
  {
    in = reader;
  }
  public csv(String csvFile)
  {
    try
    {
      in = new BufferedReader(
        new InputStreamReader(new FileInputStream(csvFile)));
    }
    catch (Exception e) 
    {
      //MsgLog.log("csv","Could not open "+csvFile,TErrorList.file_error,e,1);
    }      
  }
  public String readLine()
  {
    if (in == null) return null;
    try 
    {
      return in.readLine();
    }
    catch (IOException e)
    {
      // End of File ?...
      return null;
    }
  }
  public int readFile(csvColumn[] csvCols)
  {
    return readFile(csvCols,null);
  }
  public int readFile(csvColumn[] csvCols, RowHandler csvRowHndlr)
  {
    boolean headerRead = false, haltIteration = false;
    String s, strValue;
    int i = 0, idx = 0;
    if (in == null) return TErrorList.no_such_file;
    try
    {
      while((s=in.readLine()) != null)
      {
        if (s.length() == 0) continue;
        if (s.startsWith("#") || s.startsWith(";") || s.startsWith("%")) continue;
        if (!headerRead)
        {
          for (i=0; i<csvCols.length; i++)
          {
            if (csvCols[i] == null) continue;
            csvCols[i].setColumnPosition(findcol(s,csvCols[i].getColumnName()));
            if (csvCols[i].getColumnPosition() == -1 && csvCols[i].getColumnDefaultValue() == null)
              throw new NoSuchFieldException();
          }
          headerRead = true;
          continue;
        }
        for (i=0; i<csvCols.length; i++)
        {
          if (csvCols[i] == null) continue;
          strValue = csvCols[i].getColumnDefaultValue();
          if (csvCols[i].getColumnPosition() != -1)
            strValue = colptr(csvCols[i].getColumnPosition(),s);
          if (csvCols[i].getColumnHandler().process(strValue.trim(),idx) != 0)
          { // any individual field handler can halt the iteration
            haltIteration = true;
          }
        }
        if (haltIteration) break;
        if (csvRowHndlr != null) 
        { // halt iteration on non-zero return code
          if (csvRowHndlr.process(idx) != 0) break;
        }
        idx++;
      }
    }
    catch (NoSuchFieldException e)
    { // database corrupt ?
      MsgLog.log("csv readFile","looking for " + csvCols[i].getColumnName() + " in corrupt file !",TErrorList.no_such_column,e,0);
      return TErrorList.no_such_column;
    }
    catch (IOException e)
    {
      // End of File ?...
    }
    return 0;
  }
  public void close()
  {
    try 
    {
      if (in != null) in.close();
    }
    catch (IOException e)
    {
      MsgLog.log("csv readFile","Cannot close file!",TErrorList.file_error,e,0);
      System.out.println("Cannot close file !");
      e.printStackTrace();
    }
  }  
  public int findcol(String line,String ColumnEntry)
  {
    String subs;
    String s = line.toUpperCase();
    String e = ColumnEntry.toUpperCase();
    String altEntry1=null, altEntry2=null;
    String[] altPieces = null;
    char brkc = 0, altc = 0;
    int i, n, col;

    if (e.indexOf(' ') > 0)
    {
      brkc = ' ';
      altc = '_';
    }
    else if (e.indexOf('_') > 0) 
    {
      brkc = '_';
      altc = ' ';
    }
    if (brkc != 0) /* prepare alternates */
    {
      altEntry1 = e.replace(brkc,altc);
      altPieces = e.split(String.valueOf(brkc));
      altEntry2 = new String(altPieces[0]);
      for (i = 1; i<altPieces.length; i++) altEntry2 = altEntry2.concat(altPieces[i]);
    }    
    for (i=0, col=0; i < s.length(); col++)
    {
      if ((n = s.indexOf(',',i)) < 0) // not found
      {
        n = s.length();
      }
      subs = s.substring(i,n).trim();
      if (subs.compareTo(e) == 0) return col;
      if (brkc != 0) 
      {
        if (subs.compareTo(altEntry1) == 0) return col;
        if (subs.compareTo(altEntry2) == 0) return col;
      }
      if ((i = n + 1) >= s.length()) return -1;	  
    }
    return -1;
  }  
  public int namcmp(String targetString,String searchString,int column)
  {
    String s = colptr(column,searchString);
    if (s == null) return -1;
    return s.compareTo(targetString);
  }  
  public String colptr(int column,String line)
  {
    String subs;
    String s = line; //line.toUpperCase();
    int i, n, col;
    
    for (i=0, col=0; i < s.length(); col++)
    {
      if (s.charAt(i) == '"') 
      {
        n = s.indexOf('"',i+1); 
        if (n < 0 || n >= s.length()) break;
        subs = s.substring(i+1,n).trim();
        i = n;
        n = s.indexOf(',',i);
      }
      else
      {
        if ((n=s.indexOf(',',i)) == -1) n = s.length(); //break;
        subs = s.substring(i,n).trim();
      }
      if (col == column) return subs;
      if ((i = n + 1) >= s.length()) break;
    }
    return "";
  }  
  public static float getFloatFromString(String s)
  {
    float fval = 0;
    try
    {
      fval = Float.parseFloat(s);
    }
    catch (NumberFormatException e)
    {
      fval = 0;
    }
    return fval;
  }
  public static int getIntFromString(String s)
  {
    int ival = 0;
    try
    {
      if (s.startsWith("0x") || s.startsWith("0X"))
      {
        ival = Integer.parseInt(s.substring(2),16);
      }
      else
      {
        ival = Integer.parseInt(s);
      }
    }
    catch (NumberFormatException e)
    {
      ival = 0;
    }
    return ival;
  }
}