//: TLink.java
package de.desy.tine.headers;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.util.Arrays;

import de.desy.tine.endianUtils.Swap;
public final class TReqHdr
{
  private String username = null; 
  private short  tineProtocol;
  private short  numInPacket;
  static public final short  hdrSizeInBytes = 2 + 16 + 2 + 2; // network header size
  private short totalSizeInBytes;
  private short msgSizeInBytes;
  ByteArrayOutputStream dBuffer;
  public TReqHdr(String u,int protocolLevel)
  {
    username = u.toUpperCase();
    tineProtocol = (short)protocolLevel;
    numInPacket = 1;
    dBuffer = new ByteArrayOutputStream(hdrSizeInBytes);
  }
  public String getUserName() { return username; }
  public short getTotalSizeInBytes() { return totalSizeInBytes; }
  public short getMsgSizeInBytes() { return msgSizeInBytes;}
  public void setMsgSizeInBytes(short sizeInBytes) 
  { 
    msgSizeInBytes = sizeInBytes;
  }
  private byte[] bstr = new byte[16];
  public byte[] toByteArray()
  {
    totalSizeInBytes = (short)(msgSizeInBytes + hdrSizeInBytes);
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      ds.writeShort(Swap.Short(totalSizeInBytes));
      //byte[] bstr = new byte[16];
      Arrays.fill(bstr, (byte)0);
      byte[] tmp  = username.getBytes();
      for (int i=0; i<16 && i<tmp.length; i++) bstr[i] = tmp[i];
      ds.write(bstr,0,16);
      ds.writeShort(Swap.Short(tineProtocol));
      ds.writeShort(Swap.Short(numInPacket));
      ds.close();
      return dBuffer.toByteArray();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return null;  
  }
  void removeThisMethod(short msgsize)  // prepare outgoing ...
  {
    msgSizeInBytes = msgsize;
    totalSizeInBytes = (short)(msgsize + hdrSizeInBytes);
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      ds.writeShort(Swap.Short(totalSizeInBytes));
      //byte[] bstr = new byte[16];
      Arrays.fill(bstr, (byte)0);
      byte[] tmp  = username.getBytes();
      for (int i=0; i<16 && i<tmp.length; i++) bstr[i] = tmp[i];
      ds.write(bstr,0,16);
      ds.writeShort(Swap.Short(tineProtocol));
      ds.writeShort(Swap.Short(numInPacket));
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }	 
  }
  public TReqHdr(byte[] data,int off, int len)  // prepare incoming ...
  {
    int bytesread = 0;
    try
    {
      ByteArrayInputStream dinBuffer = new ByteArrayInputStream(data,off,len);
      DataInputStream ds = new DataInputStream(dinBuffer);
      // overall message length:
      totalSizeInBytes = Swap.Short(ds.readShort());
      bytesread = 2;
      //byte[] bstr = new byte[16];
      Arrays.fill(bstr, (byte)0);
      while (bytesread < hdrSizeInBytes)
      {
        ds.read(bstr,0,16);
        username = (new String(bstr)).trim();
        tineProtocol = Swap.Short(ds.readShort());
        numInPacket = Swap.Short(ds.readShort());
        bytesread += 20;
      }
      ds.close();
      dinBuffer.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
  }      
  public String toString()
  {
    return "User : " + username + "\nTotal size : " + totalSizeInBytes + "\nTotal number : " + numInPacket;
  }
  public short getProtocol()
  {
    return tineProtocol;
  }
  public short getTineProtocol()
  {
    return tineProtocol;
  }
  public void setTineProtocol(int protocolLevel)
  {
    tineProtocol = (short)protocolLevel;
  }
}