package de.desy.tine.histUtils;

import java.util.Vector;
import de.desy.tine.client.TLink;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.queryUtils.TQuery;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.types.*;

// TODO: finish this ...
public final class THistory
{
  private static final int MAX_HSTCFG_SIZE = 100;
  private static HstCfg[] cfgs = new HstCfg[MAX_HSTCFG_SIZE];
  private static McaCfg[] mcacfgs = new McaCfg[MAX_HSTCFG_SIZE];
  private static PrfCfg[] prfcfgs = new PrfCfg[MAX_HSTCFG_SIZE];
  private static boolean isInitialized = false;
  private static void initCfgs()
  {
    if (isInitialized) return;
    for (int i=0; i<MAX_HSTCFG_SIZE; i++)
    {
      cfgs[i] = new HstCfg();
      mcacfgs[i] = new McaCfg();
      prfcfgs[i] = new PrfCfg();
    }
    isInitialized = true;
  }
  private static boolean isCentralArchiveServer(String server)
  {
    return server.endsWith("HISTORY");
  }
  private static boolean isCentralArchiver(String server)
  {
    return server.endsWith("ARCHIVER");
  }
  private static String getQueryPropertyRoot(String property)
  {
    int slen = property.length();
    if (property.endsWith(".HIST") || property.endsWith(".ARCH")) 
      return property.substring(0, slen-5);
    if (property.endsWith(".HST") || property.endsWith(".ARC")) 
      return property.substring(0, slen-4);
    if (property.endsWith(".AR")) 
      return property.substring(0, slen-3);
    return property;
  }
  
  private static boolean isArchiveQueryProperty(String property)
  {
    if (property.endsWith(".HIST")) return true;
    if (property.endsWith(".HST")) return true;
    if (property.endsWith(".ARCH")) return true;
    if (property.endsWith(".ARC")) return true;
    if (property.endsWith(".AR")) return true;
    return false;
  }
  private static synchronized int getArchivedNumPointsInInterval(String context,String server, String property, String device, double start, double stop,int timeout)
  {
    int cc = 0;
    int[] startstop = new int[2];
    int[] npts = new int[1];
    TDataType dout, din;
    
    if (server == null && context == null) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    startstop[0] = (int)start;
    startstop[1] = (int)stop;
    din = new TDataType(startstop);
    dout = new TDataType(npts);
    
    String dname = new String("/" + context + "/" + server + "/" + device);
    String pname = new String(property);
    if (!isCentralArchiveServer(server))
    { // not a call to a central archive server
      if (!isArchiveQueryProperty(property))
        pname = pname + ".HIST"; // meta extension not given, try this ...
    }
    TLink tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
    cc = tl.execute(timeout, true);
    tl.close();
    return cc > 0 ? -cc : npts[0];
  }  
  private static synchronized int getDataArray(String context,String server, String property, String device, int index, double start, double stop, TCompoundDataObject[] data, THistorySource src, int timeout)
  {
    int idx, cc = 0;
    HstSrc[] srvs = new HstSrc[2];
    String s1 = server;
    String p1 = property;
    if (s1 == null) s1 = "HISTORY";
    if ((idx=property.indexOf(".HIST")) >= 0 ||
        (idx=property.indexOf(".HST")) >= 0 ||
        (idx=property.indexOf(".AR")) >= 0)
    { // passed with .HIST appended, remove it!
      p1 = property.substring(0, idx-1);
    }
    String s2 = s1; // server 2nd source
    if (!isCentralArchiveServer(s1)) 
    { // a specific server was passed
      p1 = property + ".ARCH"; // redirect to same channel at the central archiver
    }
    else
    { // 2nd source must redirect back to find the target server's local histories
      s2 = "ARCHIVER"; 
    }
    int npts1 = getArchivedNumPointsInInterval(context, s1, p1, device, start, stop, timeout);
    if (npts1 < 0) npts1 = 0;
    srvs[0] = new HstSrc("Central Archive",s1,npts1);
    String p2 = property + ".HIST";
    int npts2 = getArchivedNumPointsInInterval(context, s2, p2, device, start, stop, timeout);
    if (npts2 < 0) npts2 = 0;
    srvs[1] = new HstSrc("Local History",s2,npts2);
    
    boolean done = false;
    int nret = 0;
    if (npts1 > 500 || npts1 > npts2)
    { // central archive has 'enough' points or at least more than the local history
      nret = getArchivedDataArray(context,s1,p1,device,index,start,stop,data,timeout);      
      if (nret > 0)
      { // everything's okay
        srvs[0].setAsDefault(true);
        srvs[0].setNumberReturned(nret);
        done = true;
      }
      else
      {
        cc = -nret;
      }
    }
    if (!done && npts2 > 0)
    { // came up emty upstairs, so try the local history ...
      nret = getArchivedDataArray(context,s2,p2,device,index,start,stop,data,timeout);
      if (nret > 0)
      {
        srvs[1].setNumberReturned(nret);        
      }
      else
      {
        cc = -nret;
      }
      srvs[1].setAsDefault(true);
    }
    if (src != null)
    { // was passed, fill it in ...
      src.CentralArchive = srvs[0];
      src.LocalHistory = srvs[1];
      src.setNumberOfPointsReturned(nret);
    }
    return cc;
  }
  private static synchronized int getArchivedDataArray(String context,String server, String property, String device, int index, double start, double stop, TCompoundDataObject[] data, int timeout)
  {
    int cc = 0;
    int[] startstop = new int[3];
    TDataType dout, din;
    if (server == null && context == null) return TErrorList.argument_list_error;
    if (data == null) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (timeout < 100) timeout = 100;
    startstop[0] = (int)start;
    startstop[1] = (int)stop;
    startstop[2] = index;
    din = new TDataType(startstop);
    dout = new TDataType(data);
    
    String dname = new String("/" + context + "/" + server + "/" + device);
    String pname = new String(property);
    if (!isCentralArchiveServer(server))
    { // not a call to a central archive server
      if (!isArchiveQueryProperty(property))
        pname = pname + ".HIST"; // meta extension not given, try this ...
    }
    TLink tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
    cc = tl.execute(timeout, true);
    tl.close();
    if (cc != 0 && isCentralArchiver(server))
    { // special 2nd chance !
      if (isArchiveQueryProperty(property))
      {
        pname = getQueryPropertyRoot(property) + ".ARCH";
      }
      else
      {
        pname = new String(property) + ".ARCH";
      }
      tl = new TLink(dname,pname,dout,din,TAccess.CA_READ);
      cc = tl.execute(timeout, true);
      tl.close();
    }
    return cc == 0 ? dout.getCompletionLength() : -cc;
  }  
  private static int getArchivedDataArray(String context,String server, String property, String device, int index, double start, double stop,TCompoundDataObject[] data)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data,2000);
  }
  private static synchronized int getArchivedDataSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data)
  {
    int cc = 0;
    int[] startstop = new int[2];
    TDataType dout, din;

    if (server == null && context == null) return TErrorList.argument_list_error;
    if (targetTime == null || data == null) return TErrorList.argument_list_error;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    startstop[0] = (int)targetTime[0];
    startstop[1] = (int)targetTime[0];
    din = new TDataType(startstop);
    dout = new TDataType(data);
    
    String dname = new String("/" + context + "/" + server + "/" + device);
    String[] pname = new String[2];
    pname[0] = new String(property);
    int ntries = 1;
    if (!server.endsWith("HISTORY"))
    { // not a call to a central archive server
      if (!isArchiveQueryProperty(property))
      { // meta extension not given, try this ...
        pname[0] = property + ".HIST";
        pname[1] = property + ".ARCH";
        ntries = 2;
      }
    }
    for (int i=0; i<ntries; i++)
    {
      TLink tl = new TLink(dname,pname[i],dout,din,TAccess.CA_READ);
      cc = tl.execute(2000, true);
      targetTime[0] = tl.getLastDataTimeStamp();
      tl.close();
      if (cc == 0 && dout.getCompletionLength() > 0) break;
    }
    return cc;
  }  
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,FLTINT[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,FLTINT[] data)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,FLTINT[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,DBLDBL[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many ponits both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,DBLDBL[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many ponits both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,FLTINT[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many ponits both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME16I[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many ponits both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME32I[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified.  Both the central archive and local archive are 
   * queried for the number of stored points over the interval. Data are usually
   * retrieved from the central archive system if there are at least 500 points over the
   * time interval in question.  Otherwise the local archive system is used.  By examining
   * the THistorySource object one can see which source was chosen and how many ponits both
   * sources have over the interval.  For dedicated retrieval from one source or the
   * other, please use an overloaded method which does not take the THistorySource parameter!
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  The src parameter will give the
   * exact number of points actually stored over the interval specified.
   * The number of points returned in the call will often be less than array dimension and
   * less than the exact number of points stored over the interval. One can use the 
   * getNumberReturned() method of the history source object to ascertain the number
   * of archived data points returned. 
   * @param src is a reference to a THistorySource object to receive the history source information.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * @return 0 upon success or a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME64I[] data,THistorySource src,int timeout)
  {
    return getDataArray(context,server,property,device,0,start,stop,data,src,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,DBLDBL[] data)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, int index,double start, double stop,DBLDBL[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,index,start,stop,data,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME16I[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME16I[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME32I[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME32I[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data,timeout);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME64I[] data)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data);
  }
  /**
   * Retrieve an archive data set (value - timestamp pairs) for the target property and device
   * given and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param data is a reference to a (doublet) data array object to contain the returned data.
   * This can be one of FLTINT[], DBLDBL[], NAME64I[], and so on.  The array should be 
   * dimensioned to the maximum number of points desired (effectively determining a raster
   * over which the archived data will be distributed.  
   * The number of points returned in the call will often be less than array dimension
   * so one should pay attention to the return value of the call.
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getArchivedData(String context,String server, String property, String device, double start, double stop,NAME64I[] data,int timeout)
  {
    return getArchivedDataArray(context,server,property,device,0,start,stop,data,timeout);
  }
  /**
   * Retrieve the number of points stored for the give parameters
   * and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getNumPointsInInterval(String context,String server, String property, String device, double start, double stop)
  {
    return getArchivedNumPointsInInterval(context, server, property, device, start, stop, 1000);
  }
  /**
   * Retrieve the number of points stored for the give parameters
   * and over the time range specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param start is the start time in seconds (UTC double, a TINE timestamp) 
   * @param stop is the stop time in seconds (UTC double, a TINE timestamp)
   * @param timeout gives the number of milliseconds to wait for the call to complete 
   * (usually 1000 is sufficient).
   * 
   * @return if positive, the number of points returned in the archive call; if negative,
   * the negative of a TINE error code.
   */
  public static int getNumPointsInInterval(String context,String server, String property, String device, double start, double stop,int timeout)
  {
    return getArchivedNumPointsInInterval(context, server, property, device, start, stop, timeout);
  }
  /**
   * Retrieve an archive data set for the target property and device
   * given at the time specified. If server is null, an empty string, or "HISTORY",
   * the central archive server is queried for the data. Otherwise the local history 
   * subsystem of the device server specified is queried. The returned data set
   * represents the entire data array, be it a multi-channel array, trace array, or
   * scalar at a stored time nearest (but not greater than) the given time.  
   * 
   * @param context is the desired context  
   * @param server is the device server which manages the archived parameter. If null
   * then the central archive will be queried for the property and device
   * @param property is the desired archived property
   * @param device is the desired device name
   * @param targetTime is the targetted time in seconds (UTC double, a TINE timestamp) 
   * This is passed as a single-valued array, so that the targetted time can be 
   * altered by the call to reflect the time found.
   * @param data is a reference to a float data array object to contain the returned data.
   * The array should be dimensioned to the number of points in the data record.
   * 
   * @return 0 if success or a TINE error code.
   */
  public static int getArchivedSnapshot(String context,String server, String property, String device, double[] targetTime,float[] data)
  {
    return getArchivedDataSnapshot(context,server,property,device,targetTime,data);
  }
  public static String[] getArchivedProperties(String context)
  {
    return getArchivedProperties(context,null,null);
  }
  public static String[] getArchivedProperties(String context,String server)
  {
    return getArchivedProperties(context,server,null);
  }
  public static String[] getArchivedProperties(String context,String server,String subsystem)
  {
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY";
    boolean isCentral = isCentralArchiveServer(server);
    TDataType dtsubs;
    if (subsystem != null && subsystem.length() > 0 && 
        !subsystem.startsWith("ALL") && !subsystem.startsWith("All"))
    {
      dtsubs = new TDataType(subsystem);
    }
    else
    { 
      dtsubs = new TDataType();
    }
    NAME32[] nlist = null;
    int[] nprps = new int[1];
    String ntagsString, tagsString;
    TDataType dtnprps = new TDataType(nprps);
    TDataType dtprps;
    tagsString = isCentral ? "TAGS" : "HISTORIES";
    ntagsString = "N" + tagsString;
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,ntagsString,dtnprps,dtsubs,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    if (nprps[0] > 0)
    {
      nlist = new NAME32[nprps[0]];
      dtprps = new TDataType(nlist);
      tl = new TLink(dname,tagsString,dtprps,dtsubs,TAccess.CA_READ);
      cc = tl.execute(800, true);
      tl.close();
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
    }
    if (nlist == null) return null;
    String nam;
    Vector<String> v = new Vector<String>(nlist.length);
    int n = 0;
    for (int i=0; i<nlist.length; i++)
    {
      nam = nlist[i].getName();
      if (nam.endsWith(".NAM")) continue;
      v.add(nam);
      n++;
    }
    return v.toArray(new String[n]);
  }
  public static String[] getArchivedDevices(String context,String server, String property)
  {
    return TQuery.getDeviceNames(context, server, property);
  }
  public static String[] getArchivedSubsystems(String context)
  {
    return getArchivedSubsystems(context,null);
  }
  public static String[] getArchivedSubsystems(String context,String server)
  {
    if (server == null && context == null) return null;
    if (context == null) context = "DEFAULT";
    if (server == null) server = "HISTORY"; // assume the standard central archive server for the given context
    if (!isCentralArchiveServer(server)) return null;
    int cc;
    NAME32[] nlist = null;
    int[] nsubs = new int[1];
    String ntagsString, tagsString;
    TDataType dtnsubs = new TDataType(nsubs);
    TDataType dtsubs;
    tagsString = "SUBSYSTEMS";
    ntagsString = "N" + tagsString;
    
    String dname = "/" + context + "/" + server;
    TLink tl = new TLink(dname,ntagsString,dtnsubs,null,TAccess.CA_READ);
    cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    if (nsubs[0] > 0)
    {
      nlist = new NAME32[nsubs[0]];
      dtsubs = new TDataType(nlist);
      tl = new TLink(dname,tagsString,dtsubs,null,TAccess.CA_READ);
      cc = tl.execute(800, true);
      tl.close();
      if (cc != 0)
      { // TODO: throw an exception here
        return null;
      }
    }
    String nam;
    Vector<String> v = new Vector<String>(nlist.length);
    int n = 0;
    for (int i=0; i<nlist.length; i++)
    {
      nam = nlist[i].getName();
      if (nam.length() == 0) continue;
      v.add(nam);
      n++;
    }
    return v.toArray(new String[n]);
  }
  public static String[] getOperationHistoryContexts()
  {
    return TQuery.getContexts("STATE");
  }
  public static String[] getArchiveServerContexts()
  {
    return TQuery.getContexts("HISTORY");
  }
  public static String[] getOperationHistoryConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "STATE", "CONFIGURATION");
  }
  public static String[] getViewerMcaConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION.MCA");
  }
  public static String[] getViewerConfigurationNames(String context)
  {
    if (context == null) return null;
    return TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION");
  }
  public static String[] getViewerTraceConfigurationNames(String context)
  {
    if (context == null) return null;
    String[] nms = TQuery.getDeviceNames(context, "HISTORY", "CONFIGURATION.TRACE");
    if (nms == null || nms.length < 2) return nms;
    String[] nmsr = new String[nms.length-2];
    boolean hasProperties = false, hasTraces = false;
    for (int i=0, k=0; i<nms.length; i++)
    {
      if (nms[i].compareToIgnoreCase("properties") == 0) 
      {
        hasProperties = true;
        continue;
      }
      if (nms[i].compareToIgnoreCase("traces") == 0) 
      {
        hasTraces = true;
        continue;
      }
      if (k<nms.length-2) nmsr[k++] = nms[i];
    }
    return hasProperties && hasTraces ? nmsr : nms;
  }
  public static String[] getViewerConfigurationNames(String context,String group)
  {
    if (context == null) return null;
    String prp = "CONFIGURATION";
    if (group != null) prp += "." + group;
    String[] nms = TQuery.getDeviceNames(context, "HISTORY", prp);
    if (nms == null || nms.length < 2) return nms;
    String[] nmsr = new String[nms.length-1];
    boolean hasProperties = false;
    for (int i=0, k=0; i<nms.length; i++)
    {
      if (nms[i].compareToIgnoreCase("properties") == 0) 
      {
        hasProperties = true;
        continue;
      }
      if (k<nms.length-1) nmsr[k++] = nms[i];
    }
    return hasProperties ? nmsr : nms;
  }
  public static String[] getViewerConfigurationGroups(String context)
  {
    if (context == null) return null;
    NAME64[] prps = new NAME64[100];
    TDataType dtprps = new TDataType(prps);
    String dname = "/" + context + "/HISTORY";
    String dprop = "CONFIGURATION.*";
    TLink tl = new TLink(dname,dprop,dtprps,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      return null;
    }
    int len = dtprps.getCompletionLength();
    String nam;
    Vector<String> v = new Vector<String>(len);
    int n = 0;
    for (int i=0; i<len; i++)
    {
      nam = prps[i].getName();
      if (nam.compareToIgnoreCase("CONFIGURATION") == 0) continue;
      if (nam.length() == 0) continue;
      if (nam.endsWith(".NAM")) continue;
      if (nam.endsWith(".MCA")) continue;
      v.add(nam.substring(14));
      n++;
    }
    return n > 0 ? v.toArray(new String[n]) : null;
    
  }
  public static HstCfg[] getSelectableProptiesConfiguration(String context,String group)
  {
    if (context == null || group == null) return null;
    return getViewerConfiguration(context,"properties",group);
  }
  public static HstCfg[] getSelectableTracesConfiguration(String context)
  {
    if (context == null) return null;
    return getViewerConfiguration(context,"traces","TRACE");
  }
  public static HstCfg[] getViewerConfiguration(String context,String name)
  {
    return getViewerConfiguration(context,name,null);
  }
  public static synchronized HstCfg[] getViewerConfiguration(String context,String name,String group)
  {
    if (context == null || name == null) return null;
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(cfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION";
    if (group != null && group.length() > 0) dprop += "." + group;
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getViewerConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    boolean hasPropertiesList = false;
    for (int i=0; i<num; i++)
    { // is 'properties' in the list ?
      if (cfgs[i].getDescription().compareToIgnoreCase("properties") == 0)
      {
        hasPropertiesList = true;
        break;
      }
    }
    int ncfg = hasPropertiesList ? num - 1 : num;
    HstCfg[] cfg = new HstCfg[ncfg];
    for (int i=0, n=0; i<num; i++) 
    {
      if (cfgs[i].getDescription().compareToIgnoreCase("properties") != 0)
      {
        cfg[n++] = new HstCfg(cfgs[i]);        
      }
    }
    return cfg;
  }
  public static synchronized McaCfg[] getMcaViewerConfiguration(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(mcacfgs);
    String dname = "/" + context + "/HISTORY/" + name;
    String dprop = "CONFIGURATION.MCA";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getMcaViewerConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    McaCfg[] cfg = new McaCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new McaCfg(mcacfgs[i]);
    }
    return cfg;
  }
  public static synchronized McaCfg[] getOperationHistoryConfiguration(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(mcacfgs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "CONFIGURATION";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryConfiguration",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    McaCfg[] cfg = new McaCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new McaCfg(mcacfgs[i]);
    }
    return cfg;
  }
  public static synchronized PrfCfg[] getOperationHistoryProfile(String context,String name)
  {
    if (context == null || name == null) return null; 
    if (!isInitialized) initCfgs();
    TDataType dtcfg = new TDataType(prfcfgs);
    String dname = "/" + context + "/STATE/" + name;
    String dprop = "PROFILE";
    TLink tl = new TLink(dname,dprop,dtcfg,null,TAccess.CA_READ);
    int cc = tl.execute(800, true);
    tl.close();
    if (cc != 0)
    { // TODO: throw an exception here
      DbgLog.log("getOperationHistoryProfile",TErrorList.getErrorString(cc));
      return null;
    }
    int num = dtcfg.getCompletionLength();
    if (num == 0) return null;
    PrfCfg[] cfg = new PrfCfg[num];
    for (int i=0; i<num; i++) 
    {
      cfg[i] = new PrfCfg(prfcfgs[i]);
    }
    return cfg;
  }
}
