/*
 * Created on Aug 23, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.alarms;

import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.devices.TDevice;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.types.TCompoundDataObject;

// each device has a (rather dynamic) linked list of these at any given time ...
public class TAlarm
{
  public static final int ALM_SYSTEM_SYSTEM = 5000;
  public static final int ALM_SYSTEM_HARDWARE = 5001;
  public static final int ALM_SEVERITY_NONE = 0;
  public static final int ALM_SEVERITY_INFO = 3;
  public static final int ALM_SEVERITY_WARN = 8;
  public static final int ALM_SEVERITY_ERR = 12;
  public static final int ALM_SEVERITY_FATAL = 15;
  private long almTimestamp; // alarm timestamp
  private long almStarttime; // alarm start time
  private int almCode; // alarm code;
  private boolean almRemovalBlocked; // flag to keep alarm in table for CAS readout
  private byte almDescriptor; // alarm descriptor
  private byte almClearCount; // alarm's clear counter
  private byte[] almData = null;
  private TAlarmDefinition almDef = null;
  private TDevice almDevice = null;
  private boolean isTerminated = false;
  private static int almCollapseWindow = 100; // total number of alarms
  public static int getAlmCollapseWindow() { return almCollapseWindow; }
  public static void setAlmCollapseWindow(int window) { if (window > 10) almCollapseWindow = window; }
  private static int almOscillationWindow = 8; // clear counts
  public static int getAlmOscillationWindow() { return almOscillationWindow; }
  public static void setAlmOscillationWindow(int window) { if (window > 8) almOscillationWindow = window; }
  private static int almTerminationWindow = 30; // seconds
  private static int almHeartbeatWindow = 1200; // seconds
  public static int getAlmHeartbeatWindow() { return almHeartbeatWindow; }
  public static void setAlmHeartbeatWindow(int window) { almHeartbeatWindow = window; }
  public static int getTerminationWindow () { return almTerminationWindow; }
  public static void setTerminationWindow(int window) { if (window > 0) almTerminationWindow = window; }
  public int getTimeStamp() { return (int)(almTimestamp/1000); }
  public int getTimeStampUSec() { return (int)((almTimestamp%1000)*1000); }
  public int getStarttime() { return (int)(almStarttime/1000); }
  public int getStarttimeUSec() { return (int)((almStarttime%1000)*1000); }
  public void setTimeStamp(long timestamp) { if (timestamp > 0) almTimestamp = timestamp; } 
  public void setTimeStamp(int timestamp) { if (timestamp > 0) almTimestamp = ((long)timestamp)*1000; } 
  public int getCode() { return almCode; }
  public boolean isRemovalBlocked() { return almRemovalBlocked; }
  public void blockRemoval() { almRemovalBlocked = true; }
  public void allowRemoval() { almRemovalBlocked = false; }
  public byte getDescriptor() { return almDescriptor; }
  public void setDescriptor(byte flags) { almDescriptor = flags; }
  public byte getClearCount() { return almClearCount; }
  public void resetClearCount() { almClearCount = 0; }
  public void clear() 
  { 
    if ((almDescriptor & TAlarmDescriptor.TERMINATE) == TAlarmDescriptor.TERMINATE)
      almClearCount = (byte)almOscillationWindow;
    if (almClearCount < almOscillationWindow) almClearCount++; 
  }
  public byte[] getData() { return almData; }
  public void setData(byte[] data) { almData = data; }
  public void setData(Object data)
  {
    ByteArrayOutputStream baos = new  ByteArrayOutputStream(TStrings.ALARM_DATA_SIZE);
    DataOutputStream ds = new DataOutputStream(baos);
    int range = TStrings.ALARM_DATA_SIZE;
    int mrange;
    try
    {
      if (data instanceof short[])
      {
        mrange = ((short[])data).length;
        range = Math.min(range/2, mrange);
        for (int i=0; i<range; i++) ds.writeShort(Swap.Short(((short[])data)[i]));
      }
      if (data instanceof int[])
      {
        mrange = ((int[])data).length;
        range = Math.min(range/4, mrange);
        for (int i=0; i<range; i++) ds.writeInt(Swap.Long(((int[])data)[i]));
      }
      if (data instanceof float[])
      {
        mrange = ((float[])data).length;
        range = Math.min(range/4, mrange);
        for (int i=0; i<range; i++) ds.write(Swap.Float(((float[])data)[i]));
      }
      if (data instanceof double[])
      {
        mrange = ((double[])data).length;
        range = Math.min(range/8, mrange);
        for (int i=0; i<range; i++) ds.write(Swap.Double(((double[])data)[i]));
      }
      if (data instanceof char[])
      {
        mrange = ((char[])data).length;
        range = Math.min(range, mrange);
        ds.writeBytes(new String((char[])data,0,range));
      }
      if (data instanceof String)
      {
        mrange = ((String)data).length();
        range = Math.min(range, mrange);
        ds.writeBytes(((String)data).substring(0, range));
      }
      if (data instanceof StringBuffer)
      {
        mrange = ((StringBuffer)data).length();
        range = Math.min(range, mrange);
        ds.writeBytes(((StringBuffer)data).toString().substring(0, range));
      }
      if (data instanceof TCompoundDataObject[])
      {
        TCompoundDataObject[] tdo = (TCompoundDataObject[])data;
        mrange = ((TCompoundDataObject[])data).length;
        range = Math.min(range/tdo[0].getSizeInBytes(), mrange);
       for (int i=0; i<range; i++) ds.write(tdo[i].toByteArray());
      }
    }
    catch (Exception e)
    { // TODO: log or debug it      
    }
  }
  public int getSeverity() { return almDef == null ? 0 : almDef.getAlarmSeverity(); }
  public String getTag() { return almDef == null ? "" : almDef.getAlarmTag(); }
  public TAlarm(int code, byte flags, byte[] data,TDevice device) 
  {
    almTimestamp = System.currentTimeMillis();
    almClearCount = 0;
    almRemovalBlocked = false;
    almCode = code;
    almDescriptor = flags;
    almData = data;
    almDevice = device;
    TEquipmentModule eqm = null;
    if (almDevice != null && (eqm=almDevice.getEqm()) != null)
    { // try the configured definitions list
      almDef = (TAlarmDefinition)eqm.getAlarmDefinitionList().get(code);
    }
    if (almDef == null)
    { // now try the stock definitions
      almDef = (TAlarmDefinition)eqm.getEquipmentModuleFactory().getStockAlarmDefinitions().get(Integer.toString(code));
    }
    if (almDef == null)
    { // still nothing, so make a new one ...
      String as = TErrorList.getErrorString(code);
      almDef = new TAlarmDefinition(as,as,"this device",code,TAlarm.ALM_SEVERITY_WARN,0);
    }
    if ((flags & TAlarmDescriptor.TERMINATE) == TAlarmDescriptor.TERMINATE)
    {  
      isTerminated = true;
    }
    if ((flags & TAlarmDescriptor.NEW) == TAlarmDescriptor.NEW)
    { // as this is a constructor, the NEW bit almost assuredly set!
      almStarttime = almTimestamp;
    }
  }
  public TAlarmDefinition getAlmDef() { return almDef; }
  public void setAlarmDefinition(TAlarmDefinition almDef)
  {
    this.almDef = almDef;
  }
  // see XPropertyQuery for template to make an Alarm Message Structure on the fly from
  // a stock property query !
  public boolean isTerminated()
  {
    return isTerminated;
  }
  public boolean isTransient()
  {
    return (almDescriptor & TAlarmDescriptor.TRANSIENT) == TAlarmDescriptor.TRANSIENT;
  }
  public void setTerminated(boolean isTerminated)
  {
    almDescriptor &= ~TAlarmDescriptor.NEW;
    almDescriptor |= TAlarmDescriptor.TERMINATE;
    this.isTerminated = isTerminated;
  }
}
