package de.desy.tine.server.alarms;

import de.desy.tine.definitions.TStrings;
import de.desy.tine.structUtils.TTaggedStructure;

public class TAlarmDefinition extends TTaggedStructure
{
  private char[] alarmTag = new char[TStrings.ALARM_TAG_SIZE];
  private String alarmTagString;
  private int[] alarmCode = new int[1];
  private int[] alarmMask = new int[1];
  private short[] alarmSystem = new short[1];
  private short[] alarmSeverity = new short[1];
  private byte[] alarmDataFormat = new byte[1];
  private byte[] alarmDataArraySize = new byte[1];
  private byte[] reserved = new byte[2];
  private char[] alarmText = new char[TStrings.ALARM_TEXT_SIZE];
  private String alarmTextString;
  private char[] deviceText = new char[TStrings.ALARM_TEXT_SIZE];
  private String deviceTextString;
  private char[] dataText = new char[TStrings.ALARM_TEXT_SIZE];
  private String dataTextString;
  private char[] url = new char[TStrings.ALARM_TEXT_LONGSIZE];
  private String urlString;
  public int getAlarmCode() { return alarmCode[0]; }
  public int getAlmCode() { return alarmCode[0]; } // for compatibility
  public int getAlarmMask() { return alarmMask[0]; }
  public int getAlarmSystem() { return alarmSystem[0]; }
  public int getAlarmSeverity() { return alarmSeverity[0]; }
  public int getAlarmDataFormat() { return alarmDataFormat[0]; }
  public int getAlarmDataArraySize() { return alarmDataArraySize[0]; }
  private String trimmedString(char[] charArray)
  {
    int len=0;
    while (charArray[len] != 0 && len < charArray.length) len++;
    return new String(charArray,0,len);
  }
  public String getAlarmTag()
  {
    if (alarmTagString == null) alarmTagString = trimmedString(alarmTag);//new String(alarmTag).trim();
    return alarmTagString;
  }
  public String getAlarmText()
  {
    if (alarmTextString == null) alarmTextString = trimmedString(alarmText);//new String(alarmText).trim();
    return alarmTextString;
  }
  public String getAlmText() { return getAlarmText(); } // for compatibility
  public String getDeviceText()
  {
    if (deviceTextString == null) deviceTextString = trimmedString(deviceText);//new String(deviceText).trim();
    return deviceTextString;
  }
  public String getDataText()
  {
    if (alarmDataArraySize[0] == 0) dataTextString = new String("no data associated with alarm");
    if (dataTextString == null) dataTextString = trimmedString(dataText);//new String(dataText).trim();
    return dataTextString;
  }
  public String getUrl()
  {
    if (urlString == null) urlString = trimmedString(url);//new String(url).trim();
    return urlString;
  }
  private static boolean initialized = false;
  private void initStructDescription()
  {
    if (initialized) 
    {
      //TODO:  this seems stupid to repeat this over and over ....
      //return;
    }
    addField(alarmTag,"almTag");
    addField(alarmCode,"almCode");
    addField(alarmMask,"almMask");
    addField(alarmSystem,"almSystem");
    addField(alarmSeverity,"almSeverity");
    addField(alarmDataFormat,"almDataFormat");
    addField(alarmDataArraySize,"almDataArraySize");
    addField(reserved,"reserved");
    addField(alarmText,"alarmText");
    addField(deviceText,"deviceText");
    addField(dataText,"dataText");
    addField(url,"url");
    initDone();      
    initialized = true;
  }
  public TAlarmDefinition()
  {
    super("ADSr4");
    initStructDescription();
  }
  private void pushString(String s,char[] c)
  {
    if (s != null) 
    {
      int len = s.length();
      if (len > c.length) len = c.length; 
      s.getChars(0, len, c, 0);
      for (int i=len; i<c.length; i++) c[i]=0;
    }    
  }
  public TAlarmDefinition(String tag,String txt,String devTxt,int code,int severity,int system,
      int mask,short dataFormat,int dataSize,String dataTxt,String urlTxt)
  {
    super("ADSr4");
    initStructDescription();
    pushString(tag,alarmTag);
    pushString(txt,alarmText);
    pushString(devTxt,deviceText);
    setAlarmCode(code);
    setAlarmSeverity(severity);
    setAlarmSystem(system);
    setAlarmMask(mask);
    pushString(dataTxt,dataText);
    setAlarmDataFormat(dataFormat);
    setAlarmDataArraySize(dataSize);
    pushString(urlTxt,url);
  }
  public TAlarmDefinition(String tag,String txt,String devTxt,int code,int severity,int system)
  {
    super("ADSr4");
    initStructDescription();
    pushString(tag,alarmTag);
    pushString(txt,alarmText);
    pushString(devTxt,deviceText);
    setAlarmCode(code);
    setAlarmSeverity(severity);
    setAlarmSystem(system);
  }
  public void setAlarmCode(int almCode)
  {
    alarmCode[0] = almCode;
  }
  public void setAlarmDataArraySize(int almDataArraySize)
  {
    alarmDataArraySize[0] = (byte)almDataArraySize;
  }
  public void setAlarmDataFormat(int almDataFormat)
  {
    int bf = almDataFormat%512;
    if (bf > 127) bf =- 256;
    alarmDataFormat[0] = (byte)(bf);
  }
  public void setAlarmMask(int almMask)
  {
    alarmMask[0] = almMask;
  }
  public void setAlarmSeverity(int almSeverity)
  {
    alarmSeverity[0] = (short)almSeverity;
  }
  public void setAlarmSystem(int almSystem)
  {
    alarmSystem[0] = (short)almSystem;
  }
  public void setAlarmTag(String tag)
  {
    pushString(tag,alarmTag);
  }
  public void setAlarmText(String alarmTxt)
  {
    pushString(alarmTxt,alarmText);
  }
  public void setDataText(String dataTxt)
  {
    pushString(dataTxt,dataText);
  }
  public void setDeviceText(String deviceTxt)
  {
    pushString(deviceTxt,deviceText);
  }
  public void setUrl(String urlTxt)
  {
    pushString(urlTxt,url);
  }
}
