/*
 * Created on Aug 24, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.devices;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TStrings;
import de.desy.tine.server.alarms.TAlarm;
import de.desy.tine.server.alarms.TAlarmDescriptor;
import de.desy.tine.server.equipment.TEquipmentModule;
import de.desy.tine.server.logger.*;
/**
 * \brief Instances of devices served by a device server should inherit from TDevice
 *
 * @author duval
 *
 */
public class TDevice
{
  public int devNumber;
  public String devName;
  private String devDescription;
  private LinkedList<TAlarm> devAlarmList = new LinkedList<TAlarm>();
  private int devAlarmMrt = 0; // Most Recent Timestamp
  private int devAlarmMrct = 0; // Most Recent Complete Timestamp
  private short devAlarmHsv = 0;
  private int nDevAlarmsAtMrt;
  private int nDevAlarmsAtMrct;
  private int nDevAlarmsAtHsv;
  private TEquipmentModule eqm = null;
  private String redirection = null;
  private LinkedList<String> devPrpList = null;
  public LinkedList<String> getPropertyList() { return devPrpList; }
//  public void setPropertyList(LinkedList<String> prpList) { devPrpList = prpList; }
  private LinkedList<String> mapStringArrayToList(String[] prpArray)
  {
    if (prpArray == null) return null;
    LinkedList<String> ll = new LinkedList<String>();
    for (String p : prpArray) ll.add(p);
    return ll;
  }
  public void setPropertyList(String[] prpList)
  {
    devPrpList = mapStringArrayToList(prpList);
  }
  public void setPropertyList(List<String> prpList)
  {
    if (prpList instanceof LinkedList)
    {
      devPrpList = (LinkedList<String>)prpList;
    }
    else if (prpList instanceof ArrayList)
    {
      String[] prps = ((ArrayList<String>)prpList).toArray(new String[0]);
      devPrpList = mapStringArrayToList(prps);
    } 
  }
  public int getHighestAlarmSeverity() { return devAlarmHsv; }
  public int getNumberOfAlarmsAtMostRecentTimestamp() { return nDevAlarmsAtMrct; }
  public int getNumberOfAlarmsAtHighestSeverity() { return nDevAlarmsAtHsv; }
  public TAlarm[] getAlarmList()
  {
    synchronized (devAlarmList)
    {
      return (TAlarm[])devAlarmList.toArray(new TAlarm[0]);
    }
  }
  private TAlarm getAlarmFromList(int code,byte[] data,short flags)
  { // called only by setAlarm() ...
    TAlarm alm = null;
    synchronized (devAlarmList)
    {
      Iterator<TAlarm> li = devAlarmList.iterator();
      if (code == 0) DbgLog.log("getAlarmFromList","searching for alarm code == 0?");
      while (li.hasNext())
      {
        if ((alm=(TAlarm)li.next()).isTerminated()) 
        { // is marked for termination
          if (!alm.isTransient())
          { // oscillation window too small ?
            int dt = (alm.getTimeStamp()+TAlarm.getTerminationWindow()) - (int)(System.currentTimeMillis()/1000);
            int oscw = TAlarm.getAlmOscillationWindow();
            if (dt > 0 && dt < 128-oscw)
            { // adjust the oscillation window
              oscw += dt;
              TAlarm.setAlmOscillationWindow(oscw);
              DbgLog.log("getAlarmFromList","increase alarm oscillation windows to "+oscw+" clear counts");
            }
          }
          continue;
        }
        else
        { // not marked for termination
          if (alm.getCode() == code) 
          { // it's already in the list so remove NEW bit !
            byte dsc = alm.getDescriptor();
            byte[] adata = alm.getData();
            int dlen = alm.getAlmDef() == null ? 0 : alm.getAlmDef().getAlarmDataArraySize();
            if (dlen > 0 && data != null && adata != null)
            { // check for alarm data changed
              for (int n=0; n<dlen && n<adata.length && n<data.length; n++) 
                if (data[n] != adata[n])
                { // remove NEW and add DATACHANGE
                  dsc &= ~(byte)(TAlarmDescriptor.NEW);
                  dsc |= (byte)TAlarmDescriptor.DATACHANGE;
                  DbgLog.log("getAlarmFromList",alm.getTag() + " : alarm data changed");
                }
            }
            if (alm.getClearCount() > 1)
            { // check for oscillating alarm ...
              dsc &= ~(byte)(TAlarmDescriptor.NEW);
              dsc |= (byte)TAlarmDescriptor.OSCILLATION;
              DbgLog.log("getAlarmFromList",alm.getTag() + " : mark alarm as oscillating");
            }
            alm.resetClearCount();
            alm.setDescriptor(dsc); 
            return alm;
          }
        }
      }
      // not found, so this is a new one ...
      // TODO: worry about 'max_alarms_exceeded', i.e. alarm storms and the length of this list!
      flags |= TAlarmDescriptor.NEW;
      if ((flags & TAlarmDescriptor.INSTANT) != 0) flags |= TAlarmDescriptor.TERMINATE;
      alm = new TAlarm(code,(byte)flags,data,this);
      if (eqm.casReadRequired) alm.blockRemoval();
      devAlarmList.add(alm);
    }
    DbgLog.log("getAlarmFromList",devName + " : " + alm.getTag() + "(" + code + ") : add alarm to list");
    return alm;
  }
  public int getNumberOfAlarms() 
  { 
    synchronized (devAlarmList)
    {
      return devAlarmList.size();
    }   
  }
  public int getMostRecentTimestamp() { return devAlarmMrct; }
  private void refreshMrt(int ts)
  {
    if (ts > devAlarmMrt) 
    {
      devAlarmMrt = ts;
      nDevAlarmsAtMrt = 1;
    }
    else if (ts == devAlarmMrt)
    {
      nDevAlarmsAtMrt++;
    }  
  }
  private void refreshHsv(short sv)
  {
    if (sv > devAlarmHsv)
    {
      devAlarmHsv = sv;
      nDevAlarmsAtHsv = 1;
    }
    else if (sv == devAlarmHsv)
    {
      nDevAlarmsAtHsv++;
    }
  }
  public void processAlarms()
  {
    int thisTimeStamp = (int)(System.currentTimeMillis()/1000);
    TAlarm alm = null;
    synchronized (devAlarmList)
    {
      Iterator<TAlarm> li = devAlarmList.iterator();
      nDevAlarmsAtHsv = 0; devAlarmHsv = 0;
      nDevAlarmsAtMrt = 0; devAlarmMrt = 0; 
      nDevAlarmsAtMrct = 0; devAlarmMrct = 0;
      while (li.hasNext())
      {
        alm = (TAlarm)li.next();
        if (alm.isTerminated() &&  
            thisTimeStamp > alm.getTimeStamp() + TAlarm.getTerminationWindow())
        {
          if (alm.getSeverity() > 0 && alm.isRemovalBlocked())
          {
            DbgLog.log("processAlarms",alm.getTag() + " : missing CAS read -> removal blocked");
          }
          else
          {
            DbgLog.log("processAlarms",alm.getTag() + " : remove alarm");
            devAlarmList.remove(alm);
            continue;
          }
        }
        // apply termination flag if necessary ...
        if (!alm.isTerminated() &&
            alm.getClearCount() > (TAlarm.getAlmOscillationWindow()-3) && 
            thisTimeStamp > alm.getTimeStamp())
        {
          byte dsc = alm.getDescriptor();
          dsc &= ~(byte)TAlarmDescriptor.NEW;
          dsc |= (byte)TAlarmDescriptor.TERMINATE;
          alm.setDescriptor(dsc);
          alm.setTimeStamp(thisTimeStamp);
          //refreshMrt(thisTimeStamp);
          alm.setTerminated(true);
          DbgLog.log("processAlarms",alm.getTag() + " : mark alarm as terminated");
        }
        // apply heartbeat if necessary
        if (thisTimeStamp > alm.getTimeStamp() + TAlarm.getAlmHeartbeatWindow())
        {
          byte dsc = alm.getDescriptor();
          dsc &= ~(byte)TAlarmDescriptor.NEW;
          dsc |= (byte)TAlarmDescriptor.HEARTBEAT;
          alm.setDescriptor(dsc);
          alm.setTimeStamp(thisTimeStamp);
          //refreshMrt(thisTimeStamp);
        }
        refreshMrt(alm.getTimeStamp());
        refreshHsv((short)alm.getSeverity());
      }
    }
    // adjust most recent completed timestamp (if necessary)
    if (thisTimeStamp > devAlarmMrt && devAlarmMrt > devAlarmMrct)
    {
      nDevAlarmsAtMrct = nDevAlarmsAtMrt;
      devAlarmMrct = devAlarmMrt;          
    }
  }
  /**
   * \brief sets an alarm with the given code
   * 
   * @param code The alarm code to apply to the alarm
   * @return
   */
  public TAlarm setAlarm(int code)
  {
    return setAlarm(code,null,TAlarmDescriptor.NEW);    
  }
  /**
   * \brief sets an alarm with the given data set
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @return
   */
  public TAlarm setAlarm(int code,byte[] data)
  {
    return setAlarm(code,data,TAlarmDescriptor.NEW);
  }
  /**
   * \brief sets an alarm with the given data set
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @return
   */
  public TAlarm setAlarm(int code,TDataType data)
  {
    return setAlarm(code,data.getDataBuffer(),TAlarmDescriptor.NEW);
  }
  /**
   * \brief sets an alarm with the given data set
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @return
   */
  public TAlarm setAlarm(int code,float data)
  {
    float[] fdata = new float[1];
    fdata[0] = data;
    TDataType td = new TDataType(fdata);
    return setAlarm(code,td.getDataBuffer(),TAlarmDescriptor.NEW);
  }
  /**
   * \brief sets an alarm with the given data set
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @return
   */
  public TAlarm setAlarm(int code,int data)
  {
    int[] idata = new int[1];
    idata[0] = data;
    TDataType td = new TDataType(idata);
    return setAlarm(code,td.getDataBuffer(),TAlarmDescriptor.NEW);
  }
  /**
   * \brief sets an alarm with the given data set
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @return
   */
  public TAlarm setAlarm(int code,short[] data)
  {
    TDataType td = new TDataType(data);
    return setAlarm(code,td.getDataBuffer(),TAlarmDescriptor.NEW);
  }
  /**
   * \brief sets an alarm with the given data set and given flags
   * 
   * On rare occasions you might want to apply the alarm descriptor flags
   * yourself in a call to setAlarms().  
   * 
   * @param code The alarm code to apply to the alarm
   * @param data The alarm data associated with the alarm 
   * @param flags The alarm descriptor to apply to the alarm
   * @return
   */
  public TAlarm setAlarm(int code,byte[] data,short flags)
  {
    TAlarm alm = getAlarmFromList(code,data,flags);
    if (alm.isTerminated()) DbgLog.log("setAlarm",alm.getTag() + " : set alarm as terminated");
    return alm;
  }
  /**
   * \brief Clears all alarms in the local alarm server's list
   *
   */
  public void clearAlarms()
  {
    synchronized (devAlarmList)
    {
      Iterator<TAlarm> li = devAlarmList.iterator();
      while (li.hasNext()) ((TAlarm)li.next()).clear();
    }
  }
  /**
   * \brief clears the alarm with the given code in the local alarm server's list
   * 
   * The alarm's clear counter will be incremented.  A call to clearAlarm() will
   * not cause the alarm to be removed from the list until the clear counter is
   * greater than 8.
   * 
   * @param code Is the alarm code specifying which alarm to clear 
   */
  public void clearAlarm(int code)
  {
    TAlarm alm = null;
    synchronized (devAlarmList)
    {
      Iterator<TAlarm> li = devAlarmList.iterator();
      while (li.hasNext())
      {
        alm = (TAlarm)li.next();
        if ((alm.getDescriptor() & TAlarmDescriptor.TERMINATE) != 0) continue;
        if (alm.getCode() != code) continue;
        alm.clear();
      }
    }
  }
  /**
   * \brief removes the alarm with the given code from the local alarm server's list
   * 
   * @param code Is the alarm code specifying which alarm to remove from the 
   * local alarm server's list
   */
  public void removeAlarm(int code)
  {
    synchronized (devAlarmList)
    {
      Iterator<TAlarm> li = devAlarmList.iterator();
      TAlarm alm = null;
      while (li.hasNext())
      {
        alm = (TAlarm)li.next();
        if ((alm.getDescriptor() & TAlarmDescriptor.TERMINATE) != 0) continue;
        if (alm.getCode() != code) continue;
        alm.setDescriptor((byte)TAlarmDescriptor.TERMINATE);
        alm.clear();
      }
    }
  }
  /**
   * \brief Simple constructor which assigns a device name and number
   * 
   * Use this constructor if you do not make use of a '<local-names>-devices.csv'
   * but instead either read your own database to determine the device name
   * name list.  You can assign a number and name to the device in question. 
   * 
   * @param name The device name to apply to this device
   * @param number The device number to apply to this device
   */
  public TDevice(String name, int number)
  {
    TStrings.validateDeviceName(name);
    devName = name;
    devNumber = number;
  }
  /**
   * \brief Simple constructor which assigns a device number
   * 
   * Use this constructor if you make use of a '<local-names>-devices.csv' 
   * configuration file to register devices or if you intend on using the
   * setName() method to attach a name to the device.
   *  
   * @param number The device number associated with this device
   */
  public TDevice(int number)
  {
    devName = "#" + number;
    devNumber = number;
  }
  /**
   * \brief Complex constructor which assigns name, number, and equipment module
   * 
   * You should not normally need to make use of this constructor, as the
   * identity of the equipment module is usually determined elsewhere during
   * initialization.
   * 
   * @param name The device name associated with this device
   * @param number The device number associated with this device
   * @param equipmentModule The equipment module to associate with this device
   */
  public TDevice(String name, int number, TEquipmentModule equipmentModule)
  {
    TStrings.validateDeviceName(name);
    devName = name;
    devNumber = number;
    eqm = equipmentModule;
  }
  public TEquipmentModule getEqm()
  {
    return eqm;
  }
  /**
   * \brief returns the registered device name
   * 
   * @return The registered device name
   */
  public String getName() { return devName; }
  /**
   * \brief sets the device name for this device
   * 
   * @param name The device name to associate with this device
   */
  public void setName(String name) 
  {
    TStrings.validateDeviceName(name);
    devName = name;
  }
  public void setEqm(TEquipmentModule eqm)
  {
    this.eqm = eqm;
  }
  /**
   * @brief returns the redirection string associated with this device
   * @return the redirection string associated with this device
   */
  public String getRedirection()
  {
    return redirection;
  }
  /**
   * \brief sets a redirection string to apply to this device
   * 
   * If the device has a redirection string associated with it, then
   * all remote requests for properties for this device will be
   * redirected to the remote device server given in the redirection string.
   * 
   * @param redirection The redirection string of the general form 
   * '/<context>/<server>/<device name>[property]'.  In this case, ONLY
   * the '<server>' element is relevant.
   */
  public void setRedirection(String redirection)
  {
    this.redirection = redirection;
  }
  /**
   * \brief gets the assigned description for this device
   * 
   * A device description can be assigned to a specific device. This
   * method call returns the assigned description.
   * 
   */
  public String getDescription()
  {
    return devDescription;
  }
  /**
   * \brief sets a description for this device
   * 
   * A device description can be assigned to a specific device via this
   * method call.
   * 
   * @param description The description to be applied to this device. 
   */
  public void setDescription(String description)
  {
    devDescription = description;
  }
}
