package de.desy.tine.server.equipment;


/**
 * @brief Server side cycle trigger methods
 * A cycle trigger interface.
 * 
 * The update() method will be called upon incoming 
 * cycle triggers (or system error).  The current cycle number will be passed 
 * along with the status (usually 0). 
 * 
 * @author duval
 *
 */
public interface TCycleTrigger
{
  /**
   * @brief is processed following an incoming cycle number event.
   * 
   * If a CYCLER exists and a TCycleTrigger is in use, then
   * its update routine will be called following the trigger
   * event associated with the incoming cycle number or a link
   * timeout notification.
   * 
   * @param cycleNumber is the last arrived cycle number
   * @param status is the cycle link status
   */
  void update(int cycleNumber,int status);
}
