package de.desy.tine.server.equipment;

import java.io.*;
import java.util.Date;
import java.util.LinkedList;
import de.desy.tine.csvUtils.*;
import de.desy.tine.definitions.*;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.properties.TStockProperties;
import de.desy.tine.structUtils.TTaggedStructure;

public class TEquipmentManifest extends TTaggedStructure implements RowHandler
{
  private static final String mfHdr = "EXPORT, CONTEXT, EQM, FEC, PORT, VERSION, FEC_HOME, LAST_STARTED";
  private static LinkedList<TEquipmentManifest> mfLst = null;
  private static String crlf = "\n";
  private char[] exp = new char[TStrings.EXPORT_NAME_SIZE];
  public String export;
  private char[] ctx = new char[TStrings.CONTEXT_NAME_SIZE];
  public String context;
  private char[] eqmName = new char[TStrings.EQM_NAME_SIZE];
  public String eqm;
  private char[] fecName = new char[TStrings.FEC_NAME_SIZE];
  public String fec;
  private char[] ver = new char[16];
  public String version;
  private char[] fecHome = new char[32];
  public String fec_home;
  private char[] started = new char[32];
  public String last_started;
  private int[] portOffset = new int[1];
  public int port;
  private void pushString(String s,char[] c)
  {
    int len = s.length();
    if (len > c.length) len = c.length;     
    s.getChars(0, len, c, 0);
  }
  private void initStructure()
  {
    addField(exp,"export");
    addField(ctx,"context");
    addField(eqmName,"eqmName");
    addField(fecName,"fecName");
    addField(ver,"version");
    addField(fecHome,"fecHome");
    addField(started,"lastStarted");
    addField(portOffset,"port");
    initDone();    
  }
  TEquipmentManifest()
  {
    super("FECmf");
    initStructure();
  }
  TEquipmentManifest(TEquipmentManifest reference)
  {
    super("FECmf");
    initStructure();
    pushString(reference.export,exp);
    export = new String(exp).trim();
    pushString(reference.context,ctx);
    context = new String(ctx).trim();
    pushString(reference.eqm,eqmName);
    eqm = new String(eqmName).trim();
    pushString(reference.fec,fecName);
    fec = new String(fecName).trim();
    port = portOffset[0] = reference.port;
    pushString(reference.version,ver);
    version = new String(ver).trim();
    pushString(reference.fec_home,fecHome);
    fec_home = new String(fecHome).trim();
    pushString(reference.last_started,started);
    last_started = new String(started).trim();
        
  }
  private void replace(TEquipmentModule em)
  {
    TEquipmentModuleFactory tf = TEquipmentModuleFactory.getInstance();
    pushString(em.getExportName(),exp);
    export = new String(exp).trim();
    pushString(em.getContext(),ctx);
    context = new String(ctx).trim();
    pushString(em.getLocalName(),eqmName);
    eqm = new String(eqmName).trim();
    pushString(tf.getFecName(),fecName);
    fec = new String(fec).trim();
    portOffset[0] = tf.getPort();
    port = portOffset[0];
    pushString(TStockProperties.getStockSrvVersion(),ver);
    version = new String(ver).trim();
    pushString(tf.getInitializer().getFecHome(),fecHome);
    fec_home = new String(fecHome).trim();
    pushString(new Date().toString(),started);
    last_started = new String(started).trim();
  }
  private static String getEntryAsString(TEquipmentModule em)
  {
    TEquipmentModuleFactory tf = TEquipmentModuleFactory.getInstance();
    String mfEntry = em.getExportName().trim();
    mfEntry += ", " + em.getContext().trim();
    mfEntry += ", " + em.getLocalName().trim();
    mfEntry += ", " + tf.getFecName().trim();
    mfEntry += ", " + tf.getPort();
    mfEntry += ", " + TStockProperties.getStockSrvVersion().trim();
    mfEntry += ", " + tf.getInitializer().getFecHome().trim();
    mfEntry += ", " + new Date().toString().trim();
    mfEntry += crlf;
    return mfEntry;    
  }
  private static String getEntryAsString(TEquipmentManifest tem)
  {
    String mfEntry = tem.export.trim();
    mfEntry += ", " + tem.context.trim();
    mfEntry += ", " + tem.eqm.trim();
    mfEntry += ", " + tem.fec.trim();
    mfEntry += ", " + tem.port;
    mfEntry += ", " + tem.version.trim();
    mfEntry += ", " + tem.fec_home.trim();
    mfEntry += ", " + tem.last_started.trim();
    mfEntry += crlf;
    return mfEntry;    
  }
  public static synchronized int update(TEquipmentModule em)
  { // update the manifest for this equipment module
    if (mfLst != null) mfLst.clear();
    
    csvColumn[] mfCols = new csvColumn[8];
    TEquipmentManifest mfRows = new TEquipmentManifest();
    mfCols[0] = new csvColumn("EXPORT","",new StringFieldHandler(mfRows,"export"));
    mfCols[1] = new csvColumn("CONTEXT","",new StringFieldHandler(mfRows,"context"));
    mfCols[2] = new csvColumn("EQM","",new StringFieldHandler(mfRows,"eqm"));
    mfCols[3] = new csvColumn("FEC","",new StringFieldHandler(mfRows,"fec"));
    mfCols[4] = new csvColumn("PORT","",new IntFieldHandler(mfRows,"port"));    
    mfCols[5] = new csvColumn("VERSION","",new StringFieldHandler(mfRows,"version"));
    mfCols[6] = new csvColumn("FEC_HOME","",new StringFieldHandler(mfRows,"fec_home"));
    mfCols[7] = new csvColumn("LAST_STARTED","",new StringFieldHandler(mfRows,"last_started"));

    String mfFilePath;
    if (File.separatorChar == '/')
    { // unix or mac
      mfFilePath = "/var/tmp";
    }
    else if (File.separatorChar == '\\')
    { // windows
      mfFilePath = System.getenv("SystemDrive");
      if (mfFilePath == null) mfFilePath = "C:";
      crlf = "\r\n";
    }
    else
    {
      mfFilePath = ".";
    }
    mfFilePath += File.separator + "tine" + File.separator + "cache";
    
    String mfFileName = mfFilePath + File.separator + "fec.mf";
    // open it
    csv mfFile = new csv(mfFileName);
    // read it
    int rc = mfFile.readFile(mfCols,mfRows);
    // close it
    mfFile.close();  
    BufferedReader mfin;
    BufferedWriter mfout;
    if (rc != 0)
    { // create a new manifest
      File mf = new File(mfFilePath);
      mf.mkdirs();
      try
      {
        mfout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mfFileName)));
        mfout.write(mfHdr.trim() + crlf);
        mfout.write(getEntryAsString(em));
        mfout.close();
        return 0; // we're finished
      }
      catch (Exception e)
      {
        MsgLog.log("TEquipmentManifest","cannot update FEC manifest !",TErrorList.file_error,e,0);
        return TErrorList.file_error;
      }
    }
    try
    {
      String mfTempName = mfFilePath + File.separator + "fec.tmp";
      mfin = new BufferedReader(new InputStreamReader(new FileInputStream(mfFileName)));
      mfout = new BufferedWriter(new OutputStreamWriter(new FileOutputStream(mfTempName)));
      String hdr = mfin.readLine();
      mfout.write(hdr + crlf);
      boolean found = false;
      int nfecs = mfLst.size();
      TEquipmentManifest tem;
      for (int i=0; i<nfecs; i++)    
      {
        tem = mfLst.get(i);
        if (tem.export.compareToIgnoreCase(em.getExportName()) == 0 &&
            tem.context.compareToIgnoreCase(em.getContext()) == 0)
        { // this one is re-starting
          found = true;
          tem.replace(em);
        }
        mfout.write(getEntryAsString(tem));
      }
      if (!found)
      { // there's a manifest, but this one's not yet in it !
        mfout.write(getEntryAsString(em));        
      }
      mfin.close();
      mfout.close();
      File mf = new File(mfFileName);
      mf.delete();
      mf = new File(mfTempName);      
      mf.renameTo(new File(mfFileName));
    }
    catch (Exception e)
    {
    }
    return 0;
  }
  public int process(int index)
  { // if reading a file, add this one to a link list
    if (mfLst == null) mfLst = new LinkedList<TEquipmentManifest>();
    mfLst.add(new TEquipmentManifest(this));
    return 0;
  }
}
