/*
 * Created on Nov 5, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 * 
 * CHANGES
 * 2005-02-09 JW * main() removed; see de.desy.tine.client.test.sine
 * 2005-02-10 JW * Constructors made public
 *               * registerProperty(), registerDeviceName() made public
 * 				 * getExportPro[pertyNames(), getExportDeviceNames() added (wrappers)
 *               * Made (almost...) independent from TEquipmentModuleFactory-Singleton
 * 				 * SetBackgroundTask removed; belongs to EqmFactory.
 * 				 * CreateEquipmentModule removed / replaced by initialize().
 * 2004-02-14 JW * old stock properties replaced by TPropertyList
 *               * Stock Equipment function removed, replaced by property handlers
 * 2004-02-15 JW * registerStockProperties changed; register only those properties which 
 * 					are actually supported.
 * 13.7.05 * PD added getExportInformationFromFile()
 * 
 */
package de.desy.tine.server.equipment;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.StringTokenizer;
import java.util.concurrent.ConcurrentLinkedQueue;
import de.desy.tine.client.*;
import de.desy.tine.csvUtils.*;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.queryUtils.PropertyQuery;
import de.desy.tine.queryUtils.TPropertyQuery;
import de.desy.tine.queryUtils.XPropertyQuery;
import de.desy.tine.server.alarms.*;
import de.desy.tine.server.connections.TClient;
import de.desy.tine.server.connections.TClientEntry;
import de.desy.tine.server.connections.TContractTable;
import de.desy.tine.server.devices.TDevice;
import de.desy.tine.server.devices.TDeviceList;
import de.desy.tine.server.histories.THistoryRecord;
import de.desy.tine.server.histories.THistoryRecordStruct;
import de.desy.tine.server.histories.THistorySpecification;
import de.desy.tine.server.logger.*;
import de.desy.tine.server.properties.TExportProperty;
import de.desy.tine.server.properties.TMetaProperties;
import de.desy.tine.server.properties.TPropertyDescription;
import de.desy.tine.server.properties.TPropertyEGU;
import de.desy.tine.server.properties.TPropertyHandler;
import de.desy.tine.server.properties.TPropertyList;
import de.desy.tine.server.properties.TStockProperties;
import de.desy.tine.startup.TInitializer;
import de.desy.tine.startup.TInitializerFactory;
import de.desy.tine.stringUtils.StringToName;
import de.desy.tine.stringUtils.WildcardMatch;
import de.desy.tine.structUtils.TStructRegistry;
import de.desy.tine.types.*;
import de.desy.tine.xmlUtils.AlarmDefinitionCfg;
import de.desy.tine.xmlUtils.DeviceCfg;
import de.desy.tine.xmlUtils.EqmCfg;
import de.desy.tine.xmlUtils.FecCfg;
import de.desy.tine.xmlUtils.NameCfgList;
import de.desy.tine.xmlUtils.PropertyCfg;

/**
 * \brief A server handles all data requests through its equipment module.  
 * 
 *  An exported device server refers to one and only one equipment module. The
 *  equipment module will have a property list and a device list associated with it.
 *  
 * @author duval 
 */
public class TEquipmentModule
{
  private String context; // the server context (necessary here?)
  private String subsystem;
  private String exportName = null; // the device server name
  private String moduleName = null; // the 'local' name
  private String groupName = null; // the device group (if any)
  private int groupIndex = 0;  // the device group index (if any)
  private String master; // failover master (if set)
  private String slaveMaster; // failover slave master (if set and eqm is a slave)
  private int failoverType = TEquipmentModuleFactory.FAILOVER_NONE;
  public int getFailoverType() { return failoverType; }
  protected TEquipmentModuleFactory gEqmFactory = null;
  public TEquipmentModuleFactory getEquipmentModuleFactory() 
  {
    if (gEqmFactory == null) gEqmFactory = TEquipmentModuleFactory.getInstance();
    return gEqmFactory; 
  }
  protected LinkedList<String> gRegisteredUsersList = new LinkedList<String>();
  protected LinkedList<String> gRegisteredNetsList = new LinkedList<String>();
  protected LinkedList<String> gGCastNetsList = new LinkedList<String>();
  protected HashMap<Integer,TAlarmDefinition> alarmDefinitionList = new HashMap<Integer,TAlarmDefinition>();
  protected boolean gRejectAllNets = false;
  public HashMap<Integer, TAlarmDefinition> getAlarmDefinitionList() { return alarmDefinitionList; }
  public int addAlarmDefinition(int code,TAlarmDefinition adef)
  {
    if (code < 0 || adef == null) return TErrorList.argument_list_error;
    alarmDefinitionList.put(new Integer(code),adef);
    return 0;
  }
  public int addAlarmDefinition(TAlarmDefinition adef)
  {
    if (adef == null) return TErrorList.argument_list_error;
    return addAlarmDefinition(adef.getAlarmCode(),adef);
  }
  public TAlarmDefinition getAlarmDefinition(int code)
  {
    return alarmDefinitionList.get(new Integer(code));
  }
  TPropertyList stockList; // = new TPropertyList(); // a list or a query fcn
  TPropertyList propertyList = new TPropertyList(); // a list or a query fcn
  TDeviceList deviceList = new TDeviceList();
  public TDeviceList getDeviceList() { return deviceList; }
  public TPropertyList getPropertyList() { return propertyList; }
  public int getDeviceNumber(String devName)
  {
    try
    {
      if (devName.startsWith("#")) return Integer.parseInt(devName.substring(1));
      return deviceList.getDeviceNumber(devName);
    }
    catch (Exception e)
    {
      return -1;
    }
  }
  public int getDeviceNumber(String devName,String devProperty)
  {
    try
    {
      if (devName == null) return -TErrorList.argument_list_error;
      if (devName.startsWith("#")) return Integer.parseInt(devName.substring(1));
      if (deviceList.isPropertyOriented() && devProperty != null)
      {
        TExportProperty prp = propertyList.getFirstProperty(devProperty);
        if (prp != null)
        {
          ArrayList<String> lst = prp.getDeviceList();
          if (lst != null) return lst.indexOf(devName);
        }
      }
      return deviceList.getDeviceNumber(devName);
    }
    catch (Exception e)
    {
      return -1;
    }
  }
  protected TAlarmDynSet almDynSet = new TAlarmDynSet();
  protected String[] gPropertyNameList = null;
  protected LinkedList<TAlarmWatchEntry> gAlarmWatchList = new LinkedList<TAlarmWatchEntry>();
  protected LinkedList<THistoryRecord> gLclHstList = new LinkedList<THistoryRecord>();
  private boolean useMonthlyHistoryFiles = false;
  public void setUseMSecHistoryTimestamps(boolean useMsecHistoryTimestamps)
  {
    useMinimalStorage = !useMsecHistoryTimestamps;
  }
  private boolean useMinimalStorage = false;
  public boolean isUseMinimalStorage() { return useMinimalStorage; }
  public boolean isUseMSecHistoryTimestamps() { return !useMinimalStorage; }
  public void setUseMonthlyHistoryFiles(boolean useMonthlyFiles)
  {
    useMonthlyHistoryFiles = useMonthlyFiles;
  }
  public boolean isUseMonthlyHistoryFiles() { return useMonthlyHistoryFiles; }
  THistoryRecord getLocalHistoryRecord(String Property,String Device)
  {
    int listsize = gLclHstList.size();
    THistoryRecord lhr = null;
    for (int i=0; i<listsize; i++)
    {
      lhr = (THistoryRecord)gLclHstList.get(i);
      if (lhr.getPrp().compareTo(Property) != 0) continue;
      // TODO: worry about matching device name ? (vs multi-channel array)
      return lhr;
    }
    return null;
  }
  public int addLocalHistoryRecord(int recordIndex,String device,String property,int dataSize,int dataFormat)
  {
    return addLocalHistoryRecord(recordIndex,device,property,dataSize,dataFormat,TArrayType.AT_UNKNOWN);
  }
  public int addLocalHistoryRecord(int recordIndex,String device,String property,int dataSize,int dataFormat,int arrayType)
  {
    THistorySpecification spc = 
      new THistorySpecification(1000,1000,600,1,18000,10,0,"");
    return addLocalHistoryRecord(recordIndex,device,property,dataSize,dataFormat,arrayType,spc);
  }
  public int addLocalHistoryRecord(int recordIndex,String device,String property,int dataSize,int dataFormat,int arrayType,THistorySpecification histSpec)
  {
    THistoryRecord hst = new THistoryRecord(this,device,property,dataSize,dataFormat,arrayType,recordIndex,histSpec);
    if (useMinimalStorage) hst.setUseMinimalStorage(true);
    if (useMonthlyHistoryFiles) hst.setUseMonthlyHistoryFiles(true);
    gLclHstList.add(hst);
    return 0;    
  }
  private boolean isIdle = false;
  public boolean isIdleState() { return isIdle; }
  public void setIdleState(boolean value) 
  { 
    isIdle = value;
    Iterator<TEquipmentBackgroundTask> li = bkgTasks.iterator();
    while (li.hasNext())
    { // pass on to any associated background tasks ...
      li.next().setIdleState(value);
    }
  }
  public boolean hasInitialized = false;
  public boolean isRegistered; // registered with name server
  public boolean hasCasAttached = false;
  public boolean casReadRequired = false;
  public boolean grpRegistered = false;
  public boolean registerAsMaster = false;
  private String casName;
  public String getCasName() { return casName; }
  //TAlarmTable alarms; // active alarms + alarm definitions
  int numClients;
  int numContracts;
  boolean isCasInClientList()
  {
    if (casName == null) return false;
    TClientEntry[] tce = getClientList();
    for (int i=0; i<tce.length; i++) 
      if (tce[i].cln.userName.compareTo(casName) == 0)
      {
        casReadRequired = true;
        hasCasAttached = true;
        return true;
      }
    return false;
  }
  class AccessLock
  {
    static final int LOCK_UNLOCKED = 0;
    static final int LOCK_PREEMPTIVE = 1;
    static final int LOCK_PERSISTENT = 2;
    static final int LOCK_ABORT = 3;
    private TClient cln;
    private int lockType = LOCK_UNLOCKED;
    private int lockDuration = 0;
    private long lockStart;
    int getLockDuration() { return lockDuration; }
    int getLockType() { return lockType; }
    long getLockStart() { return lockStart; }
    void checkAccessLocks()
    {
      long t = System.currentTimeMillis();
      if (t > lockStart + lockDuration)
      {
        lockStart = 0;
        lockType = LOCK_UNLOCKED;
      }
    }
    boolean clientCanWrite(TClient tc)
    {
      if (lockType == LOCK_UNLOCKED) return true;
      if (tc.IPaddress == cln.IPaddress &&
          tc.userName.compareTo(cln.userName) == 0 &&
          tc.port == cln.port)
        return true;
      return false;
    }
    int setAccessLock(TClient tc,int type,int duration)
    {
      boolean passed = false;
      if (type < LOCK_UNLOCKED || type > LOCK_ABORT) return TErrorList.out_of_range;
      if (clientCanWrite(tc)) passed = true;
      else if (lockType != LOCK_PERSISTENT) passed = true;
      if (!passed) return TErrorList.access_denied;
      if (duration < 0) duration = 0;
      if (duration > TEquipmentModuleFactory.MAXIMUM_LOCK_DURATION)
        duration = TEquipmentModuleFactory.MAXIMUM_LOCK_DURATION;
      if (!TEquipmentModuleFactory.getInstance().gPutCommandsInFeclog)
      {
        String cmdString = "COMMAND ACCESSLOCK" + " called by " + tc.userName + 
          " from addr " + tc.IPaddress.getHostAddress() + " : ";
        if (lockType != type) TFecLog.log(cmdString);
      }
      if (type == LOCK_ABORT || type == LOCK_UNLOCKED)
      {
        lockStart = 0;
        lockType = LOCK_UNLOCKED;
      }
      else
      {
        cln = tc;
        lockType = type;
        lockDuration = duration;
        lockStart = System.currentTimeMillis();        
      }
      return 0;
    }
    String[] getAccessLockAsString()
    {
      String[] s = new String[3];
      if (lockType == LOCK_UNLOCKED) 
      {
        s[0] = s[1] = "";
      }
      else
      {
        int secs = (int)((lockStart + lockDuration - System.currentTimeMillis())/1000);
        if (secs < 0) secs = 0;
        s[0] = cln.userName;
        s[1] = cln.IPaddress.getHostAddress();
        s[2] = new String("" + secs + " sec remaining");
      }
      return s;
    }
  }
  public AccessLock accessLock = new AccessLock();
  // REMOVED TEquipmentFunction stkMod; // stock meta property eqp. fcn.
  //TEquipmentFunction stkFcn; // stock property eqp. fcn.
  TEquipmentFunction eqpFcn = null;
  //TEquipmentBackgroundTask eqpBkg;
  private TEquipmentInitRoutine eqpIni;
  public TEquipmentInitRoutine getInitializationRoutine() { return eqpIni; }
  public void setInitializationRoutine(TEquipmentInitRoutine initRoutine)
  {
    eqpIni = initRoutine;
  }
  private TEquipmentExitRoutine eqpExi;
  public TEquipmentExitRoutine getExitRoutine() { return eqpExi; }
  public void setExitRoutine(TEquipmentExitRoutine exitRoutine)
  {
    eqpExi = exitRoutine;
  }
  public NAME16[] getRegisteredUsers() { return StringToName.stringSetToName16(gRegisteredUsersList); }
  public int addRegisteredUser(String newUser)
  {
    newUser = newUser.toUpperCase();
    if (gRegisteredUsersList.contains(newUser)) return TErrorList.already_assigned;
    gRegisteredUsersList.add(newUser);    
    return 0;
  }
  public int removeRegisteredUser(String oldUser)
  {
    oldUser = oldUser.toUpperCase();
    if (!gRegisteredUsersList.contains(oldUser)) return TErrorList.invalid_name;    
    gRegisteredUsersList.remove(oldUser);    
    return 0;
  }
  public void removeAllRegisteredUsers()
  {
    gRegisteredUsersList.clear();
  }
  public int addRegisteredNet(String newNet)
  {
    if (gRegisteredNetsList.contains(newNet)) return TErrorList.already_assigned;
    gRegisteredNetsList.add(newNet);    
    return 0;
  }
  public int removeRegisteredNet(String oldNet)
  {
    if (!gRegisteredNetsList.contains(oldNet)) return TErrorList.invalid_name;    
    gRegisteredNetsList.remove(oldNet);
    return 0;
  }
  public void removeAllRegisteredNets()
  {
    gRegisteredNetsList.clear();
  }
  public NAME16[] getRegisteredNets() { return StringToName.stringSetToName16(gRegisteredNetsList); }
  public int addGCastNet(String newNet)
  {
    if (gGCastNetsList.contains(newNet)) return TErrorList.already_assigned;
    gGCastNetsList.add(newNet); 
    return 0;
  }
  public int removeGCastNet(String oldNet)
  {
    if (!gGCastNetsList.contains(oldNet)) return TErrorList.invalid_name;    
    gGCastNetsList.remove(oldNet);
    return 0;
  }
  public void removeAllGCastNets()
  {
    gGCastNetsList.clear();
  }
  public NAME16[] getGCastNets() { return StringToName.stringSetToName16(gGCastNetsList); }
  public void setCompletionString(String completionString)
  {
    TContractTable tct = this.getTEqmFactory().getCurrentContractEntry();
    if (tct != null) tct.compString = completionString;
  }
  public int setLocalName(String lclName)
  {
    moduleName = lclName;
    if (lclName.length() > TStrings.EQM_NAME_SHORTSIZE)
    {
      moduleName = lclName.substring(0,TStrings.EQM_NAME_SHORTSIZE);
      TFecLog.log("module name "+lclName+" too long; truncating to "+moduleName);
    }
    TStrings.validateLocalName(moduleName);
    getEquipmentModuleFactory().registrationPending = true;
    return 0;
  }
  public String getLocalName() { return moduleName; }
  public String getModuleName() { return moduleName; }
  public int setExportName(String expName)
  {
    if (exportName == null) 
    {
      exportName = new String(expName);
      TStrings.validateExportName(exportName);
    }
    return 0;
  }
  public String getExportName()
  {
    return exportName;
  }
  public void setGroupName(String grpName)
  {
    if (groupName == null) 
    {
      groupName = new String(grpName);
      TStrings.validateExportName(groupName);
      grpRegistered = false;
      TEquipmentModuleFactory.getInstance().registrationPending = true;
    }
  }
  public String getGroupName() { return groupName; }
  public void setGroupIndex(int grpIndex)
  {
    groupIndex =  grpIndex;
  }
  public int getGroupIndex() { return groupIndex; }
  public int setContext(String expContext)
  {
    if (context == null && expContext != null) 
    {
      context = expContext;
      TStrings.validateContextName(expContext);
    }
    return 0;
  }
  public String getContext()
  {
    return context;
  }
  public int setMaster(String expMaster)
  {
    if (master == null && expMaster != null)
    {
      master = expMaster;
      TStrings.validateExportName(expMaster);
      if (master.length() > 0)
      {
        failoverType = TEquipmentModuleFactory.FAILOVER_MASTER;
        registerAsMaster = true;
      }
      else
      {
        failoverType = TEquipmentModuleFactory.FAILOVER_NONE;
        registerAsMaster = false;
      }
    }
    return 0; 
  }
  public String getMaster() { return master; }
  public int setSlaveMaster(String expMaster, String slvMaster)
  {
    if (master != null && expMaster != null)
    {
      master = expMaster; slaveMaster = slvMaster; 
      TStrings.validateExportName(expMaster);
      TStrings.validateExportName(slvMaster);
      if (master.length() > 0 && slaveMaster.length() > 0)
      {
        failoverType = TEquipmentModuleFactory.FAILOVER_SLAVE;
        startFailoverMonitor();
      }
      else
      {
        failoverType = TEquipmentModuleFactory.FAILOVER_NONE;
      }
      registerAsMaster = false;
    }
    return 0; 
  }
  public String getSlaveMaster() { return slaveMaster; }
  private int foPollingInterval = 1000;
  private int foThreshold = 5;
  private boolean foMonitorStarted = false;
  class FailOverCallback implements TLinkCallback
  {
    int errcnt=0;
    int last_cc=0;
    boolean ismaster = false;
    public void callback(TLink link)
    {
      int cc = link.getLinkStatus();
      if (last_cc != cc)
      {
        TFecLog.log("failover slave link status changed from "+
            TErrorList.getErrorString(last_cc)+" to "+TErrorList.getErrorString(cc));
        last_cc = cc;
      }
      if (cc != 0)
      {
        errcnt++;
        if (errcnt > foThreshold && !ismaster)
        { // need to take over as the master !
          TEquipmentModuleFactory.getInstance().registrationPending = true;
          registerAsMaster = true;
          ismaster = true;
          ClsLog.log(getExportName()+" is taking over as "+getMaster()+" master",
              getContext(),"FAILOVER",gEqmFactory.getFecName(),ClsLog.ClogPriority.CLOG_PRIORITY_IMPORTANT,ClsLog.ClogStatus.CLOG_STATUS_WARN);
        }
        return;
      }
      errcnt = 0;
      ismaster = false;
    }    
  }
  private int startFailoverMonitor()
  {
    if (foMonitorStarted) return TErrorList.already_assigned;
    if (failoverType != TEquipmentModuleFactory.FAILOVER_SLAVE) return TErrorList.not_allowed;
    String mstr = getSlaveMaster();
    if (mstr == null || mstr.length() == 0) return TErrorList.not_allowed;
    String tgt = "/"+getContext()+"/"+mstr;
    int[] lval = new int[1];
    TDataType dout = new TDataType(lval);
    TLink lnk = new TLink(tgt,"SRVSTARTTIME",dout,null,TAccess.CA_READ);
    int id = lnk.attach(TMode.CM_TIMER, new FailOverCallback(), foPollingInterval);
    foMonitorStarted = true;
    TFecLog.log("failover slave link to "+tgt+" : "+(id < 0 ? lnk.getLastError() : "success"));
    return ( id < 0) ? -id : 0;
  }
  public TEquipmentModuleFactory getTEqmFactory() { return gEqmFactory; }
  public TClient getCaller()
  {
    TContractTable tct = gEqmFactory.getCurrentContractEntry();
    TClientEntry tce = null;
    if (tct != null)
    {
      try
      {
        tce = (TClientEntry)tct.clt.element();
      }
      catch (NoSuchElementException e)
      {
        DbgLog.log("getCaller", "contract does not have a client list");
      }
    }
    return tce != null ? tce.cln : null;
  }   
  public TClientEntry[] getClientList()
  {
    TContractTable tct = gEqmFactory.getCurrentContractEntry();
    return tct != null ? ((TClientEntry[])tct.clt.toArray(new TClientEntry[0])) : null;
  }
  protected int clearCASAlarmList()
  {
    String ctx = getContext();
    String exp = getExportName();
    if (ctx == null || exp == null) return TErrorList.invalid_name;
    String devName = "/" + ctx + "/CAS/" + exp;
    int cc = TErrorList.address_unknown;
    String fec = null;
    try
    {
      TLink lnk = new TLink(devName,"REMOVEALARMS",null,null,TAccess.CA_WRITE);
      cc = lnk.execute(200,true);
      fec = lnk.srvAddr.fecName;
      lnk.close();
    }
    catch (Exception e) {}
    if (cc == 0) 
    { // CAS exists in this context and is monitoring this FEC
      casReadRequired = true;
      casName = fec;
      TFecLog.log("clear alarm list at " + devName + "(" + casName + ")");
    }
    else
    {
      TFecLog.log("this server not monitored by a central alarm server");
    }
    return cc;
  }
  /**
   * Initializer used by constructors.
   * 
   * @param name Module name.
   * @param eqpFunction The TEquipmentFunction to attach to the equipment module.
   */
  private void initialize(String name, TEquipmentFunction eqpFunction)
  {
    gEqmFactory = TEquipmentModuleFactory.getInstance();
    if (name != null)
    {
      moduleName = name.length() > TStrings.EQM_NAME_SHORTSIZE ? name.substring(0, TStrings.EQM_NAME_SHORTSIZE) : name;
    }
    else
    {
      moduleName = "";
    }
    if (eqpFunction != null)
    {
      this.eqpFcn = eqpFunction;
      this.eqpFcn.setEquipmentModule(this);
    }
    registerStockProperties();
    getRegisteredUsersFromFile();
    getRegisteredNetsFromFile();
    getRegisteredAlarmDefinitionsFromFile();
    gEqmFactory.addEquipmentModule(this);
    if (gEqmFactory.isInitialized)
    { // this is being called late (after the normal system start)
      if (!hasInitialized)
      { // have not yet initialized -> call startup routine myself !
        startup();
      }
    }
    hasInitialized = true;
  }
  private long lastHistoryCycleTime = 0;
  private int historyCycleInterval = 1000; // milliseconds
  public void setHistoryCycleInterval(int cycleInterval) { historyCycleInterval = cycleInterval; }
  protected void historyCycle()
  {
    if (gEqmFactory.gEqpFcnBusySemaphore) return;
    long thisCycleTime = System.currentTimeMillis();
    THistoryRecord hst = null;
    THistorySpecification spc = null;
    TDataType dout = null;
    TDataType din = new TDataType();
    TAccess access = new TAccess(TAccess.CA_HIST + TAccess.CA_READ);
    int cc = 0;
    
    if (thisCycleTime-lastHistoryCycleTime < historyCycleInterval) return;
    lastHistoryCycleTime = thisCycleTime;
    int nHistoryRecords = gLclHstList.size();
    for (int i=0; i<nHistoryRecords; i++)
    {
      hst = (THistoryRecord)gLclHstList.get(i);
      spc = hst.getHspec();
      if (spc.getDepthShort() <= 0) continue;
      hst.setArchiveFileName(thisCycleTime); // will also remove outdated files if necessary
      if (thisCycleTime-hst.getTimestamp() < spc.getPollngRate()) continue;
      dout = new TDataType(hst.getArraySize(),hst.getFmt());
      dout.setDataTimeStamp(TDataTime.getDataTimeStamp());
      dout.setSystemDataStamp(TEquipmentModuleFactory.getInstance().getSystemStamp());
      gEqmFactory.gEqpFcnBusySemaphore = true;
      cc = callProperty(hst.getPrp(),hst.getDev(),dout,din,access);
      gEqmFactory.gEqpFcnBusySemaphore = false;
      if (cc != 0)
      {
        continue;
      }
      hst.writeRecordToSTS(dout);
      if (needToArchiveRecord(hst)) 
      {
        hst.writeRecordToLTS(dout);
        if (hst.isPointOfInterest()) hst.writeRecordToPIF(dout);
      }
    }
    return;
  }
  private boolean needToArchiveRecord(THistoryRecord hst)
  {
    long ts = System.currentTimeMillis();
    THistorySpecification spc;

    if (hst == null) return false;
    if((spc=hst.getHspec()) == null) return false;
    if (spc.getArchiveRate() < 0 ||
        spc.getHeartbeat() < 0   ||
        spc.getDepthLong() <= 0) return false;/* no long term storage */
    if (hst.getRngBufferSize() <= 0) return false;
    if (hst.getTimestamp() <= hst.getLastArchiveTime()) return false;/* data timestamp hasn't changed */
    if (ts < hst.getLastArchiveTime() + spc.getArchiveRate()) return false;
    if (ts >= hst.getLastArchiveTime() + spc.getHeartbeat()*1000) return true; /* archive no matter what */
    if (hst.isWithinTolerance()) return false;
    return true;    
  }
  protected void checkAlarmsInWatchTable()
  {
    if (gEqmFactory.gEqpFcnBusySemaphore) return;
    TDataType dout = null;
    TDataType din = new TDataType();
    TAccess access = new TAccess(TAccess.CA_ALARM + TAccess.CA_READ);
    TAlarmWatchEntry awe = null;
    TAlarmDefinition adef = null;
    TAlarm alm = null;
    TDevice dev = null;
    int cc = 0, cnt = 0, siz = 0, sev = 0, devNr, code = 0, asys = 0;
    float fvals[] = null;
    int ivals[] = null;
    byte[] abytes = null;
    synchronized (gAlarmWatchList)
    {
      int awlength = gAlarmWatchList.size();
      for (int i=0; i<awlength; i++)
      {
        awe = (TAlarmWatchEntry)gAlarmWatchList.get(i);
        int chk = awe.checkNormal();
        if (chk != 0)
        {
          if (chk == 1 )
          ivals = new int[awe.getSiz()];
          dout = new TDataType(ivals);
          gEqmFactory.gEqpFcnBusySemaphore = true;
          cc = callProperty(awe.getPrp(),awe.getDev(),dout,din,access);
          gEqmFactory.gEqpFcnBusySemaphore = false;        
        }
        else
        {
          fvals = new float[awe.getSiz()];
          dout = new TDataType(fvals);
          gEqmFactory.gEqpFcnBusySemaphore = true;
          cc = callProperty(awe.getPrp(),awe.getDev(),dout,din,access);
          gEqmFactory.gEqpFcnBusySemaphore = false;
        }
        if (cc != 0) continue;
        boolean hasAlarm = false;
        devNr = deviceList.getDeviceNumber(awe.getDev());
        cnt = 0; siz = awe.getSiz();
        for (int j=0; j<siz; j++)
        {
          if (awe.getAtyp() == TArrayType.AT_CHANNEL) devNr = j;
          if (awe.getAtyp() != TArrayType.AT_WAVEFORM && siz == deviceList.getNumberOfDevices()) 
            devNr = j; /* assume array of devices */
          dev = deviceList.getDevice(devNr);
          hasAlarm = false;
          if (ivals != null)
          {    
            int nval = awe.getNormal();
            if ((chk == 1 && nval != (awe.getMask() & ivals[j])) ||
                (chk == -1 && nval == (awe.getMask() & ivals[j])) )
            {
              code = awe.getCode();
              if (code == 0) code = TErrorList.invalid_data;
              sev = awe.getSev();
              asys = awe.getAsys();
              hasAlarm = true;
            }
            abytes = Swap.LongToBytes(ivals[j]);
          }
          else if (fvals != null)
          {
            if (fvals[j] > awe.getHi()) 
            {
              code = awe.getCode();
              if (code == 0) code = awe.getCodeHigh();
              if (code == 0) code = TErrorList.value_too_high;
              sev = awe.getSev();
              asys = awe.getAsys();
              hasAlarm = true;
            }
            else if (fvals[j] > awe.getHiwarn()) 
            {
              code = awe.getCode();
              if (code == 0) code = awe.getCodeHigh();
              if (code == 0) code = TErrorList.warn_too_high;          
              sev = awe.getSev()-2;
              asys = awe.getAsys();
              hasAlarm = true;
            }
            else if (fvals[j] < awe.getLo()) 
            {
              code = awe.getCode();
              if (code == 0) code = awe.getCodeLow();
              if (code == 0) code = TErrorList.value_too_low;          
              sev = awe.getSev();
              asys = awe.getAsys();
              hasAlarm = true;
            }
            else if (fvals[j] < awe.getLowarn()) 
            {
              code = awe.getCode();
              if (code == 0) code = awe.getCodeLow();
              if (code == 0) code = TErrorList.warn_too_low;          
              sev = awe.getSev()-2;
              asys = awe.getAsys();
              hasAlarm = true;
            }
            abytes = Swap.Float(fvals[j]);
          }
          if (!hasAlarm)
          { // run through clearAlarm()
            dev.clearAlarm(TErrorList.invalid_data);
            dev.clearAlarm(TErrorList.value_too_high);
            dev.clearAlarm(TErrorList.value_too_low);
            dev.clearAlarm(TErrorList.warn_too_high);
            dev.clearAlarm(TErrorList.warn_too_low);
            if (awe.getCode() > 0) dev.clearAlarm(awe.getCode()); 
            if (awe.getCodeHigh() > 0) dev.clearAlarm(awe.getCodeHigh()); 
            if (awe.getCodeLow() > 0) dev.clearAlarm(awe.getCodeLow()); 
            continue; 
          }
          if (code == 0) continue;
          cnt++;
          awe.setCnt(cnt);
          if (dev != null)
          {
            alm = dev.setAlarm(code,abytes);
            if (alm != null)
            {
              adef = alm.getAlmDef();
              if (sev != 0) adef.setAlarmSeverity(sev);
              if (asys != 0) adef.setAlarmSystem(asys);
            }
          }
        }
        /* only clear the watch table counter when a complete pass has no alarms */
        if (cnt == 0) awe.setCnt(0);
      }
    }
  }
  private void updateAlarmDynSet()
  {
    long ts = System.currentTimeMillis();
    int tss = (int)(ts/1000);
    int tsu = (int)(ts%1000)*1000;
    if (deviceList.getNumberOfAlarms() <= TAlarm.getAlmCollapseWindow())
    { // no need to collapse
      if (almDynSet.starttime == 0) return;
      if ((almDynSet.descriptor & TAlarmDescriptor.TERMINATE) == TAlarmDescriptor.TERMINATE) 
        return;
      almDynSet.descriptor = TAlarmDescriptor.TERMINATE;
      almDynSet.timestamp = tss;
      almDynSet.timestampUSec = tsu;
      return;
    }
    if (almDynSet.descriptor == 0 || 
        (almDynSet.descriptor & TAlarmDescriptor.TERMINATE) == TAlarmDescriptor.TERMINATE)
    { // either never set or last set to terminated -> this is now NEW !
      almDynSet.descriptor = TAlarmDescriptor.NEW;
      almDynSet.timestamp = tss;
      almDynSet.timestampUSec = tsu;
      almDynSet.starttime = almDynSet.timestamp;
      almDynSet.starttimeUSec = almDynSet.timestampUSec;
      return;
    }
    if (almDynSet.timestamp +  TAlarm.getAlmHeartbeatWindow() < tss) 
    { //  set the heartbeat
      almDynSet.descriptor = TAlarmDescriptor.HEARTBEAT;
      almDynSet.timestamp = tss;
      almDynSet.timestampUSec = tsu;
      return;
    }    
  }
  protected void update()
  {
    if (isIdle) return;
    accessLock.checkAccessLocks();
    checkAlarmsInWatchTable();
    deviceList.refreshAlarms();
    updateAlarmDynSet();
    historyCycle();
  }
  /**
   * Registers the equipment module's stock properties
   */
  private int prpQueryCall(String devName, TDataType dout, TDataType din, TAccess devAccess)
  { // use in calls for "PROPS" or "PROPERTIES"
    int cc = 0;
    TDevice dev = getDeviceList().getDevice(devName);
    LinkedList<String> lst = null;
    boolean checkWildCard = (devName.compareTo("*") != 0 && devName.indexOf("*") != -1);
    
    if (dev != null && (lst=dev.getPropertyList()) != null)
    {
      String[] slst = new String[lst.size()];
      lst.toArray(slst);
      if (checkWildCard) slst = StringToName.matchStringsInArray(slst,devName);
      if (dout.getFormat() == TFormat.CF_USTRING)
      {
        USTRING[] ustr = new USTRING[slst.length];
        char[] cha;
        StringBuffer sb = new StringBuffer(80);
        TExportProperty prp;
        for (int i = 0; i<slst.length; i++)
        {
          ustr[i] = new USTRING();
          if ((prp = propertyList.getFirstProperty(slst[i])) == null) continue;
          cha = prp.getName().toCharArray();
          sb.insert(0,cha);
          for (int c=cha.length; c<32; c++) sb.insert(c,(char)0);
          cha = prp.getDescription().getText().toCharArray();
          sb.insert(32,cha);
          ustr[i].setValues(prp.getOutputSize(),0,0,prp.getOutputFormat(),new String(sb));
        }
        cc = dout.putData(ustr);
      }
      else
      {
        cc = dout.putData(StringToName.stringArrayToName64(slst));
      }
    }
    else
    {
      String[] slst = new String[propertyList.countUniqueProperties()];
      propertyList.getPropertyNames().toArray(slst);
      if (checkWildCard) slst = StringToName.matchStringsInArray(slst,devName);
      if (dout.getFormat() == TFormat.CF_USTRING)
      {
        USTRING[] ustr = new USTRING[slst.length];
        char[] cha;
        StringBuffer sb = new StringBuffer(80);
        TExportProperty prp;
        for (int i = 0; i<slst.length; i++)
        {
          ustr[i] = new USTRING();
          if ((prp = propertyList.getFirstProperty(slst[i])) == null) continue;
          cha = prp.getName().toCharArray();
          sb.insert(0,cha);
          for (int c=cha.length; c<32; c++) sb.insert(c,(char)0);
          cha = prp.getDescription().getText().toCharArray();
          sb.insert(32,cha);
          ustr[i].setValues(prp.getOutputSize(),0,0,prp.getOutputFormat(),new String(sb));
        }
        cc = dout.putData(ustr);
      }
      else
      {
        cc = dout.putData(StringToName.stringArrayToName64(slst));
      }
    }
    if (cc == 0)
    {
      if (propertyList.isDeviceOriented()) 
        cc = TErrorList.has_query_function | TErrorList.CE_SENDDATA;
    }
    return cc;      
  }
  private boolean stockPropertiesRegistered = false;
  void registerStockProperties()
  {
    if (stockPropertiesRegistered) return;
    final TEquipmentModule thisEqm = this;
    if (gEqmFactory == null) gEqmFactory = TEquipmentModuleFactory.getInstance();
    stockList = new TPropertyList();
    // TODO Separate handlers for overloaded properties
    TPropertyHandler propertiesStructHandler = new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout != null)
        {
          switch (dout.dFormat)
          { // TODO: tidy this copy-and-past job up !
            case TFormat.CF_STRUCT:
              if (dout.getTag().compareTo("PRPQSr4") == 0 || (dout.getArrayLength() % TPropertyQuery.sizeInBytes) == 0)
              {
                if (din != null)
                { // looking for a specific property ...
                  LinkedList<TExportProperty> propInstances;
                  String propertyName = din.toString();
                  propertyName = propertyName.replace('\n', (char) 0).trim();
                  propInstances = propertyList.getEqualProperties(propertyName);
                  if ((propInstances == null) || (propInstances.size() == 0))
                      return TErrorList.not_exported;
                  // fill in the targeted property info
                  THistoryRecord lhr = null;
                  TPropertyQuery[] prpq = new TPropertyQuery[propInstances.size()];
                  byte[] tba = new byte[prpq.length * TPropertyQuery.sizeInBytes];
                  Iterator<TExportProperty> it = propInstances.iterator();
                  for (int i = 0; (it.hasNext()) && (i < prpq.length); i++)
                  {
                    prpq[i] = new TPropertyQuery((TExportProperty) it.next(), propInstances.size() - 1);
                    if ((lhr=getLocalHistoryRecord(prpq[i].prpName,"#0")) != null)
                    {
                      prpq[i].prpHistoryDepthLong = (short)lhr.getHspec().getDepthLong();
                      prpq[i].prpHistoryDepthShort = (short)lhr.getHspec().getDepthShort();
                    }
                    System.arraycopy(prpq[i].toByteArray(), 0, tba, i * TPropertyQuery.sizeInBytes,
                        TPropertyQuery.sizeInBytes);
                  }
                  //return (short)dout.putData(xpq[0].toByteArray(),"XPQS");
                  return dout.putData(tba, "PRPQSr4");
                }
                else
                { // want the whole list ...
                  String[] slst = new String[propertyList.countUniqueProperties()];
                  propertyList.getPropertyNames().toArray(slst);
                  TPropertyQuery[] prpq = new TPropertyQuery[slst.length];
                  if (slst.length == 0) break;
                  byte[] tba = new byte[slst.length * TPropertyQuery.sizeInBytes];
                  TExportProperty prp;
                  for (int i=0; i<slst.length; i++)
                  {
                    prp = propertyList.getFirstProperty(slst[i]);
                    prpq[i] = new TPropertyQuery(prp,1);
                    System.arraycopy(prpq[i].toByteArray(), 0, tba, i * TPropertyQuery.sizeInBytes,
                        TPropertyQuery.sizeInBytes);
                  }
                  return dout.putData(tba, "PRPQSr4");
                }                
              }
              else if (dout.getTag().compareTo("XPQS") == 0 || (dout.getArrayLength() % XPropertyQuery.sizeInBytes) == 0)
              {
                if (din != null)
                { // looking for a specific property ...
                  LinkedList<TExportProperty> propInstances;
                  String propertyName = din.toString();
                  propertyName = propertyName.replace('\n', (char) 0).trim();
                  propInstances = propertyList.getEqualProperties(propertyName);
                  if ((propInstances == null) || (propInstances.size() == 0))
                      return TErrorList.not_exported;
                  // fill in the targeted property info
                  THistoryRecord lhr = null;
                  XPropertyQuery[] xpq = new XPropertyQuery[propInstances.size()];
                  byte[] tba = new byte[xpq.length * XPropertyQuery.sizeInBytes];
                  Iterator<TExportProperty> it = propInstances.iterator();
                  for (int i = 0; (it.hasNext()) && (i < xpq.length); i++)
                  {
                    xpq[i] = new XPropertyQuery((TExportProperty) it.next(), propInstances.size() - 1);
                    if ((lhr=getLocalHistoryRecord(xpq[i].prpName,"#0")) != null)
                    {
                      xpq[i].prpHistoryDepthLong = (short)lhr.getHspec().getDepthLong();
                      xpq[i].prpHistoryDepthShort = (short)lhr.getHspec().getDepthShort();
                    }
                    System.arraycopy(xpq[i].toByteArray(), 0, tba, i * XPropertyQuery.sizeInBytes,
                        XPropertyQuery.sizeInBytes);
                  }
                  //return (short)dout.putData(xpq[0].toByteArray(),"XPQS");
                  return dout.putData(tba, "XPQS");
                }
                else
                { // want the whole list ...
                  String[] slst = new String[propertyList.countUniqueProperties()];
                  propertyList.getPropertyNames().toArray(slst);
                  XPropertyQuery[] xpq = new XPropertyQuery[slst.length];
                  if (slst.length == 0) break;
                  byte[] tba = new byte[slst.length * XPropertyQuery.sizeInBytes];
                  TExportProperty prp;
                  for (int i=0; i<slst.length; i++)
                  {
                    prp = propertyList.getFirstProperty(slst[i]);
                    xpq[i] = new XPropertyQuery(prp,1);
                    System.arraycopy(xpq[i].toByteArray(), 0, tba, i * XPropertyQuery.sizeInBytes,
                        XPropertyQuery.sizeInBytes);
                  }
                  return dout.putData(tba, "XPQS");
                }
              }
              else if (dout.getTag().compareTo("PQS") == 0 || (dout.getArrayLength() % PropertyQuery.sizeInBytes) == 0)
              { // legacy request ...
                if (din != null)
                { // looking for a specific property ...
                  LinkedList<TExportProperty> propInstances;
                  String propertyName = din.toString();
                  propertyName = propertyName.replace('\n', (char) 0).trim();
                  propInstances = propertyList.getEqualProperties(propertyName);
                  if ((propInstances == null) || (propInstances.size() == 0))
                      return TErrorList.not_exported;
                  // fill in the targeted property info
                  PropertyQuery[] pqs = new PropertyQuery[1];
                  TExportProperty prp = propertyList.getFirstProperty(propertyName);
                  byte[] tba = new byte[PropertyQuery.sizeInBytes];
                  pqs[0] = new PropertyQuery();
                  pqs[0].name = propertyName;
                  pqs[0].prpDesc = prp.getDescription().getText();
                  pqs[0].prpAccess = (byte)prp.getAccessMode();
                  pqs[0].prpFormat = (byte)(prp.getOutputFormat()%512);
                  pqs[0].prpSize = (short)prp.getOutputSize();
                  System.arraycopy(pqs[0].toByteArray(), 0, tba, 0, PropertyQuery.sizeInBytes);
                  return dout.putData(tba, "PQS");
                }
                else
                { // want the whole list ...
                  String[] slst = new String[propertyList.countUniqueProperties()];
                  propertyList.getPropertyNames().toArray(slst);
                  if (slst.length == 0) break;
                  PropertyQuery[] pqs = new PropertyQuery[slst.length];
                  byte[] tba = new byte[slst.length * PropertyQuery.sizeInBytes];
                  TExportProperty prp;
                  for (int i=0; i<slst.length; i++)
                  {
                    pqs[i] = new PropertyQuery();
                    pqs[i].name = slst[i];
                    prp = propertyList.getFirstProperty(slst[i]);
                    pqs[i].prpDesc = prp.getDescription().getText();
                    pqs[i].prpAccess = (byte)prp.getAccessMode();
                    pqs[i].prpFormat = (byte)(prp.getOutputFormat()%512);
                    pqs[i].prpSize = (short)prp.getOutputSize();
                    System.arraycopy(pqs[i].toByteArray(), 0, tba, i * PropertyQuery.sizeInBytes,
                        PropertyQuery.sizeInBytes);
                  }
                  return dout.putData(tba, "PQS");                
                }
              }
          }
        }
        /**/
        return TErrorList.illegal_format;
      }
    };
    
    TPropertyHandler alarmsStructHandler = new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout == null) return TErrorList.dimension_error;
        if (dout.dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
        int[] amrb = new int[4];
        if (din != null)
        {
          switch (din.dFormat)
          {
            case TFormat.CF_STRUCT:
              if (din.dArrayLength != 16) return TErrorList.illegal_format;
              byte[] bdata = din.getDataBuffer();
              byte[] ival = new byte[4];
              for (int i=0; i<3; i++)
              {
                System.arraycopy(bdata,i*4,ival,0,4);
                amrb[i] = Swap.Long(ival);
              }
              break;
            case TFormat.CF_LONG:
              din.getData(amrb);
              break;
            default:
              return TErrorList.illegal_format;
          }
        }
        else
        {
          amrb[0] = 0; // start time
          amrb[1] = (int)(System.currentTimeMillis()/1000); // stop time
          amrb[2] = 0; // minimum severity
          amrb[3] = 0; // as yet unused ...
        }
        boolean collapse = false;
        int dstart, dstop;
        int astart = amrb[0];
        int astop = amrb[1];
        int asev = amrb[2];
        if (devName == null || devName.length() == 0 || devName.startsWith("#") || devName.contains("*"))
        { // currently any of these is a signal that the caller want's ALL alarms
          int nalms = deviceList.getNumberOfAlarms();
          if (nalms > TAlarm.getAlmCollapseWindow()) collapse = true;
          dstart = 0; dstop = deviceList.getNumberOfDevices();
        }
        else
        {
          dstart = deviceList.getDeviceNumber(devName);
          dstop = dstart + 1;
        }
        TAlarm[] alms = null;
        TAlarmMessage[] ams = null;
        int buflength = 0;
        byte[] tba = null;
        int code = TErrorList.max_alarms_exceeded, na = 0;
        if (collapse)
        { // collapse to a single message
          na = buflength = 1;
          ams = new TAlarmMessage[buflength];
          String atag = "more than "+TAlarm.getAlmCollapseWindow()+" total alarms";
          short[] tna = new short[1];
          tna[0] = (short)deviceList.getNumberOfAlarms();
          TDataType dt = new TDataType(tna);
          if (deviceList.getNumberOfDevices() > TAlarm.getAlmCollapseWindow())
          { // all devices with same alarm code ?
            TDevice[] tdevs = deviceList.getDeviceList();
            TAlarm[] as;
            int n = 0;
            for (TDevice td : tdevs)
            { // scan thru the device list
              if ((as=td.getAlarmList()) == null || as[0] == null) continue;
              for (int i=0; i<as.length; i++)
                if (isCasInClientList()) as[i].allowRemoval();
              if (code == TErrorList.max_alarms_exceeded) code = as[0].getCode(); 
              if (code != as[0].getCode()) break;
              n++;
            }
            if (n > TAlarm.getAlmCollapseWindow())
            { // at least this many with the same code
              TAlarmDefinition ad = getAlarmDefinition(code);
              atag = ""+n+" alarms: "+ ad != null ? ad.getAlarmTag() : "<no defined alarm tag>";
            }
          }
          ams[0] = new TAlarmMessage(thisEqm.getExportName(),thisEqm.getExportName(),atag,code,
                          deviceList.getHighestAlarmSeverity(null,0),almDynSet,dt);
          tba = new byte[buflength * TAlarmMessage.sizeInBytes];
        }
        else
        {
          int ts = 0;
          buflength = deviceList.getNumberOfAlarms();
          ams = new TAlarmMessage[buflength];
          tba = new byte[buflength * TAlarmMessage.sizeInBytes];
          TDevice[] dlst = deviceList.getDeviceList();
          for (int i=dstart; i<dstop; i++)
          {
            alms = dlst[i].getAlarmList();
            if ((alms == null) || (alms.length == 0)) continue;
            for (int n=0; n<alms.length && na<buflength; n++)
            {
              if ((ts=alms[n].getTimeStamp()) < astart || ts > astop) continue;
              if (alms[n].getSeverity() < asev) continue;
              ams[na] = new TAlarmMessage(thisEqm,dlst[i].getName(),alms[n]);
              System.arraycopy(ams[na].toByteArray(), 0, tba, (na)*TAlarmMessage.sizeInBytes, TAlarmMessage.sizeInBytes);
              // CAS is configured and has now read the alarm => can remove it !
              if (casReadRequired && isCasInClientList()) alms[n].allowRemoval();
              na++;
            }
          }
        }
        if (dout.getTag().compareTo("AMSr4") == 0)
        { // modern
          for (int i=0; i<buflength && i<na; i++)
          {
            System.arraycopy(ams[i].toByteArray(), 0, tba, i*TAlarmMessage.sizeInBytes, TAlarmMessage.sizeInBytes);
          }
          if (na < dout.getArrayLength()) dout.setArrayLength(na);
          return dout.putData(tba, "AMSr4");
        }
        else 
        {  // is legacy support necessary ?
          
        }
        return TErrorList.invalid_structure_tag;
      }
    };
    TPropertyHandler alarmDefsStructHandler = new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout == null) return TErrorList.dimension_error;
        if (dout.dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
        int almdefsize = alarmDefinitionList.size();
        int adssize = TStructRegistry.getSizeInBytes("ADSr4");
        byte[] tba = new byte[almdefsize * adssize];
        TAlarmDefinition[] adef = (TAlarmDefinition[])alarmDefinitionList.values().toArray(new TAlarmDefinition[0]);

        if (dout.getTag().compareTo("ADSr4") == 0)
        { // modern
          for (int i=0; i<almdefsize; i++)
          {
            System.arraycopy(adef[i].toByteArray(), 0, tba, i*adssize, adssize);
          }
          if (almdefsize < dout.getArrayLength()) dout.setArrayLength(almdefsize);
          return dout.putData(tba, "ADSr4");
        }
        else 
        {  // is legacy support necessary ?
          
        }
        return TErrorList.invalid_structure_tag;
      }
    };
    TPropertyHandler alarmWatchStructHandler = new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout == null) return TErrorList.dimension_error;
        if (dout.dFormat != TFormat.CF_STRUCT) return TErrorList.illegal_format;
        synchronized (gAlarmWatchList)
        {
          int almwtblsize = gAlarmWatchList.size();
          int awssize = TStructRegistry.getSizeInBytes("AWSr4");
          byte[] tba = new byte[almwtblsize * awssize];
          TAlarmWatchEntry[] aws = (TAlarmWatchEntry[])gAlarmWatchList.toArray(new TAlarmWatchEntry[0]);  
          if (devAccess.isWrite())
          {
            if (din.getFormat() != TFormat.CF_STRUCT) return TErrorList.illegal_format;
            if (din.getTag().compareTo("AWSr4") == 0)
            {
              TAlarmWatchEntry[] awe = new TAlarmWatchEntry[1];
              awe[0] = new TAlarmWatchEntry();
              din.getData(awe);
              awe[0].setEquipmentModule(getLocalName());
              boolean isNew = false;
              for (int i=0; i<aws.length; i++)
              {
                if (aws[i].equals(awe[0])) 
                {
                  aws[i].adjustFrom(awe[0]);
                  isNew = true;
                  break;
                }
              }
              if (isNew)
              { // wasn't in list
                TExportProperty p = propertyList.getFirstProperty(awe[0].getPrp());
                if (p == null) return TErrorList.illegal_property;
                if (awe[0].getSiz() <= 0 || awe[0].getSiz() > p.getOutputSize())
                {
                  awe[0].setSiz(p.getOutputSize());
                }
                awe[0].setFmt(p.getOutputFormat());
                gAlarmWatchList.add(awe[0]);
              }
            }
            else
            {
              return TErrorList.invalid_structure_tag;            
            }
          }
          if (dout.getTag().compareTo("AWSr4") == 0)
          { // modern
            for (int i=0; i<almwtblsize; i++)
            {
              System.arraycopy(aws[i].toByteArray(), 0, tba, i*awssize, awssize);
            }
            if (almwtblsize < dout.getArrayLength()) dout.setArrayLength(almwtblsize);
            return dout.putData(tba, "AWSr4");
          }
        }
        return TErrorList.invalid_structure_tag;
      }
    };
    stockList.addProperty(TStockProperties.NALARMS, new TPropertyHandler()
    {
      int nalmdefs = alarmDefinitionList.size();
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int devnr = -1;
        
        if (!devName.contains("*") && !devName.startsWith("#"))
        {
          devnr = deviceList.getDeviceNumber(devName);
        }
        if (casName != null  && !hasCasAttached) isCasInClientList(); 
        int[] nalms = new int[6];
        if (devnr < 0)
        { // all of them
          nalms[0] = deviceList.getNumberOfAlarms();
          nalms[1] = deviceList.getMostRecentAlarmTimestamp(nalms,3);
          nalms[2] = deviceList.getHighestAlarmSeverity(nalms,4);
          nalms[5] = nalmdefs;
          if (nalms[0] > TAlarm.getAlmCollapseWindow())
          {
            nalms[4] = nalms[3] = nalms[0] = 1;
          }
        }
        else
        { // just the device given ...
          int mrts = 0, ts, hsv = 0, sv, nts = 0, nsv = 0;
          TDevice[] dlst = deviceList.getDeviceList();
          if (devnr >= dlst.length) return TErrorList.device_not_connected;
          TAlarm[] alms = dlst[devnr].getAlarmList();
          nalms[0] = alms.length;
          for (int i=0; i<alms.length; i++)
          {
            if ((ts=alms[i].getTimeStamp()) > mrts) 
            {
              mrts = ts;
              nts = 0;
            }
            if (ts == mrts) nts++;
            if ((sv=alms[i].getSeverity()) > hsv) 
            {
              hsv = sv;
              nsv = 0;
            }
            if (sv == hsv) nsv++;
          }
          nalms[1] = mrts;
          nalms[2] = hsv;
          nalms[3] = nts;
          nalms[4] = nsv;
          nalms[5] = nalmdefs;
        }
        if (dout.dArrayLength > 5) nalms[5] = getAlarmDefinitionList().size();
        return dout.putData(nalms);
      }
    });
    stockList.addProperty(TStockProperties.NALARMDEFS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData((short)alarmDefinitionList.size());
      }
    });
    stockList.addProperty(TStockProperties.NALMWATCH, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData((short)gAlarmWatchList.size());
      }
    });
    stockList.addProperty(TStockProperties.ALARMS, alarmsStructHandler);
    stockList.addProperty(TStockProperties.ALARMSEXT, alarmsStructHandler);
    stockList.addProperty(TStockProperties.ALARMSEXT_STRUCTIN, alarmsStructHandler);
    stockList.addProperty(TStockProperties.ALARMDEFS, alarmDefsStructHandler);
    stockList.addProperty(TStockProperties.ALMWATCHTBL, alarmWatchStructHandler);
    stockList.addProperty(TStockProperties.NHISTORIES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData((short)gLclHstList.size());
      }
    });
    stockList.addProperty(TStockProperties.HISTORIES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        THistoryRecord[] hsts =  (THistoryRecord[])gLclHstList.toArray(new THistoryRecord[0]);
        NAME64[] nams = new NAME64[hsts.length];
        for (int i=0; i<nams.length; i++) nams[i] = new NAME64(hsts[i].getPrp());
        return dout.putData(nams);
      }
    });
    stockList.addProperty(TStockProperties.ADDHISTORY, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        //TODO: finish this  
        if (devAccess.isWrite())
        {
          if (din.getFormat() != TFormat.CF_STRUCT) return TErrorList.illegal_format;
          if (din.getTag().compareTo("HRSr4") == 0)
          {
            THistoryRecordStruct[] hrs = new THistoryRecordStruct[1];
            hrs[0] = new THistoryRecordStruct();
            din.getData(hrs);
            THistoryRecord hr = getLocalHistoryRecord(hrs[0].getProperty(),hrs[0].getDevice());
            if (hr == null)             
            { // wasn't in list
              //TODO: method to determine next record index
              // call addLocalHistoryRecord()
            }
            else
            {
              THistorySpecification hsp = hr.getHspec();
              hsp.setAbsoluteTolerance(hrs[0].getAbsoluteTolerance());
              hsp.setPercentTolerance(hrs[0].getPercentTolerance());
              hsp.setArchiveRate(hrs[0].getArchiveRate());
              hsp.setPollngRate(hrs[0].getPollingRate());
              hsp.setDepthLong(hrs[0].getDepthLong());
              hsp.setDepthShort(hrs[0].getDepthShort());
              hsp.setHeartbeat(hrs[0].getHeartbeat());
            }
            return 0;
          }
          else
          {
            return TErrorList.invalid_structure_tag;            
          }
        }
        return TErrorList.illegal_read_write;
      }
             
    });
    stockList.addProperty(TStockProperties.PROPERTIES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return prpQueryCall(devName,dout,din,devAccess);
      }
    });
    stockList.addProperty(TStockProperties.PROPERTIES_USTRING, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return prpQueryCall(devName,dout,din,devAccess);
      }
    });
    stockList.addProperty(TStockProperties.PROPERTIES_STRUCT, propertiesStructHandler);
    stockList.addProperty(TStockProperties.PROPS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return prpQueryCall(devName,dout,din,devAccess);
      }
    });
    stockList.addProperty(TStockProperties.PROPS_USTRING, new TPropertyHandler()
        {
          public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
          {
            return prpQueryCall(devName,dout,din,devAccess);
          }
        });
    stockList.addProperty(TStockProperties.PROPS_STRUCT, propertiesStructHandler);
    stockList.addProperty(TStockProperties.NPROPERTIES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData(propertyList.countUniqueProperties());
      }
    });
    // Same as NPROPERTIES
    stockList.addProperty(TStockProperties.NPROPS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData(propertyList.countUniqueProperties());
      }
    });
    stockList.addProperty(TStockProperties.NDEVICES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        //return dout.putData(deviceList.size());
        return dout.putData(deviceList.getNumberOfDevices());
      }
    });
    stockList.addProperty(TStockProperties.DEVICES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        String[] namlst = deviceList.getDeviceNameList();
        if (devName.compareTo("*") != 0 && devName.indexOf("*") != -1)
        { // query for a subset of device names ...
          namlst = StringToName.matchStringsInArray(namlst,devName);
        }
        //return dout.putData(StringToName.stringSetToName16(deviceList));
        int cc = dout.putData(StringToName.stringArrayToName64(namlst));
        if (cc == 0)
        {
          if (getDeviceList().isPropertyOriented()) 
            cc = TErrorList.has_query_function | TErrorList.CE_SENDDATA;
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.DEVDESCRIPTION, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        TDevice tdv = deviceList.getDevice(devName);
        if (tdv == null) return TErrorList.illegal_device;
        String dsc = tdv.getDescription();
        return dout.putData(dsc == null ? "" : dsc);
      }
    });    
    stockList.addProperty(TStockProperties.SRVIDLE, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        int[] idleState = new int[1];
        if (din != null && devAccess.isWrite())
        {
          if ((cc=din.getData(idleState)) != 0) return cc;
          isIdle = idleState[0] == 0 ? false : true;
        }
        if (dout != null && dout.dArrayLength > 0)
        {
          idleState[0] = isIdle ? -1 : 0;
          cc = dout.putData(idleState);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.SRVINIT, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        if (!devAccess.isWrite()) return TErrorList.illegal_read_write;
        {
          if (eqpIni == null) return TErrorList.not_implemented;
          cc = eqpIni.initialize();
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.USERS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        NAME16[] n16 = getRegisteredUsers();
        if (n16 == null) 
        {
          n16 = new NAME16[1];
          n16[0] = new NAME16("");
        }
        return dout.putData(n16);
      }
    });
    stockList.addProperty(TStockProperties.NUSERS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData((short)gRegisteredUsersList.size());
      }
    });
    stockList.addProperty(TStockProperties.ADDUSER, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc;
        if (din != null && devAccess.isWrite())
        {
          NAME16[] usr = new NAME16[1];
          if ((cc=din.getData(usr)) != 0) return cc;
          return addRegisteredUser(usr.toString());
        }
        return TErrorList.illegal_read_write;
      }
    });
    stockList.addProperty(TStockProperties.DELUSER, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc;
        if (din != null && devAccess.isWrite())
        {
          NAME16[] usr = new NAME16[1];
          if ((cc=din.getData(usr)) != 0) return cc;
          return removeRegisteredUser(usr.toString());
        }
        return TErrorList.illegal_read_write;
      }
    });
    stockList.addProperty(TStockProperties.IPNETS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData(getRegisteredNets());
      }
    });
    stockList.addProperty(TStockProperties.NIPNETS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData((short)gRegisteredNetsList.size());
      }
    });
    stockList.addProperty(TStockProperties.ACCESSLOCK, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        if (din != null && devAccess.isWrite())
        {
          short[] svals = new short[2];
          if ((cc=din.getData(svals)) != 0) return cc;
          if ((cc=accessLock.setAccessLock(getCaller(),svals[0],svals[1])) != 0) return cc;
        }
        if (dout != null)
        {
          String[] s = accessLock.getAccessLockAsString();
          NAME32[] n32 = new NAME32[3];
          n32[0] = new NAME32(s[0]);
          n32[1] = new NAME32(s[1]);
          n32[2] = new NAME32(s[2]);
          cc = dout.putData(n32);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.SRVSELFTEST, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout.getFormat() != TFormat.CF_TEXT) return TErrorList.illegal_format;
        String fileName = gEqmFactory.getInitializer().getFecHome() + "/" + getLocalName() + "-selftest.csv";
        if (!TFecLog.fileExists(fileName)) return TErrorList.no_such_file;
        int nlines = dout.getArrayLength()/80;
        return (short) dout.putData(TFecLog.getLines(fileName,nlines));
      }
    });
    stockPropertiesRegistered = true;
    /* TEMPLATE */
    /*
     * stockList.setPropertyHandler("?", new TPropertyHandler() { public short
     * call(String devName, TDataType dout, TDataType din, short devAccess) {
     * return (short) 0; } });
     */
  }
  public TEquipmentModule(TEquipmentFunction eqpFunction)
  {
    String name = eqpFunction.getClass().getName();
    StringTokenizer t = new StringTokenizer(name, ".$");
    String lastToken = name;
    while (t.hasMoreTokens()) lastToken = t.nextToken();
    initialize(lastToken, eqpFunction);
  }
  public TEquipmentModule(String localName, TEquipmentFunction eqpFunction)
  {
    initialize(localName, eqpFunction);
  }
  public TEquipmentModule(String localName)
  {
    initialize(localName, null);
  }
  /**
   * \brief Principal constructor
   * 
   * Use this constructor in most cases.  This will attempt to determine
   * a suitable local name for the equipment module.
   *
   */
  public TEquipmentModule()
  {
    String name = this.getClass().getName();
    StringTokenizer t = new StringTokenizer(name, ".$");
    String localName = name;
    while (t.hasMoreTokens()) localName = t.nextToken();
    initialize(localName, null);
  }
  /**
   * Assigns an equipment function.
   * 
   * @param eqpFunction
   */
  public void setEquipmentFunction(TEquipmentFunction eqpFunction)
  {
    this.eqpFcn = eqpFunction;
    this.eqpFcn.setEquipmentModule(this);
  }
  /**
   * Registers a property without a handler. This property has to be handled by
   * the equipment function.
   * 
   * @param prpName property name
   * @param dout Provides output format specification
   * @param din Provides input format specification
   * @param acc Access information
   * @param prpDesc Property description
   * @param prpId Roperty ID
   * @return
   */
  public void registerProperty(String prpName, TDataType dout, TDataType din, short acc,
      TPropertyDescription prpDesc, int prpId)
  {
    TStrings.validatePropertyName(prpName);
    if (prpDesc == null) prpDesc = new TPropertyDescription(null);
    if (prpDesc.getArrayRowLength() == 0) prpDesc.setArrayRowLength(dout.dArrayLength);
    if (dout.dArrayLength == 1) prpDesc.setArrayType(TArrayType.AT_SCALAR);
    TExportProperty prp = new TExportProperty(prpId, prpName, acc, din, dout);
    prp.setDescription(prpDesc);
    propertyList.addProperty(this, prp, null);
    if (prpName.endsWith(".NAM")) deviceList.setPropertyOriented(true);
  }
  /**
   * Registers a property and assigns a handler.
   * 
   * @param prp
   * @param handler
   */
  public void registerProperty(TExportProperty prp, TPropertyHandler handler)
  {
    propertyList.addProperty(this, prp, handler);
    String aprpnam = gEqmFactory.getAliasFromName(prp.getName());
    if (aprpnam != null)
    { // register the alias as well !
      TExportProperty aprp = new TExportProperty(prp);
      aprp.setName(aprpnam);
      propertyList.addProperty(this,aprp, handler);
    }
    if (prp.getName().endsWith(".NAM")) deviceList.setPropertyOriented(true);
 }
  public int attachPropertyHandler(String prp,TPropertyHandler handler)
  {
    String aprp = gEqmFactory.getAliasFromName(prp);
    if (aprp != null) propertyList.setHandler(aprp,handler);
    return propertyList.setHandler(prp,handler);
  }
  /**
   * Adds a collection of device names to the module's device list.
   */
  // TODO Test if sequence remains unchanged
  public void registerDevices(Collection<String[]> deviceCollection)
  {
//    deviceList.addAll(deviceNames);
    int ndevs = deviceCollection.size();
    String[] devs = (String[])deviceCollection.toArray(new String[0]);
    for (int i=0; i<ndevs; i++) 
    { 
      TStrings.validateDeviceName(devs[i]);
      deviceList.addDevice(devs[i],this);
    }
  }
  public void registerDevices(String[] deviceNames)
  {
    if (deviceNames == null) return;
    for (int i=0; i<deviceNames.length; i++) 
    {
      TStrings.validateDeviceName(deviceNames[i]);
      deviceList.addDevice(deviceNames[i],i,this);
    }
  }
  public void registerDevices(TDevice[] devices)
  {
    if (devices == null) return;
    for (int i=0; i<devices.length; i++) 
    {
      devices[i].setEqm(this);
      deviceList.addDevice(devices[i]);
    }
  }
  public void registerDevices(int numberOfDevices)
  {
    if (numberOfDevices <= 0) return;
    TDevice[] devices = new TDevice[numberOfDevices];
    for (int i=0; i<numberOfDevices; i++) 
    {
      devices[i] = new TDevice(i);
      devices[i].setEqm(this);
      deviceList.addDevice(devices[i]);
    }
  }
  /**
   * Adds a device name to the module's device list.
   * 
   * @param devName
   * @return
   */
  public int registerDeviceName(String devName)
  {
    TStrings.validateDeviceName(devName);
    deviceList.addDevice(devName,this);
    String adev = gEqmFactory.getAliasFromName(devName);
    if (adev != null)
    { // register the alias as well !
      deviceList.addDevice(adev,this);
    }
    return deviceList.getDeviceNumber(devName);
  }
  public int registerDeviceName(String devName,int devNumber)
  {
    TStrings.validateDeviceName(devName);
    deviceList.addDevice(devName,devNumber,this);
    String adev = gEqmFactory.getAliasFromName(devName);
    if (adev != null)
    { // register the alias as well !
      deviceList.addDevice(adev,devNumber,this);
    }
    return devNumber;
  }
  /**
   * Returns all property names in a linked list.
   * 
   * @return
   */
  public LinkedList<String> getPropertyNames()
  {
    LinkedList<String> rv = new LinkedList<String>();
    rv.addAll(propertyList.getPropertyNames());
    return rv;
  }
  private void dumpList(String text,LinkedList<String> lst)
  {
    System.out.println(text);
    for (Iterator<String> it = lst.iterator(); it.hasNext();)
    {
      System.out.println((String) it.next());
    }
  }
  public void dumpProperties()
  {
    dumpList("Registered Properties :",getPropertyNames());
  }
  public void dumpDevices()
  {
    dumpList("Registered Devices :",getDeviceNames());    
  }
  /**
   * Returns all device names in a linked list.
   * 
   * @return
   */
  public LinkedList<String> getDeviceNames()
  {
    return deviceList.getDeviceLinkedList();
  }
  /**
   * Returns true if this equipment module has registered a stock property of a
   * given name.
   * 
   * @param propertyName
   * @return
   */
  public boolean hasStockProperty(String propertyName)
  {
    return stockList.hasProperty(propertyName);
  }
  public boolean hasProperty(String propertyName)
  {
    return propertyList.hasProperty(propertyName);
  }
  /**
   * Calls a stock property handler
   * 
   * @param propertyName
   * @param devName
   * @param dout
   * @param din
   * @param devAccess
   * @return
   */
  public short callStockProperty(String propertyName, String devName, TDataType dout,
      TDataType din, TAccess devAccess)
  {
    return stockList.callPropertyHandler(propertyName, devName, dout, din, devAccess);
  }
  /**
   * Finds a property and calls the property handler. If no property handler is
   * installed, the equipment function is called instead.
   * 
   * @param propertyName
   * @param devName
   * @param dout
   * @param din
   * @param devAccess
   * @return Redurn code of property handler/equipment function.
   *         TErrorList.non_existent_elem if no property handler and no
   *         equipment function could be found.
   */
  public short callProperty(String propertyName, String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    try
    {
      if (propertyList.hasProperty(propertyName))
      {
        return propertyList.callPropertyHandler(propertyName, devName, dout, din, devAccess);
      }
      else if (eqpFcn != null)
      {
        if (TEquipmentModuleFactory.debugLevel > 1) 
          DbgLog.log("callProperty","call function handler for "+getLocalName()+"/"+devName+"["+propertyName+ "]");        
        return eqpFcn.call(propertyName, devName, dout, din, devAccess.toShort());
      }
      else
      {
        if (TEquipmentModuleFactory.debugLevel > 1) 
          DbgLog.log("callProperty","no handler for "+getLocalName()+"/"+devName+"["+propertyName+ "]");        
        return TErrorList.non_existent_elem;
      }
    }
    catch (Exception e)
    {
      setCompletionString(e.getMessage());
      return TErrorList.code_failure+1000;
    }
  }
  public short callMetaProperty(String metaPropertyName, String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    try
    {
      if (!gEqmFactory.metaProps.isMetaProperty(metaPropertyName)) return TErrorList.illegal_property;
      String propertyName = gEqmFactory.metaProps.getTargetProperty(metaPropertyName);
      if (!propertyList.hasProperty(propertyName)) return TErrorList.illegal_property;
      TExportProperty prp = propertyList.getFirstProperty(propertyName);
      switch (gEqmFactory.metaProps.getIndex(metaPropertyName))
      {
        case TMetaProperties.DSC_ID:
          return (short) dout.putData(prp.getDescription().getText());        
        case TMetaProperties.EGU_ID: // to satisfy DOOCS calls ...
          switch (dout.dFormat)
          {
            case TFormat.CF_TEXT:
            case TFormat.CF_NAME16:
            case TFormat.CF_NAME32:
            case TFormat.CF_NAME64:
              return (short) dout.putData(prp.getDescription().getYRange().getUnits());
            case TFormat.CF_FLOAT:
            case TFormat.CF_DOUBLE:
            case TFormat.CF_DLONG:
            case TFormat.CF_LONG:
            case TFormat.CF_SHORT:
              float[] yrng = new float[2];
              yrng[0] = prp.getDescription().getYRange().getMinValue();
              yrng[1] = prp.getDescription().getYRange().getMaxValue();
              return (short) dout.putData(yrng);
            case TFormat.CF_USTRING:
              USTRING[] u = new USTRING[1];
              u[0] = new USTRING(prp.getDescription().getYRange().getGraphType(),
                                 prp.getDescription().getYRange().getMinValue(),
                                 prp.getDescription().getYRange().getMaxValue(),
                                 0,
                                 prp.getDescription().getYRange().getUnits());
              return (short) dout.putData(u);
            default:
              return TErrorList.illegal_format;
          }
        case TMetaProperties.HST_ID:
          THistoryRecord hst = null;
          int listsize = gLclHstList.size();
          int index = 0, sample = 0;
          long start = 0;
          long stop = System.currentTimeMillis()-1000;
          float lwr = 0, upr = 0;
          Object data = null;
          int dataSize = 0, inputDataSize = 0;
          boolean isCentralArchiveRequest = false;
          
          if (TMetaProperties.isCentralArchive(metaPropertyName))
          {
            isCentralArchiveRequest = true;
          }
          boolean isSnapshot = TMetaProperties.isHistorySnapshot(metaPropertyName);
          dataSize = dout.getArrayLength();
          start = stop - dataSize * 1000; // millisecond timestamps here!
          // check the time range and/or filtering criteria the caller has given us ...
          if (din != null && (inputDataSize=din.getArrayLength()) > 0)
          {
            switch (din.getFormat())
            {
              case TFormat.CF_LONG:
                int[] ival = (int[])din.getDataObject();
                switch (inputDataSize)
                {
                  case 4: sample = ival[3];
                  case 3: index = ival[2];
                  case 2: if (ival[1] < (int)(stop/1000)) stop = ((long)ival[1])*1000;
                  case 1: start = ((long)ival[0])*1000;
                  default: break;
                }
                if (sample < 0) sample = 0;
                break;
                
              case TFormat.CF_DLONG:
                long[] lval = (long[])din.getDataObject();
                switch (inputDataSize)
                {
                  case 4: sample = (int)lval[3];
                  case 3: index = (int)lval[2];
                  case 2: if (lval[1] < stop/1000) stop = (long)(lval[1]*1000);
                  case 1: start = (long)(lval[0]*1000);
                  default: break;
                }
                if (sample < 0) sample = 0;
                break;
                
              case TFormat.CF_DOUBLE:
                double[] dval = (double[])din.getDataObject();
                switch (inputDataSize)
                {
                  case 4: sample = (int)dval[3];
                  case 3: index = (int)dval[2];
                  case 2: if (dval[1] < stop/1000) stop = (long)(dval[1]*1000);
                  case 1: start = (long)(dval[0]*1000);
                  default: break;
                }
                if (sample < 0) sample = 0;
                break;
              case TFormat.CF_FWINDOW:
                FWINDOW[] fw = (FWINDOW[])din.getDataObject();
                start = ((FWINDOW[])fw)[0].i1val * 1000;
                stop = ((FWINDOW[])fw)[0].i2val * 1000;
                lwr = ((FWINDOW[])fw)[0].f1val;
                upr = ((FWINDOW[])fw)[0].f2val;
                if (lwr < upr) sample = -1;
                break;
              default:
                return TErrorList.illegal_format;
            }
          }
          else
          { // no input given
            if (dataSize == 1) 
            {
              start = stop;
              isSnapshot = true;
            }
          } 
          if (isSnapshot)
          {
            if (start < 1) return TErrorList.argument_list_error;
            stop = start;
            sample = 0;
            index = 0;
            lwr = upr = (float)0.0;          
          }
          // what the caller wants returned ....
          boolean literalDevice = false;
          for (int i=0; i<listsize; i++)
          { // go thru the list to see if the requested device was stored literally
            hst = (THistoryRecord)gLclHstList.get(i);
            if (hst.getPrp().compareToIgnoreCase(prp.getName()) != 0) continue;
            if (hst.getDev().compareTo(devName) != 0) continue;
            literalDevice = true;
            break;
          }
          for (int i=0; i<listsize; i++)
          {
            hst = (THistoryRecord)gLclHstList.get(i);
            if (hst.getPrp().compareToIgnoreCase(prp.getName()) == 0)
            {
              if (hst.getDev().compareTo(devName) != 0)
              { // should we look for a match in the device name ?
                if (index == 0 && !literalDevice)
                  index = getDeviceList().getDeviceNumber(devName);
                else
                  continue;
              }
              if (isCentralArchiveRequest)
              {
                String rmtsrv = hst.getRemoteHistoryServer();
                if (rmtsrv == null || rmtsrv.length() == 0) return TErrorList.not_implemented;
                return (short)gEqmFactory.setRedirectionString(rmtsrv);
              }
              // snapshots always have index = 0
              if (start == stop && dataSize > 1) index = 0; 
              switch (dout.dFormat)
              {
                case TFormat.CF_BYTE:
                  data = new byte[dataSize]; break;
                case TFormat.CF_SHORT:
                  data = new short[dataSize]; break;
                case TFormat.CF_LONG:
                  data = new int[dataSize]; break;
                case TFormat.CF_DLONG:
                  data = new long[dataSize]; break;                
                case TFormat.CF_LNGINT:
                  data = new INTINT[dataSize]; break;
                case TFormat.CF_FLOAT:
                  data = new float[dataSize]; break;
                case TFormat.CF_FLTINT:
                  data = new FLTINT[dataSize]; break;
                case TFormat.CF_INTFLTINT:
                  data = new INTFLTINT[dataSize]; break;
                case TFormat.CF_DOUBLE:
                  data = new double[dataSize]; break;
                case TFormat.CF_DBLDBL:
                  data = new DBLDBL[dataSize]; break;
                case TFormat.CF_NAME16:  
                  data = new NAME16[dataSize]; break;
                case TFormat.CF_NAME16I:  
                  data = new NAME16I[dataSize]; break;
                case TFormat.CF_NAME32:  
                  data = new NAME32[dataSize]; break;
                case TFormat.CF_NAME32I:  
                  data = new NAME32I[dataSize]; break;
                case TFormat.CF_NAME64:  
                  data = new NAME64[dataSize]; break;
                case TFormat.CF_NAME64I:  
                  data = new NAME64I[dataSize]; break;
                case TFormat.CF_HISTORY:
                  data = new HISTORY[dataSize];
                  // create the first one so as to pass the tag along
                  short f = TFormat.getFormatCode(dout.getTag());
                  if (f == TFormat.CF_NULL) return TErrorList.illegal_format;
                  TDataType d = new TDataType(1,f);
                  ((HISTORY[])data)[0] = new HISTORY(d);
                  break;
              }
              int nr = hst.getDataFromLTS(index,start,stop,lwr,upr,data,dataSize,0,sample);
              if (nr < 0) 
              {
                if (nr == -TErrorList.server_redirection)
                {
                  String rmtsrv = hst.getRemoteHistoryServer();
                  if (rmtsrv == null || rmtsrv.length() == 0) return TErrorList.not_implemented;
                  return (short)gEqmFactory.setRedirectionString(rmtsrv);
                }
                return (short)-nr;
              }
              if (nr > 0) 
              {
                if (isSnapshot || nr >= dataSize - 1) 
                { // found the snapshot in or filled the buffer from LTS 
                  dout.putData(data);
                  dout.setDataTimeStamp(hst.getLastReadTimestamp());
                  return 0;
                }
                start = hst.getLastReadTimestamp() + 1;
                if (start == stop) return 0; // we're finished
              }
              nr = hst.getDataFromSTS(index,start,stop,lwr,upr,data,dataSize,nr,sample);
              if (nr < 0) return (short)-nr;
              if (dataSize > 1 && nr < dataSize) dout.dArrayLength = nr;
              dout.putData(data);
              dout.setDataTimeStamp(hst.getLastReadTimestamp());
              return 0;
            }
          }
          return TErrorList.not_implemented;
        case TMetaProperties.NAM_ID:
          // if .NAM properties are registered they are handled with the normal property handler
          String[] slst = deviceList.getDeviceNameList();
          ArrayList<String> lst = prp.getDeviceList();
          if (lst != null && lst.size() > 0) 
          {
            slst = new String[lst.size()];
            lst.toArray(slst);
          }
          return (short)dout.putData(StringToName.stringArrayToName16(slst));
        default:  
          break;
      }
      return TErrorList.illegal_property;
    }
    catch (Exception e)
    {
      setCompletionString(e.getMessage());
      return TErrorList.code_failure + 1000;
    }
  }
  // csv file EXPORTS handlers:
  class expRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    private TExportProperty expp;
    private TPropertyDescription prpDesc;
    private String targetLocalName = null;
    private String localName = null;
    private String expName = null;
    public void setLocalName(String value) { localName = value; }
    public void setExpName(String value) { expName = value; }
    public TExportProperty getExportProperty() { return expp; } 
    public TPropertyDescription getPropertyDescription() { return prpDesc; } 
    expRowHndlr(String targetName) 
    { 
      expp = new TExportProperty();
      prpDesc = new TPropertyDescription();
      if (targetLocalName == null) targetLocalName = targetName;
    }
    public int process(int index)
    {
      if (localName == null || localName.compareTo(moduleName) != 0) return 0;
      if (targetLocalName != null && targetLocalName.compareTo(moduleName) != 0) return 0;
      // register this one
      if (expName != null) setExportName(expName);
      if ((TAccess.CA_WRITE & expp.getAccessMode()) == TAccess.CA_WRITE)
      { // a writeable property
        if (expp.getInputFormat() == TFormat.CF_NULL &&
            expp.getOutputFormat() != TFormat.CF_NULL)
        { // probably a quick registration: supply this 
          expp.setInputFormat(expp.getOutputFormat());
        }
      }
      expp.setDescription(prpDesc);
      registerProperty(expp,null);
      // allocate a new one
      expp = new TExportProperty();
      prpDesc = new TPropertyDescription();
      return 0;
    }
  }
  class expXnHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expXnHndlr(expRowHndlr rowHndlr)
    {
      rHndlr = rowHndlr; 
    }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      String nm = (strValue.length() > TStrings.EXPORT_NAME_SIZE) ? strValue.substring(0,TStrings.EXPORT_NAME_SIZE) : strValue;
      rHndlr.setExpName(nm);
      return 0;
    }
  }
  class expCtxHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      setContext(strValue);
      String fctx = gEqmFactory.getFecContext();
      if (fctx == null || fctx.length() == 0)
      { // make sure the general context is set
        gEqmFactory.setFecContext(strValue);
      }
      return 0;
    }
  }
  class expMstrHndlr implements csvHandler
  {
    public int process(String strValue, int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      setMaster(strValue);
      return 0;
    }
  }
  class expSlavHndlr implements csvHandler
  {
    public int process(String strValue, int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      String mstr = getMaster();
      if (mstr != null && mstr.length() > 0) setSlaveMaster(mstr,strValue);
      return 0;
    }
  }
  class expSubHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      setSubsystem(strValue);
      return 0;
    }
  }
  class expGrpHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      setGroupName(strValue);
      return 0;
    }
  }
  class expGrpIdxHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      setGroupIndex(Integer.valueOf(strValue).intValue());
      return 0;
    }
  }
  class expLnHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    private String targetLocalName = null;
    expLnHndlr(expRowHndlr rowHndlr, String targetName)
    {
      if (targetLocalName == null) targetLocalName = targetName;
      rHndlr = rowHndlr; 
      //localName = targetName;
    }
    public int process(String strValue,int index)
    {
      rHndlr.setLocalName(strValue);
      if (strValue == null || strValue.length() == 0) return 0;
      if (targetLocalName != null && targetLocalName.compareTo(strValue) != 0) return 0;
      String nm = (strValue.length() > TStrings.EQM_NAME_SHORTSIZE) ? strValue.substring(0,TStrings.EQM_NAME_SHORTSIZE) : strValue;
      setLocalName(nm);
      return 0;
    }
  }
  class expPrpHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      rHndlr.getExportProperty().setName(strValue);
      return 0;
    }
  }
  class expPrpDescHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpDescHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      // parse the csv file text description according to convention
      rHndlr.getPropertyDescription().parseComplexString(strValue);
      return 0;
    }
  }
  class expPrpSizeHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpSizeHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      rHndlr.getExportProperty().setOutputSize(new Integer(strValue).intValue());
      return 0;
    }
  }
  class expPrpFmtHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpFmtHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      // parse the format string according to convention
      int stop = strValue.indexOf(".");
      String fmtString = stop != -1 ? strValue.substring(0,stop) : strValue;
      short fmt = TFormat.getFormatCode(fmtString);
      rHndlr.getExportProperty().setOutputFormat(fmt);
      if (stop != -1)
      {
        String fss = strValue.substring(stop+1);
        if (fmt == TFormat.CF_STRUCT)
        {
          rHndlr.getExportProperty().setOutputTag(fss);
        }
        else
        {
          rHndlr.getPropertyDescription().setArrayType(TArrayType.getType(fss));
        }
      }
      return 0;
    }
  }
  class expPrpInSizeHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpInSizeHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      rHndlr.getExportProperty().setInputSize(new Integer(strValue).intValue());
      return 0;
    }
  }
  class expPrpInFmtHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpInFmtHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      int stop = strValue.indexOf(".");
      String fmtString = stop != -1 ? strValue.substring(0,stop) : strValue;
      short fmt = TFormat.getFormatCode(fmtString);
      rHndlr.getExportProperty().setInputFormat(fmt);
      if (fmt == TFormat.CF_STRUCT)
      {
        rHndlr.getExportProperty().setInputTag(strValue.substring(stop+1));
      }
      return 0;
    }
  }
  class expPrpIdHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpIdHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      int idx = index;
      try
      {
        if (strValue != null && strValue.length() > 0) 
          idx = new Integer(strValue).intValue();
      }
      catch (NumberFormatException e)
      {
        idx = index;
      }
      rHndlr.getExportProperty().setId(idx);
      return 0;
    }
  }
  class expPrpAccessHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpAccessHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      // parse access mode according to convention
      String accessString = strValue.trim().toUpperCase();
      short ta = 0;
      if (accessString.indexOf("WRITE") >= 0) ta |=  TAccess.CA_WRITE;
      if (accessString.indexOf("READ") >= 0) ta |=  TAccess.CA_READ;
      rHndlr.getExportProperty().setAccessMode(ta);
      return 0;
    }
  }
  class expPrpRdrHndlr implements csvHandler
  {
    private expRowHndlr rHndlr;
    expPrpRdrHndlr(expRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      // TODO: parse the csv file redirection string according to convention
      rHndlr.getPropertyDescription().setRedirection(strValue);
      return 0;
    }
  }
  // TODO: call this from some strategic location 
  // (currently tested inside of SineEquipmentModule) 
  public int getExportInformationFromFile()
  {
    return getExportInformationFromFile(moduleName);
  }
  public int getExportInformationFromFile(String eqmName)
  {
    TEquipmentModuleFactory.getInstance().getFecXmlDocument();
    FecCfg cfg = gEqmFactory.getFecXmlCfg();
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<EqmCfg> el = cfg.getEqmList();
      LinkedList<String> dv = null;
      String grp, ctx, sub, exp, mstr, slav;
      String fctx = gEqmFactory.getFecContext();
      for( EqmCfg ec : el ) 
      {
        if (ec.getName().compareTo(eqmName) == 0) 
        {
          LinkedList<PropertyCfg> pl = ec.getPropertyList();
          for (PropertyCfg pc : pl)
          {
            TExportProperty expp = new TExportProperty(
              pc.getId().getValue(),pc.getName(),pc.getDescription(),
              pc.getSizeOut().getValue(),(short)pc.getDTypeOut().getValue(),pc.getDTagOut(),
              pc.getSizeIn().getValue(),(short)pc.getDTypeIn().getValue(),pc.getDTagIn());
            expp.getDescription().setArrayType((short)pc.getDArrayTypeOut());
            expp.getDescription().setXRange(
                new TPropertyEGU(pc.getXEgu(),pc.getXmin().getValue(),pc.getXmax().getValue()));
            expp.getDescription().setYRange(
                new TPropertyEGU(pc.getEgu(),pc.getMin().getValue(),pc.getMax().getValue()));
            // TODO:
            //expp.getDescription().setArrayNumRows(pc.getNumRows().getValue());
            expp.getDescription().setRedirection(pc.getRedirection());
            registerProperty(expp,null);
            
            dv = pc.getDevices();
            if (dv != null && dv.size() > 0) 
            {
              int nr = 0;
              ArrayList<String> lst = new ArrayList<String>(dv.size());
              for (String d : dv)
              {
                lst.add(nr++,d);
              }
              expp.setDeviceList(lst);
              getDeviceList().setPropertyOriented(true);
            }
          }
          grp = ec.getGroup();
          if (grp != null && grp.length() > 0)
          {
            setGroupName(grp);
            if (ec.getGroupIndex().isValid()) 
              setGroupIndex(ec.getGroupIndex().getValue());
          }
          exp = ec.getServer();
          if (exp != null && exp.length() > 0) setExportName(exp);
          ctx = ec.getContext();
          if (ctx == null || ctx.length() == 0) 
          { // not specified for the eqm : use the default 
            ctx = cfg.getContext();
            if (ctx == null || ctx.length() == 0) ctx = fctx; // hope for the best
          }
          if (fctx == null || fctx.length() == 0)
          { // make sure this is set
            gEqmFactory.setFecContext(ctx);
            fctx = ctx;
          }
          if (ctx != null && ctx.length() > 0) setContext(ctx);
          sub = ec.getSubsystem();
          if (sub == null || sub.length() == 0) 
          { // not specified for the eqm : use the default 
            sub = cfg.getSubsystem();
          }
          if (sub != null && sub.length() > 0) setSubsystem(sub);
          mstr = ec.getMaster();
          if (mstr != null && mstr.length() > 0) 
          {
            setMaster(mstr);
            slav = ec.getSlaveMaster();
            if (slav != null && slav.length() > 0) setSlaveMaster(mstr, slav);
          }
          break;
        }
      }    
      return 0;
    }
    // no fec.xml : look for exports.csv ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] expCols = new csvColumn[17];
    expRowHndlr expRows = new expRowHndlr(getLocalName());
    expCols[0] = new csvColumn("EXPORT_NAME","",new expXnHndlr(expRows));
    expCols[1] = new csvColumn("LOCAL_NAME","",new expLnHndlr(expRows,getLocalName()));
    expCols[2] = new csvColumn("PROPERTY","",new expPrpHndlr(expRows));
    expCols[3] = new csvColumn("DESCRIPTION","",new expPrpDescHndlr(expRows));
    expCols[4] = new csvColumn("PROPERTY_SIZE","",new expPrpSizeHndlr(expRows));
    expCols[5] = new csvColumn("FORMAT","",new expPrpFmtHndlr(expRows));
    expCols[6] = new csvColumn("PROPERTY_INSIZE","",new expPrpInSizeHndlr(expRows));
    expCols[7] = new csvColumn("INFORMAT","",new expPrpInFmtHndlr(expRows));
    expCols[8] = new csvColumn("PROPERTY_ID","",new expPrpIdHndlr(expRows));
    expCols[9] = new csvColumn("ACCESS","READ",new expPrpAccessHndlr(expRows));
    expCols[10] = new csvColumn("REDIRECTION","",new expPrpRdrHndlr(expRows));
    expCols[11] = new csvColumn("GROUP","",new expGrpHndlr());
    expCols[12] = new csvColumn("GROUP_INDEX","",new expGrpIdxHndlr());
    expCols[13] = new csvColumn("CONTEXT","",new expCtxHndlr());
    expCols[14] = new csvColumn("SUBSYSTEM","",new expCtxHndlr());
    expCols[15] = new csvColumn("MASTER","",new expMstrHndlr());
    expCols[16] = new csvColumn("SLAVE_MASTER","",new expSlavHndlr());
    // open it
    csv expFile = new csv(initializer.getExportsResource(eqmName));
    // read it
    int rc = expFile.readFile(expCols,expRows);
    // close it
    expFile.close();  
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get registered exports and properties from exports.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv file USERS handlers
  class usrNameHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      addRegisteredUser(strValue.toUpperCase());
      return 0;
    }
  }
  public int getRegisteredUsersFromFile()
  {
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] usrCols = new csvColumn[1];
    //usrRowHndlr usrRows = new usrRowHndlr(getLocalName());
    usrCols[0] = new csvColumn("USER_NAME","",new usrNameHndlr());
    // open it
    csv usrFile = new csv(initializer.getUsersResource());
    // read it
    int rc = usrFile.readFile(usrCols);
    // close it
    usrFile.close();
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get registered users from users.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv file IPNETS handlers
  class subnetNameHndlr implements csvHandler
  {
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      addRegisteredNet(strValue);
      return 0;
    }
  }
  public int getRegisteredNetsFromFile()
  {
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] netsCols = new csvColumn[1];
    //usrRowHndlr usrRows = new usrRowHndlr(getLocalName());
    netsCols[0] = new csvColumn("SUBNET","",new subnetNameHndlr());
    // open it
    csv netsFile = new csv(initializer.getNetsResource());
    // read it
    int rc = netsFile.readFile(netsCols);
    // close it
    netsFile.close(); 
    if (rc == 0 && gRegisteredNetsList.size() == 0) gRejectAllNets = true;
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get registered ip nets from ipnets.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv file <EQM>-alarms handlers:
  class almDefRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    private TAlarmDefinition adef;
    public String tag; // 16 char
    public int code;   // 4 bytes
    public int mask;   // 4 bytes
    public short severity; // 2 bytes;
    public short dataFormat; // 1 byte
    public int dataSize; // 1 byte
    public String text;      // 40 chars
    public String deviceText; // 38 chars
    public int system;   // 2 bytes;
    public String dataText; // 40 chars
    public String url; // 40 chars
    public TAlarmDefinition getAlarmDefinition() { return adef; } 
    almDefRowHndlr() 
    {
      tag = "";
      text = "";
      deviceText = "";
      dataText = "";
      url = "";
    }
    public int process(int index)
    {
      // make a new one
      adef = new TAlarmDefinition(tag,text,deviceText,code,severity,system,
          mask,dataFormat,dataSize,dataText,url);
      // add this to the list
      addAlarmDefinition(code,adef);
      return 0;
    }
  }
  public int getRegisteredAlarmDefinitionsFromFile()
  {
    FecCfg cfg = gEqmFactory.getFecXmlCfg();
    String eqmName = getLocalName(); 
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<EqmCfg> el = cfg.getEqmList();
      for( EqmCfg ec : el ) 
      {
        if (ec.getName().compareTo(eqmName) == 0) 
        {          
          LinkedList<AlarmDefinitionCfg> al = ec.getAlarmDefinitions();
          for (AlarmDefinitionCfg ac : al)
          {
            addAlarmDefinition(ac.getAlarmCode().getValue(),
                new TAlarmDefinition(ac.getTag(),ac.getAlarmText(),ac.getDataText(),ac.getAlarmCode().getValue(),ac.getSeverity().getValue(),ac.getAlarmSystem().getValue()));
          }
          break;
        }
      }    
      return 0;
    }
    // no fec.xml : look for alarms.csv ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] almCols = new csvColumn[11];
    almDefRowHndlr almRows = new almDefRowHndlr();
    almCols[0] = new csvColumn("ALARM_TAG","",new StringFieldHandler(almRows,"tag"));
    almCols[1] = new csvColumn("ALARM_CODE","",new IntFieldHandler(almRows,"code"));
    almCols[2] = new csvColumn("ALARM_MASK","",new IntFieldHandler(almRows,"mask"));
    almCols[3] = new csvColumn("SEVERITY","",new ShortFieldHandler(almRows,"severity"));
    almCols[4] = new csvColumn("DATA_FORMAT","",new FormatFieldHandler(almRows,"dataFormat"));
    almCols[5] = new csvColumn("DATA_ARRAYSIZE","",new IntFieldHandler(almRows,"dataSize"));
    almCols[6] = new csvColumn("ALARM_TEXT","READ",new StringFieldHandler(almRows,"text"));
    almCols[7] = new csvColumn("DEVICE_TEXT","",new StringFieldHandler(almRows,"deviceText"));
    almCols[8] = new csvColumn("DATA_TEXT","",new StringFieldHandler(almRows,"dataText"));
    almCols[9] = new csvColumn("URL","",new StringFieldHandler(almRows,"url"));
    almCols[10] = new csvColumn("ALARM_SYSTEM","",new IntFieldHandler(almRows,"system"));
    // open it
    csv almDefsFile = new csv(initializer.getAlmDefinitionResource(getLocalName()));
    // read it
    int rc = almDefsFile.readFile(almCols,almRows);
    // close it
    almDefsFile.close();  
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get alarm definitions from " + getLocalName() + "/alarms.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv file <EQM>-devices.csv and <property>-names.csv handlers:
  class devNameRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    public String deviceName;
    public String deviceNameAlt;
    public int deviceNumber;
    public int deviceNumberAlt;
    private boolean isPrpOriented = false;
    public String deviceRedirection;
    public String deviceDescription;
    private TExportProperty prp = null;
    public String prpListFile = null;
    devNameRowHndlr() 
    {
      deviceNumber = 0;
      deviceName = "#0";
    }
    devNameRowHndlr(TExportProperty property) 
    {
      deviceNumber = 0;
      deviceName = "#0";
      deviceRedirection = null;
      deviceDescription = null;
      prp = property;
      isPrpOriented = prp == null ? false : true;
    }
    public int process(int index)
    {
      TDeviceList dlst = getDeviceList();
      if (dlst == null) return TErrorList.device_not_connected;
      int ndevs = dlst.getNumberOfDevices();
      if (ndevs == 0) return TErrorList.device_not_connected;
      if (deviceNumber < 0 && deviceNumber >= ndevs) return TErrorList.illegal_equipment_number;
      if (!isPrpOriented)
      {
        TDevice tdv = dlst.getDevice(deviceNumber);
        if (tdv == null) return TErrorList.device_not_connected;
        String oldName = tdv.getName();
        if (deviceName.compareTo(oldName) != 0)
        {
          dlst.renameDevice(oldName,deviceName);
        }
        if (deviceRedirection != null && deviceRedirection.length() > 0)
        {
          tdv.setRedirection(deviceRedirection);
          deviceRedirection = null; // reset this !
        }
        if (deviceDescription != null && deviceDescription.length() > 0)
        {
          tdv.setDescription(deviceDescription);
          deviceDescription = null; // reset this !
        }
        if (prpListFile != null && prpListFile.length() > 0)
        {
          getRegisteredPropertyListFromFile(tdv,prpListFile);
          propertyList.setDeviceOriented(true);
          prpListFile = null; // reset this !
        }
      }
      else
      { 
        if (prp != null) 
        {
          int prplen = prp.getOutputSize();
          ArrayList<String> lst = prp.getDeviceList();
          if (lst == null) 
          {
            lst = new ArrayList<String>(prplen);
            prp.setDeviceList(lst);  
          }
          if (deviceName.length() == 0) 
          {
            deviceName = deviceNameAlt;
            if (deviceNumber == 0) deviceNumber = deviceNumberAlt;
          }
          if (deviceNumber < prplen) lst.add(deviceNumber,deviceName);
          dlst.setPropertyOriented(true);
        }
      }
      return 0;
    }
  }
  public int getRegisteredDevicesFromFile()
  {
    FecCfg cfg = gEqmFactory.getFecXmlCfg();
    String eqmName = getLocalName();
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<EqmCfg> el = cfg.getEqmList();
      String deviceRedirection;
      String deviceDescription;
      String prpSet;
      for( EqmCfg ec : el ) 
      {
        if (ec.getName().compareTo(eqmName) == 0) 
        {    
          LinkedList<NameCfgList> nl = ec.getNameCfgList();
          TDeviceList dlst = getDeviceList();
          if (dlst == null) return TErrorList.device_not_connected;
          int ndevs = dlst.getNumberOfDevices();
          if (ndevs == 0) return TErrorList.device_not_connected;
          LinkedList<DeviceCfg> dl = ec.getDeviceList();
          for (DeviceCfg dc : dl)
          {
            TDevice tdv = dlst.getDevice(dc.getNumber().getValue());
            if (tdv == null) return TErrorList.device_not_connected;
            prpSet = dc.getPropertySet();
            if (prpSet != null && prpSet.length() > 0)
            { // is there a device-specific property list ?
              for (NameCfgList nc : nl)
              { // yes ! find it
                if (nc.getName().compareToIgnoreCase(prpSet) == 0)
                { // assign the names list
                  tdv.setPropertyList(nc.getMembers());
                }
              }
              propertyList.setDeviceOriented(true);
            }
            String oldName = tdv.getName();
            if (dc.getName().compareTo(oldName) != 0)
            {
              dlst.renameDevice(oldName,dc.getName());
            }
            deviceRedirection = dc.getRedirection();
            if (deviceRedirection != null && deviceRedirection.length() > 0)
            {
              tdv.setRedirection(deviceRedirection);
              deviceRedirection = null; // reset this !
            }
            deviceDescription = dc.getDescription();
            if (deviceDescription != null && deviceDescription.length() > 0)
            {
              tdv.setDescription(deviceDescription);
              deviceDescription = null; // reset this !
            }
          }
          break;
        }
      }    
      return 0;
    }
    // no fec.xml : look for devices.csv ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    csv devNamesFile = new csv(initializer.getDevicesResource(getLocalName()));
    int rc = devNamesFile == null ? 
        TErrorList.file_error : 
        getRegisteredDevicesFromFile(devNamesFile,new devNameRowHndlr());
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get device names from " + getLocalName() + "-devices.csv : " + TErrorList.errorString[rc]);
    return rc;
  }
  public int getRegisteredNamesFromPropertyFile(String propertyName)
  {
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    TExportProperty prp = propertyList.getFirstProperty(propertyName);
    String devFileName = propertyName + "-names.csv";
    csv devNamesFile = new csv(initializer.getFecHome() + "/" + getLocalName() + "/" + devFileName);
    int rc = devNamesFile == null ? 
        TErrorList.file_error : 
        getRegisteredDevicesFromFile(devNamesFile,new devNameRowHndlr(prp));
    if (rc == TErrorList.no_such_file)
    {
      devNamesFile = new csv(initializer.getFecHome() + "/" + devFileName);
      rc = devNamesFile == null ? 
          TErrorList.file_error : 
          getRegisteredDevicesFromFile(devNamesFile,new devNameRowHndlr(prp));
    }
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get device names from " + devFileName + " : " + TErrorList.errorString[rc]);
    return rc;
  }
  public int getRegisteredDevicesFromFile(csv csvFile,devNameRowHndlr devRows)
  {
    // register the csv database structure: 
    csvColumn[] devCols = new csvColumn[7];
    devCols[0] = new csvColumn("DEVICE_NUMBER","",new IntFieldHandler(devRows,"deviceNumber"));
    devCols[1] = new csvColumn("DEVICE_NAME","",new StringFieldHandler(devRows,"deviceName"));
    devCols[2] = new csvColumn("MODULE_NUMBER","",new IntFieldHandler(devRows,"deviceNumberAlt")); // alias for DEVICE_NAME
    devCols[3] = new csvColumn("MODULE_NAME","",new StringFieldHandler(devRows,"deviceNameAlt"));
    devCols[4] = new csvColumn("REDIRECTION","",new StringFieldHandler(devRows,"deviceRedirection"));
    devCols[5] = new csvColumn("PROPERTY_LIST","",new StringFieldHandler(devRows,"prpListFile"));
    devCols[6] = new csvColumn("DEVICE_DESCRIPTION","",new StringFieldHandler(devRows,"deviceDescription"));
    // read the file
    int rc = csvFile.readFile(devCols,devRows);
    // close it
    csvFile.close();  
    return rc;
  }
  class devPrpLstHndlr implements csvHandler
  {
    protected LinkedList<String> prpLst = null;
    public int process(String strValue,int index)
    {
      if (strValue == null || strValue.length() == 0) return 0;
      if (prpLst == null) prpLst = new LinkedList<String>();
      prpLst.add(strValue);
      return 0;
    }
  }
  public int getRegisteredPropertyListFromFile(TDevice device, String filename)
  {
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    csv csvFile = new csv(initializer.getFecHome() + "/" + filename);
    devPrpLstHndlr namhdlr = new devPrpLstHndlr();
    csvColumn[] devCols = new csvColumn[1];
    devCols[0] = new csvColumn("PROPERTY_NAME","",namhdlr);
    // read the file
    int rc = csvFile.readFile(devCols,null);
    // close it
    csvFile.close();
    if (rc == 0) device.setPropertyList(namhdlr.prpLst);
    return rc;
  }
  public boolean isDeviceSetLocal(String devName, String propertyName)
  {
    TExportProperty prp = propertyList.getFirstProperty(propertyName);
    WildcardMatch wc = new WildcardMatch(devName);    
    String tgt;
    TDevice[] dlst = getDeviceList().getDeviceList();
    for (int i=0; i<dlst.length; i++)
    {
      if ((tgt=dlst[i].getRedirection()) != null) 
      {
        if (wc.matches(tgt)) return false;
      }
    }    
    if ((tgt=prp.getDescription().getRedirection()) != null)
    {
      if (wc.matches(tgt)) return false;      
    }
    return true; /* appears to be local so-far */    
  }
  public String getSubsystem()
  {
    return subsystem;
  }
  public void setSubsystem(String subsystem)
  {
    TStrings.validateSubSystemName(subsystem);
    this.subsystem = subsystem;
  }
  /**
   * \brief An equipment module can schedule a property or set of properties for immediate 
   * execution and delivery  
   * 
   *  \param devPropertyList A single string argument containing one or more
   *  properties to be scheduled for immediate execution.  This should be a
   *  single string.  If more than one property is to be scheduled, then the
   *  property list string should be a set of properties separated by commas or
   *  blanks
   *  
   *  
   * @author duval 
   */
  public int scheduleProperty(String devPropertyList)
  {
    if (devPropertyList == null) return TErrorList.argument_list_error;
    String[] prps  = null;
    String tok = " "; // try a space first
    if (!devPropertyList.contains(tok)) tok = ";";
    if (!devPropertyList.contains(tok)) tok = ",";
    if ((prps = devPropertyList.split(tok)) == null) return TErrorList.code_failure;
    for (int i=0; i<prps.length; i++)
    {
      if (propertyList.hasProperty(prps[i])) gEqmFactory.systemScheduleProperty(this,prps[i].trim());
    }
    return 0;
  }
  ConcurrentLinkedQueue<TEquipmentBackgroundTask> bkgTasks = new ConcurrentLinkedQueue<TEquipmentBackgroundTask>();
  public int addEquipmentBackgroundTask(TEquipmentBackgroundTask tbkg)
  {
    if (tbkg == null) return TErrorList.invalid_parameter;
    bkgTasks.add(tbkg);
    return getEquipmentModuleFactory().addEquipmentBackgroundTask(tbkg);
  }
  public int startup()
  {
    int rc = 0;
    getRegisteredDevicesFromFile();
    clearCASAlarmList();
    if (getInitializationRoutine() != null)
    {
      try
      {
        rc = getInitializationRoutine().initialize();
        TFecLog.log(getLocalName()+" initialization routine called : "+TErrorList.getErrorString(rc));
      }
      catch (Exception e)
      {
        TFecLog.log(getLocalName()+" initialization routine called : unhandled exception "+e.getMessage());          
      }
    }
    // is there a general alarm watch file ?
    gEqmFactory.getAlarmWatchListFromFile(getLocalName());
    // is there a local history file ?
    gEqmFactory.getHistoryListFromFile(getLocalName());
    return rc;
  }
  public int shutdown()
  {
    int rc = 0;
    if (!hasInitialized) return TErrorList.not_initialized;
    hasInitialized = false;
    if (getExitRoutine() != null) 
    { // there is a registered exit routine !
      try
      {
        rc = getExitRoutine().terminate();
        TFecLog.log(getLocalName()+" exit routined called : "+TErrorList.getErrorString(rc));
      }
      catch (Exception e)
      {
        TFecLog.log(getLocalName()+" exit routined called : unhandled exception "+e.getMessage());        
      }
    }
    alarmDefinitionList.clear();
    removeAllGCastNets();
    removeAllRegisteredNets();
    removeAllRegisteredUsers();
    return rc;
  }
  /**
   * @brief Registers a cycle trigger callback dispatch function
   * 
   * If a CYCLER is running in a server's context, then the server will receive 'Cycle Number' events
   * scheduled by the designated CYCLER server.
   * The cycle number will make use of the 'System Data Stamp' to tag all data sets obtained from the
   * server.  A server can also register a trigger function dispatch routine (or routines) to be called when a 'Cycle
   * Number' event occurs.  The dispatch routines will be called prior to setting the 'System Stamp' to the new
   * Cycle Number, which will be set following the execution of all dispatch routines.  Optionally, the server can
   * provide a property (or list of properties) to be scheduled following the dispatch execution.  This will ensure
   * that such properties will be called immediately following dispatch execution AND contain the most recent Cycle Number
   * as the 'System Data Stamp'.
   *
   * @param trigger is a reference to the dispatch handler to be called when a new cycle number arrives.
   * 
   * @return 0 upon success or TErrorList.already_assigned
   * 
   * \b Example:
   * @include eg_cycleTrigger.java
   */
  public int registerCycleTrigger(TCycleTrigger trigger)
  {
    return getEquipmentModuleFactory().registerCycleTrigger(trigger);
  }
  /**
   * @brief Unregisters a previously registered cycle trigger callback dispatch function
   * 
   * If a cycle trigger event dispatch routine is no longer required, it can be unregistered using this 
   * routine.
   * 
   * @param trigger is a reference to the dispatch handler to be removed
   * 
   * @return 0 upon success or return TErrorList.un_allocated
   */
  public int unregisterCycleTrigger(TCycleTrigger trigger)
  {
    return getEquipmentModuleFactory().unregisterCycleTrigger(trigger);
  }  
}