/*
 * Created on Nov 9, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 * 
 * CHANGES
 * 2005-02-10 JW * registerStockProperties() put into constructor
 * 13.7.05 PD * added getFecInformationFromFile() and overloaded method SystemInit() 
 */
package de.desy.tine.server.equipment;

import java.io.*;
import java.net.*;
import java.nio.channels.ServerSocketChannel;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import org.w3c.dom.Document;
import de.desy.tine.xmlUtils.*;
import de.desy.tine.addrUtils.*;
import de.desy.tine.bitfieldUtils.TBitfield;
import de.desy.tine.bitfieldUtils.TBitfieldRegistry;
import de.desy.tine.bitfieldUtils.TField;
import de.desy.tine.client.*;
import de.desy.tine.csvUtils.*;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.*;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.headers.*;
import de.desy.tine.io.*;
import de.desy.tine.queryUtils.*;
import de.desy.tine.server.*;
import de.desy.tine.server.alarms.*;
import de.desy.tine.server.connections.*;
import de.desy.tine.server.devices.TDevice;
import de.desy.tine.server.histories.*;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.MsgLog;
import de.desy.tine.server.logger.TFecLog;
import de.desy.tine.server.properties.*;
import de.desy.tine.startup.*;
import de.desy.tine.stringUtils.*;
import de.desy.tine.structUtils.*;
import de.desy.tine.types.*;

//import java.io.*;
/**
 * \internal
 * 
 * @author duval 
 */
public class TEquipmentModuleFactory
{
  public static final int FAILOVER_NONE = 0x00;
  public static final int FAILOVER_MASTER = 0x01;
  public static final int FAILOVER_SLAVE = 0x02;
  private static final String FAILOVER_MASTER_DESC = "failover MASTER";
  private static final String FAILOVER_SLAVE_DESC = "failover SLAVE";
  private String gFecContext;
  private String gFecName;
  private String gFecDescription = "unknown";
  private String gFecLocation = "unknown";
  private String gFecSubsystem = "TEST";
  private String gFecHardware = "none";
  private String gFecResponsible = "unknown";
  private FECAddr gFecAddr;
  private FECInfo gFecInfo;
  private TLinkFactory gLinkFactory = null;
  private LinkedList<clnInputData> clnInputDataTable = new LinkedList<clnInputData>();
  private LinkedList<AliasTableEntry> gAliasList = new LinkedList<AliasTableEntry>();
  private static TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
  public TInitializer getInitializer() { return initializer;}
  private static TEquipmentModuleFactory instance = new TEquipmentModuleFactory();
  private TPacket srvp = null; // udp server socket for handling requests
  private int sckRcvBufferSize = initializer.getSrvRcvBufferSize(); //0x10000;
  private int sckTimeToLive = initializer.getSckTimeToLive(); //16;
  private ServerSocket srvs = null; // tcp server socket for handling requests
  private LinkedList<TContractBucket> bucketList = new LinkedList<TContractBucket>(); // do we need this ?
  private int gPortOffset = 0;
  public int getPort() { return gPortOffset; }
  static final int maxNumEqmTableEntries = 16;
  private int numEqmTableEntries = 0;
  private TEquipmentModule[] eqmTable = new TEquipmentModule[maxNumEqmTableEntries];
  private int numExpiredContracts = 0;
  private ArrayList<TContractTable> conTable = new ArrayList<TContractTable>();
  private LinkedList<TContractTable> conTableAdd = new LinkedList<TContractTable>();
  private ArrayList<TClient> clnTable = new ArrayList<TClient>();
  private LinkedList<TClient> clnTableAdd = new LinkedList<TClient>();
  private TContractTable currentContractEntry = null;
  protected HashMap<String, TAlarmDefinition> stkAlarmDefinitions = new HashMap<String, TAlarmDefinition>();
  public HashMap<String, TAlarmDefinition> getStockAlarmDefinitions() { return stkAlarmDefinitions; }
  private int gSystemTick = 10;
  private int gStaleData = 0;
  private int gBurstLimit = 1024;
  private int gCycleDelay = STD_CYCLE_INTERVAL;
  private int gBurstLimitReachedCount = 0;
  private int gClientReconnects = 0;
  private int gClientRetries = 0;
  private int gContractMisses = 0;
  private int gContractDelays = 0;
  private int gClientMisses = 0;
  private int gSingleLinkCount = 0;
  private long gDataTimeStamp = 0;
  private long gDataTimeStampOffset = 0;
  private long gSrvStartupTime = 0;
  private long gAppCompilationTime = 0;
  private int gAppVersionMajor = 1;
  private int gAppVersionMinor = 0;
  private int gAppVersionRevision = 0;
  private int[] gSrvStatsBuffer = new int[16];
  private String gLastErrorString;
  private String gLastMessage;
  private long gLastWriteAccessTime = 0;
  private String gLastWriteAccessUser = null;
  private InetAddress gLastWriteAccessAddr = null;
  private String gLastWriteAccessPrp = null;
  private String gLastWriteAccessDev = null;
  private String gLastWriteAccessEqm = null;
  protected static int debugLevel = 0;
  public static void setDebugLevel(int level) { debugLevel = level; }
  private int gTotalUdpPackets = 0;
  private int gTotalTcpPackets = 0;
  private TAcceptorThread acceptThrd;
  private TCycleThread cycleThrd;
  private boolean gLinkTablesAccessed = false;
  private boolean gSystemExitCondition = false;
  private boolean isInsideCycle = false;
  public boolean isInitialized = false;
  public boolean registrationPending = false;
  private long registrationTime = 0;
  public boolean gSynchronizeContracts = true;
  public boolean gRequireAcknowledgments = true;
  public boolean gEqpFcnBusySemaphore = false;
  public boolean gSystemPresetMemory = false;
  public boolean gRetardSingleContractRemoval = true;
  public boolean gPutCommandsInFeclog = true;
  private static boolean gServerWaiting = false;
  private static boolean gSystemRunningStandAlone = false;
  public boolean isRunningStandAlone() { return gSystemRunningStandAlone; }
  private int gSystemStamp = 0;
  public int getSystemStamp() { return gSystemStamp; }
  public void setSystemStamp(int value) { gSystemStamp = value; }
  /**
   * \brief A server can be toggled in and out of a waiting state with this method
   * 
   * During initialization it might be prudent to set the waiting state to 'true'
   * until all initialization has been completed and the server is ready to server, 
   * at which point the waiting state can then be set to 'false'.
   * 
   * If the server is in a waiting state, all requests will receive the error code
   * 'not initialized' as the link status.
   * 
   * @param value is the wait state.
   */
  public static void setServerWaiting(boolean value) { gServerWaiting = value; }
  public static boolean getServerWaiting() { return gServerWaiting; }
  private static final short ACK_PENDING = 0x7fff;
  private static final short SUB_PENDING = 0x3fff;
  private static final short BCAST_ID = 255;
  private static final short PRP_SCHEDULE_SIGNAL = 0;
  private static final short PRP_REQUEST_SIGNAL = 1;
  private static final short PRP_CANCEL_SIGNAL = 2;
  private static final short RENEWAL = 60;
  private static final short REMINDER = 5;
  private static final short ACK_REQUEST = (2 * REMINDER);
  private static final short HEARTBEAT = 60;
  private static final short REPLY_PENDING = 3;
  private static final int STD_CYCLE_INTERVAL = 1;
  public static final int MAXIMUM_LOCK_DURATION = 60; // seconds
  private static int srvPacketMtu = initializer.getSrvPacketMtu();//1472;
  public static int getServerPacketMTU() { return srvPacketMtu; }
  private int gMinPollingRate = 10;
  private int gMaxPollingRate = 60000;
  private int gNumBkgTasks = 0;
  private int gNumEnsErrors = 0;
  private TEquipmentModuleHook tEqmHook = null;
  private FecCfg gFecXmlCfg = null;
  public FecCfg getFecXmlCfg() { return gFecXmlCfg; }
  public void setTEquipmentModuleHook(TEquipmentModuleHook hook) 
  { 
    tEqmHook = hook;
    // if this is called after the factory's systemInit() has been called ...
    if (isInitialized) hook.SystemInit();
  }
  public TEquipmentModuleHook getTEquipmentModuleHook() { return tEqmHook; }
  private String tRedirectedServer;
  private boolean isDeviceRedirected(String redirString,String devName)
  {
    if (redirString == null || redirString.length() == 0) return false;
    if (devName == null || devName.length() == 0) return true;
    int idx = redirString.lastIndexOf('/');
    if (idx == -1) return true; // not given in redirection string
    int edx = redirString.indexOf('[');
    if (edx == -1) edx = redirString.length(); else edx--;
    if (redirString.substring(idx+1,edx+1).compareTo(devName) == 0) return true;
    return false;
  }
  private byte[] redirectionStringToBytes(String redirString)
  { // parse strings of form "/<context>/<server>/<device>[<property>] into byte array ...
    byte[] b = new byte[TStrings.STATUS_SIZE];
    String s, ctx = "", srv, dev = "", prp = "";    
    int indx = redirString.indexOf('[');
    if (indx != -1)
    { // property given !
      if (indx == 0) return b; // invalid string
      s = redirString.substring(0,indx);
      prp = redirString.substring(indx+1);
      if ((indx=prp.indexOf(']')) == -1) return b; // invalid string
      prp = indx > 0 ? prp.substring(0,indx) : "";
    }
    else 
    {
      s = redirString;
    }
    indx = s.indexOf('/');
    if (indx == 0)  
    { // context given, peel it off
      indx = s.substring(1).indexOf('/');
      if (indx == -1) return b; // invalid string 
      ctx = s.substring(1, indx+1);
      s = s.substring(indx+2);
      indx = s.indexOf('/');
    }
    if (indx == -1)
    { // s is now of the form <server>/<device>
      srv = s;
      dev = "";
    }
    else
    {
      srv = indx > 0 ? s.substring(0,indx) : "";
      dev = s.substring(indx+1);
    }
    // this is the way it's done in release 3 ...
    int len, off = 0;
    len = ctx.length(); if (len > TStrings.CONTEXT_NAME_SIZE) len = TStrings.CONTEXT_NAME_SIZE;
    System.arraycopy(ctx.getBytes(),0,b,off,len);    
    off += TStrings.CONTEXT_NAME_SIZE;
    len = srv.length(); if (len > TStrings.EXPORT_NAME_SIZE) len = TStrings.EXPORT_NAME_SIZE;
    System.arraycopy(srv.getBytes(),0,b,off,len);    
    off += TStrings.EXPORT_NAME_SIZE;
    len = dev.length(); if (len > TStrings.DEVICE_NAME_SIZE) len = TStrings.DEVICE_NAME_SIZE;
    System.arraycopy(dev.getBytes(),0,b,off,len);    
    off += TStrings.DEVICE_NAME_SIZE;
    len = prp.length(); if (len > TStrings.PROPERTY_NAME_SIZE) len = TStrings.PROPERTY_NAME_SIZE;
    System.arraycopy(prp.getBytes(),0,b,off,len);    
    return b;
  }
  private TServerStatistics serverStatistics = new TServerStatistics(this);
  public int setRedirectionString(String redirectionString) 
  { 
    tRedirectedServer = redirectionString;
    return TErrorList.server_redirection;
  }
  public TContractTable getCurrentContractEntry() { return currentContractEntry; }
  TPropertyList stockList = new TPropertyList(); // a list or a query fcn
  public TPropertyList getStockList() { return stockList; }
  TMetaProperties metaProps = new TMetaProperties();
  class AliasTableEntry 
  {
    private String name;
    private String alias;
    AliasTableEntry(String thisName,String thisAlias)
    {
      name = thisName;
      alias = thisAlias;
    }
    public String getTag() { return name; }
    public String getAlias() { return alias; }
  }
  public int addAliasToList(String name, String alias)
  {
    if (name == null || name.length() == 0) return TErrorList.argument_list_error;
    if (alias == null || alias.length() == 0) return TErrorList.argument_list_error;
    gAliasList.add(new AliasTableEntry(name,alias));
    return 0;
  }
  /**
   * Register common stock properties.
   */
  void registerStockProperties()
  {
    /*
     * HOW TO ADD A STOCK PROPERTY 1. Create Stock Property (see
     * TStockProperties) 2. Add a property handler here
     */
    stockList.addProperty(TStockProperties.NSTOCKPROPS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(TStockProperties.getNumStockProperties());
      }
    });
    // Returns all known stock properties, if available or not.
    stockList.addProperty(TStockProperties.STOCKPROPS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(StringToName.stringSetToName32(TStockProperties.getPropertyNames()));
      }
    });
    stockList.addProperty(TStockProperties.STOCKPROPS_USTRING, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        boolean checkWildCard = (devName.compareTo("*") != 0 && devName.indexOf("*") != -1);
        TPropertyList lst = TStockProperties.getPropertyList();            
        String[] slst = new String[lst.countUniqueProperties()];
        lst.getPropertyNames().toArray(slst);
        if (checkWildCard) slst = StringToName.matchStringsInArray(slst,devName);
        USTRING[] ustr = new USTRING[slst.length];
        char[] cha;
        StringBuffer sb = new StringBuffer(80);
        TExportProperty prp;
        for (int i = 0; i<slst.length; i++)
        {
          ustr[i] = new USTRING();
          if ((prp = lst.getFirstProperty(slst[i])) == null) continue;
          cha = prp.getName().toCharArray();
          sb.insert(0,cha);
          for (int c=cha.length; c<32; c++) sb.insert(c,(char)0);
          cha = prp.getDescription().getText().toCharArray();
          sb.insert(32,cha);
          ustr[i].setValues(prp.getOutputSize(),0,0,prp.getOutputFormat(),new String(sb));
        }
        return (short) dout.putData(ustr);
      }
    });
    stockList.addProperty(TStockProperties.STOCKPROPS_STRUCT, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyStockProps(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.SRVSTARTTIME, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(new Date(getServerStartTime()).toString());
      }
    });
    stockList.addProperty(TStockProperties.SRVSTARTTIME_LONG, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData((int) (getServerStartTime() / 1000));
      }
    });
    stockList.addProperty(TStockProperties.STRUCTFORMAT_INTINT, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyStructFormat(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.STRUCTFORMAT_NDD, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyStructFormat(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.STRUCTFORMAT, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyStructFormat(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.BITFIELDFORMAT, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyBitfieldFormat(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.CONTRACTS, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyContracts(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.CLIENTS, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyClients(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.ACTIVITY, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return callPropertyActivity(devName, dout, din, devAccess);
      }
    });
    stockList.addProperty(TStockProperties.SRVDESC, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(getFecDescription());
      }
    });
    stockList.addProperty(TStockProperties.SRVLOCATION, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(getFecLocation());
      }
    });
    stockList.addProperty(TStockProperties.SRVALIASLIST, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (gAliasList == null) return TErrorList.not_implemented;
        int len = gAliasList.size()*2;
        if (len == 0) return TErrorList.not_defined;
        NAME32 n32[] = new NAME32[len];
        AliasTableEntry ate;
        for (int i=0; i<len/2; i++)
        {
          ate = (AliasTableEntry)gAliasList.get(i);
          n32[2*i] = new NAME32(ate.name);
          n32[2*i+1] = new NAME32(ate.alias);
        }
        return (short) dout.putData(n32);
      }
    });
    stockList.addProperty(TStockProperties.SRVLASTACCESS, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        NAME32[] n32 = new NAME32[dout.dArrayLength];
        String s = null;
        switch (dout.dArrayLength)
        {
          case 6:
            s = getLastAccessEqm();
            n32[5] = new NAME32(s != null ? s : "");
          case 5: 
            n32[4] = new NAME32(new Date(getLastAccessTime()).toString());
          case 4:
            s = getLastAccessDev();
            n32[3] = new NAME32(s != null ? s : "");
          case 3: 
            s = getLastAccessPrp();
            n32[2] = new NAME32(s != null ? s : "");
          case 2: 
            InetAddress iaddr = getLastAccessAddress();
            s = iaddr != null ? getLastAccessAddress().getHostAddress() : "";
            n32[1] = new NAME32(s);
          case 1: 
            s = getLastAccessUser();
            n32[0] = new NAME32(s != null ? s : "");
            break;
          default:
            return TErrorList.dimension_error;
        }
        return (short) dout.putData(n32);
      }
    });
    stockList.addProperty(TStockProperties.SRVLASTACCESS_LONG, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData((int)(getLastAccessTime()/100));
      }
    });
    stockList.addProperty(TStockProperties.SRVSTATS, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(getServerStats());
      }
    });
    stockList.addProperty(TStockProperties.SRVEXIT, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (devAccess.isWrite())
        {
          int[] level = new int[]{0};
          if (din != null) din.getData(level);
          TFecLog.log("Server exit (level "+level[0]+") from remote caller");
          shutdown(level[0]);
          return 0;
        }
        return TErrorList.illegal_read_write;
      }
    });
    stockList.addProperty(TStockProperties.SRVOS, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData(TStockProperties.getStockSrvOS());
      }
    });
    stockList.addProperty(TStockProperties.SRVVERSION, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return dout.putData(TStockProperties.getStockSrvVersion());
      }
    });
    stockList.addProperty(TStockProperties.DEBUGLEVEL, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        short[] dbglvl = new short[1]; 
        if (din != null && devAccess.isWrite())
        {
          if  ((cc=din.getData(dbglvl)) == 0) 
          {
            debugLevel = dbglvl[0];
            TLinkFactory.setOutputDebugLevel(debugLevel);
          }
        }
        if (cc == 0 && dout != null)
        {
          dbglvl[0] = (short)TLinkFactory.getOutputDebugLevel();
          cc = dout.putData(dbglvl);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.LOGCOMMANDS, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        short[] logcmds = new short[1]; 
        if (din != null && devAccess.isWrite())
        {
          if  ((cc=din.getData(logcmds)) == 0) 
            gPutCommandsInFeclog = logcmds[0] != 0 ? true : false;
        }
        if (cc == 0 && dout != null)
        {
          logcmds[0] = (short)(gPutCommandsInFeclog ? -1 : 0);
          cc = dout.putData(logcmds);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.MESSAGE, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        char[] msg = new char[128]; 
        if (din != null && devAccess.isWrite())
        {
          if  ((cc=din.getData(msg)) == 0)
          {
            gLastMessage = new String(msg).substring(0, din.dCompletionLength);
            String s = "MSG: " + gLastMessage;
            System.out.println(s);
            TFecLog.log(s);
          }
        }
        if (cc == 0 && dout != null)
        {
          cc = dout.putData(gLastMessage);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.LOGFILE, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        if (dout.getFormat() != TFormat.CF_TEXT) return TErrorList.illegal_format;
        int nlines = dout.getArrayLength()/80;
        return (short) dout.putData(TFecLog.getLines(nlines));
      }
    });
    stockList.addProperty(TStockProperties.LOGDEPTH, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        int cc = 0;
        int[] depth = new int[1]; 
        if (din != null && devAccess.isWrite())
        {
          if  ((cc=din.getData(depth)) == 0) 
          {
            TFecLog.setLogDepth(depth[0]);
          }
        }
        if (cc == 0 && dout != null)
        {
          depth[0] = TFecLog.getLogDepth();
          cc = dout.putData(depth);
        }
        return cc;
      }
    });
    stockList.addProperty(TStockProperties.APPDATE_STRING, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(new Date(getAppCompilationTime()).toString());
      }
    });
    stockList.addProperty(TStockProperties.APPDATE, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData((int) (getAppCompilationTime() / 1000));
      }
    });
    stockList.addProperty(TStockProperties.APPVERSION, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(getAppVersion());
      }
    });
    stockList.addProperty(TStockProperties.SRVLOGFILES, new TPropertyHandler()
    {
      public int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        return (short) dout.putData(StringToName.stringArrayToName64(TFecLog.getFiles()));
      }
    });
    stockList.addProperty(TStockProperties.SRVLOGFILE, new TPropertyHandler()
    {
      protected int call(String devName, TDataType dout, TDataType din, TAccess devAccess)
      {
        String tgtLogFile = null;
        if (devName.length() > 0 && !devName.startsWith("#"))
        {
          tgtLogFile = devName;
        }
        else
        {
          tgtLogFile = "fec.log.0";          
        }
        if (din != null && din.getArrayLength() > 0)
        {
          if (devAccess.isWrite())
          {
            char[] txt = new char[din.getArrayLength()];
            din.getData(txt);
            TFecLog.writeTextFile(tgtLogFile, txt);
          }
        }
        if (dout.getFormat() != TFormat.CF_TEXT) return TErrorList.illegal_format;
        int nlines = dout.getArrayLength()/80;
        return (short) dout.putData(TFecLog.getLines(tgtLogFile,nlines));
      }
    });

  }
  private short callPropertyActivity(String devName, TDataType dout, TDataType din,TAccess devAccess)
  {
    short cc = 0;
    ByteArrayOutputStream bsout = new ByteArrayOutputStream(1024);
    TDataOutputStream dsout = new TDataOutputStream(bsout);
    try
    {
      serverStatistics.setNumTotalContracts((short)(conTable.size()));
      serverStatistics.setNumTotalClients((short)clnTable.size());
      serverStatistics.setNumTargetClients((short)getCurrentContractEntry().eqm.numClients);
      serverStatistics.setNumTargetContracts((short)getCurrentContractEntry().eqm.numContracts);
      serverStatistics.setNumBkgTsks((short)gNumBkgTasks);
      serverStatistics.setNumConnectionTimeouts(TLinkFactory.getInstance().getTotalLinkTimeouts());
      serverStatistics.setNumConnectionArrivals(TLinkFactory.getInstance().getTotalConnectionArrivals());
      serverStatistics.setNumUdpPkts(gTotalUdpPackets);
      serverStatistics.setNumTcpPkts(gTotalTcpPackets);
      serverStatistics.setSystemPollingRate(gMinPollingRate);
      serverStatistics.writeData(dsout);
      dsout.close();
      cc = (short) dout.putData(bsout.toByteArray(),TServerStatistics.getStructDescription().getTagName());
    }
    catch (IOException e)
    {
      cc = TErrorList.code_failure;
    }
    return cc;
  }
  private short callPropertyStockProps(String devName, TDataType dout, TDataType din,
      TAccess devAccess)
  {
    LinkedList<TExportProperty> matchingProps;
    ByteArrayOutputStream bsout = new ByteArrayOutputStream(1024);
    TDataOutputStream dsout = new TDataOutputStream(bsout);
    Iterator<TExportProperty> it;

    if (dout.getTag().compareTo("PRPQSr4") == 0 || (dout.getArrayLength() % TPropertyQuery.sizeInBytes) == 0)
    {
      if (din != null)
      { // target property specified ...
        String tgt = din.toString();
        tgt = tgt.replace('\n', (char) 0).trim();
        matchingProps = TStockProperties.getPropertyList().getEqualProperties(tgt);
        if (matchingProps == null || matchingProps.size() == 0) return TErrorList.not_exported;
        // fill in the targeted property info
        try
        {
          TPropertyQuery pqs = null;
          for (it = matchingProps.iterator(); it.hasNext();)
          {
            pqs = new TPropertyQuery((TExportProperty) it.next(), matchingProps.size() - 1);
            pqs.writeData(dsout);
          }
          dsout.close();
          return (short) dout.putData(bsout.toByteArray(), "PRPQSr4");
        }
        catch (Exception e)
        {
          return TErrorList.code_failure;
        }
      }
      else
      { // want the whole list ...
        TPropertyList lst = TStockProperties.getPropertyList();
        String[] slst = new String[lst.countUniqueProperties()];
        lst.getPropertyNames().toArray(slst);
        TPropertyQuery[] pqs = new TPropertyQuery[slst.length];
        if (slst.length == 0) return 0;
        byte[] tba = new byte[slst.length * XPropertyQuery.sizeInBytes];
        TExportProperty prp;
        for (int i=0; i<slst.length; i++)
        {
          prp = lst.getFirstProperty(slst[i]);
          pqs[i] = new TPropertyQuery(prp,1);
          System.arraycopy(pqs[i].toByteArray(), 0, tba, i * TPropertyQuery.sizeInBytes,
              XPropertyQuery.sizeInBytes);
        }
        return (short) dout.putData(tba, "PRPQSr4");
      }
    }
    if (dout.getTag().compareTo("XPQS") == 0 || (dout.getArrayLength() % XPropertyQuery.sizeInBytes) == 0)
    {
      if (din != null)
      { // target property specified ...
        String tgt = din.toString();
        tgt = tgt.replace('\n', (char) 0).trim();
        matchingProps = TStockProperties.getPropertyList().getEqualProperties(tgt);
        if (matchingProps == null || matchingProps.size() == 0) return TErrorList.not_exported;
        // fill in the targeted property info
        try
        {
          XPropertyQuery xpq = null;
          for (it = matchingProps.iterator(); it.hasNext();)
          {
            xpq = new XPropertyQuery((TExportProperty) it.next(), matchingProps.size() - 1);
            xpq.writeData(dsout);
          }
          dsout.close();
            return (short) dout.putData(bsout.toByteArray(), "XPQS");
        }
        catch (Exception e)
        {
          return TErrorList.code_failure;
        }
      }
      else
      { // want the whole list ...
        TPropertyList lst = TStockProperties.getPropertyList();
        String[] slst = new String[lst.countUniqueProperties()];
        lst.getPropertyNames().toArray(slst);
        XPropertyQuery[] xpq = new XPropertyQuery[slst.length];
        if (slst.length == 0) return 0;
        byte[] tba = new byte[slst.length * XPropertyQuery.sizeInBytes];
        TExportProperty prp;
        for (int i=0; i<slst.length; i++)
        {
          prp = lst.getFirstProperty(slst[i]);
          xpq[i] = new XPropertyQuery(prp,1);
          System.arraycopy(xpq[i].toByteArray(), 0, tba, i * XPropertyQuery.sizeInBytes,
              XPropertyQuery.sizeInBytes);
        }
        return (short) dout.putData(tba, "XPQS");
      }
    }
    return TErrorList.illegal_format;
  }
  private short callPropertyStructFormat(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    if( din == null )  //added by Karol, 18.02.2008
    {
        MsgLog.log("callPropertyStructFormat", "callPropertyStructFormat, din == null",TErrorList.code_failure,null,0);
        return TErrorList.code_failure;
    }
    short cc = 0;
    String tag = din.toString();
    tag = tag.replace('\n', (char) 0).trim();
    NAME16II[] structData = TStructRegistry.getRegisteredStructure(tag);
    if (structData != null && structData.length > 0)
    {
      switch (dout.getFormat())
      {
        case TFormat.CF_INTINT:
          INTINT[] ii = new INTINT[structData.length];
          for (int i=0; i<structData.length; i++)
            ii[i] = new INTINT(structData[i].i1val,structData[i].i2val);
          cc = (short) dout.putData(ii);
          break;
        case TFormat.CF_NAME16II:
          cc = (short) dout.putData(structData);
          break;
        case TFormat.CF_NAME64DBLDBL:
          // TODO: let the registry return NAME64DBLDBL
          NAME64DBLDBL[] sd64 = new NAME64DBLDBL[structData.length];
          for (int i=0; i<structData.length; i++) 
            sd64[i] = new NAME64DBLDBL(structData[i].name,structData[i].i1val,structData[i].i2val);
          cc = (short) dout.putData(sd64);
          break;
        default:
          cc = TErrorList.illegal_format;
      }
    }
    else
    {
      cc = TErrorList.invalid_structure_tag;
    }
    return cc;
  }
  private short callPropertyBitfieldFormat(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    if( din == null )  //added by Karol, 18.02.2008
    {
        MsgLog.log("callPropertyBitfieldFormat", "callPropertyBitfieldFormat, din == null",TErrorList.code_failure,null,0);
        return TErrorList.code_failure;
    }
    short cc = 0;
    String tag = din.toString();
    tag = tag.replace('\n', (char) 0).trim();
    TBitfield tbf = TBitfieldRegistry.getBitfield(tag);
    if (tbf == null) return TErrorList.invalid_structure_tag; 
    TField[] tbff = tbf.getFields();
    NAME16I[] ni = new NAME16I[tbff.length+1];
    for (int i=0; i<tbff.length; i++) 
      ni[i] = new NAME16I(tbff[i].getName(),tbff[i].getBitmask());
    ni[tbff.length] = new NAME16I(tag,tbf.getFormat());
    switch (dout.getFormat())
    {
      case TFormat.CF_NAME16I:
        cc = (short) dout.putData(ni);
        break;
      case TFormat.CF_NAME64DBL:
        // TODO: let the registry return NAME64DBL
        NAME64DBL[] ni64 = new NAME64DBL[ni.length];
        for (int i=0; i<ni.length; i++) 
          ni64[i] = new NAME64DBL(ni[i].name,ni[i].ival);
        cc = (short) dout.putData(ni64);
        break;
      default:
        cc = TErrorList.illegal_format;
    }
    return cc;
  }
  private short callPropertyContracts(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    short cc = 0;
    String dTag = dout.getTag();
    if (dTag.length() == 0)
    {
      int ssz = TStructRegistry.getSizeInBytes("CONQS");
      if ((dout.getArrayLength() % ssz) != 0)
      {
        return TErrorList.invalid_structure_size;
      }
      dTag = "CONQS";
    }
    if (dTag.compareTo("CONQS") != 0 && dTag.compareTo("CTQSr4") != 0)
    {
      return TErrorList.invalid_structure_tag;
    }
    synchronized (conTable)
    {
      Iterator<TContractTable> it = conTable.iterator();
      ByteArrayOutputStream bsout = new ByteArrayOutputStream(1024);
      TDataOutputStream sdout = new TDataOutputStream(bsout);
      TContractTable contract = null;
      try
      {
        while (it.hasNext())
        {
          contract = (TContractTable) it.next();
          contract.writeData(dTag,sdout);
        }
        sdout.close();
        dout.putData(bsout.toByteArray(), dTag);
      }
      catch (Exception e)
      {
        cc = TErrorList.code_failure;
      }
    }
    return cc;
  }
  private short callPropertyClients(String devName, TDataType dout, TDataType din, TAccess devAccess)
  {
    short cc = 0;
    synchronized (clnTable)
    {
      Iterator<TClient> it = clnTable.iterator();
      ByteArrayOutputStream bsout = new ByteArrayOutputStream(1024);
      TDataOutputStream sdout = new TDataOutputStream(bsout);
      TClient element = null;
      try
      {
        while (it.hasNext())
        {
          element = (TClient) it.next();
          element.writeData(sdout);
        }
        sdout.close();
        dout.putData(bsout.toByteArray(), "CLNQS");
      }
      catch (Exception e)
      {
        cc = TErrorList.code_failure;
      }
    }
    return cc;
  }
  private InetAddress mcastInetAddress = null;
  /**
   * ???
   * 
   * @return
   */
  private InetAddress getMCastAddr()
  {
    if (mcastInetAddress != null) return mcastInetAddress;
    try
    {
      String[] mcaddr = initializer.getMCastAddress().split("\\.");
      InetAddress myip = InetAddress.getLocalHost();
      String[] ipaddr = myip.getHostAddress().split("\\.");
      String ip = mcaddr[0] + "." + mcaddr[1] + "." + ipaddr[2] + "." + ipaddr[3];
      mcastInetAddress = InetAddress.getByName(ip);
      return mcastInetAddress;
    }
    catch (Exception e)
    {
      return null;
    }
  }
  /**
   * ???
   * 
   * @param ia
   * @return
   */
  private InetAddress getBCastAddr(InetAddress ia)
  {
    String strAddr = ia.getHostAddress();
    strAddr = strAddr.substring(0,strAddr.lastIndexOf('.')).concat("255");
    try 
    {
      return InetAddress.getByName(strAddr);
    }
    catch (Exception e)
    {
      return null;
    }
  }
  /**
   * Finds the best polling interval which satisfies 2 different client requests. 
   * For instance, if client A wants an interval of 2000 msecs and client B wants 
   * an interval of 3000 msecs, then polling every 1000 msecs can satisfy both. 
   * 
   * @param p1 is polling interval number 1
   * @param p2 is polling interval number 2
   * @return a polling interval which can satisfy the given intervals
   */
  private int findpoll(int p1, int p2)
  {
    int p3;
    if (gMinPollingRate > gMaxPollingRate) gMinPollingRate = gMaxPollingRate;
    if (p1 < gMinPollingRate) return p2 <= gMinPollingRate ? gMinPollingRate : p2;
    if (p2 < gMinPollingRate || p1 == p2) return p1;
    if (p2 < p1)
      while ((p3 = p1 % p2) >= gMinPollingRate)
      {
        p1 = p2;
        p2 = p3;
      }
    else if (p1 < p2) while ((p3 = p2 % p1) >= gMinPollingRate)
    {
      p2 = p1;
      p1 = p3;
    }
    return p1 > p2 ? p2 : p1;
  }
  /**
   * TODO keep bcast nets list in factory (not specific to an eqm); follow isMemberControlNets
   * logic
   * 
   * @param tc the Tine Client to check against the list
   * @return
   */
  private boolean isMemberBcastNets(TClient tc)
  {
    if (debugLevel > 0) DbgLog.log("IsMemberBcastNets","non yet implemented");
    return true;
  }
  private boolean isMemberControlNets(TEquipmentModule eqm, TClient tc)
  {
    if (eqm.gRejectAllNets) return false;
    if (eqm.gRegisteredNetsList == null || 
        eqm.gRegisteredNetsList.size() == 0) return true; // everyone is !
    // check vs subnet, vs specified address ...
    String tgtAddr = tc.IPaddress.getHostAddress();
    if (eqm.gRegisteredNetsList.contains(tgtAddr)) return true;
    String s[] = tgtAddr.split("\\.");
    if (s.length < 4) return false;
    String netAddr = s[0] + "." + s[1] + "." + s[2] + ".255";
    if (eqm.gRegisteredNetsList.contains(netAddr)) return true;
    netAddr = s[0] + "." + s[1] + ".255.255";
    if (eqm.gRegisteredNetsList.contains(netAddr)) return true;
    return false;
  }
  private boolean isMemberUsersList(TEquipmentModule eqm, String userName)
  {
    if (eqm.gRegisteredUsersList == null || 
        eqm.gRegisteredUsersList.size() ==  0) return true; // everyone is !
    if (eqm.gRegisteredUsersList.contains(userName.toUpperCase())) return true;
    return false;
  }
  private boolean isMemberLockSet(TEquipmentModule eqm, TClient tc)
  {
    if (eqm.accessLock.clientCanWrite(tc)) return true;
    return false;
  }
  /**
   * Find the equipment module associated with the contract request.
   * TeqmFactory manages all registered equipment modules, and in most cases
   * this is a list of 1.  The value of numEqmTableEntries very rarely is
   * as large as 5 or 6, so a simple name check will always yield fast results.
   * 
   * @param tc is the contract for which the associated equipment module is 
   * desired.
   * @return the equipment module found or a null pointer.
   */
  private TEquipmentModule LocateEquipmentModule(TContract tc)
  {
    for (int i = 0; i < numEqmTableEntries; i++)
    {
      if (eqmTable[i] == null) continue;
      if (eqmTable[i].getModuleName().length() == 0) continue;
      if (eqmTable[i].getModuleName().compareTo(tc.eqmName) != 0) continue;
      return eqmTable[i];
    }
    return null;
  }
  /**
   * \brief finds a client entry within the contract table for the client 
   * address given
   * 
   * "Client Entries" and "Contract Entries" are two different things.  Furthermore,
   * a Client Entry is composed of a Contract-independent Client Address plus a 
   * Contract-specific Client Status.  A "Contract Entry" is composed of a Client-
   * independent Contract plus a client entry list.  This routine checks to see if 
   * the input Client Address is found within the client entry list belonging to the
   * input contract entry.  If not, it creates a new Client Entry based on the
   * input client address and appends this entry to the contract table's list.   
   * It also increments counters on the Client Entry (number of contracts the client 
   * has) and on the Contract Entry (number of clients the contract is servicing).
   * 
   * @param tct is the Contract Table Entry to be scanned
   * @param tc is the Client Address for which the Client Entry is desired
   * 
   * @return a TClientEntry object
   */
  private TClientEntry LocateClientInContract(TContractTable tct, TClient tc)
  {
    TClientEntry tce = null;
    Iterator<TClientEntry> li = tct.clt.iterator();
    while (li.hasNext())
    { // scan thru the contract's current client list
      tce = (TClientEntry) li.next();
      //if (tce.cln.IPaddress.equals(tc.IPaddress)) {
      if (tce.cln.equals(tc))
      { // client is already in this contract's client list
        if (tc.ncontracts == 0)
        { // client has no contracts (?) << should not be possible !
          DbgLog.log("LocateClientInContract","found client in contract list, but client has no contracts!");
          tct.pending = 0;
          if (tct.nclients == 0) tct.nclients++;
          tc.ncontracts++;
          tc.addContract(tct);
          if (tc.tineProtocol == 0)
          {
            tc.tineProtocol = TLinkFactory.DEFAULT_PROTOCOL_LEVEL;
          }
        }
        if (debugLevel > 2)
          DbgLog.log("LocateClientInContract","found client " + tc.IPaddress.getHostAddress()
                + " in contract list");
        return tce;
      }
    }
    tce = new TClientEntry();
    tce.cln = tc;
    tce.cln.ncontracts++;
    tce.cln.addContract(tct);
    tce.sts = new TClientStatus(tc.inetProtocol);
    tce.sts.statusCode = (short) tct.returnCode;
    tce.sts.counter = 1;
    tct.clt.add(tce);
    tct.nclients++;
    if (debugLevel > 2)
      DbgLog.log("LocateClientInContract","added client " + tc.IPaddress.getHostAddress() + " to contract list");
    return tce;
  }
  private TContractTable LocateContractInList(TContract con, TDataType din, boolean[] isNew)
  {
    TContractTable tct = null;
    int freeContractID = 1;
    isNew[0] = false;
    synchronized (conTable)
    {
      for (int i = 0; i < conTable.size(); i++)
      {
        tct = (TContractTable) conTable.get(i);
        if (tct.contractID == freeContractID) freeContractID++;
        if (!tct.contract.equals(con)) continue;
        if (tct.din != null && !tct.din.equals(din)) continue;
        if (debugLevel > 2) DbgLog.log("LocateContractInList","found contract in list");
        return tct;
      }
    }
    synchronized (conTableAdd)
    {
      Iterator<TContractTable> li = conTableAdd.iterator();
      while (li.hasNext())
      {
        tct = (TContractTable) li.next();
        if (tct.contractID == freeContractID) freeContractID++;
        if (!tct.contract.equals(con)) continue;
        if (tct.din != null && !tct.din.equals(din)) continue;
        if (debugLevel > 2) DbgLog.log("LocateContractInList","found contract in list");
        return tct;
      }
      if ((tct = new TContractTable()) != null)
      {
        int sizeOut = con.dataSizeOut;
        String tagOut = con.dataTagOut;
        if (con.dataFormatOut == TFormat.CF_STRING)
        {
          sizeOut = Integer.parseInt(tagOut);
          tagOut = "";
        }
        tct.contract = con;
        tct.contractID = freeContractID;
        tct.din = din;
        tct.dout = new TDataType(sizeOut, con.dataFormatOut);
        tct.dout.setTag(tagOut);
        tct.drb = new TDataType(sizeOut, con.dataFormatOut);
        tct.drb.setTag(tagOut);
        tct.eqm = LocateEquipmentModule(con);
        if (debugLevel > 2) DbgLog.log("LocateContractInList","added contract to list (ID = " + tct.contractID + ")");
        conTableAdd.add(tct);
        isNew[0] = true;
      }
    }
    return tct;
  }
  private TClient LocateClientInList(InetAddress ia, int port)
  {
    TClient tc = null;
    if (debugLevel > 3)
      DbgLog.log("LocateClientInList","incoming request from client " + ia.getHostAddress());
    synchronized (clnTableAdd)
    { // first check the not yet appended list ...
      Iterator<TClient> li = clnTableAdd.iterator();
      while (li.hasNext())
      {
        tc = (TClient) li.next();
        if (!tc.IPaddress.equals(ia)) continue;
        if (tc.port != port) continue;
        if (debugLevel > 2)
          DbgLog.log("LocateClientInList","found client (not yet appended)" + ia.getHostAddress() + " in client list");
        return tc;        
      }
      synchronized (clnTable)
      { // now check the working list
        for (int i = 0; i < clnTable.size(); i++)
        {
          tc = (TClient) clnTable.get(i);
          if (!tc.IPaddress.equals(ia)) continue;
          if (tc.port != port) continue;
          if (debugLevel > 2)
            DbgLog.log("LocateClientInList","found client" + ia.getHostAddress() + " in client list");
          return tc;
        }
      }
      if ((tc = new TClient()) != null)
      { // seems to be a new one ...
        tc.IPaddress = ia;
        tc.port = port;
        clnTableAdd.add(tc);
        if (debugLevel > 2)
            DbgLog.log("LocateClientInList","add client" + ia.getHostAddress() + " to client list");
      }
      else
      {
        throw new RuntimeException("code error : cannot make new instance of TClient!");
      }
    }
    return tc;
  }
  private TEquipmentModuleFactory()
  {
    debugLevel = Integer.parseInt(System.getProperty("debug.level", "0"));
    cycleThrd = new TCycleThread();
    cycleThrd.start();
    acceptThrd = new TAcceptorThread(TTransport.UDP);
    acceptThrd.start();
    acceptThrd = new TAcceptorThread(TTransport.TCP);
    acceptThrd.start();
    this.registerStockProperties();
    getAliasListFromFile();
  }
  public long getServerStartTime()
  {
    return gSrvStartupTime;
  }
  public int[] getServerStats()
  {
    //TODO: put in the cyclecounts logic ...
    gSrvStatsBuffer[0] = 0; // ave busy time
    gSrvStatsBuffer[1] = 0; // gCycleCounts
    gSrvStatsBuffer[2] = 0; // gMaxCycleCounts
    gSrvStatsBuffer[3] = gSingleLinkCount;
    gSrvStatsBuffer[4] = gClientMisses;
    gSrvStatsBuffer[5] = gClientReconnects;
    gSrvStatsBuffer[6] = gClientRetries;
    gSrvStatsBuffer[7] = gContractMisses;
    gSrvStatsBuffer[8] = gContractDelays;
    gSrvStatsBuffer[9] = gBurstLimitReachedCount;
    gSrvStatsBuffer[10] = (int)gDataTimeStampOffset * 1000;
    
    return gSrvStatsBuffer;
  }
  public long getAppCompilationTime()
  {
    return gAppCompilationTime;
  }
  public String getAppVersion()
  {
    StringBuffer appv = new StringBuffer(32); 
    appv.append("" + gAppVersionMajor);
    if (gAppVersionMinor < 10)  appv.append(".0" + gAppVersionMinor);
    else appv.append("." + gAppVersionMinor);
    if (gAppVersionRevision < 10) appv.append(".000" + gAppVersionRevision);  
    else if (gAppVersionRevision < 100) appv.append(".00" + gAppVersionRevision);  
    else if (gAppVersionRevision < 1000) appv.append(".0" + gAppVersionRevision);  
    else appv.append("." + gAppVersionRevision);
    return appv.toString();  
  }
  public int getAppVersionMajor() { return gAppVersionMajor; }
  public int getAppVersionMinor() { return gAppVersionMinor; }
  public int getAppVersionRevision() { return gAppVersionRevision; }
  public void setAppVersionMajor(int major) { gAppVersionMajor = major; }
  public void setAppVersionMinor(int minor) { gAppVersionMinor = minor; }
  public void setAppVersionRevision(int revision) { gAppVersionRevision = revision; }
  public static TEquipmentModuleFactory getInstance()
  {
    if (instance == null) instance = new TEquipmentModuleFactory();
    return instance;
  }
  public int setFecName(String name)
  {
    if (name.length() <= TStrings.FEC_NAME_SIZE)
    {
      gFecName = new String(name);
    }
    else
    {
      gFecName = new String(name.substring(0, TStrings.FEC_NAME_SIZE));
      TFecLog.log("FEC name "+name+" too long; truncating to "+gFecName);
      TStrings.validateFecName(gFecName);
    }
    return 0;
  }
  public String getFecName() { return gFecName; }
  public int setFecContext(String context)
  {
    TStrings.validateContextName(context);
    gFecContext = new String(context);
    return 0;
  }
  public String getFecContext() { return gFecContext; }
  public void setFecDescription(String description) { if (description != null) gFecDescription = description; }
  public String getFecDescription() { return gFecDescription; }
  public void setFecLocation(String location) { if (location != null) gFecLocation = location; }
  public String getFecLocation() { return gFecLocation; }
  public void setFecSubsystem(String subsystem) 
  { 
    if (subsystem != null) 
    {
      TStrings.validateSubSystemName(subsystem);
      gFecSubsystem = subsystem; 
    }
  }
  public String getFecSubsystem() { return gFecSubsystem; }
  public void setFecHardware(String hardware) { if (hardware != null) gFecHardware = hardware; }
  public String getFecHardware() { return gFecHardware; }
  public void setFecResponsible(String responsible) { if (responsible != null) gFecResponsible = responsible; }
  public String getFecResponsible() { return gFecResponsible; }
  public long getLastAccessTime() { return gLastWriteAccessTime; }
  public String getLastAccessUser() { return gLastWriteAccessUser; }
  public InetAddress getLastAccessAddress() { return gLastWriteAccessAddr; }
  public String getLastAccessEqm() { return gLastWriteAccessEqm; }
  public String getLastAccessPrp() { return gLastWriteAccessPrp; }
  public String getLastAccessDev() { return gLastWriteAccessDev; }
  public String makeModuleName(String suggestedModuleName)
  {
    String eqmName;
    int n;
    eqmName = suggestedModuleName;
    if (eqmName.length() > 6)
    {
      eqmName = eqmName.substring(0,6);
      TFecLog.log("Module Name "+suggestedModuleName+" too long; truncating to "+eqmName);
    }
    TStrings.validateLocalName(eqmName);
    for (n=0; n < numEqmTableEntries && eqmTable[n].getModuleName().compareTo(eqmName) != 0; n++);
    if (n == numEqmTableEntries) return eqmName;
    DbgLog.log("makeModuleName","equipment module name conflict !");
    DbgLog.log("makeModuleName","Please use a different name for " + suggestedModuleName + " and try again!");
    return null;
  }
  public int addEquipmentBackgroundTask(TEquipmentBackgroundTask tbkg)
  {
    TBackgroundThread tbt = new TBackgroundThread(tbkg);
    if (tbt == null) return -1;
    TFecLog.log("add Equipment Background task " + tbkg.getClass().getSimpleName());
    tbkg.startup();
    tbt.start();
    gNumBkgTasks++;
    return 0;
  }
  public TEquipmentModule getEquipmentModule(String EqmName)
  {
    String eqm;
    for (int i = 0; i<numEqmTableEntries; i++)
    {
      if ((eqm=eqmTable[i].getLocalName()) == null) continue; // should not be possible
      if (eqm.compareTo(EqmName) == 0) return eqmTable[i];
    }
    return null;
  }
  public int addEquipmentModule(TEquipmentModule teqm)
  {
    int i;
    for (i = 0; i < numEqmTableEntries; i++)
    {
      if (eqmTable[i] == teqm) return i;
    }
    if (i < maxNumEqmTableEntries)
    {
      TFecLog.log("add Equipment Module (" + i + ") " + teqm.getClass().getSimpleName());
      numEqmTableEntries++;
      eqmTable[i] = teqm;
      if (teqm.getLocalName().length() > 0) registrationPending = true;
      return i;
    }
    return -1;
  }
  private int maxDeadTime(short mode, int pollingRate)
  {
    if (mode == TMode.CM_REFRESH) return HEARTBEAT + pollingRate;
    return 2 * pollingRate;
  }
  /**
   * Geht alle Contracts durch und schickt Daten an die Clients. Wird aufgerufen
   * aus TEquipmentModuleFactory.systemSrvCycle. TODO Unklar, was im Einzelnen passiert.
   * 
   * @return
   */
  private byte[] deliveryPayload = new byte[TPacket.MAX_DATAGRAM_SIZE];
  private int doDelivery()
  {
    int i, k, n, extptr, extsize, msgsize, stepsize, packetmtu,ipos;
    int fsize, dsize, dptr;
    boolean passed = false, snddata = false, actionTaken = false, isLegacy = false;
    int totalsize = 0, basemode;
    TContractTable tct = null;
    TClientEntry tce = null;
    TClient tc = null;
    TPHdr pHdr = null;
    //Arrays.fill(deliveryPayload, (byte)0);
    byte[] payload = deliveryPayload;
    long t_now = System.currentTimeMillis();
    int pkts_sent = 0;
    int hdrsize, extlimit;
    int npkts=0,ncnts=0;
    TContractTable[] ctbl = null;
    if (debugLevel > 0 && clnTable.size() == 0)
        DbgLog.log("doDelivery","no client table in deliver data !");
    for (k=0, ipos=0; k < clnTable.size(); k++)
    { // run thru the client table; collect outgoing data for each
      if ((tc = (TClient) clnTable.get(k)) == null) continue;
      isLegacy = tc.tineProtocol < 6;
      hdrsize = isLegacy ? TPHdr.hdrSizeInBytesLegacy : TPHdr.hdrSizeInBytes;
      extlimit = isLegacy ? TStrings.STATUS_SHORTSIZE : TStrings.STATUS_SIZE;
      passed = true;
      totalsize = 2;
      ctbl = tc.getContractList();
      if (ctbl == null || ctbl.length == 0)
      {
        MsgLog.log("doDelivery","remove dangling client " + tc.userName,TErrorList.non_existent_client,null,1);
        removeClient(k--);
        continue;
      }      
      for (i=ipos, n=0; i < ctbl.length && n < tc.ncontracts; i++)
      { // run thru this client's contract list 
        tct = ctbl[i];
        Iterator<TClientEntry> li = tct.clt.iterator();
        boolean found = false;
        while (li.hasNext())
        {
          tce = (TClientEntry) li.next();
          if (tc.equals(tce.cln)) { found = true; break; }
        }
        if (!found)
        { // no match
          tct.pending = 0;
          continue;
        }
        if (debugLevel > 1) DbgLog.log("doDelivery","delivering data");
        if (tc.inetProtocol == TTransport.TCP && tc.output == null) tce.sts.counter = 0;
        n++;
        snddata = true;
        extptr = extsize = 0;
        basemode = tce.sts.mode;
        if (tce.sts.Stale != 0)
        { // marked as stale
          actionTaken = true;
          stepsize = hdrsize;
          packetmtu = 2 + stepsize + tce.sts.mtu;
          snddata = (tce.sts.statusCode == 0) || (tce.sts.statusCode & TErrorList.CE_SENDDATA) != 0;
          if (snddata)
          {
            dsize = tce.sts.nextDataSize;
            dptr = tce.sts.nextDataPtr;
          }
          else
          {
            extsize = extlimit;
            dsize = extsize;
            dptr = 0;
          }
          if ((msgsize = dsize + stepsize) > packetmtu)
          {
            tce.sts.counter = 0;
            continue;
          }
          //if (msgsize % 2 != 0) msgsize++; // make sure of even boundaries
          if ((totalsize + msgsize) > packetmtu)
          { // msgsize too big to fit in this packet !
            passed = false;
            ipos = i;
            break;
          }
          else
          {
            ncnts++;
            pHdr = new TPHdr(msgsize, tce, tct);
            System.arraycopy(pHdr.dBuffer.toByteArray(), 0, payload, totalsize, stepsize);
            fsize = TFormat.formatSizeOf(tct.contract.dataFormatOut);
            if (snddata && fsize > 0)
            {
              if ((tce.sts.statusCode & 0xfff) != 0)
              {
                extsize = (int) (tce.sts.nextDataPtr + tce.sts.nextDataSize - tct.contract.dataSizeOut * fsize);
                // shouldn't be possible ...
                if (extsize > extlimit) extsize = extlimit;
                if (extsize > dsize) /* err string remainder */
                {
                  extsize = dsize;
                  extptr = extlimit - extsize;
                }
                if (extsize > 0) dsize -= (short)extsize; else extsize = 0;
              }
              if (dsize > 0)
              {
                byte[] b = tct.dout.getDataBuffer();
                int thissize = b.length - dptr;
                if (thissize < 0) thissize = 0;
                if (dsize > thissize) dsize = thissize;
                if (debugLevel > 1)
                  DbgLog.log("doDelivery","dout size :" + b.length + " payload size : " + payload.length
                    + "  data size : " + dsize);
                try
                {
                  if (b != null && dsize > 0) 
                    System.arraycopy(b, dptr, payload, totalsize + stepsize, dsize);
                }
                catch (ArrayIndexOutOfBoundsException e)
                {
                  e.printStackTrace();
                  MsgLog.log("doDelivery","b len " + b.length + " dptr " + dptr + 
                    " payload " + payload.length + " offset " + totalsize + stepsize + " size " + dsize,TErrorList.dimension_error,e,0);
                }
              }
            }
          }
          if (passed)
          {
            totalsize += msgsize;
            if (extsize > 0) 
            { // insert err string
              byte b[] = null;
              if (tce.sts.statusCode == TErrorList.server_redirection)
              { // the property handler has handled the call and is returning this information!
                gLastErrorString = tct.compString;
                tce.sts.counter = 1;
                b = redirectionStringToBytes(tct.compString);
              }
              else if ((tce.sts.statusCode & 0xfff) < TErrorList.errorString.length)
              {
                gLastErrorString = TErrorList.getErrorString(tce.sts.statusCode & 0xff);
                b = gLastErrorString.getBytes();
              }
              else if (tct.compString != null)
              {
                gLastErrorString = tct.compString;
                b = gLastErrorString.getBytes();
              }
              if (b == null) b = "error unknown".getBytes();
              // only copy size of gLastErrorString (not extsize !)             
              Arrays.fill(payload, totalsize-extsize, totalsize-1, (byte)0);
              System.arraycopy(b, extptr, payload, totalsize - extsize,b.length);
            }
            if (debugLevel > 2)
            {
              String dbgstr = "" + totalsize + " bytes to " + tce.cln.userName + 
                " (" + tce.sts.counter + ") <" + tce.sts.statusCode + ">";
              DbgLog.log("doDelivery",dbgstr);
            }
            if (!snddata || tce.sts.numblks == tce.sts.blknum)
            { // datablock finished and/or error code
              if (debugLevel > 2) DbgLog.log("doDelivery","data delivery finished");
              tce.sts.Stale = 0;
              if (tce.sts.counter > 0) tce.sts.counter--;
              if (tce.sts.mode == TMode.CM_CANCEL) tce.sts.counter = 0;
              tce.sts.blknum = 1;
              tce.sts.nextDataPtr = 0;
              tce.sts.BytesRemaining = tce.sts.BytesTotal;
              tct.pending--;
              if ((t_now - tce.sts.lasttime) > maxDeadTime(tce.sts.mode,tce.sts.PollingRate))
              {
                tce.sts.misses++;
                gClientMisses++;
              }
            }
            else
            { // just send out this piece ...
              tce.sts.blknum++;
              tce.sts.nextDataPtr += tce.sts.nextDataSize;
              tce.sts.BytesRemaining = tce.sts.BytesTotal - tce.sts.nextDataPtr;
              gStaleData++;
            }
            tce.sts.nextDataSize = tce.sts.mtu < tce.sts.BytesRemaining ? 
                                   tce.sts.mtu : tce.sts.BytesRemaining;
          }
        }
        if (passed && tce.sts.counter <= 0)
        {
          if (debugLevel > 0)
            DbgLog.log("doDelivery","remove " + tce.cln.userName + " from contract "
                + tct.contract.eqmProperty);
          actionTaken = true;
          tce.cln.ncontracts--;
          tce.cln.rmvContract(tct);
          tct.clt.remove(tce);
          tct.nclients--;
          n--;
          if (debugLevel > 1)
            DbgLog.log("doDelivery"," contract " + tct.contract.eqmProperty + " now has "
                + tct.nclients + " clients");
        }
        if (passed && tct.nclients == 0)
        {
          actionTaken = true;
          if (gRetardSingleContractRemoval && basemode == TMode.CM_SINGLE)
          {
            if (debugLevel > 1)
              DbgLog.log("doDelivery","mark contract " + tct.contract.eqmProperty + " as expired");
            tct.expired = true;
            tct.contract.dataAccess = 
              TAccess.removeBits(tct.contract.dataAccess,(short)(TAccess.CA_FIRST | TAccess.CA_LAST));
            numExpiredContracts++;
          }
          else
          {
            if (debugLevel > 1) DbgLog.log("doDelivery","remove contract " + tct.contract.eqmProperty);
            removeContract(tct);
          }
        }
      }
      if (totalsize > 2 && tc != null)
      { // something ready to send out
        actionTaken = true;
        //System.arraycopy(Swap.ShortToBytes(totalsize), 0, payload, 0, 2);
        // this is quicker :
        payload[0] = (byte) (totalsize & 0xff);
        payload[1] = (byte) (totalsize >> 8);
        sendToPeer(tc, payload, totalsize);
        pkts_sent++;
        npkts++;
        if (debugLevel > 2) DbgLog.log("doDelivery","client "+k+" sent packet "+npkts+" ncontracts: "+ncnts);
      }
      ncnts=0;
      if (passed && tc.ncontracts == 0)
      {
        if (debugLevel > 1) DbgLog.log("doDelivery","remove client " + tc.userName);
        removeClient(k--);
      }
      if (!passed) 
      { // still data to send to this client => back up (cancel the increment on k)
        k--; 
      }
      else 
      { // finished with this client, reset counters ...
        ipos = 0;
        npkts = 0;
      }
    }
    if (gStaleData > 0) gStaleData--; else gStaleData = 0;
    if (!actionTaken || pkts_sent == 0) 
    {
      if (debugLevel > 2) 
        DbgLog.log("doDelivery",actionTaken ? "no packets sent in doDelivery()" : "empty pass through doDelivery()");
      gStaleData = 0;
    }
    if (debugLevel > 1) DbgLog.log("doDelivery","sent "+pkts_sent+" packets in "+(System.currentTimeMillis()-t_now)+" msecs");
    return gStaleData;
  }
  /**
   * Schickt Daten an einen Client. Wird aufgerufen aus deliverData().
   * 
   * @param tc
   * @param payload
   * @param length
   */
  private void sendToPeer(TClient tc, byte[] payload, int length)
  {
    if (debugLevel > 1)
        DbgLog.log("sendToPeer",tc.IPaddress.getHostAddress() + ":" + tc.port);
    try
    {
      switch (tc.inetProtocol)
      {
        case TTransport.UDP:
          srvp.dpOut = new DatagramPacket(payload, length, tc.IPaddress, tc.port);
          srvp.getSocket().send(srvp.dpOut);
          break;
        case TTransport.TCP:
          //if (!tc.sck.isConnected()) break;
          byte[] buf = new byte[length];
          System.arraycopy(payload,0,buf,0,length);
          //tc.output.write(payload,0,length);
          //tc.output.write(buf); // more efficient ?
          tc.output.write(buf); // more efficient ?
          //tc.sck.getOutputStream().write(buf);
          break;
      }
    }
    catch (Exception e)
    {
      DbgLog.log("sendToPeer",e.toString());
      // TODO: TCP -> close the socket if the endpoint has closed !!
    }
  }
  public void removeClient(int i)
  {
    synchronized (clnTable)
    {
      clnTable.remove(i);
      serverStatistics.setNumTotalClients((short) clnTable.size());
    }
  }
  public void removeContract(int i)
  {
    synchronized (conTable)
    {
      conTable.remove(i);
    }
  }
  public void removeContract(TContractTable tct)
  {
    synchronized (conTable)
    {
      conTable.remove(tct);
      serverStatistics.setNumTotalContracts((short) conTable.size());
    }    
  }
  public void addContract(TContractTable tct)
  {
    synchronized (conTable)
    {
      conTable.add(tct);
    }        
  }
  public void dumpContractTable()
  {
    System.out.println("\nCurrent Contract Table");
    synchronized (conTable)
    {
      TContractTable tct;
      for (int i=0; i<conTable.size(); i++)
      {
        tct = conTable.get(i);
        System.out.println(tct.toString());
      }
    }
    System.out.println("\n(inside scheduler : "+gIsInsideScheduler+")");
    System.out.println("");
  }
  public void dumpClientTable()
  {
    System.out.println("\nCurrent Client Table");
    synchronized (clnTable)
    {
      TClient tc;
      for (int i=0; i<clnTable.size(); i++)
      {
        tc = clnTable.get(i);
        System.out.println(tc.toString());
      }
    }
    System.out.println("");    
  }
  private void updateTables()
  {
    TContractTable tct = null;
    TClient tc = null;
    synchronized (clnTableAdd)
    { 
      if (clnTableAdd.size() > 0) synchronized (clnTable)
      { // add any new clients to the client table
        Iterator<TClient> li = clnTableAdd.iterator();
        while (li.hasNext())
        {
          tc = (TClient) li.next();
          clnTable.add(tc);
        }
        clnTableAdd.clear();
      }
    }
    serverStatistics.setNumTotalClients((short) clnTable.size());
    synchronized (conTableAdd)
    { // add any new contract to the contract table
      if (conTableAdd.size() > 0) synchronized (conTable)
      {
        Iterator<TContractTable> li = conTableAdd.iterator();
        while (li.hasNext())
        {
          tct = (TContractTable) li.next();
          conTable.add(tct);
        }
      }
      conTableAdd.clear();
    }
    serverStatistics.setNumTotalContracts((short) conTable.size());
    for (int i=0; i<clnTable.size(); i++)
    { // make sure this is also synchronized
      tc = (TClient) clnTable.get(i);
      tc.appendAddedContracts();
    }
  }
  private void assertClientTableConsistent(TContractTable tct)
  {
    try
    {
      if (tct.clt.size() == 0) return;
      Iterator<TClientEntry> li = tct.clt.iterator();
      TClientEntry tce;
      while (li.hasNext())
      {
        tce = (TClientEntry)li.next();
        for (int i=0; i<clnTable.size(); i++)
        {
          if (tce.cln.equals(clnTable.get(i))) return;
        }
        li.remove(); // wasn't found
        if (tct.nclients > 0) tct.nclients--;
      }
    }
    catch (Exception e) {};
  }
  private boolean gIsInsideScheduler = false;
  public boolean isInsideScheduler() { return gIsInsideScheduler; }
  private synchronized void doScheduler()
  {
    gIsInsideScheduler = true;
    try
    {
      TContractTable tct = null;
      TClientEntry tce = null;
      int cc = 0, datasize = 0, msec = 0, dss, dus;
      boolean isScheduled=false, isCanceled=false, stale=false, dirty=false, hasData=false, markedStale=false;
      long dts, delta_t = 0, t_now = System.currentTimeMillis(), t_stamp = 0;
      
      updateTables();
      
      for (int i = 0; i < conTable.size(); i++)
      {
        tct = (TContractTable) conTable.get(i);
        if (debugLevel > 3)
            DbgLog.log("doScheduler","looking at contract " + tct.contract.toString() + " with "
                + tct.nclients + " clients @ " + new Date(t_now).toString());
        if (tct.clt.size() == 0 && !tct.expired)
        {
          if (debugLevel > 3) DbgLog.log("doScheduler","contract has no clients");
          //if ((((short) tct.contract.dataAccess) & TAccess.CA_LAST) != 0)
          if (TAccess.isLast(tct.contract.dataAccess))
          { // going out of scope
            doContract(tct);
            removeContract(i--);
          }
          tct.expired = true;
          tct.pending = 0;
          continue;
        }
        assertClientTableConsistent(tct);
        isScheduled = false;
        if (tct.isStale) 
        {
          gStaleData++;
          markedStale = true;
        }
        cc = 0;
        datasize = tct.contract.getOutputDataSize();
        if ((delta_t = t_now - tct.lasttime) < 0) tct.lasttime = t_now;
        if (delta_t <= (long) tct.pollingRate - gSystemTick) continue;
        if (debugLevel > 2) DbgLog.log("doScheduler","contract needs attention : polling interval " + tct.pollingRate +  " last accessed : " + tct.lasttime);
        if (tct.expired)
        {
          if (delta_t > 3 * tct.pollingRate && !tct.isStale)
          {
            if (debugLevel > 1) DbgLog.log("doScheduler","removing expired contract");
            removeContract(i--);
            continue;
          }
          if (tct.clt.size() == 0) continue; // expired, no clients
        }
        if (delta_t < gMaxPollingRate && delta_t >= 2 * tct.pollingRate)
        {
          tct.misses++;
          gContractMisses++;
          if (debugLevel > 1)
            DbgLog.log("doScheduler","missed schedule on "+tct.contract.eqmProperty
              +" "+tct.contract.eqmDeviceName+" : "+delta_t+" vs "+tct.pollingRate);
        }
        if (tct.pending > 0 && clnTable.size() > 0)
        { // still sending out the last round
          if (debugLevel > 2) DbgLog.log("doScheduler","still sending out contract to clients");
          tct.delays++;
          gContractDelays++;
          if (gStaleData == 0) gStaleData++;
          Iterator<TClientEntry> li = tct.clt.iterator();
          while (li.hasNext())
          {
            tce = (TClientEntry) li.next();
            tce.sts.Stale = (byte) tct.pending;
          }
          continue; 
        }
        if (tct.starttime == 0) tct.starttime = t_now;
        t_stamp = t_now;
        msec = (int) (t_now % 100);
        // Synchronize on 100 msec boundaries if possible
        if (gSynchronizeContracts && (tct.pollingRate % 100 == 0)
            && msec < (100 - (int) gSystemTick)) t_stamp = t_now - msec;
        dts = gDataTimeStamp = t_now + gDataTimeStampOffset;
        dss = gSystemStamp;
        dus = 0;
        if (tct.nclients == 1)
        { // one client left 
          tce = (TClientEntry) tct.clt.element();
          if (tce.sts.counter == 1 && !tct.isRepeat)
          { // dangling client going out of scope ?
            tct.contract.dataAccess |= TAccess.CA_LAST;
          }
        }
        hasData = true; // be optimistic !
        if (debugLevel > 2) DbgLog.log("doScheduler","processing contract");
        switch (cc = doContract(tct))
        {
          case 0:
            break;
          case TErrorList.not_ready:
            // call returned not_ready or still not_signalled
            tct.contract.dataAccess = 
              TAccess.removeBits(tct.contract.dataAccess,TAccess.CA_LAST);
            tct.isRepeat = true;
            if (!gEqpFcnBusySemaphore
                || ((t_now - tct.starttime) < tct.pollingRate + 500 && gEqpFcnBusySemaphore))
            { // don't send out
              tct.setStale((byte) 0);
              tct.pending = 0;
              continue; 
            }
            break;
          case TErrorList.server_redirection:
            // the property handler is handling the call and has filled in tRedirectedServer
            tct.compString = new String(tRedirectedServer);
          default:
            if ((cc & 0xf000) != 0) 
            { // CE_RENEGOTIATE or CE_SENDDATA
              int rc = 0;
              if ((rc = tct.renegotiateContract((short) (cc & 0xfff))) != 0)
                cc = rc;
              else
                cc &= ~TErrorList.CE_RENEGOTIATE;
              datasize = tct.contract.getOutputDataSize();
            }
            else
            {
              hasData = false;
            }
        }
        if ((tct.contract.dataAccess & TAccess.CA_WRITE) == TAccess.CA_WRITE )
        {
          gLastWriteAccessTime = System.currentTimeMillis();
          gLastWriteAccessUser = ((TClientEntry)tct.clt.element()).cln.userName;
          gLastWriteAccessAddr = ((TClientEntry)tct.clt.element()).cln.IPaddress;
          gLastWriteAccessPrp = tct.contract.eqmProperty;
          gLastWriteAccessDev = tct.contract.eqmDeviceName;
          gLastWriteAccessEqm = tct.contract.eqmName;
        }
        if (hasData)
        {
          if (debugLevel > 3) DbgLog.log("doScheduler","Contract has data");
          dts = tct.drb.dTimestamp;
          dss = tct.drb.sysDataStamp;
          dus = tct.drb.usrDataStamp;
          if (datasize > 0 && !tct.dout.equals(tct.drb))
          { // Data have changed; the cached data are dirty
            tct.dout.copy(tct.drb);
            stale = dirty = true;
          }
          else
          {
            if (t_now - tct.lastmarked > (HEARTBEAT - 1) * 1000) stale = true;
            dirty = false; // Cached data have not changed
          }
        }
        if (cc != 0) stale = true;
        if (stale) tct.lastmarked = t_stamp; // Last marked as Stale
        if (debugLevel > 2)
            DbgLog.log("doScheduler","contract data is " + (stale ? "stale" : "not stale"));
        if (tct.lasttime == PRP_SCHEDULE_SIGNAL) isScheduled = true;
        if (tct.lasttime == PRP_CANCEL_SIGNAL) 
        {
          stale = true;
          isCanceled = true;
        }
        if (cc == TErrorList.not_ready)
        { // was 'not_ready' long enough ! -> send 'busy' back to caller
          cc = TErrorList.operation_busy;
        }
        else
        { // normal case, call is finished ...
          tct.starttime = 0;
          if (!isCanceled) tct.lasttime = t_stamp; // Last polled
          tct.compStatus = TErrorList.not_signalled;
        }
        tct.sysstamp = dss;
        tct.dtimestamp = dts;
        tct.usrstamp = dus;
        tct.returnCode = cc;
        Iterator<TClientEntry> li = tct.clt.iterator();
        while (li.hasNext())
        {
          tce = (TClientEntry) li.next();
          tce.sts.statusCode = (short) cc;
          if (isScheduled) tce.sts.lasttime = PRP_SCHEDULE_SIGNAL;
          if (cc != 0) tce.sts.Stale = 1; // stale because of error code
          if (tce.sts.mode == TMode.CM_REFRESH || tce.sts.mode == TMode.CM_EVENT)
          {
            if (stale) 
            { // stale because of data change or system heartbeat
              tce.sts.Stale = 1;
              // force acknowledgment
              if (dirty && gRequireAcknowledgments) tce.sts.counter = ACK_REQUEST; 
            }
          }
          else // honor polling rate
          { // stale because of polling rate ?
            if ((t_now - tce.sts.lasttime) > tce.sts.PollingRate - gSystemTick) tce.sts.Stale = 1; 
          }
          if (tce.sts.Stale > 0) // client for this contract is to be refreshed
          { 
            if (debugLevel > 2) DbgLog.log("doScheduler","refresh client with new data");
            tct.pending++;
            tce.sts.lasttime = t_stamp;
            if (gStaleData == 0) gStaleData++; // ensure pass thru doDelivery
          }
          if (tce.sts.mode > TMode.CM_SINGLE && tce.sts.counter == ACK_REQUEST) 
          { // get serious until an acknowledgment comes
            tce.sts.mode = ACK_PENDING; // reset by renewal
            tce.sts.PollingRate = tct.pollingRate;
          }
          if (tce.sts.mode == TMode.CM_CANCEL) 
          { // ensure pass thru doDelivery so that it's removed !
            gStaleData++;
          }
        }
      }
      if (debugLevel > 2) 
      {
        if (conTable.size() > 0)
          DbgLog.log("doScheduler","processed "+conTable.size()+" contracts in "+(System.currentTimeMillis()-t_now)+" msecs");
      }
      for (int i = 0; gStaleData > 0 && i < gBurstLimit; i++)
      { // send data to clients ...
        doDelivery();
      }
      if (gStaleData > 0) 
      { // caught by burst limit ...
        if (debugLevel > 2) DbgLog.log("doScheduler","deliverData() : burst limit exceeded");
        gBurstLimitReachedCount++;
      }
      if (markedStale)
      { // marked stale via getRequest : clear the flag
        for (int i = 0; i < conTable.size(); i++)
        {
          tct = (TContractTable) conTable.get(i);
          tct.isStale = false;
        }
      }
    }
    catch (Exception e)
    {
      TFecLog.log("doScheduler", "unhandled execption "+e.getMessage());
    }
    gIsInsideScheduler = false;
    return;
  }
  private boolean isWildcardContract(String deviceString)
  {
    if (deviceString == null) return false;
    return deviceString.indexOf('*') != -1;
  }
  private int doContractWithWildcardDevice(TContractTable tct)
  {
    TDataType drdb;
    int start = 0, len = 0;
    boolean isCont = false;
    int cc = TErrorList.not_allowed, i, k;
    TAccess acc = new TAccess(tct.contract.dataAccess);
    int devlistsize = 0;   
    String[] devlist = null;
    int atype;
    NAME16FI[] nfi;
    NAME16II[] nii;

    TEquipmentModule eqm = tct.eqm;
    String devprp = tct.contract.eqmProperty;
    String devnam = tct.contract.eqmDeviceName;
    TExportProperty prp = eqm.propertyList.getFirstProperty(devprp);
    if ((atype=prp.getDescription().getArrayType()) == TArrayType.AT_SPECTRUM) 
      return TErrorList.not_allowed;
    if (tct.dout == null || tct.dout.dArrayLength < 1) return TErrorList.not_allowed;
    if (!eqm.isDeviceSetLocal(devnam,devprp)) return TErrorList.data_not_local;

    if (prp.getDeviceList() != null)
    {
      devlistsize = prp.getDeviceList().size();
      devlist = new String[devlistsize];
      prp.getDeviceList().toArray(devlist);
    }
    if (devlistsize == 0)
    {
      devlist = eqm.getDeviceList().getDeviceNameList();
      //eqm.getDeviceNames().toArray(devlist);
      devlistsize = devlist.length;
    }
    if ((atype & TArrayType.AT_CHANNEL) == TArrayType.AT_CHANNEL)
    {
      len = devlistsize;
      int endpoints[] = StringToName.getContiguousEndpoints(devlist,devnam);
      if (endpoints != null)
      {
        isCont = true;
        start = endpoints[0];
        len = endpoints[1] - endpoints[0] + 1;
      }
    }
    WildcardMatch wc = new WildcardMatch(devnam);

    switch (tct.dout.dFormat)
    {
      case TFormat.CF_NAME16DBLDBL:
      case TFormat.CF_NAME32DBLDBL:
      case TFormat.CF_NAME64DBLDBL:
      case TFormat.CF_USTRING:
      case TFormat.CF_NAME16FI:
        if (isCont)
        {
          int dsiz = tct.contract.dataSizeOut > len ? len : tct.contract.dataSizeOut;
          nfi = new NAME16FI[dsiz];
          float[] frb = new float[dsiz];
          drdb = new TDataType(frb);
          cc = tct.eqm.callProperty(devprp, devlist[start], drdb, tct.din, acc);
          if (cc != 0) return cc;
          for (i=0; i<drdb.dArrayLength; i++)
          {
            nfi[i] = new NAME16FI(devlist[start+i],0,frb[i]);
          }
          tct.drb.dArrayLength = drdb.dArrayLength;
        }
        else
        {
          int dsiz = tct.contract.dataSizeOut;
          nfi = new NAME16FI[dsiz];
          float[] frb = new float[1];
          drdb = new TDataType(frb);
          for (i=0, k=0; i<devlistsize && k<tct.dout.dArrayLength; i++)
          {
            if (!wc.matches(devlist[i])) continue;
            cc = tct.eqm.callProperty(devprp, devlist[i], drdb, tct.din, acc);
            if (cc == TErrorList.server_redirection) return TErrorList.data_not_local;
            nfi[k] = new NAME16FI(devlist[i],cc,frb[0]);
            k++;
          }
          tct.drb.dArrayLength = k;
        }
        tct.drb.putData(nfi);
        return 0;
      case TFormat.CF_NAME16II:
        if (isCont) 
        {
          int dsiz = tct.contract.dataSizeOut > len ? len : tct.contract.dataSizeOut;
          nii = new NAME16II[dsiz];
          int[] irb = new int[dsiz];
          drdb = new TDataType(irb);
          cc = tct.eqm.callProperty(devprp, devlist[start], drdb, tct.din, acc);
          if (cc != 0) return cc;
          for (i=0; i<drdb.dArrayLength; i++)
          {
            nii[i] = new NAME16II(devlist[start+i],0,irb[i]);
          }
          tct.drb.dArrayLength = drdb.dArrayLength;
        }
        else
        {
          int dsiz = tct.contract.dataSizeOut;
          nii = new NAME16II[dsiz];
          int[] irb = new int[1];
          drdb = new TDataType(irb);
          for (i=0, k=0; i<devlistsize && k<tct.dout.dArrayLength; i++)
          {
            if (!wc.matches(devlist[i])) continue;
            cc = tct.eqm.callProperty(devprp, devlist[i], drdb, tct.din, acc);
            if (cc == TErrorList.server_redirection) return TErrorList.data_not_local;
            nii[k] = new NAME16II(devlist[i],irb[0],cc);
            k++;
          }
          tct.drb.dArrayLength = k;
        }
        tct.drb.putData(nii);
        return 0;
      default:
        if (tct.dout.dFormat <= TFormat.CF_NAME32 && devnam.compareTo("*") == 0)
        { /* primitive format allowed in this case */
          if (isCont) 
          {
            if (tct.dout.dArrayLength > len) tct.dout.dArrayLength = len;
            return tct.eqm.callProperty(devprp, devlist[0], tct.dout, tct.din, acc);
          }
          drdb = new TDataType(1,tct.dout.dFormat);
          for (i=0; i<devlistsize && i<tct.dout.dArrayLength; i++)
          {
            cc = tct.eqm.callProperty(devprp, devlist[i], drdb, tct.din, acc);
            if (cc != 0) break;
          }
          if (cc == TErrorList.server_redirection) return TErrorList.data_not_local;
          tct.drb.dArrayLength = i;
          tct.drb.putData(drdb);
          return cc;
        }
    }
    return cc;
  }
  private int doContractWithWildcardProperty(TContractTable tct)
  {
    LinkedList<String> lst = null;
    TDataType drdb;
    int cc = TErrorList.not_allowed, i, k, devnr;
    TAccess acc = new TAccess(tct.contract.dataAccess);
    int prplistsize = 0;   
    String[] prplist = null;
    NAME32DBLDBL[] ndd;
   
    TEquipmentModule eqm = tct.eqm;
    String devprp = tct.contract.eqmProperty;
    String devnam = tct.contract.eqmDeviceName;
    TExportProperty prp;
    if (tct.dout == null || tct.dout.dArrayLength < 1) return TErrorList.not_allowed;
    if ((devnr=eqm.getDeviceList().getDeviceNumber(tct.contract.eqmDeviceName)) < 0)
      return TErrorList.illegal_equipment_number;

    lst = eqm.getDeviceList().getDevice(devnr).getPropertyList();
    if (lst != null) 
    {
      prplistsize = lst.size();
      prplist = new String[prplistsize];
      lst.toArray(prplist);
    }
    else 
    {
      if (eqm.gPropertyNameList == null)
      { // just do this once
        if ((lst = eqm.getPropertyNames()) == null) return TErrorList.code_failure;
        prplistsize = lst.size();
        prplist = new String[prplistsize];
        lst.toArray(prplist);
        eqm.gPropertyNameList = prplist;
      }
      prplistsize = eqm.gPropertyNameList.length;
      prplist = eqm.gPropertyNameList;
    }
    
    WildcardMatch wc = new WildcardMatch(devprp);

    switch (tct.dout.dFormat)
    {
      case TFormat.CF_NAME16DBLDBL:
      case TFormat.CF_NAME32DBLDBL:
      case TFormat.CF_NAME64DBLDBL:
      case TFormat.CF_USTRING:
      case TFormat.CF_NAME16FI:
      case TFormat.CF_NAME16II:
        int dsiz = tct.contract.dataSizeOut;
        ndd = new NAME32DBLDBL[dsiz];
        double[] drb = new double[1];
        drdb = new TDataType(drb);
        for (i=0, k=0; i<prplistsize && k<tct.dout.dArrayLength; i++)
        {
          if (!wc.matches(prplist[i])) continue;
          prp = eqm.propertyList.getFirstProperty(prplist[i]);
          if (prp.getDescription().getArrayType() == TArrayType.AT_SPECTRUM) continue;
          cc = tct.eqm.callProperty(prplist[i], devnam, drdb, tct.din, acc);
          if (cc == TErrorList.server_redirection) return TErrorList.data_not_local;
          ndd[k] = new NAME32DBLDBL(prplist[i],drb[0],cc);
          k++;
        }
        tct.drb.dArrayLength = k;
        tct.drb.putData(ndd);
        return 0;
      default:
        break;
    }
    return cc;
  }
  
  private int doContract(TContractTable tct)
  {
    if (tct.expired) return tct.returnCode;
    int cc = TErrorList.non_existent_elem; // return code
    String devnam = new String(tct.contract.eqmDeviceName);
    String devprp = new String(tct.contract.eqmProperty);
    short lclAccess = tct.contract.dataAccess;
    if (tct.isRepeat) lclAccess |= TAccess.CA_REPEAT;
    TAccess acc = new TAccess(lclAccess);
    int datasize = tct.contract.getOutputDataSize();
    boolean isRegisteredProperty = false;
    int gate = 0, dsize = tct.contract.dataSizeOut;
    if (tct.contract.dataFormatOut == TFormat.CF_STRING)
      dsize = Integer.parseInt(tct.contract.dataTagOut);
    tct.dout.dArrayLength = tct.drb.dArrayLength = dsize;
    tct.drb.dTimestamp = gDataTimeStamp + gDataTimeStampOffset;
    tct.drb.sysDataStamp = gSystemStamp;
    if (debugLevel > 0) 
      DbgLog.log("doContract","process " + tct.contract.eqmName + "/" + 
        tct.contract.eqmDeviceName + "[" + tct.contract.eqmProperty + "]");
    currentContractEntry = tct;
    tct.compString = null;
    if (tct.contract.extStringSize > 0)
    { // long device name !
      devnam = tct.contract.getExtendedDeviceName();
    }
    if (devprp.indexOf("*") != -1) 
    {
      if (TFormat.isSimpleFormat(tct.contract.dataFormatOut))
      {
        devnam = devprp;
        devprp = new String("PROPERTIES");
      }
      else
      {
        isRegisteredProperty = true; // wildcard call 
      }
    }
    if (stockList.hasProperty(devprp))
    { // Check common stock properties
      if (debugLevel > 1) DbgLog.log("doContract","Common Stock Property found: " + tct.contract.eqmProperty);
      gDataTimeStamp = gSrvStartupTime + gDataTimeStampOffset;
      if (datasize > 0) tct.drb.dataFill(0);
      cc = stockList.callPropertyHandler(devprp, devnam, tct.drb, tct.din, acc);
    }
    else if (tct.eqm == null)
    {
      DbgLog.log("doContract","No Equipment Module found.");
      cc = TErrorList.non_existent_elem;
    }
    else if (tct.eqm.hasStockProperty(devprp))
    { // check stock properties of equipment module ...
      if (debugLevel > 1) DbgLog.log("doContract","Stock Property found: " + tct.contract.eqmProperty);
      gDataTimeStamp = gSrvStartupTime + gDataTimeStampOffset;
      if (datasize > 0) tct.drb.dataFill(0);
      cc = tct.eqm.callStockProperty(devprp, devnam, tct.drb, tct.din, acc);
    }
    else if (tct.eqm.propertyList.hasProperty(devprp) == false)
    { // not a registered property; is it a meta-property ?
      if ((gate=metaProps.getGate(tct.contract.eqmProperty)) > 0)
      {
        if (debugLevel > 1) DbgLog.log("doContract","Gated Property found: " + tct.contract.eqmProperty);
        devprp = metaProps.getTargetProperty(tct.contract.eqmProperty);
        isRegisteredProperty = true;
      }
      else
      {
        if (debugLevel > 1) DbgLog.log("doContract","Meta Property found: " + tct.contract.eqmProperty);
        cc = tct.eqm.callMetaProperty(devprp, devnam, tct.drb, tct.din, acc);
      }
    }
    else if (tct.compStatus == TErrorList.not_signalled && gEqpFcnBusySemaphore)
    { // eqm is busy doing something, but not this contract ...
      if (debugLevel > 1) DbgLog.log("doContract","equipment module is busy : " + tct.contract.eqmProperty);
      cc = TErrorList.operation_busy;
    }
    else 
    { // a registered property, give the 'go-ahead'
      if (debugLevel > 1) DbgLog.log("doContract","registered Property found: " + tct.contract.eqmProperty);
      devprp = tct.contract.eqmProperty;
      isRegisteredProperty = true;
    }
    if (isRegisteredProperty)
    { // Call Equipment module's properties
      if (tct.eqm.isIdleState())
      { // server is marked idle -> jump out
        gate = 0;
        cc = TErrorList.server_idle;
      }
      else if (isWildcardContract(devnam))
      {
        if (debugLevel > 1) DbgLog.log("doContract","wildcard device: "+tct.contract.eqmDeviceName+" ["+tct.contract.eqmProperty+"]");
        cc = doContractWithWildcardDevice(tct);
      }
      else if (isWildcardContract(devprp))
      {
        if (debugLevel > 1) DbgLog.log("doContract","wildcard property: "+ tct.contract.eqmProperty);
        cc = doContractWithWildcardProperty(tct);
      }
      else
      {
        if (debugLevel > 1) 
          DbgLog.log("doContract","call property handler for "+tct.contract.eqmName+"/"+tct.contract.eqmDeviceName+"["+tct.contract.eqmProperty+ "]");
        gEqpFcnBusySemaphore = true;
        cc = tct.eqm.callProperty(devprp, devnam, tct.drb, tct.din, acc);
        gEqpFcnBusySemaphore = false;
      }
      if (gate > 0)
      {
        gateOutputData(tct.drb,gate,metaProps.getIndex(tct.contract.eqmProperty) != TMetaProperties.GAT_ID);
      }
    }
    if (tct.eqm != null && tct.eqm.propertyList == null)
    {
      if (debugLevel > 1) DbgLog.log("doContract","Module has no property list");
    }
    if (tct.dout.dArrayLength > tct.drb.dArrayLength)
    {
      if (!gSystemPresetMemory) 
      { // zero out the remainder for later data comparisons
        tct.dout.dataFill(0, tct.drb.dArrayLength);
      }
      tct.dout.dArrayLength = tct.drb.dArrayLength;
    }
    tct.dout.dTimestamp = tct.drb.dTimestamp;
    tct.dout.sysDataStamp = tct.drb.sysDataStamp;
    tct.dout.usrDataStamp = tct.drb.usrDataStamp;
    if ((tct.contract.dataAccess & TAccess.CA_FIRST) != 0)
      tct.contract.dataAccess = TAccess.removeBits(tct.contract.dataAccess,TAccess.CA_FIRST);
    if (cc < 0) cc = TErrorList.ambiguous;
    if (cc != TErrorList.not_ready) tct.isRepeat = false;
    // record possible new data output length
    tct.dataSizeOut = tct.dout.dArrayLength;
    if (tct.contract.dataSizeOut != tct.dout.dArrayLength)
    {
      cc |= TErrorList.CE_RENEGOTIATE;
    }
    if (tct.contract.dataFormatOut == TFormat.CF_STRING) cc |= TErrorList.CE_RENEGOTIATE;
    tct.compStatus = cc;
    currentContractEntry = null;
    return cc;
  }
  private void gateOutputData(TDataType dout, int gate, boolean outputBoolean)
  {
    int i;
    Object hDataObject = dout.getDataObject();
    switch (dout.getFormat())
    {
      case TFormat.CF_BYTE:
      {
        byte[] data = (byte[])hDataObject; 
        for (i=0; i<dout.getArrayLength(); i++) 
        {
          data[i] &= gate;
          if (outputBoolean) data[i] = (byte)(data[i] == (byte)gate ? 1 : 0);
        }
        dout.putData(data);
        break;
      }
      case TFormat.CF_SHORT:
      {
        short[] data = (short[])hDataObject; 
        for (i=0; i<dout.getArrayLength(); i++) 
        {
          data[i] &= gate;
          if (outputBoolean) data[i] = (short)(data[i] == (short)gate ? 1 : 0);
        }
        dout.putData(data);
        break;
      }
      case TFormat.CF_LONG:
      {
        int[] data = (int[])hDataObject; 
        for (i=0; i<dout.getArrayLength(); i++) 
        {
          data[i] &= gate;
          if (outputBoolean) data[i] = (int)(data[i] == (int)gate ? 1 : 0);
        }
        dout.putData(data);
        break;
      }
    }
  }
  public static TClient getCaller()
  {
    TContractTable tct = getInstance().getCurrentContractEntry();
    return tct != null ? ((TClientEntry)tct.clt.element()).cln : null;
  }   
  public static TClientEntry[] getClientList()
  {
    TContractTable tct = getInstance().getCurrentContractEntry();
    return tct != null ? ((TClientEntry[])tct.clt.toArray(new TClientEntry[0])) : null;
  }
  private class clnInputData
  {
    TDataType din;
    TClient tc;
    int id;
    int starttime;
  }
  private boolean assertRedirectionValid(String rdr)
  {
    int indx = rdr.indexOf('/');
    String cntName = gFecContext, expName = null;
    switch (indx)
    {
      case -1: // just the server name ?
        expName = rdr; 
        break;
      case  0: // context + server name (+ device name)
        rdr = rdr.substring(1);
        if ((indx=rdr.indexOf('/')) == -1) break;
        cntName = rdr.substring(0,indx);
        rdr = rdr.substring(indx+1);
        indx = rdr.indexOf('/');
      default:
        expName = indx > 0 ? rdr.substring(0,indx) : rdr;
        break;
    }
    if (expName == null) return false;
    for (int i = 0; i<numEqmTableEntries; i++)
    {
      if (eqmTable[i].getContext().compareToIgnoreCase(cntName) != 0) continue;
      if (eqmTable[i].getExportName().compareToIgnoreCase(expName) != 0) continue;
      return false; // oops!  redirecting to myself !
    }
    return true;
  }
  private short isIllegalProperty(TClient tc, TSubscription ts,byte[] errorData)
  { // note: wildcard properties such as "PROP*" no longer allowed !
    TEquipmentModule eqm = LocateEquipmentModule(ts.contract);
    String tgtProperty = ts.contract.eqmProperty;
    String tgtDevice = ts.contract.eqmDeviceName;
    String rdirStr = null;
    if (eqm == null) 
    { // subscription must point to something being managed
      if (TStockProperties.isFecProperty(tgtProperty)) return 0;
      return TErrorList.non_existent_elem;
    }
    if (!eqm.hasInitialized) return TErrorList.not_initialized;
    if (ts.contract.eqmProperty.indexOf(".") >= 0)
    {
      tgtProperty = metaProps.getTargetProperty(ts.contract.eqmProperty);
    }
    // check if device is categorically redirected ...
    TDevice td = eqm.getDeviceList().getDevice(tgtDevice);
    if (td != null && (rdirStr=td.getRedirection()) != null)
    {
      byte[] rdb = redirectionStringToBytes(rdirStr);
      System.arraycopy(rdb,0,errorData,0,rdb.length);
      return TErrorList.server_redirection;      
    }
    // check if registered
    if (eqm.propertyList.hasProperty(tgtProperty) == true)
    { // check for property redirection
      TPropertyDescription dsc = eqm.propertyList.getFirstProperty(tgtProperty).getDescription();
      rdirStr = dsc.getRedirection();
      if (rdirStr != null && isDeviceRedirected(rdirStr,ts.contract.eqmDeviceName))
      { // registered as a redirected property !
        if (assertRedirectionValid(rdirStr))
        {
          byte[] rdb = redirectionStringToBytes(rdirStr);
          System.arraycopy(rdb,0,errorData,0,rdb.length);
          return TErrorList.server_redirection;      
        }
        else
        {
          dsc.setRedirection(null);
          TFecLog.log("redirection to "+rdirStr+" is the local process : rejected !");
        }
      }
    }
    else 
    { // check if stock
      if (tgtProperty.indexOf("*") == -1 && // mapped as "PROPERTIES"
          stockList.hasProperty(tgtProperty) == false &&
          eqm.stockList.hasProperty(tgtProperty) == false) return TErrorList.illegal_property;
    }
    if (((ts.contract.dataAccess & TAccess.CA_WRITE) == TAccess.CA_WRITE) &&
        tgtProperty.compareTo("MESSAGE") != 0)
    { // check if WRITE (and not MESSAGE)
      String cmdString = "COMMAND " + tgtProperty + " called by " + tc.userName + 
                         " from addr " + tc.IPaddress.getHostAddress() + " : ";
      if (!isMemberUsersList(eqm,tc.userName))
      {
        TFecLog.log(cmdString + "user refused access");
        return TErrorList.access_denied;
      }
      if (!isMemberControlNets(eqm,tc))
      {
        TFecLog.log(cmdString + "network refused access");
        return TErrorList.access_denied;
      }
      if (tgtProperty.compareTo("ACCESSLOCK") != 0 && !isMemberLockSet(eqm,tc))
      {
        TFecLog.log(cmdString + "access lock refused access");
        return TErrorList.access_denied;
      }
      if (gPutCommandsInFeclog) TFecLog.log(cmdString + "accepted");
    }
    return 0;
  }
  private short fixOutputDataTypeAndSize(TSubscription ts)
  {
    TEquipmentModule eqm = LocateEquipmentModule(ts.contract);
    if (eqm == null) return TErrorList.non_existent_elem;
    short df = ts.contract.dataFormatOut >= 0 ? 
               ts.contract.dataFormatOut : (short)(256 + ts.contract.dataFormatOut); 
    if (df == TFormat.CF_DEFAULT)
    { // caller wants default settings
      TExportProperty prop = eqm.propertyList.getFirstProperty(ts.contract.eqmProperty);
      ts.contract.dataFormatOut = (byte) prop.getOutputFormat();
      int fmtsize = TFormat.formatSizeOf(prop.getOutputFormat());
      int datasize = prop.getOutputSize() * fmtsize;
      int hdrsize = TFormat.getFormatHeaderSize(ts.contract.dataFormatOut);
      datasize += hdrsize;
      // EqpSizeOut is in bytes ...
      if (ts.contract.dataSizeOut < datasize) return TErrorList.buffer_too_small;
      ts.contract.dataSizeOut = datasize;
      if (fmtsize != 0)
      {
        ts.contract.dataSizeOut -= hdrsize;
        ts.contract.dataSizeOut -= ts.contract.dataSizeOut % fmtsize;
        ts.contract.dataSizeOut /= fmtsize;
      }
    }
    return 0;
  }
  private void shrinkInputDataTypeRepository()
  {
    clnInputData cip = null;
    for (int i = 0; i < clnInputDataTable.size(); i++)
    {
      cip = (clnInputData) clnInputDataTable.get(i);
      if (cip.din.blksin == cip.din.numblks)
      {
        clnInputDataTable.remove(i--);
      }
    }
  }
  private TDataType getInputDataTypeRepository(TClient tc, TSubscription ts)
  {
    int sizeIn = ts.contract.dataSizeIn;
    String tagIn = ts.contract.dataTagIn;
    if (ts.contract.dataFormatIn == TFormat.CF_STRING)
    {
      sizeIn = Integer.parseInt(tagIn);
      tagIn = "";
    }
    if (ts.numblks == 1) 
    { // just this one ...
      TDataType tdt;
      tdt = new TDataType(sizeIn, ts.contract.dataFormatIn);
      tdt.setTag(tagIn);
      return tdt;
    }
    if (debugLevel > 1) DbgLog.log("getInputDataTypeRepository","in-coming long datagram ...");
    clnInputData cip = null;
    for (int i = 0; i < clnInputDataTable.size(); i++)
    {
      cip = (clnInputData) clnInputDataTable.get(i);
      if (!cip.tc.IPaddress.equals(tc.IPaddress)) continue;
      if (cip.id != ts.id) continue;
      if (cip.starttime != ts.starttime) continue;
      if (debugLevel > 1) DbgLog.log("getInputDataTypeRepository","found it in list ...");
      return cip.din;
    }
    cip = new clnInputData();
    cip.id = ts.id;
    cip.starttime = ts.starttime;
    cip.tc = tc;
    cip.din = new TDataType(sizeIn, ts.contract.dataFormatIn);
    cip.din.setTag(tagIn);
    clnInputDataTable.add(cip);
    if (debugLevel > 1) DbgLog.log("getInputDataTypeRepository","add it to list ...");
    return cip.din;
  }
  public String getNameFromAlias(String alias)
  {
    if (gAliasList == null) return null;
    // TODO: is this faster than a hash map?
    AliasTableEntry ate;
    for (int i=0; i<gAliasList.size(); i++)
    {
      ate = (AliasTableEntry)gAliasList.get(i);
      if (ate.alias.compareToIgnoreCase(alias) == 0) return ate.name;
    }
    return null;
  }
  public String getAliasFromName(String name)
  { // TODO: as is, this only allows 1 alias per name
    if (gAliasList == null) return null;
    // TODO: is this faster than a hash map?
    AliasTableEntry ate;
    for (int i=0; i<gAliasList.size(); i++)
    {
      ate = (AliasTableEntry)gAliasList.get(i);
      if (ate.name.compareToIgnoreCase(name) == 0) return ate.alias;
      if (metaProps.isMetaProperty(ate.name))
      { // there's an alias for a meta property ...
        if (metaProps.getTargetProperty(ate.name).compareToIgnoreCase(name) == 0)
          return ate.alias;
      }
    }
    return null;
  }
  private static boolean[] isNewContract = new boolean[1];
  private byte[] requestBytes = new byte[256];
  private void getRequest(TClient tc, byte[] data)
  {
    Arrays.fill(requestBytes,(byte)0);
    byte[] d = requestBytes; //new byte[256];
    TDataType din = null;
    int offset = 0, datasize = 0;
    int renewalMultiplier;
    int cc = 0;
    boolean isTrueMCast = false;
    boolean isLegacy = false;
    TClientEntry tce = null;
    String astr = null;
    // peel off the Theader
    TReqHdr hdr = new TReqHdr(data, 0, TReqHdr.hdrSizeInBytes);
    tc.userName = hdr.getUserName();
    int tineProtocol = hdr.getProtocol();    
    if (tineProtocol < 5) 
    {
      SendMessageToCaller(tc, null, TErrorList.illegal_protocol,null);
    }
    tc.tineProtocol = (short)tineProtocol;
    if (tineProtocol < 6) isLegacy = true; 
    int totalsize = hdr.getTotalSizeInBytes();
    int hdrSize = isLegacy ? TSubscription.hdrSizeInBytesLegacy : TSubscription.hdrSizeInBytes;
    int conSize = isLegacy ? TContractP5.hdrSizeInBytes : TContract.hdrSizeInBytes; 
    offset += TReqHdr.hdrSizeInBytes;
    if (debugLevel > 3) DbgLog.log("getRequest",hdr.toString());
    while (offset < totalsize)
    { // peel off the TSubscription
      TSubscription sub = new TSubscription(tineProtocol, data, offset, hdrSize);
      // Is the caller using alias names ?
      if ((astr=getNameFromAlias(sub.contract.eqmDeviceName)) != null) sub.contract.eqmDeviceName = astr;
      if ((astr=getNameFromAlias(sub.contract.eqmProperty)) != null) sub.contract.eqmProperty = astr;
      offset += hdrSize + conSize + sub.contract.extStringSize;
      if (debugLevel > 3) DbgLog.log("getRequest",sub.toString());
      // prepare incoming data ...
      if ((datasize = sub.getInputDataSize()) > 0)
      {
        din = getInputDataTypeRepository(tc, sub);
        din.update(data, offset, datasize, sub.blknum, sub.blkid);
        din.bytesin += datasize;
        offset += datasize + (datasize % 2);
        if (din.blksin < din.numblks) continue;
        // when finished ...
        din.getData();
        if (debugLevel > 3) DbgLog.log("getRequest","input data : \n" + din.toString());
      }
      // requested output data ...
      fixOutputDataTypeAndSize(sub); // default format ?
      // is contract in list ? if no -> add to list
      TContractTable tct = LocateContractInList(sub.contract, din, isNewContract);
      // finished with TDataType repository entry ?
      shrinkInputDataTypeRepository();
      if (tct == null)
      { // have I got a contract table entry ?
        SendMessageToCaller(tc, sub, TErrorList.resources_exhausted, null);
        continue;
      }
      if (gServerWaiting)
      { // is the server flagged as 'waiting?'
        SendMessageToCaller(tc, sub, TErrorList.not_initialized, d);
        continue;        
      }
      if ((cc = isIllegalProperty(tc, sub, d)) != 0)
      { // can caller access the property ?
        SendMessageToCaller(tc, sub, cc, d);
        continue;
      }
      // is a new link attaching to an expired contract ?
      if (tct.expired && ((sub.mode & TMode.CM_RETRY) == 0))
      {
        tct.lasttime = 0;
        tct.expired = false;
        isNewContract[0] = true;
        if (numExpiredContracts > 0) numExpiredContracts--;
        if (debugLevel > 1) 
          DbgLog.log("getRequest",tct.contract.eqmName+"/"+tct.contract.eqmDeviceName+"["+tct.contract.eqmProperty+ "] attaching to expired contract");
      }
      // is broadcast or multicast request ?
      if ((sub.mode & (TMode.CM_BCAST | TMode.CM_MCAST)) != 0)
      {
        isTrueMCast = (sub.mode & TMode.CM_MCAST) != 0;
        sub.mode = TMode.getBaseMode(sub.mode); // strip off
        if (sub.mode == TMode.CM_CANCEL) continue; // don't do anything ...
        if (isTrueMCast || isMemberBcastNets(tc))
        {
          // send subscription ID to caller
          ByteArrayOutputStream boas = new ByteArrayOutputStream(2);
          DataOutputStream dos = new DataOutputStream(boas);
          try
          {
            dos.writeShort((short) tct.contractID);
          }
          catch (Exception e) {}
          byte[] subid = boas.toByteArray();
          Swap.Bytes(subid, TFormat.CF_SHORT);
          SendMessageToCaller(tc, sub, TErrorList.get_subscription_id, subid);
          // add network to client list and replace cln
          tc.IPaddress = isTrueMCast ? getMCastAddr() : getBCastAddr(tc.IPaddress);
          tc.port = initializer.getMCastPort();
          tc = LocateClientInList(tc.IPaddress, tc.port);
          tc.userName = "NETWORK";
          sub.id = (short) tct.contractID;
          sub.starttime = BCAST_ID;
        }
      }
      // is client in contractList?
      tce = LocateClientInContract(tct, tc);
      if (tce == null)
      {
        SendMessageToCaller(tc, sub, TErrorList.resources_exhausted, null);
        continue;
      }
      // fill in renewed fields ...
      tce.sts.reset(sub);
      renewalMultiplier = 1;
      if (sub.pollingRate < 500) renewalMultiplier = 3;
      // Force Contract execution
      if (tce.sts.mode != ACK_PENDING)
      {
        if ((tce.sts.counter == 0 && sub.starttime != BCAST_ID) || sub.mode == TMode.CM_REFRESH)
        {
          tce.sts.Stale = 1;
          tce.sts.lasttime = System.currentTimeMillis();
        }
        // update counters ? ...
        if (sub.mode != TMode.CM_CANCEL && sub.mode != TMode.CM_REFRESH
            && tce.sts.counter > ACK_REQUEST * renewalMultiplier
            && tce.sts.counter < RENEWAL * renewalMultiplier)
        {
          tce.sts.reconnects++;
          gClientReconnects++;
        }
      }
      // is retry ?
      if ((sub.mode & TMode.CM_RETRY) != 0)
      { // if expired, DeliverData() will send the last payload again
        tce.sts.Stale = 1;
        tct.isStale = true;
        tce.sts.isRetry = true;
        gClientRetries++;
        sub.mode = TMode.getBaseMode(sub.mode);
        if (debugLevel > 1) 
          DbgLog.log("getRequest",tct.contract.eqmName+"/"+tct.contract.eqmDeviceName+"["+tct.contract.eqmProperty+ "] is retrying reqeust");
      }
      if (tct.nclients == 0) 
      { // just turned off, turn it back on (client must have still been in the local list)
        tct.nclients++;
        tct.contract.dataAccess = 
          TAccess.removeBits(tct.contract.dataAccess,TAccess.CA_LAST);
      }
      tct.contract.dataAccess = 
        TAccess.removeBits(tct.contract.dataAccess,TAccess.CA_REPEAT);// don't allow this from the caller
      if (isNewContract[0]) tct.contract.dataAccess |= TAccess.CA_FIRST; // turn on the FIRST bit
      // switch on mode
      int SubsciptionPacketOffset = 0; // for staggering the POLLING guys ...
      switch (tce.sts.mode = sub.mode)
      {
        case TMode.CM_SINGLE:
          gSingleLinkCount++;
          tce.sts.counter = 1;
          if (!(tct.expired && tce.sts.isRetry))
          { // normal case ...
            tct.lasttime = PRP_REQUEST_SIGNAL;
          }
          else  
          { // a RETRY marked as expired should not be called
            tct.lasttime = System.currentTimeMillis();
            if (tct.dataSizeOut < tct.contract.dataSizeOut)
              tct.renegotiateContract((short)tct.compStatus);
          }
          if (tct.pollingRate == 0) tct.pollingRate = tce.sts.PollingRate;
          if (tct.nclients == 1)
          { // turn on LAST bit
            tct.contract.dataAccess |= TAccess.CA_LAST;
          }
          break;
        case TMode.CM_POLL:
          SubsciptionPacketOffset = (int)(100.0 * Math.random());
          if (sub.pollingRate > 2000) SubsciptionPacketOffset /= 10; 
        case TMode.CM_REFRESH:
        case TMode.CM_EVENT:
          tce.sts.counter = (short) (RENEWAL * renewalMultiplier + SubsciptionPacketOffset);
          tct.pollingRate = findpoll(sub.pollingRate, tct.pollingRate);
          break;
        default:
        case TMode.CM_CANCEL:
          if (tce.cln.userName.compareTo("NETWORK") == 0) break;
          if (tct.nclients == 1)
          {
            if (isNewContract[0])
            { // canceling a brand new contract makes no sense !
              SendMessageToCaller(tc, sub, TErrorList.invalid_link, null);
              continue;              
            }
            if (tce.sts.counter > 0) // was active (turn on the LAST bit
                tct.contract.dataAccess |= TAccess.CA_LAST;
            tct.lasttime = PRP_CANCEL_SIGNAL;
            tct.pending = 0;
          }
          tce.sts.counter = 0;
          break;
      }
    }
  }
  private void validateContextAndSubsystem(TEquipmentModule eqm)
  {
    String ctx = eqm.getContext();
    String sub = eqm.getSubsystem();
    String s;
    int idx;
    if (ctx == null) 
    {
      eqm.setCompletionString("TEST");
      TFecLog.log("Context not given: set to \"TEST\"");
      return;
    }
    if ((idx=ctx.indexOf('.')) == -1) return;
    s = ctx.substring(idx+1);
    if (s.endsWith("TEST")) return;
    if (sub.length() == 0 || sub.compareToIgnoreCase(s) == 0)
    {
      sub = s;
      ctx = ctx.substring(0, idx);
    }
    if (sub.length() == 0) sub = gFecSubsystem;
    if (ctx.compareTo(eqm.getContext()) != 0)
    {
      TFecLog.log("given context " + eqm.getContext() + 
          " will be treated as the " + sub + " subsystem decorated context " + ctx);
    }
  }
  private void SendRegisteredExportToENS(TEquipmentModule eqm)
  {
    SendRegisteredExportToENS(eqm,false);
  }
  private void SendRegisteredExportToENS(TEquipmentModule eqm,boolean isMaster)
  {
    if (eqm == null) return;
    if (eqm.getLocalName() == null || eqm.getLocalName().length() == 0) 
    {
      DbgLog.log("SendRegisteredExportToENS","attempt to register equipment module with empty local name !");
      eqm.isRegistered = true;
      eqm.registerAsMaster = false;
      return;
    }
    if ( eqm.getExportName() == null || eqm.getExportName().length() == 0) 
    {
      DbgLog.log("SendRegisteredExportToENS","attempt to register equipment module with empty export name !");
      eqm.isRegistered = true;
      eqm.registerAsMaster = false;
      return;
    }
    if (gFecName == null) return; // not yet known !
    if (gFecAddr == null)
    {
      new TSrvEntry();
      gFecAddr = new FECAddr();
      gFecAddr.setFecName(gFecName);
      try
      {
        String ip;
        if (gSystemRunningStandAlone)
        {
          ip = FECAddr.LOOPBACKADDR;
        }
        else
        {
          ip = InetAddress.getLocalHost().getHostAddress().toString();
        }
        gFecAddr.setIpAddr(ip);
      }
      catch (UnknownHostException e) 
      {
        MsgLog.log("SendRegisteredExportToENS",gFecName+" cannot get local host!",TErrorList.host_not_resolved,e,0);
      }
      gFecAddr.setPortOffset(gPortOffset);
    }
    if (gFecInfo == null)
    {
      gFecInfo = new FECInfo("JAVA",getFecDescription(),getFecLocation(),
          getFecHardware(),getFecResponsible());
    }
    FECInfo inf = gFecInfo;
    String srvName = eqm.getExportName();
    if (isMaster)
    { // secondary registration as a failover master
      if (!eqm.isRegistered) 
      {
        DbgLog.log("SendRegisteredExportToENS","premature attempt to register equipment module as master!");
        return;
      }
      String dsc = eqm.getFailoverType() == TEquipmentModuleFactory.FAILOVER_MASTER ?
                   FAILOVER_MASTER_DESC : FAILOVER_SLAVE_DESC;
      inf = new FECInfo("JAVA",dsc,getFecLocation(),getFecHardware(),getFecResponsible());
      srvName = eqm.getMaster();
    }
    if (eqm.getContext() == null || eqm.getContext().length() == 0)
    { // fill this in if it hasn't been set ...
      eqm.setContext(getFecContext());
    }
    if (eqm.getSubsystem() == null || eqm.getSubsystem().length() == 0)
    { // fill this in if it hasn't been set ...
      eqm.setSubsystem(getFecSubsystem());
    }
    validateContextAndSubsystem(eqm);
    SrvAddr da = new SrvAddr(gFecAddr);
    da.expName = srvName;
    da.eqmContext = eqm.getContext(); // getFecContext();
    da.eqmName = eqm.getModuleName();
    da.subSystem = eqm.getSubsystem();
    if (gSystemRunningStandAlone)
    {
      SrvAddr srv = 
        new SrvAddr(gFecAddr.getFecName(),FECAddr.LOOPBACKADDR,
                    gFecAddr.getPortOffset(),eqm.getModuleName(),
                    srvName,eqm.getContext());
      TSrvEntry.addAddressToCacheFile(srv);
      return; 
    }
    // prepare it and send it out ...
    // devAddr-- + FecAddr + FecInfo
    byte[] expAddr = new byte[SrvAddr.sizeInBytes + FECInfo.getSizeInBytes()];
    System.arraycopy(da.toByteArray(), 0, expAddr, 0, SrvAddr.sizeInBytes);
    System.arraycopy(inf.toByteArray(), 0, expAddr, SrvAddr.sizeInBytes, FECInfo.getSizeInBytes());
    TDataType din = new TDataType(expAddr, "");
    TDataType dout = new TDataType(expAddr, "");
    if (debugLevel > 0)
    {
      DbgLog.log("SendRegisteredExportToENS","fec -> " + gFecAddr.toString());
      DbgLog.log("SendRegisteredExportToENS","dev -> " + da.toString());
      DbgLog.log("SendRegisteredExportToENS","fec info -> " + inf.toString());
    }
    // this should always point to the primary ENS
    if (TSrvEntry.currentConfiguredNameServerTag != null)
    {
      if (debugLevel > 2)
          DbgLog.log("SendRegisteredExportToENS","Using ENS : " + TSrvEntry.currentConfiguredNameServerTag);
      TLink expLnk = new TLink(TSrvEntry.currentConfiguredNameServerTag,"EXPORT",dout,din,TAccess.CA_WRITE);
      if (gLinkFactory == null) gLinkFactory = expLnk.tf;
      int rc = expLnk.execute(500, true);
      if (rc == 0) 
      {
        eqm.isRegistered = true;
        if (isMaster) eqm.registerAsMaster = false;
        TFecLog.log("FEC " + getFecName() + ", Server " + 
            da.expName + "(" + da.eqmName + ") registered with equipment name server");
        TFecLog.log("Context is " + da.eqmContext);
        gNumEnsErrors = 0;
      }
      else
      {
        if (gNumEnsErrors++ > 30) eqm.isRegistered = true;
        TFecLog.log("FEC " + getFecName() + ", Server " + 
            da.expName + "(" + da.eqmName + ") failed to register with equipment name server " + 
            TSrvEntry.currentConfiguredNameServerTag + " : " + TErrorList.getErrorString(rc));
      }
    }
    else
    {
      DbgLog.log("SendRegisteredExportToENS","No configured ENS !");
    }
    // update the manifest
    TEquipmentManifest.update(eqm);
  }
  private void joinEnsGroup(TEquipmentModule eqm)
  {
    if (gFecName == null) return; // not yet known !
    if (eqm == null) return; // huh ?
    if (!eqm.isRegistered) return;
    String grp = eqm.getGroupName();
    if (grp == null || grp.length() == 0) 
    {
      eqm.grpRegistered = true; // mark it
      return;
    }
    if (eqm.grpRegistered) return;
    String exp = eqm.getExportName(); 
    if (exp == null) 
    { // this can't work out !
      eqm.grpRegistered = true; // mark it
      return;
    }
    String cnt = eqm.getContext(); if (cnt == null) cnt = new String("");
    String sub = eqm.getSubsystem();
    if (sub == null || sub.length() == 0)
    { // fill this in if it hasn't been set ...
      sub = eqm.getEquipmentModuleFactory().getFecSubsystem();
    }
    NAME32[] n32 = new NAME32[5];
    n32[0] = new NAME32(exp);
    n32[1] = new NAME32(grp);
    n32[2] = new NAME32(cnt);
    n32[3] = new NAME32(String.valueOf(eqm.getGroupIndex()));
    n32[4] = new NAME32(sub);
    TDataType din = new TDataType(n32);
    TDataType dout = new TDataType();
    if (debugLevel > 0)
    {
      String dbgstr = "/" + cnt + "/" + exp + " -> ";
      DbgLog.log("joinEnsGroup",dbgstr+"joining group " + grp + " (index " + eqm.getGroupIndex() + ")");
    }
    // this should always point to the primary ENS
    if (TSrvEntry.currentConfiguredNameServerTag != null)
    {
      TLink expLnk = new TLink("/SITE/GENS","JOIN",dout,din,TAccess.CA_WRITE);
      if (gLinkFactory == null) gLinkFactory = expLnk.tf;
      int rc = expLnk.execute(500, true);
      expLnk.cancel();
      switch (rc)
      {
        case 0:
          eqm.grpRegistered = true;
          TFecLog.log(exp + " joined group " + grp);
          return;
        case TErrorList.link_not_open:
        case TErrorList.connection_timeout:
          // try, try again
          return;
        default:
          eqm.grpRegistered = true;
          TFecLog.log(exp + " joining group failed : " + TErrorList.getErrorString(rc));
          return;
      }    
    }
    else
    {
      DbgLog.log("joinEnsGroup","No configured GENS !");
    }
  }
  private byte[] msgBytes = new byte[TTransport.UDP_BUFFER_SIZE];
  private void SendMessageToCaller(TClient tc, TSubscription s, int cc, byte[] data)
  {
    TClientEntry tce = new TClientEntry();
    TContractTable tct = new TContractTable();
    short totalsize = 2;
    short hdrsize = TPHdr.hdrSizeInBytes;
    short msgsize = TStrings.STATUS_SIZE;
    if (s.isLegacy)
    {
      hdrsize = TPHdr.hdrSizeInBytesLegacy;
      msgsize = TStrings.STATUS_SHORTSIZE;
    }
    Arrays.fill(msgBytes, (byte)0);
    byte[] payload = msgBytes; //new byte[1500];
    int len;
    tct.dtimestamp = System.currentTimeMillis() + gDataTimeStampOffset;
    tce.sts = new TClientStatus(tc.inetProtocol);
    if (s != null)
    {
      tct.contract = s.contract;
      tce.sts.starttime = ((long)s.starttime) * 1000;
      tce.sts.id = s.id;
    }
    tce.sts.statusCode = (short)cc;
    tce.sts.numblks = 1;
    tce.cln = tc;
    // tce contains the protocol level in the following call
    TPHdr tpd = new TPHdr(msgsize+hdrsize, tce, tct);
    System.arraycopy(tpd.dBuffer.toByteArray(),0,payload,totalsize,hdrsize);
    totalsize += hdrsize;
    switch (cc)
    {
      case TErrorList.get_subscription_id:
       if (data == null) return;
       len = 2;
       break;
      case TErrorList.call_redirection:
      case TErrorList.server_redirection:
        if (data == null) return;
        len = data.length;
        break;
      default:
        data = TErrorList.getErrorString(cc).getBytes();
        len = data.length;
        break;
    }
    System.arraycopy(data,0,payload,totalsize,len);
    totalsize += msgsize; //TErrorList.ERROR_SIZE;
    if (debugLevel > 0) 
    {
      switch (cc)
      {
        case TErrorList.illegal_property:
        case TErrorList.illegal_equipment_name:
        case TErrorList.non_existent_elem:
        case TErrorList.call_redirection:
        case TErrorList.server_redirection:
        case TErrorList.access_denied:
          DbgLog.log("SendMessageToCaller","MSG : " + s.contract.eqmName + " " + s.contract.eqmProperty + 
                             " -> " + TErrorList.getErrorString((short) cc));
          break;
        default: 
          DbgLog.log("SendMessageToCaller","MSG : " + TErrorList.getErrorString((short) cc));
          break;
      }
    }
    payload[0] = (byte) (totalsize & 0xff);
    payload[1] = (byte) (totalsize >> 8);
    sendToPeer(tc,payload,totalsize); 
  }
  public class TAcceptorThread extends Thread
  {
    private int transportProtocol = TTransport.UDP; // default
    TAcceptorThread(int transport)
    {
      transportProtocol = transport;
      this.setName("Equipment Module " + TTransport.toString((short)transport) + " acceptor");
    }
    TAcceptorThread()
    {
      transportProtocol = TTransport.UDP;
      this.setName("Equipment Module UDP acceptor");
    }
    public synchronized void run()
    {
      DbgLog.log("TAcceptorThread","Acceptor Thread "+getName()+" started ...");
      while (!gSystemExitCondition)
      {
        try
        {
          if (!isInitialized) sleep(200);
          switch (transportProtocol)
          {
            case TTransport.UDP:
              systemAcceptDataGramRequest();
              gTotalUdpPackets++;
              break;
            case TTransport.TCP:
              systemAcceptDataStreamRequest();
              systemCleanupDataStreams();  // do this as a matter of course (kills inactive streams)
              gTotalTcpPackets++;
              break;
            default:
              MsgLog.log("TAcceptorThread","transportProtocol "+TTransport.toString((short)transportProtocol)+" not supported",TErrorList.illegal_protocol,null,0);
              gSystemExitCondition = true;
          }
        }
        catch (InterruptedException e)
        {
          if (debugLevel > 2) DbgLog.log("TAcceptorThread","interrupted"); 
        }
        catch (Exception e) 
        {
          MsgLog.log("TAcceptorThread", "unhandled exception: "+e.getMessage(),TErrorList.code_failure,e,0);
        }
      }
      MsgLog.log("TAcceptorThread","Acceptor Thread terminating ...",0,null,0);
    }
  }
  public class TCycleThread extends Thread
  {
    public synchronized void run()
    {
      this.setName("Equipment Module Cycle");
      DbgLog.log("TCycleThread","Cycle Thread started ...");
      while (!gSystemExitCondition)
      {
        try
        {
          sleep(systemSrvCycle());
        }
        catch (Exception e)
        {
          e.printStackTrace();
          DbgLog.log("TCycleThread","cycle thread execution error");
        }
      }
      TFecLog.log("System halted");
    }
  }
  public class TBackgroundThread extends Thread
  {
    private TEquipmentBackgroundTask tbkg;
    TBackgroundThread(TEquipmentBackgroundTask tBkgTask)
    {
      tbkg = tBkgTask;
      this.setName("Equipment Module background " + tbkg.getClass().getSimpleName());
   }
    public synchronized void run()
    {
      long interim = 0, deadTimeInMsec = 0;
      TFecLog.log("Background Thread "+getName()+" started ...");
      while (!gSystemExitCondition)
      {
        try
        {
          interim = System.currentTimeMillis();
          if (!tbkg.isIdleState()) tbkg.call();
          deadTimeInMsec = tbkg.getBackgroundTaskInterval() - (System.currentTimeMillis() - interim);
          if (deadTimeInMsec < 0) deadTimeInMsec = 0;
          if (debugLevel > 3) 
            DbgLog.log("TBackgroundThread","sleeping " + deadTimeInMsec + " msec in background task");
          sleep(deadTimeInMsec);
        }
        catch (Exception e)
        {
          e.printStackTrace();
          DbgLog.log("TBackgroundThread","thread execution error");
        }
      }
    }
  }
  private void systemCleanupDataStreams()
  {
    for (int i=0; i<bucketList.size(); i++)
    {
      if (((TContractBucket)bucketList.get(i)).active == false)
        bucketList.remove(i);
    }
  }
  private void systemAcceptDataStreamRequest()
  {
    if (!isInitialized) return;
    try
    {
      Socket sck = srvs.accept();
      sck.setTcpNoDelay(true);
      sck.setTrafficClass(0x18);
      // TODO: there's got to be a more efficient way than waking the stupid thread up every
      // 10 msec ! a .notify() from SendToPeer() doesn't work since it's not the owner of the 
      // thread. BUT this can't block waiting for input when the scheduler needs to send 
      // something out !!!
      sck.setSoTimeout(10); 
      bucketList.add(new TContractBucket(sck));
    }
    catch (Exception e)
    {
      if (debugLevel > 0) DbgLog.log("systemAcceptDataStreamRequest","interrupted IO");
    }
  }  
  private void systemAcceptDataGramRequest()
  {
    if (!isInitialized) return;
    try
    {
      srvp.getSocket().receive(srvp.dpIn);
      gLinkTablesAccessed = true;
      TClient tc = LocateClientInList(srvp.dpIn.getAddress(), srvp.dpIn.getPort());
      tc.inetProtocol = TTransport.UDP;
      // pass it on ...
      getRequest(tc, srvp.dpIn.getData());
      gLinkTablesAccessed = false;
    }
    catch (IOException e)
    {
      if (debugLevel > 0) DbgLog.log("systemAcceptDataGramRequest","IO exception: "+e.getMessage());
    }
    catch (Exception e)
    {
      MsgLog.log("systemAcceptDataGramRequest","unhandled exception",TErrorList.code_failure,e,0);
    }
  }
  private long systemSrvCycle()
  {
    long cycletime = System.currentTimeMillis();
    long msecToSleep = 10;
    if (!isInitialized) return STD_CYCLE_INTERVAL*200;
    if (gLinkTablesAccessed) return STD_CYCLE_INTERVAL;
    if (isInsideCycle) return STD_CYCLE_INTERVAL;
    isInsideCycle = true;
    try
    {
      if (debugLevel > 25) DbgLog.log("systemSrvCycle","SystemCycle called at " + new Date().toString());
      // basic server service engine (called at the polling rate)
      // call each module's update routine ...
      for (int i=0; i<numEqmTableEntries; i++) eqmTable[i].update();
      // process all external requests ...
      doScheduler(); // check contract list ...
      if (registrationPending && cycletime - registrationTime > 1000)
      { // anything still need to be sent out ?
        int doneCounter=0;
        for (int i=0; i<numEqmTableEntries; i++) 
        {
          if (eqmTable[i] == null) continue;
          if (!eqmTable[i].isRegistered) SendRegisteredExportToENS(eqmTable[i]);
          else if (eqmTable[i].registerAsMaster) SendRegisteredExportToENS(eqmTable[i],true);
          else if (!eqmTable[i].grpRegistered) joinEnsGroup(eqmTable[i]);
          else doneCounter++;        
        }
        if (doneCounter == numEqmTableEntries) registrationPending = false;
        registrationTime = cycletime;
      }
    }
    catch (Exception e)
    {
      TFecLog.log("systemSrvCycle", "unhandled exception "+e.getMessage());      
    }
    if (gCycleDelay > 10) msecToSleep = gCycleDelay;
    isInsideCycle = false;
    return msecToSleep;
  }
  public int systemInit(String fecName, int portOffset)
  {
    return systemInit(fecName,portOffset,null); 
  }
  public int systemInit(String fecName, int portOffset, String expName)
  {
    String thDisplay = instance.getInitializer().getTineHome();
    String fhDisplay = instance.getInitializer().getFecHome();
    String wd = new File(".").getAbsolutePath();
    if (thDisplay.compareTo(".") == 0) thDisplay = ". (" + wd + ")";
    if (fhDisplay.compareTo(".") == 0) fhDisplay = ". (" + wd + ")";
    TFecLog.log("VERSION   : " + TStockProperties.getStockSrvVersion());
    TFecLog.log("OS        : JAVA");    
    TFecLog.log("TINE HOME : " + thDisplay);
    TFecLog.log("FEC HOME  : " + fhDisplay);

    String lclName = null;
    if (expName == null)
    { // export name not given so try it this way ...
      for (int i = 0; lclName == null && i < numEqmTableEntries; i++)
      { // if there's already an entry then find it's eqm name
        if (eqmTable[i] == null) continue;
        lclName = eqmTable[i].getModuleName();
        expName = eqmTable[i].getExportName();
        String cntName = eqmTable[i].getContext();
        if (cntName != null && cntName.length() > 0)
          expName = "/" + cntName + "/" + expName;
      }
    }
    if (fecName != null) setFecName(fecName);
    int rc = getFecInformationFromFile(expName,lclName);
    if (rc != 0 && (fecName == null || fecName.length() == 0))
    {
      MsgLog.log("systemInit","Needed to read fecid.csv and couldn't!",TErrorList.name_unknown,null,0);
    }
    if (portOffset > 0) gPortOffset = portOffset;
    int port = initializer.getSrvPort() + gPortOffset;
    srvp = new TPacket(port,sckRcvBufferSize,sckTimeToLive);
    port = initializer.getTCPPort() + gPortOffset;
    if (gFecName != null) 
    { // everything looks okay now ...
      TLinkFactory.getInstance().setUserName(getFecName());
    }
    else
    {
      MsgLog.log("systemInit","FEC Name unknown at start time !",TErrorList.name_unknown,null,0);
    }
    try
    {   
      ServerSocketChannel srvsChannel = ServerSocketChannel.open(); 
      srvs = srvsChannel.socket();
      srvs.setReuseAddress(true);
      srvs.setReceiveBufferSize(sckRcvBufferSize);
      srvs.bind(new InetSocketAddress(port));
    }
    catch (Exception e) 
    {
      e.toString();
      MsgLog.log("systemInit","Cannot start server; closing down!",TErrorList.runtime_error,null,0);
      return TErrorList.not_running;
    }
    String sas = System.getenv("TINE_STANDALONE");
    if (sas != null && sas.compareToIgnoreCase("TRUE") == 0)
    {
      gSystemRunningStandAlone = true;
      TFecLog.log("FEC is running in stand-alone mode");
    }
    // is there a hook ?
    if (tEqmHook != null) tEqmHook.SystemInit();
    // register the stock alarm definitions ...
    registerStockAlarmDefinitions();
    // send what we've got to the ENS ...
    for (int i = 0; i < numEqmTableEntries; i++)
    { // is there a devices file (needs to be done late) ?
      eqmTable[i].startup();
      if (!eqmTable[i].isRegistered) SendRegisteredExportToENS(eqmTable[i]);
    }
    gSrvStartupTime = System.currentTimeMillis();
    // the server starttime is systematically a 4-byte integer (UTC) across ALL platfroms ...
    serverStatistics.setStarttime((int) (gSrvStartupTime/1000));
    // start global synchronization if required
    if (useGlobalSynchronization) systemStartGlobalSynchronization();
    if (useCycleTrigger) systemStartCycleTrigger();
    // signal the 'all clear' :
    isInitialized = true;
    return 0;
  }
  public int systemInit(int portOffset)
  {
    return systemInit(null, portOffset, null);
  }
  public int systemInit()
  {
    return systemInit(null, 0, null);
  }
  public void systemWait(long milliseconds)
  {
    try
    {
      if (milliseconds < 0)
        while (!gSystemExitCondition)
          Thread.sleep(1000);
      else
        Thread.sleep(milliseconds);
    }
    catch (Exception e)
    {
    }
  }
  private boolean exitOnShutdown = true;
  public void shutdown() { shutdown(0); }
  public void shutdown(int status)
  {
    TLinkFactory.getInstance().shutdown();
    gSystemExitCondition = true;
    for (int i=0; i<eqmTable.length; i++)
    { // call any termination routines ...
      if (eqmTable[i] == null) continue;
      eqmTable[i].shutdown();
    }
    try
    { // give the other threads a chance to close gracefully ...
      Thread.sleep(gSystemTick*2);
    }
    catch (InterruptedException e) {};
    if (exitOnShutdown) System.exit(status);
  }
  // csv file FECID handlers:
  class tgtHndlr implements csvHandler
  {
    String target;
    tgtHndlr(String name) { target = name; }
    public int process(String strValue,int index)
    {
      if (target == null ||  // target not given 
          strValue.length() == 0) // empty string is as good as a match
      { // not given, halt iteration !
        TFecLog.log("accept first entry in fecid.csv as valid!");
        return -1; 
      }
      if (strValue.compareTo(target) == 0) // got it !
      { // found it, halt iteration !
        TFecLog.log("found target " + target + " in fecid.csv!");
        return -1; 
      }
      return 0;
    }
  }
  class FecRowHndlr implements RowHandler
  {
    private boolean latchExpName = false;
    private boolean latchEqmName = false;
    protected boolean namefound = false;
    private String target;
    public String ctx;
    public String sub;
    public String rsp;
    public String hdw;
    public String dsc;
    public String nam;
    public String loc;
    public String exp;
    public String eqm;
    public String hstHome;
    public int port;
    public int process(int index)
    {
      if (latchExpName)
      { // looking for a specific export name (and context name ?)
        String tctx = null;
        String texp = target;
        if (target.startsWith("/"))
        {
          String[] parts = target.split("/");
          if (parts.length < 3) return 0;
          tctx = parts[1];
          texp = parts[2];
        }
        if (exp == null || exp.length() == 0) 
        {
          TFecLog.log("no exports column : accept " + nam + " as fecname");
          namefound = true; return -1;
        }
        if ((ctx != null && ctx.length() > 0) &&
            (tctx != null && tctx.length() > 0))
        { // contexts must match !
          if (ctx.compareToIgnoreCase(tctx) != 0) return 0;
        }
        if (exp.compareToIgnoreCase(texp) == 0) 
        { 
          namefound = true; return -1; // that's it!
        }
      }
      else if (latchEqmName)
      { // looking for a specific eqm name
        if (eqm == null || eqm.length() == 0) 
        {
          TFecLog.log("no local name column : accept " + nam + " as fecname");
          namefound = true; return -1;
        }
        if (eqm.compareToIgnoreCase(target) == 0)
        { 
          namefound = true; return -1; // that's it!
        }
      }
      else
      { // not latching anything -> stop on the first entry
        namefound = true; return -1;
      }
      return 0;
    }    
  }
  int getFecXmlDocument()
  {
    if (gFecXmlCfg != null) return 0; // been there, done that ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    File fecXml = null;
    try
    {
      fecXml = new File(initializer.getFecHome() + "/fec.xml");
      if (fecXml != null)
      {
        DocumentBuilderFactory f = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder;
        builder = f.newDocumentBuilder();
        Document doc = builder.parse(fecXml);
        TineXMLparser p = new TineXMLparser( doc );
        gFecXmlCfg = p.parse();
        setFecName(gFecXmlCfg.getName());
        setFecContext(gFecXmlCfg.getContext());
        setFecSubsystem(gFecXmlCfg.getSubsystem());
        setFecDescription(gFecXmlCfg.getDescription());
        setFecHardware(gFecXmlCfg.getHardware());
        setFecResponsible(gFecXmlCfg.getResponsible());
        
        LinkedList<EqmCfg> el = gFecXmlCfg.getEqmList();
        for( EqmCfg ec : el )
        {
          if (getFecContext() == null || getFecContext().length() == 0)
            setFecContext(ec.getContext());
          if (getFecDescription() == null || getFecDescription().length() == 0)
            setFecDescription(ec.getDescription());
          if (getFecHardware() == null || getFecHardware().length() == 0)
            setFecHardware(ec.getHardware());
          if (getFecSubsystem() == null || getFecSubsystem().length() == 0)
            setFecSubsystem(ec.getSubsystem());
          if (getFecResponsible() == null || getFecResponsible().length() == 0)
            setFecResponsible(ec.getResponsible());
          break;
        }            
        gPortOffset = (short)(gFecXmlCfg.getPortOffset().getValue());
        TFecLog.log("fec.xml read");
      }
    }
    catch (FileNotFoundException e)
    {
      gFecXmlCfg = null;
    }
    catch (Exception e)
    {
        e.printStackTrace();
        TFecLog.log("error reading fec.xml : " + e.getLocalizedMessage());
    }    
    return gFecXmlCfg == null ? TErrorList.no_such_file : 0;      
  }
  int getFecInformationFromFile(String targetExportName,String targetLocalName)
  {
    if (getFecXmlDocument() == 0) return 0; 
    TFecLog.log("no fec.xml found, try fecid.csv ...");
    // register the csv database structure: 
    FecRowHndlr fHndlr = new FecRowHndlr();
    if (targetExportName != null)
    {
      fHndlr.latchExpName = true;
      fHndlr.target = targetExportName;
      TFecLog.log("trying to latch export name " + targetExportName + " in fecid.csv");
    }
    else
    {
      fHndlr.latchEqmName = true;
      fHndlr.target = targetLocalName;
      TFecLog.log("trying to latch local name " + targetLocalName + " in fecid.csv");
    }
    csvColumn[] fecCols = new csvColumn[11];
    fecCols[0] = new csvColumn("FEC_NAME",null,new StringFieldHandler(fHndlr,"nam"));
    fecCols[1] = new csvColumn("PORT_OFFSET","0",new IntFieldHandler(fHndlr,"port"));
    fecCols[2] = new csvColumn("CONTEXT","",new StringFieldHandler(fHndlr,"ctx"));
    fecCols[3] = new csvColumn("EXPORT_NAME","",new StringFieldHandler(fHndlr,"exp"));
    fecCols[4] = new csvColumn("LOCAL_NAME","",new StringFieldHandler(fHndlr,"eqm"));
    fecCols[5] = new csvColumn("DESCRIPTION","",new StringFieldHandler(fHndlr,"dsc"));
    fecCols[6] = new csvColumn("LOCATION","",new StringFieldHandler(fHndlr,"loc"));
    fecCols[7] = new csvColumn("HARDWARE","none",new StringFieldHandler(fHndlr,"hdw"));
    fecCols[8] = new csvColumn("SUBSYSTEM","TST",new StringFieldHandler(fHndlr,"sub"));
    fecCols[9] = new csvColumn("RESPONSIBLE","",new StringFieldHandler(fHndlr,"rsp"));
    fecCols[10] = new csvColumn("HISTORY_HOME","",new StringFieldHandler(fHndlr,"hstHome"));
    // open it
    csv fecFile = new csv(initializer.getFecIdResource());
    // read it
    int rc = fecFile.readFile(fecCols,fHndlr);
    // close it
    fecFile.close();  
    if (rc != TErrorList.no_such_file)
    {
      TFecLog.log("get fec information from fecid.csv : " + TErrorList.errorString[rc]);
    }
    if (!fHndlr.namefound)
    { // file was found, but no matching entry !
      TFecLog.log("get fec information from fecid.csv : could not find matching entry for " + fHndlr.target);
      rc = TErrorList.non_existent_elem;
    }
    if (rc == 0)
    {
      setFecName(fHndlr.nam);
      setFecContext(fHndlr.ctx);
      setFecDescription(fHndlr.dsc);
      setFecHardware(fHndlr.hdw);
      setFecLocation(fHndlr.loc);
      setFecSubsystem(fHndlr.sub);
      setFecResponsible(fHndlr.rsp);
      
      gPortOffset = (short)fHndlr.port;
    }
    return rc;
  }
  // csv ALIAS list handlers:
  class aliasListRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    public String aName;
    public String aAlias;
    public int process(int index)
    {
      AliasTableEntry ate = new AliasTableEntry(aName,aAlias);
      gAliasList.add(ate);
      return 0;
    }
  }
  int getAliasListFromFile()
  {
    FecCfg cfg = getFecXmlCfg();
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<AliasCfg> al = cfg.getAliasList();
      for( AliasCfg ac : al ) 
      {
        gAliasList.add(new AliasTableEntry(ac.getTarget(),ac.getName()));
      }    
      return 0;
    }
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] aliasCols = new csvColumn[2];
    aliasListRowHndlr aliasRows = new aliasListRowHndlr();
    aliasCols[0] = new csvColumn("NAME","",new StringFieldHandler(aliasRows,"aName"));
    aliasCols[1] = new csvColumn("ALIAS","0",new StringFieldHandler(aliasRows,"aAlias"));
    // open it
    csv aliasFile = new csv(initializer.getAliasResource(null));
    // read it
    aliasFile.readFile(aliasCols,aliasRows);
    // close it
    aliasFile.close();  
    //TFecLog.log("get alias information from alias.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv ALMWATCH handlers:
  class almwRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    private TEquipmentModule eqm;
    public String dev;
    public String prp;
    public int siz;
    public int fmt;
    public int atyp;              /* array type */
    public int asys;              /* alarm system id */
    public int sev;               /* of hi and low (warn is sev - 2) */
    public int hisev;
    public int losev;
    public int hiwarnsev;
    public int lowarnsev;
    public int cnt;
    public int mask;
    public int normal;
    public float hi;               /* high threshold */
    public float lo;               /* low threshold */
    public float hiwarn;           /* high warn threshold */
    public float lowarn;           /* low warn threshold */
    public int code;
    public int codeHigh;
    public int codeLow;
    public boolean normalIsAlarm = false;
    public int process(int index)
    {
      if (eqm == null) return TErrorList.non_existent_elem;
      TAlarmWatchThreshold awt;
      if (mask == 0)
      {
        awt = new TAlarmWatchThreshold(cnt,hi,hiwarn,lo,lowarn);
      }
      else
      {
        awt = new TAlarmWatchThreshold(cnt,mask,normal,normalIsAlarm);
      }
      TAlarmWatchEntry awe = new TAlarmWatchEntry(eqm,dev,prp,siz,fmt,atyp,asys,sev,code,codeHigh,codeLow,awt);
  
      eqm.gAlarmWatchList.add(awe);
      return 0;
    }
  }
  class almwEqmHndlr implements csvHandler
  {
    private almwRowHndlr rHndlr;
    almwEqmHndlr(almwRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      for (int i=0; i<numEqmTableEntries; i++) 
        if (eqmTable[i].getLocalName().compareTo(strValue) == 0) 
        {
          rHndlr.eqm = eqmTable[i]; 
          return 0;
        }
      return 0; //TErrorList.non_existent_elem;
    }
  }
  class almwNormalValueHndlr implements csvHandler
  {
    private almwRowHndlr rHndlr;
    almwNormalValueHndlr(almwRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      String s;
      if (strValue.startsWith("!"))
      {
        rHndlr.normalIsAlarm = true;
        s = strValue.substring(1);
      }
      else
      {
        rHndlr.normalIsAlarm = false;
        s = strValue;        
      }
      rHndlr.normal = csv.getIntFromString(s);
      return 0; //TErrorList.non_existent_elem;
    }
  }
  private boolean stockAlarmDefinitionsRegistered = false;
  public void registerStockAlarmDefinitions()
  {
    if (stockAlarmDefinitionsRegistered) return;
    // register some stock alarm definitions ...
    stkAlarmDefinitions.put(Integer.toString(TErrorList.invalid_data), 
      new TAlarmDefinition("invalid data", "value not at set point","device",TErrorList.invalid_data,TAlarm.ALM_SEVERITY_ERR,0, 
          -1,TFormat.CF_LONG,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.value_too_high), 
      new TAlarmDefinition("value too high", "value exceeds threshold","device",TErrorList.value_too_high,TAlarm.ALM_SEVERITY_ERR,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.warn_too_high), 
      new TAlarmDefinition("warn too high", "value is near threshold","device",TErrorList.warn_too_high,TAlarm.ALM_SEVERITY_WARN,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.value_too_low), 
      new TAlarmDefinition("value too low", "value exceeds threshold","device",TErrorList.value_too_low,TAlarm.ALM_SEVERITY_ERR,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.warn_too_low), 
      new TAlarmDefinition("warn too low", "value is near threshold","device",TErrorList.warn_too_low,TAlarm.ALM_SEVERITY_WARN,0,
          -1,TFormat.CF_FLOAT,1,"last measured value",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.max_alarms_exceeded), 
        new TAlarmDefinition("more than 100 total alarms", "more than 100 total alarms (collapse alarm messages)","device",TErrorList.max_alarms_exceeded,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_SHORT,1,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.io_error), 
        new TAlarmDefinition("I/O error", "hardware input/ouput error","device",TErrorList.io_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.file_error), 
        new TAlarmDefinition("File error", "file input/ouput error","device",TErrorList.file_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_TEXT,64,"file/path name which caused the alarm",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.hardware_error), 
        new TAlarmDefinition("Hardware error", "general hardware error","device",TErrorList.hardware_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware device description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.net_read_error), 
        new TAlarmDefinition("Network read error", "network read error","device",TErrorList.net_read_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.net_write_error), 
        new TAlarmDefinition("Network write error", "network write error","device",TErrorList.net_write_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.link_error), 
        new TAlarmDefinition("Link error", "link to dependent server is not open","device",TErrorList.link_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_SYSTEM,
            -1,TFormat.CF_TEXT,64,"link which has the problem",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.driver_not_installed), 
        new TAlarmDefinition("Driver not installed", "hardware driver not installed or not initialized","device",TErrorList.driver_not_installed,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"name of hardware driver",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.port_not_available), 
        new TAlarmDefinition("Port not available", "hardware port not avaliable","device",TErrorList.port_not_available,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"Port Id",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.not_running), 
        new TAlarmDefinition("Not Running", "hardware device is not running","device",TErrorList.not_running,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_NULL,0,"",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.sedac_error), 
        new TAlarmDefinition("Sedac error", "sedac module has an error","device",TErrorList.sedac_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_SHORT,3,"line/crate/subaddress of defective module",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.gpib_error), 
        new TAlarmDefinition("GPIB error", "GPIB bus error","device",TErrorList.gpib_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.rs232_error), 
        new TAlarmDefinition("RS232 error", "RS232 bus communication error","device",TErrorList.rs232_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"hardware module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.device_error), 
        new TAlarmDefinition("Device error", "general device error/device defective","device",TErrorList.device_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_TEXT,64,"module description",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.canbus_error), 
        new TAlarmDefinition("CAN bus error", "CAN module has an error","device",TErrorList.canbus_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_LONG,4,"address of CAN module",""));
    stkAlarmDefinitions.put(Integer.toString(TErrorList.profibus_error), 
        new TAlarmDefinition("PROFIBUS error", "PROFIBUS module has an error","device",TErrorList.profibus_error,TAlarm.ALM_SEVERITY_ERR,TAlarm.ALM_SYSTEM_HARDWARE,
            -1,TFormat.CF_LONG,4,"address of PROFIBUS module",""));
    stockAlarmDefinitionsRegistered = true;
  }
  int getAlarmWatchListFromFile(String eqmName)
  {
    FecCfg cfg = getFecXmlCfg();
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<EqmCfg> el = cfg.getEqmList();
      for( EqmCfg ec : el ) 
      {
        if (ec.getName().compareTo(eqmName) == 0) 
        {
          TEquipmentModule eqm = getEquipmentModule(eqmName);
          LinkedList<PropertyCfg> pl = ec.getPropertyList();
          AlarmCfg ac;
          TAlarmWatchThreshold awt;
          int mask;
          short fmt;
          for (PropertyCfg pc : pl)
          {
            fmt = (short)pc.getDTypeOut().getValue();
            ac = pc.getAlarm();            
            mask = ac.getValueMask().getValue();
            if (mask == 0)
            {
              awt = new TAlarmWatchThreshold(ac.getCountThreshold().getValue(),
                  ac.getValueHigh().getValue(),ac.getValueWarnHigh().getValue(),
                  ac.getValueLow().getValue(),ac.getValueWarnLow().getValue());
            }
            else
            {
              awt = new TAlarmWatchThreshold(ac.getCountThreshold().getValue(),
                  mask,ac.getValueNormal().getValue(),ac.isNormalAlarm());
            }
            TAlarmWatchEntry awe = new TAlarmWatchEntry(eqm,ac.getDeviceName(),
                pc.getName(),pc.getSizeOut().getValue(),fmt,
                pc.getDArrayTypeOut(),ac.getSystem().getValue(),
                ac.getSeverity().getValue(),ac.getCode().getValue(),
                ac.getCodeHigh().getValue(),ac.getCodeLow().getValue(),awt);        
            eqm.gAlarmWatchList.add(awe);
          }
          break;
        }
      }    
      return 0;
    }
    // no fec.xml : look for exports.csv ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] almwCols = new csvColumn[21];
    almwRowHndlr almwRows = new almwRowHndlr();
    almwCols[0] = new csvColumn("LOCAL_NAME","",new almwEqmHndlr(almwRows));
    almwCols[1] = new csvColumn("DEVICE_NAME","#0",new StringFieldHandler(almwRows,"dev"));
    almwCols[2] = new csvColumn("PROPERTY","",new StringFieldHandler(almwRows,"prp"));
    almwCols[3] = new csvColumn("SIZE","0",new IntFieldHandler(almwRows,"siz"));
    almwCols[4] = new csvColumn("FORMAT","null",new FormatFieldHandler(almwRows,"fmt"));
    almwCols[5] = new csvColumn("SEVERITY","0",new IntFieldHandler(almwRows,"sev"));
    almwCols[6] = new csvColumn("ALARM_SYSTEM","0",new IntFieldHandler(almwRows,"asys"));
    almwCols[7] = new csvColumn("COUNT_THRESHOLD","3",new IntFieldHandler(almwRows,"cnt"));
    almwCols[8] = new csvColumn("HIGH","1.0",new FloatFieldHandler(almwRows,"hi"));
    almwCols[9] = new csvColumn("LOW","0.0",new FloatFieldHandler(almwRows,"lo"));
    almwCols[10] = new csvColumn("HIGH_WARN","0.5",new FloatFieldHandler(almwRows,"hiwarn"));
    almwCols[11] = new csvColumn("LOW_WARN","0.1",new FloatFieldHandler(almwRows,"lowarn"));
    almwCols[12] = new csvColumn("MASK","0",new IntFieldHandler(almwRows,"mask"));
    almwCols[13] = new csvColumn("NORMAL","0",new almwNormalValueHndlr(almwRows));
    almwCols[14] = new csvColumn("SEVERITY_HIGH","0",new IntFieldHandler(almwRows,"hisev"));
    almwCols[15] = new csvColumn("SEVERITY_LOW","0",new IntFieldHandler(almwRows,"losev"));
    almwCols[16] = new csvColumn("SEVERITY_HIGHWARN","0",new IntFieldHandler(almwRows,"hiwarnsev"));
    almwCols[17] = new csvColumn("SEVERITY_LOWWARN","0",new IntFieldHandler(almwRows,"lowarnsev"));
    almwCols[18] = new csvColumn("ALARM_CODE","0",new IntFieldHandler(almwRows,"code"));
    almwCols[19] = new csvColumn("ALARM_CODE_HIGH","0",new IntFieldHandler(almwRows,"codeHigh"));
    almwCols[20] = new csvColumn("ALARM_CODE_LOW","0",new IntFieldHandler(almwRows,"codeLow"));
    
    // open it
    csv almwFile = new csv(initializer.getAlmWatchResource(eqmName));
    // read it
    int rc = almwFile.readFile(almwCols,almwRows);
    // close it
    almwFile.close(); 
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get Alarm Watch information from almwatch.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // csv HISTORY handlers:
  class hstRowHndlr implements RowHandler
  { // the Row Handler will be called when all columns have been read in
    private TEquipmentModule eqm;
    public String dev;
    public String prp;
    public int  siz;
    public int  fmt;
    public int  atyp;              /* array type */
    public int  pollingRate;               /* polling rate in sec */
    public int  archiveRate;               /* archive rate in sec */
    public int  depthShort;             /* for short term storage */
    public int  depthLong;               /* for long term storage */
    public int  heartbeat;            /* archive heartbeat in sec */
    public float pTolerance;                 /* percent tolerance */
    public float aTolerance;                /* absolute tolerance */
    public int  recordIndex;            /* local index identifier */
    public String redirString;
    private boolean skip = false;
    public int process(int index)
    {
      if (skip) return 0;
      if (eqm == null) return TErrorList.non_existent_elem;
      THistorySpecification spc = 
        new THistorySpecification(pollingRate,archiveRate,depthShort,depthLong,heartbeat,pTolerance,aTolerance,redirString);
      if (!THistoryRecord.isAllowableFormat((short)fmt))
      {
        TFecLog.log("history format " + TFormat.toString((short)fmt) + " read from database not archiveable");
        return 0;
      }
      THistoryRecord hst = new THistoryRecord(eqm,dev,prp,siz,fmt,atyp,recordIndex,spc);
      if (eqm.isUseMSecHistoryTimestamps()) hst.setUseMinimalStorage(false); 
      if (eqm.isUseMonthlyHistoryFiles()) hst.setUseMonthlyHistoryFiles(true);
      eqm.gLclHstList.add(hst);
      return 0;
    }
  }
  class hstEqmHndlr implements csvHandler
  {
    private hstRowHndlr rHndlr;
    hstEqmHndlr(hstRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      for (int i=0; i<numEqmTableEntries; i++) 
        if (eqmTable[i].getLocalName().compareTo(strValue) == 0) 
        {
          rHndlr.eqm = eqmTable[i]; 
          return 0;
        }
      return TErrorList.non_existent_elem;
    }
  }
  class hstExpHndlr implements csvHandler
  {
    private String target;
    private hstRowHndlr rHndlr;
    hstExpHndlr(hstRowHndlr rowHndlr,String name) { rHndlr = rowHndlr; target = name; }
    public int process(String strValue,int index)
    {
      if (target == null || target.length() == 0 || strValue.compareTo(target) == 0) 
      { // found it, halt iteration !
        rHndlr.skip = false; 
      }
      else
      {
        rHndlr.skip = true;
      }
      return 0;
    }
  }
  class hstPollHndlr implements csvHandler
  {
    private hstRowHndlr rHndlr;
    hstPollHndlr(hstRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      try
      {
        rHndlr.pollingRate = Integer.parseInt(strValue);
        if (rHndlr.pollingRate < 20) rHndlr.pollingRate *= 1000; // was thinking in secs 
      }
      catch (NumberFormatException e)
      {
        rHndlr.pollingRate = 1000;
      }
      return 0;
    }
  }
  class hstArchHndlr implements csvHandler
  {
    private hstRowHndlr rHndlr;
    hstArchHndlr(hstRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      try
      {
        rHndlr.archiveRate = Integer.parseInt(strValue);
        if (rHndlr.archiveRate < 20) rHndlr.archiveRate *= 1000; // was thinking in secs 
      }
      catch (NumberFormatException e)
      {
        rHndlr.archiveRate = 10000;
      }
      return 0;
    }
  }
  class hstTolHndlr implements csvHandler
  {
    private hstRowHndlr rHndlr;
    hstTolHndlr(hstRowHndlr rowHndlr) { rHndlr = rowHndlr; }
    public int process(String strValue,int index)
    {
      int i=0;
      try
      {
        if ((i=strValue.indexOf('%')) == -1)
        { // absolute tolerance !
          rHndlr.aTolerance = Float.parseFloat(strValue);
          rHndlr.pTolerance = 0;
        }
        else
        {
          rHndlr.aTolerance = 0;
          rHndlr.pTolerance = Float.parseFloat(strValue.substring(0,i-1));
        }
      }
      catch (Exception e)
      {
        rHndlr.pTolerance = 0;
        rHndlr.aTolerance = 0;
      }
      return 0;
    }
  }
  int getHistoryListFromFile(String eqmName)
  {
    FecCfg cfg = getFecXmlCfg();
    if (cfg != null) 
    { // has parsed a fec.xml config file ...
      LinkedList<EqmCfg> el = cfg.getEqmList();
      for( EqmCfg ec : el ) 
      {
        if (ec.getName().compareTo(eqmName) == 0) 
        {
          TEquipmentModule eqm = getEquipmentModule(eqmName);
          LinkedList<PropertyCfg> pl = ec.getPropertyList();
          HistoryCfg hc;
          THistoryRecord hst;
          THistorySpecification spc;
          short fmt;
          for (PropertyCfg pc : pl)
          {
            fmt = (short)pc.getDTypeOut().getValue();
            if (!THistoryRecord.isAllowableFormat(fmt))
            {
              TFecLog.log("history format " + TFormat.toString(fmt) + " read from database not archiveable");
              continue;
            }
            hc = pc.getHistory();            
            spc = new THistorySpecification(
               hc.getPollRate().getValue(),hc.getArchiveRate().getValue(),
               hc.getDepthShort().getValue(),hc.getDepthLong().getValue(),
               hc.getHeartbeat().getValue(),
               hc.getTolerance().getTolerancePercent(), 
               hc.getTolerance().getToleranceAbsolute(),
               hc.getRedirection());
            hst = new THistoryRecord(eqm,hc.getDeviceName(),pc.getName(),
                pc.getSizeOut().getValue(),fmt,pc.getDArrayTypeOut(),
                hc.getIndex().getValue(),spc);
            if (eqm.isUseMinimalStorage()) hst.setUseMinimalStorage(true);
            if (eqm.isUseMonthlyHistoryFiles()) hst.setUseMonthlyHistoryFiles(true);
            eqm.gLclHstList.add(hst);
          }
          break;
        }
      }    
      return 0;
    }
    // no fec.xml : look for exports.csv ...
    TInitializer initializer = TInitializerFactory.getInstance().getInitializer();
    // register the csv database structure: 
    csvColumn[] hstCols = new csvColumn[13];
    hstRowHndlr hstRows = new hstRowHndlr();
    hstCols[0] = new csvColumn("INDEX","1",new IntFieldHandler(hstRows,"recordIndex"));
    hstCols[1] = new csvColumn("LOCAL_NAME","",new hstEqmHndlr(hstRows));
    hstCols[2] = new csvColumn("PROPERTY","",new StringFieldHandler(hstRows,"prp"));
    hstCols[3] = new csvColumn("DEVICE","#0",new StringFieldHandler(hstRows,"dev"));
    hstCols[4] = new csvColumn("DATA_LENGTH","1",new IntFieldHandler(hstRows,"siz"));
    hstCols[5] = new csvColumn("FORMAT","float",new FormatFieldHandler(hstRows,"fmt"));
    hstCols[6] = new csvColumn("HEARTBEAT","900",new IntFieldHandler(hstRows,"heartbeat"));
    hstCols[7] = new csvColumn("POLLING_RATE","1000",new hstPollHndlr(hstRows));
    hstCols[8] = new csvColumn("ARCHIVE_RATE","1000",new hstArchHndlr(hstRows));
    hstCols[9] = new csvColumn("TOLERANCE","0%",new hstTolHndlr(hstRows));
    hstCols[10] = new csvColumn("SHORT_DEPTH","600",new IntFieldHandler(hstRows,"depthShort"));
    hstCols[11] = new csvColumn("LONG_DEPTH","1",new IntFieldHandler(hstRows,"depthLong"));
    hstCols[12] = new csvColumn("REMOTE_SERVER","",new StringFieldHandler(hstRows,"redirString"));
    // open it
    csv hstFile = new csv(initializer.getHistoryResource(eqmName));
    // read it
    int rc = hstFile.readFile(hstCols,hstRows);
    // close it
    hstFile.close();  
    if (rc != TErrorList.no_such_file)
      TFecLog.log("get local history information from history.csv : " + TErrorList.errorString[rc]);
    return 0;
  }
  // subclass this here in the server factory as well as in the client factory
  public class TEqmFactoryBucketThread extends Thread
  {
    private TContractBucket tb;
    boolean isWaiting = false;
    TEqmFactoryBucketThread(TContractBucket tbucket)
    {
      tb = tbucket;
      tb.active = true;
      String s = null;
      Socket sck = tb.getSocket();
      if (sck != null) s = new String("tcp port " + sck.getLocalPort());
      else s = new String("tcp unbound socket ?");
      this.setName("Equipment Module Factory " + s);
    }
    public synchronized void run()
    {
      byte[] wa = new byte[TTransport.UDP_BUFFER_SIZE];
      byte[] msgsizb = new byte[2];
      int nread, nleft, n, waptr;
      TFecLog.log("EQM Factory Bucket Thread started ...");
      try
      {
        while (tb.active)
        {
          if (gSystemExitCondition) break;
          isWaiting = true;
          try
          { // the socket blocks (both is and os) for 10 msec
            nread = tb.getInputStream().read(wa,0,TTransport.UDP_BUFFER_SIZE);
          }
          catch (InterruptedIOException e) 
          {
            continue;
          }
          isWaiting = false;
          if (nread  == -1) 
          {
            tb.isDeactivating = true;
            break;
          }
          nleft = nread;
          waptr = 0;
          int ptr = tb.getBucketPointer();
          int siz = tb.getBucketSize();
          byte[] buf = tb.getBucketBuffer();
          while (nleft > 0)
          {                    
            if (ptr == 0) 
            {
              System.arraycopy(wa,waptr,msgsizb,0,2);
              siz = Swap.Short(new DataInputStream(new ByteArrayInputStream(msgsizb)).readShort());
            }
            n = siz - ptr;
            if (nleft >= n)  // at least enough
            {
              System.arraycopy(wa,waptr,buf,ptr,n);
              gLinkTablesAccessed = true;
              TClient tc = LocateClientInList(tb.getBucketEndpoint(),tb.getBucketPort());
              tc.inetProtocol = TTransport.TCP;
              tc.output = tb.getOutputStream();
              tc.sck = tb.getSocket();
              getRequest(tc, buf);
              gLinkTablesAccessed = false;
              waptr += n;                 // move the read buffer pointer along
              nleft -= n;                 // decrement what's remaining to be read
              ptr = 0;                 // reset the bucket pointer
            }
            else
            {
              System.arraycopy(wa,waptr,buf,ptr,nleft);
              ptr += nleft;
              nleft = 0;
            }
            tb.setBucketPointer(ptr);
          }                    
        }
        tb.getInputStream().close();
        tb.getOutputStream().close();
      }
      catch (IOException e)
      {
        if (debugLevel > 1) DbgLog.log("TEqmFactoryBucketThread","EQM Bucket IOException : " + e);
      }
      catch (Exception e)
      {
        if (debugLevel > 1) e.printStackTrace();
      }
      finally
      {
        tb.active = false;  // definitely !
      }
    }
  }
  public class TContractBucket extends TBucket
  {
    TContractBucket(Socket socket)
    {
      initBucket(socket,new TEqmFactoryBucketThread(this),TPacket.MAX_DATAGRAM_SIZE);
      activate();
    }
  }
  public long getDataTimeStampOffset()
  {
    return gDataTimeStampOffset;
  }
  public void setDataTimeStampOffset(long offset) 
  { 
    gDataTimeStampOffset = offset; 
  }
  public int systemScheduleProperty(TEquipmentModule eqm,String prp)
  {
    int i, listsize;
    boolean needsHistoryCycle = false;
    boolean needsScheduler = false;
    TContractTable tct = null;
    synchronized (conTable)
    {
      listsize = conTable.size();
      if (debugLevel > 2) DbgLog.log("systemScheduleProperty","Property " + prp + " is being scheduled");
      for (i=0; i<listsize; i++)
      {
        tct = (TContractTable) conTable.get(i);
        if (tct.expired) continue;
        if (tct.nclients > 0 && tct.contract.eqmProperty.compareToIgnoreCase(prp) == 0)
        { /* guarantee immediate execution of any contract using this property */
          tct.lasttime = PRP_SCHEDULE_SIGNAL;
          needsScheduler = true;
        }
      }
    }
    if (needsScheduler) doScheduler();
    THistoryRecord lhr = null;
    listsize = eqm.gLclHstList.size();
    for (i=0; i<listsize; i++)
    {
      lhr = (THistoryRecord)eqm.gLclHstList.get(i);
      if (lhr.getPrp().compareToIgnoreCase(prp) == 0)
      {
        needsHistoryCycle = true;
        lhr.setScheduled(true);
      }
    }
    if (needsHistoryCycle) eqm.update(); 
    return 0;
  }
  private LinkedList<TCycleTrigger> gCycTrgLst = new LinkedList<TCycleTrigger>();
  public TCycleTrigger[] getCycleTriggerList()
  {
    return (TCycleTrigger[]) gCycTrgLst.toArray(new TCycleTrigger[0]);
  }
  /**
   * @brief Registers a cycle trigger callback dispatch function
   * 
   * If a CYCLER is running in a server's context, then the server will receive 'Cycle Number' events
   * scheduled by the designated CYCLER server.
   * The cycle number will make use of the 'System Data Stamp' to tag all data sets obtained from the
   * server.  A server can also register a trigger function dispatch routine (or routines) to be called when a 'Cycle
   * Number' event occurs.  The dispatch routines will be called prior to setting the 'System Stamp' to the new
   * Cycle Number, which will be set following the execution of all dispatch routines.  Optionally, the server can
   * provide a property (or list of properties) to be scheduled following the dispatch execution.  This will ensure
   * that such properties will be called immediately following dispatch execution AND contain the most recent Cycle Number
   * as the 'System Data Stamp'.
   *
   * @param trigger is a reference to the dispatch handler to be called when a new cycle number arrives.
   * 
   * @return 0 upon success or TErrorList.already_assigned
   * 
   * \b Example:
   * @include eg_cycleTrigger.java
   */
  public int registerCycleTrigger(TCycleTrigger trigger)
  {
    if (gCycTrgLst.contains(trigger)) return TErrorList.already_assigned;
    gCycTrgLst.add(trigger);
    return 0;
  }
  public int unregisterCycleTrigger(TCycleTrigger trigger)
  {
    if (!gCycTrgLst.contains(trigger)) return TErrorList.un_allocated;
    gCycTrgLst.remove(trigger);
    return 0;
  }
  private boolean useGlobalSynchronization = true;
  private boolean useCycleTrigger = true;
  public void setUseGlobalSynchronization(boolean value)
  {
    useGlobalSynchronization = value;
  }
  private boolean gSynchronizationStarted = false;
  private double[] gSyncTimeStamp = new double[1];
  public double getSyncTimeStamp() { return gSyncTimeStamp[0]; }
  private int systemStartGlobalSynchronization()
  {
    if (gSynchronizationStarted) return 0;
    TDataType sdt = new TDataType(gSyncTimeStamp);
    TLink tl = new TLink("/SITE","SYSTIME",sdt,null,TAccess.CA_READ);
    int id = tl.attach(TMode.CM_GLOBAL, new TSyncCallback(), 1000);
    if (id < 0) return -id;
    gSynchronizationStarted = true;
    return 0;
  }
  private boolean gCycleTriggerStarted = false;
  private int[] gCycleNumber = new int[1];
  public int getCycleNumber() { return gCycleNumber[0]; }
  private boolean assertCycleTriggerExists()
  {
    if (gFecContext == null || gFecContext.length() == 0) return false;
    for (int i = 0; i<numEqmTableEntries; i++)
    {
      if (eqmTable[i].getLocalName() == null) continue; // should not be possible
      if (eqmTable[i].getExportName().compareToIgnoreCase("CYCLER") == 0)
      {
        TFecLog.log("this server is the machine cycler for context " + 
            gFecContext + " : not listening for machine cycles");
        return false;
      }
      TSrvEntry trgsrv = new TSrvEntry("CYCLER",gFecContext);
      if (trgsrv.fecAddr != null) 
      {
        TFecLog.log("a machine cycler exists for context " + gFecContext);
        return true;
      }
    }
    return false;
  }
  private static int cycleTriggerDeadBand = 1000;
  public int getCycleTriggerDeadBand() { return cycleTriggerDeadBand; }
  public void setCycleTriggerDeadBand(int value) 
  {
    if (value < 100) value = 100;
    cycleTriggerDeadBand = value;
  }
  private int systemStartCycleTrigger()
  {
    if (gCycleTriggerStarted) return 0;
    if (!assertCycleTriggerExists()) return TErrorList.non_existent;
    TDataType sdt = new TDataType(gCycleNumber);
    String srv = "/"+gFecContext+"/CYCLER";
    TLink tl = new TLink(srv,"CycleNumber",sdt,null,TAccess.CA_READ);
    int id = tl.attach(TMode.CM_GLOBAL, new TCycleTriggerCallback(), cycleTriggerDeadBand);
    if (id < 0) return -id;   
    gCycleTriggerStarted = true;
    return 0;
  }
  public int getBurstLimit()
  {
    return gBurstLimit;
  }
  public void setBurstLimit(int burstLimit)
  {
    gBurstLimit = burstLimit;
  }
  public boolean isExitOnShutdown()
  {
    return exitOnShutdown;
  }
  public void setExitOnShutdown(boolean exitOnShutdown)
  {
    this.exitOnShutdown = exitOnShutdown;
  }
}