package de.desy.tine.server.equipment;

import de.desy.tine.client.TLink;
import de.desy.tine.client.TLinkCallback;
import de.desy.tine.dataUtils.TDataTime;
import de.desy.tine.server.logger.DbgLog;
import de.desy.tine.server.logger.TFecLog;

public class TSyncCallback implements TLinkCallback
{
  private static TEquipmentModuleFactory f = TEquipmentModuleFactory.getInstance();
  private static final int TIMESYNC_DEPTH = 6;
  private static final double gSyncTolerance = 0.2; 
  private static int last_cc = -1, last_ptr = 0, ncorrections = 0;
  private static double last_timestamp = 0;
  private static double last_logged = 0;
  private static double last_dataTimeOffset = 0, largest_correction = 0, last_correction = 0;
  private static double[] last_dataTimeRbDeltas = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeDtDeltas = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeStamps = new double[TIMESYNC_DEPTH];
  private static double[] last_dataTimeRbTimes = new double[TIMESYNC_DEPTH];
  private void checkTimestamp(double gDataTimeStamp,double gSyncTimeStamp)
  {
    boolean applyCorrection = false;
    int i,k;
    double correction = 0, delta_t = 0, last_dt = 0, last_rb;
    last_rb = last_dataTimeRbTimes[last_ptr];
    last_dt = last_dataTimeStamps[last_ptr];
    last_ptr = (last_ptr + 1) % TIMESYNC_DEPTH;
    last_dataTimeRbTimes[last_ptr] = gDataTimeStamp;            /* my clock's current timestamp */
    last_dataTimeRbDeltas[last_ptr] = gDataTimeStamp - last_rb; /* time difference based on my clock */
    last_dataTimeStamps[last_ptr] = gSyncTimeStamp;             /* remote clock's current timestamp */
    last_dataTimeDtDeltas[last_ptr] = gSyncTimeStamp - last_dt; /* time difference based on remote clock */

    for (i=0; i<TIMESYNC_DEPTH; i++)
    {
      k = (last_ptr + i) % TIMESYNC_DEPTH;
      if (last_dataTimeStamps[k] < 1000) return;
      delta_t = last_dataTimeRbDeltas[k] - last_dataTimeDtDeltas[k];
      if (Math.abs(delta_t) > 0.01)
      { /* these deltas should be really close if the data is coming in and being read at the correct rate */
        return;
      }
      correction += last_dataTimeStamps[k] - last_dataTimeRbTimes[k];
    }
    correction /= TIMESYNC_DEPTH;
    if (correction > gSyncTimeStamp)
    {
      TFecLog.log("invalid time correction " + correction + 
                  " greater than time server clock " + gSyncTimeStamp);
      return;
    }

    if (correction > gSyncTolerance || correction < -gSyncTolerance) applyCorrection = true;
    else if (last_correction < gSyncTolerance && correction > gSyncTolerance/2) applyCorrection = true;
    else if (last_correction > gSyncTolerance && correction < -gSyncTolerance/2) applyCorrection = true;
    if (!applyCorrection) return;
    double tsOffset = (double)(f.getDataTimeStampOffset()) / 1000.0;
    tsOffset += correction;
    f.setDataTimeStampOffset((long)(tsOffset*1000));
    last_correction = correction;
    if (TEquipmentModuleFactory.debugLevel > 0) 
      DbgLog.log("checkTimestamp","global synchronization : " + correction + " sec (current offset " + tsOffset + " sec)");
    ncorrections++;
    if (Math.abs(correction) > largest_correction) largest_correction = Math.abs(correction);
    if (gDataTimeStamp < last_logged + 3600) return;
    last_logged = gDataTimeStamp;
    if (Math.abs(tsOffset-last_dataTimeOffset) < 0.1) return;
    TFecLog.log("global synchronization offset : " + tsOffset + "sec");
    TFecLog.log("largest correction : " + largest_correction + " sec  (nr. corrections " + ncorrections + ")");
    last_logged = gDataTimeStamp; ncorrections = 0; largest_correction = 0;
    last_dataTimeOffset = tsOffset;      
  }
  public void callback(TLink link)
  {
    int cc = link.getLinkStatus();
    if (cc != last_cc)
    {
      last_cc = cc;
      TFecLog.log("global synchronization : " + link.getLastError());
    }
    if (cc != 0) return;
    double this_timestamp = f.getSyncTimeStamp();
    if (last_timestamp == this_timestamp)
    { // multiple callbacks (due to extra broadcasting of the same thing)
      if (TEquipmentModuleFactory.debugLevel > 3) 
        DbgLog.log("checkTimestamp","discard extra synchronization timestamp");
      return;
    }
    last_timestamp = this_timestamp;
    checkTimestamp(TDataTime.getDataTimeStamp(), this_timestamp);
  }    
}
