/*
 * Created on Nov 5, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.server.properties;

import java.util.ArrayList;
import de.desy.tine.dataUtils.TDataType;
import de.desy.tine.definitions.TAccess;
import de.desy.tine.definitions.TArrayType;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.definitions.TStrings;

/**
 * @brief device servers must export the properties they wish to handle.
 * 
 * @author duval 
 */
public class TExportProperty
{
  private String name = ""; // property name
  private String alias = ""; // property alias name
  private int id = -1; // assigned property id
  private int outputSize = 0; // maxValue output data size
  private int inputSize = 0; // maxValue input data size
  private short outputFormat = (short) 0; // from TFormat
  private short inputFormat = (short) 0; // from TFormat
  private short accessMode = (short) 0; // from TAccess
  // Formerly NAME8, but not in stockProperty.
  private String outputTag = ""; // output data structure tag if non
  // zero-length
  private String inputTag = ""; // input data structure tag if non zero-length
  // private NAME16[] devNames; // device list if not NULL
  // private int numDevices; // number of devices in device list
  // private int numOverloads = 0; // number of overloaded properties with the
  // same name
  private TPropertyDescription description = null;
  private ArrayList<String> prpDeviceList = null;
  /** @brief sets the property's device list to the ArrayList given */
  public void setDeviceList(ArrayList<String> deviceList) { prpDeviceList = deviceList; }
  /** @brief returns the property's device list */
  public ArrayList<String> getDeviceList() { return prpDeviceList; }
  public TExportProperty()
  {
    //TODO initialize defaults
  }
  /**
   * Constructor. Creates a property with an ID.
   * 
   * @param prpId The property identifier
   * @param prpName The property name.
   * @param accessMode The access mode associating with the property (CA_READ and/or CA_WRITE).
   * @param din Specifies input format information
   * @param dout Specifies output format information
   * @param prpDscr gives a property description.
   */
  public TExportProperty(int prpId, String prpName, short accessMode, TDataType din, TDataType dout,
      String prpDscr)
  {
    TStrings.validatePropertyName(prpName);
    this.name = prpName;
    this.outputFormat = dout.dFormat;
    this.outputSize = dout.dArrayLength;
    this.outputTag = dout.getTag();
    this.inputFormat = din.dFormat;
    this.inputSize = din.dArrayLength;
    this.inputTag = din.getTag();
    this.accessMode = accessMode;
    this.description = new TPropertyDescription(prpDscr);
    this.id = prpId;
  }
  /**
   * Constructor. Creates a TExportProperty for read access.
   * 
   * @param prpName
   * @param outputSize
   * @param outputFormat
   */
  public TExportProperty(String prpName, int outputSize, short outputFormat)
  {
    initialize(prpName, TAccess.CA_READ, outputSize, outputFormat, 0, TFormat.CF_NULL);
  }
  /**
   * Constructor. Creates a TExportProperty fo read/write access.
   * 
   * @param prpName
   * @param outputSize
   * @param outputFormat
   * @param inputSize
   * @param inputFormat
   */
  public TExportProperty(String prpName, int outputSize, short outputFormat, int inputSize, short inputFormat)
  {
    initialize(prpName, TAccess.CA_READ, outputSize, outputFormat, inputSize, inputFormat);
  }
  /**
   * Constructor. Creates a property with an ID.
   * 
   * @param prpId The property identifier
   * @param prpName The property name.
   * @param accessMode The access mode associating with the property (CA_READ and/or CA_WRITE).
   * @param din Specifies input format information
   * @param dout Specifies output format information
   */
  public TExportProperty(int prpId, String prpName, short accessMode, TDataType din, TDataType dout)
  {
    this(prpId, prpName, accessMode, din, dout, "");
  }
  private void initialize(String prpName, short accessMode, int outputSize, short outputFormat,
      int inputSize, short inputFormat)
  {
    TStrings.validatePropertyName(prpName);
    this.name = prpName;
    this.id = 0;
    this.accessMode = accessMode;
    this.outputSize = outputSize;
    this.outputFormat = outputFormat;
    this.outputTag = "";
    this.inputSize = inputSize;
    this.inputFormat = inputFormat;
    this.inputTag = "";
    this.description = new TPropertyDescription("");
  }
  /**
   * A Copy Constructor. NOTE: This implementation SIMULATES a copy constructor.
   * It does not copy everything. Useful only for TStockProperty, alias properties !!!.
   * 
   * @param source
   * 
   */
  public TExportProperty(TExportProperty source)
  {
    this.id = source.id;
    this.name = source.name;
    this.outputFormat = source.outputFormat;
    this.outputSize = source.outputSize;
    this.outputTag = source.outputTag;
    this.inputFormat = source.inputFormat;
    this.inputSize = source.inputSize;
    this.inputTag = source.inputTag;
    this.accessMode = source.accessMode;
    // stockProperty formerly had no prpDescription.
    this.description = new TPropertyDescription(source.description.getText());
  }
  private void initialize(int prpId, String prpName, String prpDescription, int prpSizeOut,
      short prpFormatOut, String prpTagOut, int prpSizeIn, short prpFormatIn, String prpTagIn)
  {
    TStrings.validatePropertyName(prpName);
    this.id = prpId;
    this.name = prpName;
    this.description = new TPropertyDescription(prpDescription);
    this.outputFormat = prpFormatOut;
    this.inputFormat = prpFormatIn;
    this.outputSize = prpSizeOut;
    this.inputSize = prpSizeIn;
    this.outputTag = prpTagOut;
    this.inputTag = prpTagIn;
  }
  /**
   * Constructor. Creates a property with an ID.
   * 
   * @param prpId The property identifier
   * @param prpName The property name.
   * @param prpDescription gives a property description.
   * @param prpSizeOut gives the property's output size
   * @param prpFormatOut gives the property's output format
   * @param prpTagOut gives the property's output data tag
   * @param prpSizeIn gives the property's input size
   * @param prpFormatIn gives the property's input format
   * @param prpTagIn gives the property's input data tag
   */
  public TExportProperty(int prpId, String prpName, String prpDescription, int prpSizeOut,
      short prpFormatOut, String prpTagOut, int prpSizeIn, short prpFormatIn, String prpTagIn)
  {
    initialize(prpId, prpName, prpDescription, prpSizeOut, prpFormatOut, prpTagOut, prpSizeIn, prpFormatIn,
        prpTagIn);
  }
  /**
   * Constructor. Creates a property with an ID.
   * 
   * @param prpId The property identifier
   * @param prpName The property name.
   * @param prpDescription gives a property description.
   * @param prpSizeOut gives the property's output size
   * @param prpFormatOut gives the property's output format
   * @param prpSizeIn gives the property's input size
   * @param prpFormatIn gives the property's input format
   */
  public TExportProperty(int prpId, String prpName, String prpDescription, int prpSizeOut,
      short prpFormatOut, int prpSizeIn, short prpFormatIn)
  {
    initialize(prpId, prpName, prpDescription, prpSizeOut, prpFormatOut, "", prpSizeIn, prpFormatIn, "");
  }
  /**
   * Constructor. Creates a property with an ID.
   * 
   * @param prpId The property identifier
   * @param prpName The property name.
   * @param prpDescription gives a property description.
   * @param prpSizeOut gives the property's output size
   * @param prpFormatOut gives the property's output format
   */
  public TExportProperty(int prpId, String prpName, String prpDescription, int prpSizeOut, short prpFormatOut)
  {
    initialize(prpId, prpName, prpDescription, prpSizeOut, prpFormatOut, "", 0, TFormat.CF_NULL, "");
  }
  /**
   * @return Returns the accessMode.
   */
  public short getAccessMode()
  {
    return accessMode;
  }
  /**
   * @param accessMode
   *          The accessMode to set.
   */
  public void setAccessMode(short accessMode)
  {
    this.accessMode = accessMode;
  }
  /**
   * @return Returns the alias.
   */
  public String getAlias()
  {
    return alias;
  }
  /**
   * @param alias
   *          The alias to set.
   */
  public void setAlias(String alias)
  {
    this.alias = alias;
  }
  /**
   * @return Returns the id.
   */
  public int getId()
  {
    return id;
  }
  /**
   * @param id
   *          The id to set.
   */
  public void setId(int id)
  {
    this.id = id;
  }
  /**
   * @return Returns the inputFormat.
   */
  public short getInputFormat()
  {
    return inputFormat;
  }
  /**
   * @param inputFormat
   *          The inputFormat to set.
   */
  public void setInputFormat(short inputFormat)
  {
    this.inputFormat = inputFormat;
  }
  /**
   * @return Returns the inputSize.
   */
  public int getInputSize()
  {
    return inputSize;
  }
  /**
   * @param inputSize
   *          The inputSize to set.
   */
  public void setInputSize(int inputSize)
  {
    this.inputSize = inputSize;
  }
  /**
   * @return Returns the inputTag.
   */
  public String getInputTag()
  {
    return inputTag;
  }
  /**
   * @param inputTag
   *          The inputTag to set.
   */
  public void setInputTag(String inputTag)
  {
    this.inputTag = inputTag;
  }
  /**
   * @return Returns the name.
   */
  public String getName()
  {
    return name;
  }
  /**
   * @param name
   *          The name to set.
   */
  public void setName(String name)
  {
    this.name = name;
  }
  /**
   * @return Returns the numOverloads.
   */
  //public int getNumOverloads() {
  //  return numOverloads;
  //}
  /**
   * @param numOverloads
   *          The numOverloads to set.
   */
  //public void setNumOverloads(int numOverloads) {
  //  this.numOverloads = numOverloads;
  //}
  /**
   * @return Returns the outputFormat.
   */
  public short getOutputFormat()
  {
    return outputFormat;
  }
  /**
   * @param outputFormat
   *          The outputFormat to set.
   */
  public void setOutputFormat(short outputFormat)
  {
    this.outputFormat = outputFormat;
  }
  /**
   * @return Returns the outputSize.
   */
  public int getOutputSize()
  {
    return outputSize;
  }
  /**
   * @param outputSize
   *          The outputSize to set.
   */
  public void setOutputSize(int outputSize)
  {
    this.outputSize = outputSize;
  }
  /**
   * @return Returns the outputTag.
   */
  public String getOutputTag()
  {
    return outputTag;
  }
  /**
   * @param outputTag
   *          The outputTag to set.
   */
  public void setOutputTag(String outputTag)
  {
    this.outputTag = outputTag;
  }
  /**
   * @return Returns the description.
   */
  public TPropertyDescription getDescription()
  {
    return description;
  }
  /**
   * @param description
   *          The description to set.
   */
  public void setDescription(TPropertyDescription description)
  {
    TPropertyEGU rng = new TPropertyEGU(); 
    TPropertyEGU xrng = new TPropertyEGU("",0,0); 
    if (description.getText() == null) description.setText("description unavailable");
    if (description.getYRange() == null) description.setYRange(rng);
    if (description.getXRange() == null) description.setXRange(xrng);
    if (description.getArrayNumRows() == 0) description.setArrayNumRows(1);
    if (outputSize == 1 && description.getArrayType() == 0)
    {
      // can only be a scalar ...
      description.setArrayType(TArrayType.AT_SCALAR);
    }
    this.description = description;
  }
}