/*
 * Created on Nov 5, 2004
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 * CHANGES
 * 2005-02-10 JW Constructor made public
 * 2005-02-10 JW * New Constructor
 *               * egu,xegu renamed to yRange,xRange 
 *               * units made private; get/set methods added.
 * 				 * description and redirection javaized
 *               * description renamed to "text", to avoid things like property.getDescription().getDescription()
 * 				 * array-properties javaized
 * 13.7.05 * PD parseComplexString() 
 */
package de.desy.tine.server.properties;

import de.desy.tine.definitions.TArrayType;
import de.desy.tine.definitions.TGraphType;

/**
 * @brief Provides meta information about an exported property
 * 
 * @author duval 
 */
public class TPropertyDescription
{
  private short arrayType;
  private int arrayRowLength;
  private int arrayNumRows;
  private String text;
  private TPropertyEGU yRange;
  private TPropertyEGU xRange;
  private String redirection;
  /**
   * @brief constructs a TPropertyDescription
   * 
   * This constructor only establishes the text description of the property
   * 
   * @param prpDescription is a string description of the property.
   * 
   */
  public TPropertyDescription(String prpDescription)
  {
    if (prpDescription == null || prpDescription.length() == 0)
      text = "description not given";
    else
      text = prpDescription;
    arrayType = TArrayType.AT_UNKNOWN;
    arrayNumRows = 1;
    arrayRowLength = 0;
    yRange = new TPropertyEGU();
    xRange = new TPropertyEGU();
  }
  /**
   * @brief constructs a TPropertyDescription
   * 
   * @param newArrayType is the array type of the property
   * @param newArrayRowLength is the row length (for multiply dimensioned properties)
   * @param newNumRows is the number of rows (for multiply dimensioned properties)
   */
  public TPropertyDescription(short newArrayType, int newArrayRowLength, int newNumRows)
  {
    this.arrayType = newArrayType;
    this.arrayRowLength = newArrayRowLength;
    this.arrayNumRows = newNumRows;
  }
  /**
   * @brief constructs a TPropertyDescription
   * 
   */
  public TPropertyDescription() {}
  /**
   * @return Returns the xUnits.
   */
  public TPropertyEGU getXRange()
  {
    return xRange;
  }
  /**
   * @param units
   *          The xUnits to set.
   */
  public void setXRange(TPropertyEGU units)
  {
    // set array type to TRACE if still unset. 
    if (units.getMaxValue() != units.getMinValue() &&
        arrayType == 0) arrayType = TArrayType.AT_TRACE;
    xRange = units;
  }
  /**
   * @return Returns the yUnits.
   */
  public TPropertyEGU getYRange()
  {
    return yRange;
  }
  /**
   * @param units
   *          The yUnits to set.
   */
  public void setYRange(TPropertyEGU units)
  {
    yRange = units;
  }
  /**
   * @return Returns the description.
   */
  public String getText()
  {
    return text;
  }
  /**
   * @param description
   *          The description text to set.
   */
  public void setText(String description)
  {
    this.text = description;
  }
  /**
   * @return Returns the redirection.
   */
  public String getRedirection()
  {
    return redirection;
  }
  /**
   * @param redirection
   *          The redirection to set.
   */
  public void setRedirection(String redirection)
  {
    this.redirection = redirection;
  }
  /**
   * @return Returns the arrayNumRows.
   */
  public int getArrayNumRows()
  {
    return arrayNumRows;
  }
  /**
   * @param arrayNumRows
   *          The arrayNumRows to set.
   */
  public void setArrayNumRows(int arrayNumRows)
  {
    this.arrayNumRows = arrayNumRows;
  }
  /**
   * @return Returns the arrayRowLength.
   */
  public int getArrayRowLength()
  {
    return arrayRowLength;
  }
  /**
   * @param arrayRowLength
   *          The arrayRowLength to set.
   */
  public void setArrayRowLength(int arrayRowLength)
  {
    this.arrayRowLength = arrayRowLength;
  }
  /**
   * @return Returns the arrayType.
   */
  public short getArrayType()
  {
    return arrayType;
  }
  /**
   * @param arrayType
   *          The arrayType to set.
   */
  public void setArrayType(short arrayType)
  {
    this.arrayType = arrayType;
  }
  /**
   * @param parseableDescription
   *    A property description string of the form "[ymin:ymax yegu][xmin:xmax xegu]<description>"
   */
  public void parseComplexString(String parseableDescription)
  {
    int start, i, k, mid, stop = -1;
    if (parseableDescription == null) return;
    if ((start=parseableDescription.indexOf('[')) != -1)
    {
      if ((stop=parseableDescription.indexOf(']')) == -1) return; // mal-formed, spit it back out
      TPropertyEGU yegu = new TPropertyEGU();
      mid = stop;
      if ((i=parseableDescription.indexOf('!')) != -1)
      {
        yegu.setGraphType(TGraphType.getType(parseableDescription.substring(i+1,stop)));
        mid = i;
      }
      if ((i=parseableDescription.indexOf(':')) != -1) 
      {
        if ((k=parseableDescription.substring(i+1,mid).indexOf(' ')) == -1) k = mid; else k += i + 1;
        yegu.setMinValue(new Float(parseableDescription.substring(start+1,i)).floatValue());
        yegu.setMaxValue(new Float(parseableDescription.substring(i+1,k)).floatValue());
        if (k != mid) yegu.setUnits(parseableDescription.substring(k+1,mid));
      }
      setYRange(yegu);
      if ((start=parseableDescription.substring(stop+1).indexOf('[')) != -1)
      {
        start += stop + 1;
        if ((stop=parseableDescription.substring(start).indexOf(']')) == -1) return; // mal-formed
        TPropertyEGU xegu = new TPropertyEGU();
        stop += start;
        mid = stop;
        if ((i=parseableDescription.substring(start).indexOf('!')) != -1) mid = i + start;
        if ((i=parseableDescription.substring(start).indexOf(':')) != -1) 
        {
          i += start;
          if ((k=parseableDescription.substring(i+1,mid).indexOf(' ')) == -1) k = mid; else k += i + 1;
          xegu.setMinValue(new Float(parseableDescription.substring(start+1,i)).floatValue());
          xegu.setMaxValue(new Float(parseableDescription.substring(i+1,k)).floatValue());
          if (k != mid) xegu.setUnits(parseableDescription.substring(k+1,mid));
        }
        setXRange(xegu);
      }
    }
    setText(parseableDescription.substring(stop+1));   
  }
}