package de.desy.tine.startup;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.File;
import java.util.Properties;
import java.util.logging.FileHandler;
import de.desy.tine.server.logger.DbgLog;

public class DefaultTInitializerFactory extends TInitializerFactory 
{
  private static final String ENS="131.169.120.41";
  private static final int SRV_PORT=9003;
  private static final int TCP_PORT=8503;
  private static final int STREAM_PORT=10003;
  private static final String GCAST="238.1.1.0";
  private static final int GCAST_PORT=8004;
  private static final String MCAST="238.1.1.1";
  private static final int MCAST_PORT=9503;
  private static final String NETCAST="238.1.1.2"; 
  private static final int NETCAST_PORT=9002;
  private static final int UDP_CLN_PORT = 8050;
  private static int srvRcvBufferSize = 0x10000;
  private static int clnRcvBufferSize = 0x10000;
  private static int sckTimeToLive = 16;
  private static int srvPacketMtu = 1472;
  private static DefaultTInitializer instance = null;
  private class DefaultTInitializer implements TInitializer 
  {
    private String tineHome = null;
    private String fecHome = null;
    private String historyHome = null;
    private String logHome = null;
    private String addrFileLocation = null;
    private String eqpFileLocation = null;
    private String cshostsFileLocation = null;
    private String fecidFileLocation = null;
    private String exportsFileLocation = null;
    private String usersFileLocation = null;
    private String aliasFileLocation = null;
    private String netsFileLocation = null;
    private String alarmwFileLocation = null;
    private String historyFileLocation = null;
    private String logFileLocation = null;
    private String casHostFileLocation = null;
    private Properties tineProperties = new Properties(); 
    public DefaultTInitializer() 
    {
      initialize();
    }
    private boolean hasInitialzed = false;
    private synchronized void initialize() 
    {
      if (hasInitialzed) return;
//    Read properties file.
      try 
      {
        String cp = System.getProperty("java.class.path");
        String cpSeparator = System.getProperty("path.separator");
        String[] cplibs = cp.split(cpSeparator);
        String jarHome = null;
        int stopindex = -1;
        boolean found = false;
        for (int i=0; i<cplibs.length; i++)
        {
          if ((stopindex=cplibs[i].indexOf("tine.jar")) > 0)
          {
            jarHome = cplibs[i].substring(0,stopindex-1);
            break;
          }
        }
        if (stopindex == -1) jarHome = ".";
        String prpsFileName = jarHome + "/tine.properties";
        if (!(new File(prpsFileName).exists())) 
        { // simple check on this directory didn't find anything
          prpsFileName = System.getProperty("user.dir") + "/tine.properties";
        }
        else
        {
          DbgLog.log("initializer","found tine.properties on the class path");
          found = true;
        }
        if (!found)
        {
          if (!(new File(prpsFileName).exists())) 
          { // simple check on home directory didn't find anything
            prpsFileName = System.getProperty("java.ext.dirs") + "/tine.properties";
          }
          else
          {
            DbgLog.log("initializer","found tine.properties in java.ext.dirs directory");
            found = true;
          }
        }
        if (!found)
        {
          if (!(new File(prpsFileName).exists())) 
          { // simple check on ext dirs directory didn't find anything
            prpsFileName = System.getProperty("java.home") + "/tine.properties";
          }
          else
          {
            DbgLog.log("initializer","found tine.properties in java.home directory");
            found = true;
          }
        }
        tineProperties.load(new FileInputStream(prpsFileName));
      } 
      catch (IOException e) 
      {
        DbgLog.log("initializer","tine.properties file not found!");
        DbgLog.log("initializer","use local environment to find repositories!");
      }
      addrFileLocation = getTineHome() + "/fecaddr.csv";
      eqpFileLocation = getTineHome() + "/eqpdbase.csv";
      cshostsFileLocation = getTineHome() + "/cshosts.csv";
      fecidFileLocation = getFecHome() + "/fecid.csv";
      exportsFileLocation = getFecHome() + "/exports.csv";
      usersFileLocation = getFecHome() + "/users.csv";
      aliasFileLocation = getFecHome() + "/alias.csv";
      netsFileLocation = getFecHome() + "/ipnets.csv";
      alarmwFileLocation = getFecHome() + "/almwatch.csv";
      historyFileLocation = getFecHome() + "/history.csv";
      logFileLocation = getLogHome() + "/fec.log";
      casHostFileLocation = getFecHome() + "/cashosts.csv";

      int bs = Integer.parseInt(System.getProperty("srv.RECV_BUFFER_SIZE", "0"));
      if (bs > 0x1000) srvRcvBufferSize = bs;
      bs = Integer.parseInt(System.getProperty("cln.RECV_BUFFER_SIZE", "0"));
      if (bs > 0x1000) clnRcvBufferSize = bs;
      bs = Integer.parseInt(System.getProperty("srv.TIME_TO_LIVE", "0"));
      if (bs > 1) sckTimeToLive = bs;
      bs = Integer.parseInt(System.getProperty("srv.PACKET_MTU", "0"));
      if (bs > 512 && bs < 32000) srvPacketMtu = bs;
      
      hasInitialzed = true;
    }
    public synchronized String getTineHome() 
    {
      if (tineHome == null) 
      {
        String home = System.getProperty("tine.home");
        if (home != null) DbgLog.log("initializer","tine home set from JVM argument");
        else 
        {
          home = tineProperties.getProperty("tine.home");
          if (home != null) DbgLog.log("initializer","tine home set from tine properties");
          else 
          {
            DbgLog.log("initializer","trying to determine tine home from local environment");
            home = System.getenv("TINE_HOME");
            if (home != null) DbgLog.log("initializer","tine home set from TINE_HOME environment");
            else 
            {
              home = System.getenv("CONTROLDB");
              if (home != null) DbgLog.log("initializer","tine home set from legacy environment");
              else DbgLog.log("initializer","unable to determine tine home from local environment");
            }
          }
        }
        if (home != null)
        {
          if (home.endsWith("/") || home.endsWith("\\")) 
            home = home.substring(0,home.length()-1); // will be added later ...
        }
        if (home == null) 
        {
          home = ".";
          File f = new File(home);
          DbgLog.log("initializer","using working directory : " + f.getAbsolutePath() + " as tine home");
          
        }
        //home = "."; // test: force an unresolved ens
        tineHome = new String(home);
      }
      return tineHome;
    }
    public synchronized String getFecHome()
    {
      if (fecHome == null)
      {
        String home = System.getProperty("fec.home");
        if (home == null) home = tineProperties.getProperty("fec.home");
        if (home == null) home = System.getenv("FEC_HOME");
        if (home == null) home = System.getenv("FECDB");
        if (home != null)
        {
          if (home.endsWith("/") || home.endsWith("\\")) 
            home = home.substring(0,home.length()-1); // will be added later ...
        }
        fecHome = home != null ? new String(home) : getTineHome();
      }
      return fecHome;
    }
    public synchronized String getHistoryHome()
    {
      if (historyHome == null)
      {
        String home = System.getProperty("history.home");
        if (home == null) home = tineProperties.getProperty("history.home");
        if (home == null) home = System.getenv("HISTORY_HOME");
        if (home == null) home = System.getenv("HISTORYDB");
        if (home != null)
        {
          if (home.endsWith("/") || home.endsWith("\\")) 
            home = home.substring(0,home.length()-1); // will be added later ...
        }
        else
        {
          home = getFecHome() + "/../HISTORY";
        }
        historyHome = home;
      }
      return historyHome;
    }
    public synchronized String getLogHome()
    {
      if (logHome == null)
      {
        String home = System.getProperty("log.home");
        if (home == null) home = tineProperties.getProperty("log.home");
        if (home == null) home = System.getenv("FEC_LOG");
        if (home == null) home = System.getenv("FECLOG");
        if (home == null) home = System.getenv("LOG_HOME");
        if (home != null)
        {
          if (home.endsWith("/") || home.endsWith("\\")) 
            home = home.substring(0,home.length()-1); // will be added later ...
        }
        logHome = home != null ? new String(home) : getFecHome();
      }
      return logHome;
    }
    private synchronized BufferedReader getCfgResource(String fileLocation)
    {
      if (fileLocation == null) return null;
      try 
      {
        return new BufferedReader(
          new InputStreamReader(new FileInputStream(fileLocation))
        );
      } 
      catch (Exception e) 
      {
        if (fileLocation == cshostsFileLocation)
          DbgLog.log("initializer","Could not open configuration file : " + e);
        return null;
      }      
    }
    public synchronized BufferedWriter getFecLogWriter()
    {
      try 
      {
        return new BufferedWriter(
          new OutputStreamWriter(new FileOutputStream(logFileLocation,true))
        );
      } 
      catch (Exception e) 
      {
        DbgLog.log("initializer","Could not open fec log file " + logFileLocation + " for writing : " + e);
        return null;
      }                  
    }
    public synchronized FileHandler getFecLogResource()
    {
      try 
      {       
        return new FileHandler(logFileLocation,50000,3,true);
      } 
      catch (Exception e) 
      {
        DbgLog.log("initializer","Could not open fec log file " + logFileLocation + " for reading : " + e);
        return null;
      }            
    }
    public BufferedReader getFecIdResource() 
    {
      return getCfgResource(fecidFileLocation);
    }
    public BufferedReader getExportsResource(String eqmName) 
    {
      BufferedReader br;
      if (eqmName != null)
        exportsFileLocation = getFecHome() + "/" + eqmName + "/exports.csv";
      if ((br = getCfgResource(exportsFileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        exportsFileLocation = getFecHome() + "/exports.csv";
        br = getCfgResource(exportsFileLocation);
      }
      return br;
    }
    public BufferedReader getUsersResource() 
    {
      return getCfgResource(usersFileLocation);
    }
    public BufferedReader getNetsResource() 
    {
      return getCfgResource(netsFileLocation);
    }
    public BufferedReader getAliasResource(String eqmName) 
    {
      BufferedReader br;
      if (eqmName != null)
        aliasFileLocation = getFecHome() + "/" + eqmName + "/alias.csv";
      if ((br = getCfgResource(aliasFileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        aliasFileLocation = getFecHome() + "/alias.csv";
        br = getCfgResource(aliasFileLocation);
      }
      return br;
    }
    public BufferedReader getHistoryResource(String eqmName) 
    {
      BufferedReader br;
      if (eqmName != null)
        historyFileLocation = getFecHome() + "/" + eqmName + "/history.csv";
      if ((br = getCfgResource(historyFileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        historyFileLocation = getFecHome() + "/history.csv";
        br = getCfgResource(historyFileLocation);
      }
      return br;
    }
    public BufferedReader getAlmWatchResource(String eqmName) 
    {
      BufferedReader br;
      if (eqmName != null)
        alarmwFileLocation = getFecHome() + "/" + eqmName + "/almwatch.csv";
      if ((br = getCfgResource(alarmwFileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        alarmwFileLocation = getFecHome() + "/almwatch.csv";
        br = getCfgResource(alarmwFileLocation);
      }
      return br;
    }
    public BufferedReader getCasHostResource() 
    {
      return getCfgResource(casHostFileLocation);
    }
    /**
     * @see de.desy.tine.startup.TInitializer#getAddressResource()
     */
    public BufferedReader getAddressResource() 
    {
      return getCfgResource(addrFileLocation);
    }
    /**
     * @see de.desy.tine.startup.TInitializer#getENSAddress()
     */
    public String getENSAddress() { return ENS; }
    /**
     * @see de.desy.tine.startup.TInitializer#getEquipmentResource()
     */
    public BufferedReader getEquipmentResource() 
    {
      return getCfgResource(eqpFileLocation);
    }
    /**
     * @see de.desy.tine.startup.TInitializer#getHostsResource()
     */
    public BufferedReader getHostsResource() 
    {
      return getCfgResource(cshostsFileLocation);
    }
    /**
     * @see de.desy.tine.startup.TInitializer#getGCastAddress()
     */
    public String getGCastAddress() { return GCAST; }
    /**
     * @see de.desy.tine.startup.TInitializer#getGCastPort()
     */
    public int getGCastPort() { return GCAST_PORT; }

    public String getNetCastAddress() { return NETCAST; }
    public int getNetCastPort() { return NETCAST_PORT; }

    /**
     * @see de.desy.tine.startup.TInitializer#getMCastAddress()
     */
    public String getMCastAddress() { return MCAST; }

    /**
     * @see de.desy.tine.startup.TInitializer#getMCastPort()
     */
    public int getMCastPort() { return MCAST_PORT; }

    /**
     * @see de.desy.tine.startup.TInitializer#getSrvPort()
     */
    public int getSrvPort() { return SRV_PORT; }

    /**
     * @see de.desy.tine.startup.TInitializer#getTCPPort()
     */
    public int getTCPPort() { return TCP_PORT; }

    public int getClnPort() { return UDP_CLN_PORT; }
    public BufferedReader getAlmDefinitionResource(String eqmName)
    {
      BufferedReader br;
      if (eqmName == null) return null;
      String fileLocation = getFecHome() + "/" + eqmName + "/" + "alarms.csv";
      if ((br = getCfgResource(fileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        fileLocation = getFecHome() + "/" + eqmName + "-alarms.csv";
        br = getCfgResource(fileLocation);
        if ((br = getCfgResource(fileLocation)) == null)
        { // last chance
          fileLocation = getFecHome() + "/" + "alarms.csv";
          br = getCfgResource(fileLocation);
        }
      }
      return br;
    }
    public BufferedReader getDevicesResource(String eqmName)
    {
      BufferedReader br;
      if (eqmName == null) return null;
      String fileLocation = getFecHome() + "/" + eqmName + "/" + "devices.csv";
      if ((br = getCfgResource(fileLocation)) == null)
      { // first try local name subdirectory then try the standard home
        fileLocation = getFecHome() + "/" + eqmName + "-devices.csv";
        br = getCfgResource(fileLocation);
        if ((br = getCfgResource(fileLocation)) == null)
        { // last chance
          fileLocation = getFecHome() + "/" + "devices.csv";
          br = getCfgResource(fileLocation);
        }
      }
      return br;
    }
    public int getClnRcvBufferSize()
    {
      return clnRcvBufferSize;
    }
    public void setClnRcvBufferSize(int bufferSize)
    {
      if (bufferSize < 16000) bufferSize = 16000;
      DefaultTInitializerFactory.clnRcvBufferSize = bufferSize;
    }
    public int getSckTimeToLive()
    {
      return sckTimeToLive;
    }
    public void setSckTimeToLive(int ttl)
    {
      if (ttl < 1) ttl = 1;
      DefaultTInitializerFactory.sckTimeToLive = ttl;
    }
    public int getSrvPacketMtu()
    {
      return srvPacketMtu;
    }
    public void setSrvPacketMtu(int mtu)
    {
      if (mtu < 512) mtu = 512;
      if (mtu > 64000) mtu = 64000;
      DefaultTInitializerFactory.srvPacketMtu = mtu;
    }
    public int getSrvRcvBufferSize()
    {
      return srvRcvBufferSize;
    }
    public void setSrvRcvBufferSize(int bufferSize)
    {
      if (bufferSize < 16000) bufferSize = 16000;
      DefaultTInitializerFactory.srvRcvBufferSize = bufferSize;
    }
    @Override
    public int getStreamPort()
    {
      return STREAM_PORT;
    }
  }
  /**
   * @see de.desy.tine.startup.TInitializerFactory#getInitializer()
   */
  public synchronized TInitializer getInitializer() 
  {
    if (instance == null) instance = new DefaultTInitializer(); 
    return instance; //new DefaultTInitializer();
  }
}
