/*
 * Created on Jan 28, 2005
 *
 * To change the template for this generated file go to
 * Window&gt;Preferences&gt;Java&gt;Code Generation&gt;Code and Comments
 */
package de.desy.tine.structUtils;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.SwapUtil;
import de.desy.tine.types.*;

/**
 * \internal
 * 
 * This class describes a tagged structure, or rather an array of tagged
 * structures. A TStructDescription contains a set of TStructDefinition.Field
 * which are referenced by tagName. It also contains an arraySize which
 * indicates how often this structure occurs in a sequence. Replaces
 * TStruct.StructStruct TODO Rename to TStructArrayDescription?
 * 
 * @author jwlg
 * @version 2005-02-01
 */
public class TStructDescription
{
  private String tagName = ""; /* structure tag */
  private int rawLength = 0; /* network data size */
  private int size = 0; /* local structure size */
  /*
   * TODO What is this good for? SEE StructStruct.fill(,CF_STRUCT,) . Should
   * size really be the sum of all array sizes of all fields (as the code
   * calculates)?
   */
  private int arraySize; /*
                           * Number of sequential occurrences of this structure
                           * in an array. Doesn't belong here - design flaw for
                           * practical reasons?
                           */
  private ArrayList<Field> myFields = new ArrayList<Field>();
  boolean fieldsComplete = false; // True if all fields are added.
  /**
   * \internal
   * 
   * Describes a field of a tagged structure. Replaces TStruct.StructFormat TODO
   * Document attributes
   */
  public class Field
  {
    int arraySize; /* Number of subsequent occurrences in the structure */
    int format; /* TODO TFormat (short) casted to int intentionally(?) */
    int offset; /* Same as address? */
    int address; // TODO Seems to be unused - always identical to offset
    String name;
    /**
     * Default Constructor. Creates a Field of Format CF_NULL.
     */
    public Field()
    {
      this("",0, TFormat.CF_NULL, 0, 0);
    }
    /**
     * Constucts a Field.
     * 
     * @param newSize
     *          The number of following elements of this type.
     * @param newFormat
     *          The format of the field.
     * @param newOffset
     *          The offset.
     * @param newAddress
     *          The address.
     */
    public Field(int newSize, short newFormat, int newOffset, int newAddress)
    {
      this("",newSize,newFormat,newOffset,newAddress);
    }
    public Field(String name, int newSize, short newFormat, int newOffset, int newAddress)
    {
      arraySize = newSize;
      format = (int) newFormat;
      offset = newOffset;
      address = newAddress;
      this.name = name;
    }
    /**
     * @return Returns the address.
     */
    public int getAddress()
    {
      return address;
    }
    /**
     * Sets the address.
     * 
     * @param address
     *          The address to set.
     */
    public void setAddress(int address)
    {
      this.address = address;
    }
    /**
     * @return Returns the format.
     */
    public short getFormat()
    {
      return (short) format;
    }
    /**
     * Sets the format.
     * 
     * @param format
     *          The format to set.
     */
    public void setFormat(short format)
    {
      this.format = (int) format;
    }
    /**
     * @return Returns the offset.
     */
    public int getOffset()
    {
      return offset;
    }
    public String getName()
    {
      return name;
    }
    /**
     * Sets the offset.
     * 
     * @param offset
     *          The offset to set.
     */
    public void setOffset(int offset)
    {
      this.offset = offset;
    }
    /**
     * @return Returns the size.
     */
    public int getArraySize()
    {
      return arraySize;
    }
    /**
     * Sets the size.
     * 
     * @param size
     *          The size to set.
     */
    public void setArraySize(int size)
    {
      this.arraySize = size;
    }
  }
  /**
   * Constructs a TStructDescription
   * 
   * @param newTagName
   *          Name of the new structure
   */
  public TStructDescription(String newTagName)
  {
    this.tagName = newTagName;
  }
  /**
   * Add a Field
   * 
   * @param format
   *          The TFormat as specified in TFormat
   * @param arraySize
   *          The number of sequential occurrences of this field
   * @see TFormat
   */
  public void addField(short format, int arraySize)
  {
    addField("",format,arraySize);
  }
  public void addField(String name, short format, int arraySize)
  {
    if (!fieldsComplete)
    { // TODO Is this necessary?
      Field newField = new Field(name, arraySize, format, this.rawLength, this.rawLength);
      this.rawLength += arraySize * TFormat.formatSizeOf(format);
      // in java, we're only ever dealing with a byte stream, so
      // size = rawLength = size in bytes always
      this.size = this.rawLength;
      myFields.add(newField);
    }
  }
  /**
   * Add a field multiple times.
   * 
   * @param num
   *          Multiplicity of the field
   * @param format
   *          The TFormat as specified in TFormat
   * @param arraySize
   *          The number of sequential occurrences of this field
   */
  public void addField(int num, short format, int arraySize)
  {
    for (int i = 0; i < num; i++)
    {
      addField(format, arraySize);
    }
  }
  /**
   * Erases the contained information
   */
  private void clear()
  {
    this.arraySize = 0;
    this.rawLength = 0;
    this.fieldsComplete = false;
    this.size = 0;
    this.myFields.clear();
  }
  /**
   * Clears everything and permits definition of fields. TODO Is this necessary?
   */
  public void beginDefinition()
  {
    clear();
  }
  /**
   * Closes definition of fields. TODO Is this necessary?
   */
  public void endDefinition()
  {
    // addField(TFormat.CF_NULL,1);
    fieldsComplete = true;
  }
  /**
   * @return Returns the number of sequential occurrences of this structure.
   */
  public int getArraySize()
  {
    return arraySize;
  }
  /**
   * Sets the number of sequential occurrences of this structure.
   * 
   * @param arraySize
   *          The arraySize to set.
   */
  public void setArraySize(int arraySize)
  {
    this.arraySize = arraySize;
  }
  /**
   * @return Returns the tagName.
   */
  public String getTagName()
  {
    return tagName;
  }
  /**
   * @param tagName
   *          The tagName to set.
   */
  public void setTagName(String name)
  {
    this.tagName = name;
  }
  /**
   * @return Returns the rawLength.
   */
  public int getRawLength()
  {
    return rawLength;
  }
  /**
   * @return Returns the registered.
   */
  public boolean isFieldsComplete()
  {
    return fieldsComplete;
  }
  /**
   * @return Returns the size.
   */
  public int getSize()
  {
    return size;
  }
  /**
   * Returns a HasMap of all contained fields. The hash map contains keys of
   * type String and entries of type Field.
   */
  public ArrayList<Field> getFields()
  {
    return myFields;
  }
  /**
   * Returns an iterator over all fields. The iterator refers to an entry of
   * type Field.
   * <P>
   * <B>Note: </B>Don't use the iterator to remove Fields!
   * </P>
   * 
   * @return
   */
  public Iterator<Field> getFieldIterator()
  {
    return myFields.iterator();
  }
  public int getNFields()
  {
    return myFields.size();
  }
  public Field getField(int i)
  {
    Field rv = null;
    if ((i >= 0) && (i < myFields.size()))
    {
      rv = (Field) myFields.get(i);
    }
    return rv;
  }
  /**
   * Returns the field sizes and formats in a differently coded form used by
   * de.desy.tine.client.
   * 
   * @return LongInt Array containing type/format pairs.
   */
  public INTINT[] toLongIntArray()
  {
    int nFields = myFields.size();
    INTINT[] li = new INTINT[nFields + 1];
    Field currentField;
    for (int i = 0; i < nFields; i++)
    {
      currentField = getField(i);
      li[i] = new INTINT(currentField.getArraySize(), currentField.getFormat() + 512);
    }
    // append terminating NULL
    li[nFields] = new INTINT(1, TFormat.CF_NULL + 512);
    return li;
  }
  public NAME16II[] toNameIntIntArray()
  {
    int nFields = myFields.size();
    NAME16II[] li = new NAME16II[nFields + 1];
    Field currentField;
    for (int i = 0; i < nFields; i++)
    {
      currentField = getField(i);
      li[i] = new NAME16II(currentField.getName(),currentField.getArraySize(), currentField.getFormat() + 512);
    }
    // append terminating NULL
    li[nFields] = new NAME16II("",1, TFormat.CF_NULL + 512);
    return li;
  }
  /**
   * Converts an array of bytes into a string. This method uses the field
   * descriptions to convert an array of bytes into string using a special
   * format. TODO Describe the special format.
   * 
   * @param data
   *          A byte array containing the data
   * @param offset
   *          Offset in the byte array.
   * @return Result of conversion.
   */
  public String convertBytesToString(byte[] data, int offset)
  {
    Field currentField;
    int len;
    byte[] bstr;
    StringBuffer buffer = new StringBuffer();
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(data, offset, this.rawLength);
      DataInputStream ds = new DataInputStream(dis);
      for (int n = 0; n < myFields.size(); n++)
      {
        currentField = getField(n);
        len = currentField.getArraySize();
        if (len > 0)
        {
          buffer.append("[" + currentField.getName() + "] ->");
          switch ((currentField.getFormat()) % 512)
          {
            case TFormat.CF_BYTE:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                buffer.append(ds.readByte());
              }
              break;
            case TFormat.CF_TEXT:
              bstr = new byte[len];
              ds.read(bstr, 0, len);
              int nlen;
              for (nlen = 0; nlen < len; nlen++) if (bstr[nlen] == 0) break;
              buffer.append(new String(bstr,0,nlen).trim());
              break;
            case TFormat.CF_SHORT:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                buffer.append(SwapUtil.swapShort(ds.readShort()));
              }
              break;
            case TFormat.CF_LONG:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                buffer.append(SwapUtil.swapInt(ds.readInt()));
              }
              break;
            case TFormat.CF_DLONG:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                buffer.append(SwapUtil.swapLong(ds.readLong()));
              }
              break;              
            case TFormat.CF_FLOAT:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                // avoid NaN !
                //buffer.append(SwapUtil.swapFloat(ds.readFloat()));
                buffer.append(SwapUtil.swapFloat(ds.readInt()));
              }
              break;
            case TFormat.CF_DOUBLE:
              for (int i = 0; i < len; i++)
              {
                buffer.append(i > 0 ? "," : " ");
                // avoid NaN !
                //buffer.append(SwapUtil.swapDouble(ds.readDouble()));
                buffer.append(SwapUtil.swapDouble(ds.readLong()));
              }
              break;
          } // Type Switch
          buffer.append("\n");
        } // if len > 0
      } // for
    }
    catch (IOException e)
    {
      e.printStackTrace();
    }
    return buffer.toString();
  }
  /**
   * Overloaded for Convenience.
   * 
   * @param data
   * @see convertBytesToString(byte[],int)
   */
  public String convertBytesToString(byte[] data)
  {
    return convertBytesToString(data, 0);
  }
}