package de.desy.tine.types;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import de.desy.tine.definitions.TErrorList;
import de.desy.tine.definitions.TFormat;
import de.desy.tine.endianUtils.Swap;
import de.desy.tine.server.logger.MsgLog;

public final class NAME64DBL implements TCompoundDataObject
{
  public String name;
  public double dval;
  public double getDoubleValue() { return (double)dval; }
  public int getIntegerValue() { return (int)dval; }
  public static final int sizeInBytes = 72;
  public static final short dFormat = TFormat.CF_NAME64DBL;
  public void setValues(String newName, double newDValue)
  {
    name = newName; dval = newDValue;
  }
  public void setName(String newName) { name = newName; }
  public String getName() { return name; }
  public int getSizeInBytes() { return sizeInBytes; }
  public short getFormat() { return dFormat; }    
  ByteArrayOutputStream dBuffer = new ByteArrayOutputStream(sizeInBytes);
  public byte[] hByteArray;
  public void setByteArray(byte[] byteArray) { hByteArray = byteArray; }
  public byte[] getByteArray() { return hByteArray; }
  public NAME64DBL() {name = new String(new StringBuffer(64)); dval = 0;}  
  public NAME64DBL(String Name) { name = Name; dval = 0;}  
  public NAME64DBL(String Name,double d) { name = Name; dval = d;}  
  public NAME64DBL(NAME64DBL instance) 
  { 
    name = instance.name == null ? new String(new StringBuffer(64)) : new String(instance.name); 
    dval = instance.dval; 
  }
  
  @Override
  public String toString()
  {
      return "["+ name + ", " + dval + "]";
  } 
  public byte[] toByteArray()
  {
    if (name == null) name = new String(new StringBuffer(64));
    try
    {
      dBuffer.reset();
      DataOutputStream ds = new DataOutputStream(dBuffer);
      byte[] bname = name.getBytes();
      byte[] b = new byte[64];
      for (int i=0; i<64 && i<bname.length; i++) b[i] = bname[i];
      ds.write(b,0,b.length);
      byte[] db = Swap.Double(dval);
      ds.write(db,0,8);
      ds.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("NAME64DBL","could not prepare byte stream",TErrorList.code_failure,e,0);
      return null;
    }
    hByteArray = dBuffer.toByteArray();
    return hByteArray;
  }  
  public void toStruct()
  {
    try
    {
      ByteArrayInputStream dis = new ByteArrayInputStream(hByteArray);
      DataInputStream ds = new DataInputStream(dis);
      byte[] b = new byte[sizeInBytes];
      ds.read(b,0,64);
      int len = 0; for (int i=0; i<64 && b[i] != 0; i++, len++);
      name = new String(b,0,len);
      byte[] db = new byte[8];
      ds.read(db,0,8);
      dval = Swap.Double(db);
      ds.close(); dis.close();
    }
    catch (IOException e)
    {
      e.printStackTrace();
      MsgLog.log("NAME64DBL","could not prepare byte stream",TErrorList.code_failure,e,0);
    }  
  }
  @Override
  public int hashCode()
  {
    final int PRIME = 31;
    int result = 1;
    long temp;
    temp = Double.doubleToLongBits(dval);
    result = PRIME * result + (int) (temp ^ (temp >>> 32));
    result = PRIME * result + ((name == null) ? 0 : name.hashCode());
    return result;
  }
  @Override
  public boolean equals(Object obj)
  {
    if (this == obj) return true;
    if (obj == null) return false;
    if (getClass() != obj.getClass()) return false;
    final NAME64DBL other = (NAME64DBL) obj;
    if (Double.doubleToLongBits(dval) != Double.doubleToLongBits(other.dval)) return false;
    if (name == null)
    {
      if (other.name != null) return false;
    }
    else if (!name.equals(other.name)) return false;
    return true;
  }
}